/*****************************************************************************/
/*
                                 Config.c

This module performs three functions.

  1.  It reads the HTTPD configuration file, converting the directives
      into internal representations.

  2.  Provides functions, accessed by other modules, returning the
      original directive information as appropriate.

  3.  Provides HTML form-based configuration and reports.

Configuration can only be performed at server startup.  To reset the 
configuration the server must be restarted.

WASD HTTPd Configuration is a relevant subset of the CERN configuration 
directives (sort-of, well it started that way anyhow!)  Although there are
some  differences (beware!) the CERN documentation can be consulted for a
general discussion of the purpose and functioning some of these directives. 

The file formats may be IN ONE OF TWO possible formats (not mixed).  The
older format contains one directive at the start of a line, followed by a
mandatory parameter.  The newer format (in part to support forms-based
configuration) has the directive contained within square brackets (e.g.
"[accept]") followed on the same line, or by one or more lines containing
comments, containing parameter information, or empty.  The directive applies
to all lines following up until another directive (square bracket in column
one) is encountered.

Directives are:

  o  Accept             <host/domain>
  o  ActivityDays       <integer>
  o  AddType            <suffix> <content-type> -|</script> [<description>]
  o  AddIcon            <URL> <alternative-text> <content-type-template>
  o  AddDirIcon         <URL> <alternative-text>
  o  AddBlankIcon       <URL> <alternative-text>
  o  AddUnknownIcon     <URL> <alternative-text>
  o  AddParentIcon      <URL> <alternative-text>
  o  AuthBasic          DISABLED | ENABLED
  o  AuthDigest         DISABLED | ENABLED
  o  AuthDigestGetLife  <integer>
  o  AuthDigestPutLife  <integer>
  o  AuthVms            DISABLED | ENABLED
  o  Busy               <integer>
  o  CommentedInfo      DISABLED | ENABLED
  o  DclHardLimit       <integer>
  o  DclSoftLimit       <integer>
  o  DclCgiPlusLifeTime <integer>
  o  DclSpawnAuthPriv   DISABLED | ENABLED
  o  DclFullRequest     DISABLED | ENABLED
  o  DclZombieLifeTime  <integer>
  o  DirAccess          DISABLED | ENABLED | SELECTIVE
  o  DirLayout          <string>
  o  DirOwner           DISABLED | ENABLED
  o  DirPreExpired      DISABLED | ENABLED
  o  DirReadMe          DISABLED | TOP | BOTTOM
  o  DirReadMeFile      <file.suffix>
  o  DirWildcard        DISABLED | ENABLED
  o  DNSLookup          DISABLED | ENABLED
  o  ErrorInfo          <string>
  o  ErrorSysAdmin      <string>
  o  ErrorRecommend     DISABLED | ENABLED
  o  InputTimeout       <integer>
  o  KeepAliveTimeout   <integer>
  o  Logging            DISABLED | ENABLED
  o  Monitor            DISABLED | ENABLED
  o  OutputTimeout      <integer>
  o  Port               <integer>
  o  PutMaxKbytes       <integer>
  o  Reject             <host/domain>
  o  RequestHistory     <integer>
  o  Search             <absolute-path-to-search-script>
  o  Ssi | Shtml                      DISABLED | ENABLED
  o  SsiAccesses | ShtmlAccesses      DISABLED | ENABLED
  o  SsiExec | ShtmlExec              DISABLED | ENABLED
  o  StreamLf           <integer>
  o  Welcome            <file.suffix>


VERSION HISTORY
---------------
01-AUG-97  MGD  v4.3 activity number of days, DCL supply full request,
                bugfix; annoying repeated instances of "unknown" content type
01-JUN-97  MGD  v4.2 required additional DCL/scripting parameters
01-FEB-97  MGD  HTTPd version 4;
                major changes for form-based config;
                added error message URLs
01-OCT-96  MGD  HTML configuration report
01-AUG-96  MGD  StreamLF conversion directive;
                AuthBasic and AuthDigest directives;
12-APR-96  MGD  file record/binary now determined by record format
                (has removed the need of the text/binary, etc., in "AddType")
01-DEC-95  MGD  HTTPd version 3
27-SEP-95  MGD  added auto-scripting funtionality
07-AUG-95  MGD  added support for VMS-style directory structure and file sizes
16-JUN-95  MGD  file contents description in 'AddType' (also see dir.c module)
20-DEC-94  MGD  developed for multi-threaded HTTP daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

/* VMS related header files */
#include <rmsdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>

/* application related header */
#include "wasd.h"
#include "config.h"
#include "error.h"
#include "support.h"
#include "httpd.h"

/******************/
/* global storage */
/******************/

struct ConfigStruct  Config;

char  ConfigContentTypeBlank [] = "x-internal/blank",
      ConfigContentTypeDir [] = "x-internal/directory",
      ConfigContentTypeParent [] = "x-internal/parent",
      ConfigContentTypeUnknown [] = "x-internal/unknown",
      ConfigDefaultContentType [] = "x-internal/unknown",
      ConfigAddTypeUnknown [] = "* x-internal/unknown -";

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern int  ServerPort;
extern char  ClientHostName[];
extern char  ClientInternetAddress[];
extern char  ErrorStringSize[];
extern char  HtmlSgmlDoctype[];
extern char  ServerHostName[];
extern char  SoftwareID[];
extern char  Utility[];

/*****************************************************************************/
/*
Open the configuration file defined by the logical name HTTPD$CONFIG.  Read 
each line, passing to the appropriate interpretation function according to the 
directive on the line.  Close the file.  Function is either called before
AST-driven  processing or called at AST delivery level which makes processing
atomic.  Allow 32 extra characters on line buffer and parameter allocated
memory for concatenating additional string.
*/

int Configure (struct ConfigStruct *cfptr)

{
#  define PARAMETER_ELBOW_ROOM 32

   register char  *cptr, *sptr, *zptr;

   boolean  DebugBuffer,
            BracketDirectives,
            EquateDirectives;
   int  status,
        ParemeterLength,
        SetPrvStatus,
        LineLength;
   void  *RequestPtrBuffer;
   char  ExpandedFileName [256],
         Line [256 + PARAMETER_ELBOW_ROOM];
   struct FAB  ConfigFileFab;
   struct NAM  ConfigFileNam;
   struct RAB  ConfigFileRab;
   struct XABDAT  ConfigFileXabDat;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "Configure()\n");

   /* not interested anymore in seeing debug information for config load! */
   DebugBuffer = Debug;
   Debug = false;

   RequestPtrBuffer = cfptr->RequestPtr;
   memset (cfptr, 0, sizeof(struct ConfigStruct));
   cfptr->RequestPtr = RequestPtrBuffer;

   BracketDirectives = EquateDirectives = false;
   sys$gettim (&cfptr->LoadBinTime);

   /************************/
   /* open the config file */
   /************************/

   if (Debug) fprintf (stdout, "ConfigFileName |%s|\n", ConfigFileName);

   ConfigFileFab = cc$rms_fab;
   ConfigFileFab.fab$b_fac = FAB$M_GET;
   ConfigFileFab.fab$l_fna = ConfigFileName;
   ConfigFileFab.fab$b_fns = strlen(ConfigFileName);
   ConfigFileFab.fab$l_nam = &ConfigFileNam;
   ConfigFileFab.fab$b_shr = FAB$M_SHRGET;
   ConfigFileFab.fab$l_xab = &ConfigFileXabDat;

   ConfigFileNam = cc$rms_nam;
   ConfigFileNam.nam$l_esa = ExpandedFileName;
   ConfigFileNam.nam$b_ess = sizeof(ExpandedFileName)-1;
   ConfigFileNam.nam$l_rsa = cfptr->LoadFileName;
   ConfigFileNam.nam$b_rss = sizeof(cfptr->LoadFileName)-1;

   /* initialize the date extended attribute block */
   ConfigFileXabDat = cc$rms_xabdat;

   /* turn on SYSPRV to allow access to possibly protected file */
   EnableSysPrv ();
   status = sys$open (&ConfigFileFab, 0, 0);
   DisableSysPrv ();

   /* status from sys$open() */
   if (VMSnok (status))
   {
      if (Debug) fprintf (stdout, "sys$open() %%X%08.08X\n", status);
      if (cfptr->RequestPtr)
      {
         ConfigReportProblem (cfptr, NULL, status);
         return;
      }
      else
         ErrorExitVmsStatus (status, ConfigFileName, __FILE__, __LINE__);
   }

   cfptr->LoadFileName[ConfigFileNam.nam$b_rsl] = '\0';
   if (Debug)
      fprintf (stdout, "cfptr->LoadFileName |%s|\n", cfptr->LoadFileName);

   /* get the configuration file revision date and time */
   memcpy (&cfptr->RevBinTime,
           &ConfigFileXabDat.xab$q_rdt,
           sizeof(cfptr->RevBinTime));

   /* record access block */
   ConfigFileRab = cc$rms_rab;
   ConfigFileRab.rab$l_fab = &ConfigFileFab;
   /* try and optimise sequential read (next 3 lines) */
   ConfigFileRab.rab$b_mbc = 6;
   ConfigFileRab.rab$b_mbf = 2;
   ConfigFileRab.rab$l_rop = RAB$M_RAH;
   ConfigFileRab.rab$l_ubf = Line;
   ConfigFileRab.rab$w_usz = sizeof(Line)-PARAMETER_ELBOW_ROOM-1;

   if (VMSnok (status = sys$connect (&ConfigFileRab, 0, 0)))
   {
      sys$close (&ConfigFileFab, 0, 0);
      if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
      if (cfptr->RequestPtr)
      {
         ConfigReportProblem (cfptr, NULL, status);
         return;
      }
      else
         ErrorExitVmsStatus (status, ConfigFileName, __FILE__, __LINE__);
   }

   /************************/
   /* read the config file */
   /************************/

   cfptr->DirPreExpired = PRE_EXPIRE_INDEX_OF;
   cfptr->DefaultPost = "/ht_root/script/post";
   cfptr->HtmlContentType = "text/html";
   cfptr->IsMapContentType = "text/x-ismap";
   cfptr->MenuContentType = "text/x-menu";
   cfptr->PlainTextContentType = "text/plain";
   cfptr->SsiContentType = "text/x-shtml";
   strcpy (cfptr->DefaultDirLayout, DefaultDirListingLayout);
   cfptr->ErrorInfoPtr = cfptr->ErrorSysAdminPtr =
      cfptr->AcceptHostsPtr = cfptr->RejectHostsPtr = 
      cfptr->ContentTypeListHeadPtr = cfptr->IconListHeadPtr = NULL;
   cfptr->ContentTypeStructOverhead = sizeof(struct ContentTypeStruct);
   cfptr->IconStructOverhead = sizeof(struct IconStruct);

   while (VMSok (status = sys$get (&ConfigFileRab, 0, 0)))
   {
      cfptr->LineNumber++;
      Line[LineLength = ConfigFileRab.rab$w_rsz] = '\0';
      if (Debug) fprintf (stdout, "Line %d |%s|\n", cfptr->LineNumber, Line);

      if (Line[0] == '#') continue;
      cptr = Line;
      while (*cptr && isspace(*cptr) && *cptr == '\n') cptr++;
      if (!*cptr) continue;

      if (Line[0] == '[')
      {
         if (EquateDirectives)
         {
            ConfigReportProblem (cfptr, "mixed directive formats", 0);
            break;
         }
         BracketDirectives = true;
         cptr++;
         zptr = (sptr = cfptr->DirectiveName) + sizeof(cfptr->DirectiveName);
         while (*cptr && *cptr != ']' && sptr < zptr) *sptr++ = *cptr++;
         if (sptr >= zptr)
         {
            ConfigReportProblem (cfptr, "missing closing bracket?", 0);
            continue;
         }
         *sptr = '\0';
         if (Debug)
            fprintf (stdout, "DirectiveName |%s|\n", cfptr->DirectiveName);
         if (*cptr) cptr++;
         while (*cptr && isspace(*cptr)) cptr++;
      }
      else
      if (!BracketDirectives)
      {
         EquateDirectives = true;
         zptr = (sptr = cfptr->DirectiveName) + sizeof(cfptr->DirectiveName);
         while (*cptr && !isspace(*cptr) && *cptr != '=' && sptr < zptr)
             *sptr++ = *cptr++;
         if (sptr >= zptr)
         {
            ConfigReportProblem (cfptr, "missing parameter");
            continue;
         }
         *sptr = '\0';
         while (*cptr && isspace(*cptr)) cptr++;
         if (!*cptr)
         {
            ConfigReportProblem (cfptr, "missing parameter", 0);
            continue;
         }
      }

      if (BracketDirectives && !*cptr) continue;

      if (strsame (cfptr->DirectiveName, "Accept", -1))
         ConfigSetAccept (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ActivityDays", -1))
         cfptr->ActivityNumberOfDays = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AddType", -1))
         ConfigAddType (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AddIcon", -1))
         ConfigAddIcon (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AddBlankIcon", -1))
      {
         /* use up the PARAMETER_ELBOW_ROOM :^) */
         for (sptr = cptr; *sptr; sptr++);
         *sptr++ = ' ';
         strcpy (sptr, ConfigContentTypeBlank);
         ConfigAddIcon (cfptr, cptr);
      }
      else
      if (strsame (cfptr->DirectiveName, "AddDirIcon", -1))
      {
         /* use up the PARAMETER_ELBOW_ROOM :^) */
         for (sptr = cptr; *sptr; sptr++);
         *sptr++ = ' ';
         strcpy (sptr, ConfigContentTypeDir);
         ConfigAddIcon (cfptr, cptr);
      }
      else
      if (strsame (cfptr->DirectiveName, "AddParentIcon", -1))
      {
         /* use up the PARAMETER_ELBOW_ROOM :^) */
         for (sptr = cptr; *sptr; sptr++);
         *sptr++ = ' ';
         strcpy (sptr, ConfigContentTypeParent);
         ConfigAddIcon (cfptr, cptr);
      }
      else
      if (strsame (cfptr->DirectiveName, "AddUnknownIcon", -1))
      {
         /* use up the PARAMETER_ELBOW_ROOM :^) */
         for (sptr = cptr; *sptr; sptr++);
         *sptr++ = ' ';
         strcpy (sptr, ConfigContentTypeUnknown);
         ConfigAddIcon (cfptr, cptr);
      }
      else
      if (strsame (cfptr->DirectiveName, "AuthBasic", -1))
         cfptr->AuthBasicEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AuthDigestGetLife", -1))
         cfptr->AuthDigestNonceGetLifeTime =
            ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AuthDigestPutLife", -1))
         cfptr->AuthDigestNoncePutLifeTime =
            ConfigSetInteger (cfptr, cptr);
      else
      /* "AuthDigest" MUST come after other AuthDigests for obvious reasons */
      if (strsame (cfptr->DirectiveName, "AuthDigest", -1))
         cfptr->AuthDigestEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AuthVMS", -1))
         cfptr->AuthVmsEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Busy", -1))
         cfptr->Busy = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CommentedInfo", -1))
         cfptr->IncludeCommentedInfo = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclHardLimit", -1))
         cfptr->DclSubprocessHardLimit = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclSoftLimit", -1))
         cfptr->DclSubprocessSoftLimit = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclCgiPlusLifeTime", -1))
         cfptr->DclCgiPlusLifeTime = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclZombieLifeTime", -1))
         cfptr->DclZombieLifeTime = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclSpawnAuthPriv", -1))
         cfptr->DclSpawnAuthPriv = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclFullRequest", -1))
         cfptr->DclFullRequest = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirAccess", -1))
         ConfigSetDirAccess (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirLayout", -1))
         ConfigSetDirLayout (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirOwner", -1))
         cfptr->DirOwnerEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirPreExpired", -1))
         cfptr->DirPreExpired = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirReadMeFile", -1))
         ConfigSetDirReadMeFile (cfptr, cptr);
      else
      /* "ReadMe" MUST come after the other ReadMes for obvious reasons */
      if (strsame (cfptr->DirectiveName, "DirReadMe", -1))
         ConfigSetDirReadMe (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirWildcard", -1))
         cfptr->DirWildcardEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DNSLookup", -1))
         cfptr->DnsLookup = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ErrorInfo", -1) ||
          strsame (cfptr->DirectiveName, "ErrorSysAdmin", -1))
         ConfigSetError (cfptr, cptr);
      else
      /* "Recommend" for backward compatibility */
      if (strsame (cfptr->DirectiveName, "ErrorRecommend", -1) ||
          strsame (cfptr->DirectiveName, "Recommend", -1))
         cfptr->ErrorRecommend = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "InputTimeout", -1))
         cfptr->InputTimeoutMinutes = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "KeepAliveTimeout", -1))
         cfptr->KeepAliveTimeoutSeconds = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Logging", -1))
         cfptr->LoggingEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Monitor", -1))
         cfptr->MonitorEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "OutputTimeout", -1))
         cfptr->OutputTimeoutMinutes = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Port", -1))
         cfptr->ServerPort = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "PutMaxKbytes", -1))
         cfptr->PutMaxKbytes = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "PutVersionLimit", -1))
         cfptr->PutVersionLimit = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Reject", -1))
         ConfigSetReject (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "RequestHistory", -1))
         cfptr->RequestHistory = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Search", -1))
         ConfigSetSearch (cfptr, cptr);
      else
      /* "ShtmlAccesses" for backward compatibility */
      if (strsame (cfptr->DirectiveName, "SSIAccesses", -1) ||
          strsame (cfptr->DirectiveName, "ShtmlAccesses", -1))
         cfptr->SsiAccessesEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      /* "ShtmlExec" for backward compatibility */
      if (strsame (cfptr->DirectiveName, "SSIExec", -1) ||
          strsame (cfptr->DirectiveName, "ShtmlExec", -1))
         cfptr->SsiExecEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      /* "SSI" MUST come after the other SSIs for obvious reasons */
      /* "Shtml" for backward compatibility */
      if (strsame (cfptr->DirectiveName, "SSI", -1) ||
          strsame (cfptr->DirectiveName, "Shtml", -1))
         cfptr->SsiEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "StreamLF", -1))
         cfptr->StreamLfConversionMaxKbytes =
            ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Welcome", -1))
         ConfigSetWelcome (cfptr, cptr);
      else
         ConfigReportProblem (cfptr, "?", 0);
   }

   if (status == RMS$_EOF) status = SS$_NORMAL;

   /*************************/
   /* close the config file */
   /*************************/

   sys$close (&ConfigFileFab, 0, 0); 

   if (!cfptr->PutMaxKbytes)
      cfptr->PutMaxKbytes = DefaultPutMaxKbytes;
   if (!cfptr->PutVersionLimit)
      cfptr->PutVersionLimit = DefaultPutVersionLimit;
   if (cfptr->ErrorSysAdminPtr == NULL) cfptr->ErrorSysAdminPtr = "";
   if (cfptr->ErrorInfoPtr == NULL) cfptr->ErrorInfoPtr = "";

   if (VMSnok (status))
   {
      if (cfptr->RequestPtr)
      {
         ConfigReportProblem (cfptr, NULL, status);
         return;
      }
      else
         ErrorExitVmsStatus (status, ConfigFileName, __FILE__, __LINE__);
   }

   Debug = DebugBuffer;

   return (status);
}

/*****************************************************************************/
/*
Return true or false values equivalent to the string pointed to by 'pptr'.
*/

boolean ConfigSetBoolean
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   /*********/
   /* begin */
   /*********/

   if (strsame (pptr, "YES", 3) ||
       strsame (pptr, "ON", 2) ||
       strsame (pptr, "ENABLED", 7))
      return (true);

   if (strsame (pptr, "NO", 2) ||
       strsame (pptr, "OFF", 3) ||
       strsame (pptr, "DISABLED", 8))
      return (false);

   ConfigReportProblem (cfptr, NULL, 0);
   return (false);
}

/*****************************************************************************/
/*
Return an integer equivalent to the numeric string pointed to by 'pptr'.
*/

int ConfigSetInteger
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   /*********/
   /* begin */
   /*********/

   if (isdigit(*pptr)) return (atoi(pptr));
   ConfigReportProblem (cfptr, NULL, 0);
   return (-1);
}

/*****************************************************************************/
/*
This function formats an error report.  All lines are concatenated onto a
single string of dynamically allocated memory that (obviously) grows as
reports are added to it.  This string is then output if loading the server
configuration or is available for inclusion in an HTML page.
*/

int ConfigReportProblem
(
struct ConfigStruct *cfptr,
char *Explanation,
int StatusValue
)
{
   static $DESCRIPTOR (DirectiveFaoDsc,
                       "%HTTPD-W-CONFIG, \"!AZ\" directive !AZ at line !UL\n");
   static $DESCRIPTOR (ExplanationFaoDsc,
                       "%HTTPD-W-CONFIG, \"!AZ\" !AZ at line !UL\n");

   int  status;
   unsigned short  Length;
   char  Buffer [256],
         HtmlBuffer [512];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigReportProblem()\n");

   cfptr->ProblemCount++;

   if (StatusValue)
   {
      if (VMSok (status = sys$getmsg (StatusValue, &Length, &BufferDsc, 0, 0)))
         Buffer[Length++] = '\n';
   }
   else
   if (Explanation == NULL)
      status = sys$fao (&DirectiveFaoDsc, &Length, &BufferDsc,
                        cfptr->DirectiveName, "problem", cfptr->LineNumber);
   else
   if (Explanation[0] == '?')
      status = sys$fao (&DirectiveFaoDsc, &Length, &BufferDsc,
                        cfptr->DirectiveName, "unknown", cfptr->LineNumber);
   else
      status = sys$fao (&ExplanationFaoDsc, &Length, &BufferDsc,
                        Explanation, cfptr->LineNumber);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
      ErrorExitVmsStatus (status, "sys$fao()", __FILE__, __LINE__);
   else
      Buffer[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Buffer);

   if (cfptr->RequestPtr == NULL) fputs (Buffer, stdout);

   if ((Length = CopyToHtml (HtmlBuffer, sizeof(HtmlBuffer), Buffer, -1)) < 0)
      ErrorExitVmsStatus (SS$_BUFFEROVF, "CopyToHtml()", __FILE__, __LINE__);

   if ((cfptr->ProblemReportPtr =
        realloc (cfptr->ProblemReportPtr, cfptr->ProblemReportLength+Length+1))
       == NULL)
      ErrorExitVmsStatus (vaxc$errno, "realloc()", __FILE__, __LINE__);
   /* include the terminating null */
   memcpy (cfptr->ProblemReportPtr+cfptr->ProblemReportLength,
           HtmlBuffer, Length+1);
   cfptr->ProblemReportLength += Length;
}

/*****************************************************************************/
/*
Add a mime content type corresponding to the file suffix (extension, type).

Format: "AddType <suffix> <content-type> <-|/script> [<description>]"

Examples: "AddType .html text/html 8bit - HTML"
          "AddType .txt text/plain 7bit - plain text"
          "AddType .hlb text/x-script binary /conan VMS help library"
          "AddType .gif image/gif binary - GIF image"

The directive is stored as an singly-linked list of structures of dynamically 
allocated memory containing a reformatted version of the directive string.  
Function ConfigContentType() searches this list to get pointers to the content 
type, any auto-script, and description, into the client request data structure.
*/ 

ConfigAddType
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register char  *cptr, *sptr;
   register struct ContentTypeStruct  *clptr;

   int  status,
        Count = 0;
   char  *DescriptionPtr = "",
         *ContentTypePtr = "",
         *AutoScriptNamePtr = "",
         *SuffixPtr = "";
   char  HtmlDescription [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigAddType() |%s|\n", pptr);

   /* suffix (file type/extension) */
   SuffixPtr = pptr;
   while (*pptr && !isspace(*pptr))
   {
      *pptr = toupper(*pptr);
      pptr++;
   }
   Count += pptr - SuffixPtr;
   if (*pptr) *pptr++ = '\0';

   /* "mime" content type (representation) */
   while (*pptr && isspace(*pptr)) pptr++;
   ContentTypePtr = pptr;
   while (*pptr && !isspace(*pptr)) pptr++;
   Count += pptr - ContentTypePtr;
   if (*pptr) *pptr++ = '\0';

   /* (optional) script activation */
   while (*pptr && isspace(*pptr)) pptr++;
   AutoScriptNamePtr = pptr;
   while (*pptr && !isspace(*pptr)) pptr++;
   Count += pptr - AutoScriptNamePtr;
   if (*pptr) *pptr++ = '\0';

   /* (optional) description, to end of line */
   while (*pptr && isspace(*pptr)) pptr++;
   DescriptionPtr = pptr;
   /* find the end-of-line */
   while (*pptr) pptr++;
   if (*pptr) *pptr++ = '\0';
   /* escape any HTML-forbidden characters */
   Count += CopyTextIntoHtml (HtmlDescription, DescriptionPtr, -1);

   if (!(*SuffixPtr && *ContentTypePtr))
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   /* plus structure overhead, plus 4 x null terminators */
   Count += cfptr->ContentTypeStructOverhead + 4;
   if ((clptr = (struct ContentTypeStruct*)malloc(Count)) == NULL)
      ErrorExitVmsStatus (vaxc$errno, "malloc()", __FILE__, __LINE__);

   if (cfptr->ContentTypeListHeadPtr == NULL)
      cfptr->ContentTypeListHeadPtr = clptr; 
   else
      cfptr->ContentTypeListTailPtr->NextPtr = clptr; 
   cfptr->ContentTypeListTailPtr = clptr;
   clptr->NextPtr = NULL;

   sptr = (char*)clptr + cfptr->ContentTypeStructOverhead;

   clptr->SuffixPtr = sptr;
   cptr = SuffixPtr;
   /* ensure suffix is in upper-case so we don't have to do it matching */
   while (*cptr) *sptr++ = toupper(*cptr++);
   *sptr++ = '\0';

   clptr->ContentTypePtr = sptr;
   cptr = ContentTypePtr;
   while (*cptr) *sptr++ = *cptr++;
   *sptr++ = '\0';

   clptr->AutoScriptNamePtr = sptr;
   cptr = AutoScriptNamePtr;
   while (*cptr) *sptr++ = *cptr++;
   *sptr++ = '\0';

   clptr->ContentDescriptionPtr = sptr;
   cptr = HtmlDescription;
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';
}

/*****************************************************************************/
/*
This function searches (linear, so put the most frequently hit suffixes at the
start) the list of content type directives, pointing the appropriate content
type method in the client request data structure.   Attempts to improve
efficiency by buffering a pointer the last content-type matched.  This yields
benefits only where repeated calls are made to this function by the one
thread, e.g. during directory listings, where there are often series of the
same file type.

If 'rqptr' is NULL returns a pointer to the content-type of the supplied
'Suffix', if non-NULL updates some thread data structure members.  Always
returns a pointer to the content type.
*/ 

char* ConfigContentType
(
struct ContentTypeStruct *ctptr,
char *Suffix
)
{
   register char  *cptr, *sptr;
   register struct ContentTypeStruct  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigContentType() |%s|\n", Suffix);

   if (ctptr == NULL)
      clptr = Config.ContentTypeListHeadPtr;
   else
   if ((clptr = ctptr->LastFoundContentTypePtr) == NULL)
      clptr = Config.ContentTypeListHeadPtr;
    
   while (clptr != NULL)
   {
      /* 'clptr->SuffixPtr' has already been converted to upper case! */
      cptr = clptr->SuffixPtr;
      /** if (Debug) fprintf (stdout, "|%s|\n", cptr); **/
      sptr = Suffix;
      while (*cptr && *sptr && *sptr != ';' && *cptr == toupper(*sptr))
         { cptr++; sptr++; }

      if (*cptr || (*sptr && *sptr != ';'))
      {
         if (ctptr == NULL)
         {
            clptr = clptr->NextPtr;
            continue;
         }
         else
         if (ctptr->LastFoundContentTypePtr == NULL)
         {
            clptr = clptr->NextPtr;
            continue;
         }
         clptr = Config.ContentTypeListHeadPtr;
         ctptr->LastFoundContentTypePtr = NULL;
         continue;
      }

      /**********/
      /* match! */
      /**********/

      /* if just getting a pointer to a content type then just return that */
      if (ctptr == NULL) return (clptr->ContentTypePtr);

      ctptr->ContentTypePtr = clptr->ContentTypePtr;
      ctptr->AutoScriptNamePtr = clptr->AutoScriptNamePtr;
      ctptr->ContentDescriptionPtr = clptr->ContentDescriptionPtr;
      /* note the location of this icon information */
      ctptr->LastFoundContentTypePtr = clptr;

      if (!Debug) return (clptr->ContentTypePtr);

      fprintf (stdout, "|%s|%s|%s|\n",
               ctptr->ContentTypePtr, ctptr->AutoScriptNamePtr,
               ctptr->ContentDescriptionPtr);
      return (clptr->ContentTypePtr);
   }

   /*************/
   /* no match! */
   /*************/

   /* if just getting a pointer to a content type then just return that */
   if (ctptr == NULL) return (ConfigDefaultContentType);

   ctptr->ContentTypePtr = ConfigDefaultContentType;
   ctptr->AutoScriptNamePtr = ctptr->ContentDescriptionPtr = "";
   return (ConfigDefaultContentType);
}

/*****************************************************************************/
/*
Add a URL for an icon corresponding to the mime content type.

Format:   "AddIcon <URL> <alternative-text> <content-type-template>"

Examples: "AddIcon /icon/httpd/text.xbm [TXT] text/plain"
          "AddIcon /icon/httpd/doc.xbm [HTM] text/html"
          "AddIcon /icon/httpd/image.xbm [IMG] image/jpeg"

The template can contain a wildcard in place of a string in either side of the 
template slash.  In this case the wildcard matches any string on that side.

The directive is stored as an list of structures of dynamically allocated 
memory containing a reformatted version of the directive string, including a 
complete HTML anchor for the URL.  Function ConfigIconFor() searches this 
list to return a pointer to the anchor.
*/ 

ConfigAddIcon
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register struct IconStruct  *ilptr;

   int  Count = 0;
   char  *AltTextPtr,
         *IconUrlPtr,
         *ContentTypePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigAddIcon()\n");

   IconUrlPtr = pptr;
   while (*pptr && !isspace(*pptr)) pptr++;
   Count = pptr - IconUrlPtr;
   if (*pptr) *pptr++ = '\0';

   /* alternative text */
   while (*pptr && isspace(*pptr)) pptr++;
   AltTextPtr = pptr;
   while (*pptr && !isspace(*pptr))
   {
      /* underscores represent spaces! (especially for the "blank" icon) */
      if (*pptr == '_') *pptr = ' ';
      pptr++;
   }
   Count += pptr - AltTextPtr;
   if (*pptr) *pptr++ = '\0';

   /* template */
   while (*pptr && isspace(*pptr)) pptr++;
   ContentTypePtr = pptr;
   while (*pptr && !isspace(*pptr))
   {
      /* force to upper case to save a toupper() in ConfigIconFor() */
      *pptr = toupper(*pptr);
      pptr++;
   }
   Count += pptr - ContentTypePtr;
   if (*pptr) *pptr++ = '\0';

   if (!(*IconUrlPtr && *AltTextPtr && *ContentTypePtr))
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   /* plus structure overhead, plus the HTML IMG element used below */
   Count += cfptr->IconStructOverhead + 31;
   if ((ilptr = (struct IconStruct*)malloc(Count)) == NULL)
      ErrorExitVmsStatus (vaxc$errno, "malloc()", __FILE__, __LINE__);

   if (cfptr->IconListHeadPtr == NULL)
      cfptr->IconListHeadPtr = ilptr; 
   else
      cfptr->IconListTailPtr->NextPtr = ilptr; 
   cfptr->IconListTailPtr = ilptr;
   ilptr->NextPtr = NULL;

   sprintf ((char*)ilptr + cfptr->IconStructOverhead,
            "%s <IMG ALIGN=top SRC=\"%s\" ALT=\"%s\">",
            ContentTypePtr, IconUrlPtr, AltTextPtr);
   if (Debug)
      fprintf (stdout, "|%s|\n", (char*)ilptr + cfptr->IconStructOverhead);
}

/*****************************************************************************/
/*
This function searches the list of icons for an one corresponding to the 
supplied mime content type.  The directive content type and associated HTML 
anchor (containing the original configuration-supplied URL) is stored as an 
list of structures dynamically allocated memory containing a string.  Attempts 
to improve efficiency by buffering a pointer the last content-type matched.  
This yields benefits only where repeated calls are made to this function by 
the one thread, e.g. during directory listings, where there are often series 
of the same file type.
*/

char* ConfigIconFor
(
struct IconStruct *icptr,
char *ContentType
)
{
   register char  *cptr, *sptr;
   register struct IconStruct  *ilptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigIconFor() |%s|\n", ContentType);

   if ((ilptr = (struct IconStruct*)icptr->LastFoundIconPtr) == NULL)
      ilptr = Config.IconListHeadPtr;

   while (ilptr != NULL)
   {
      cptr = (char*)ilptr + Config.IconStructOverhead;
      /** if (Debug) fprintf (stdout, "|%s|\n", cptr); **/
      sptr = ContentType;
      while (*cptr && *cptr != '*' && *sptr && *cptr == toupper(*sptr))
         { cptr++; sptr++; }
      if (*cptr == ' ' && !*sptr)
      {
         /* hit!  note the location of this icon information */
         icptr->LastFoundIconPtr = (unsigned char*)ilptr;
         /* return a pointer to the HTML IMG element for the icon */
         if (*cptr) cptr++;
         return (cptr);
      }
      if (*cptr == '*')
      {
         /* template contained a wildcard on the left side of the slash */
         while (*cptr && *cptr != '/' && !isspace(*cptr)) cptr++;
         if (*cptr) cptr++;
         /* find the right side of the content type */
         while (*sptr && *sptr != '/') sptr++;
         if (*sptr) sptr++;
         /* compare the right side of the slash */
         while (*cptr && *cptr != '*' && *sptr && *cptr == toupper(*sptr))
            { cptr++; sptr++; }
         if (*cptr == '*')
         {
            while (*cptr == '*') cptr++;
            while (*sptr) sptr++;
         }
         if (*cptr == ' ' && !*sptr)
         {
            /* hit!  note the location of this icon information */
            icptr->LastFoundIconPtr = (unsigned char*)ilptr;
            /* return a pointer to the HTML IMG element for the icon */
            if (*cptr) cptr++;
            return (cptr);
         }
      }

      if (icptr->LastFoundIconPtr == NULL)
      {
         ilptr = ilptr->NextPtr;
         continue;
      }
      ilptr = Config.IconListHeadPtr;
      icptr->LastFoundIconPtr = NULL;
   }
   /* nothing found, essentially an error! */
   return ("");
}

/*****************************************************************************/
/*
*/

ConfigSetAccept
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register char  *sptr;

   int  Comma;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetAccept()\n");

   if (!*pptr)
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   for (sptr = pptr; *sptr; *sptr++);
   if (cfptr->AcceptHostsPtr == NULL)
      Comma = 0;
   else
      Comma = 1;
   if ((cfptr->AcceptHostsPtr =
        realloc (cfptr->AcceptHostsPtr,
                 cfptr->AcceptHostsLength+sptr-pptr+Comma+1)) == NULL)
      ErrorExitVmsStatus (vaxc$errno, "memory reallocation",
                          __FILE__, __LINE__);
   if (Comma) cfptr->AcceptHostsPtr[cfptr->AcceptHostsLength++] = ',';
   memcpy (cfptr->AcceptHostsPtr+cfptr->AcceptHostsLength, pptr, sptr-pptr+1);
   cfptr->AcceptHostsLength += sptr-pptr;
   if (Debug)
      fprintf (stdout, "cfptr->AcceptHostsPtr |%s|\n", cfptr->AcceptHostsPtr);
}

/*****************************************************************************/
/*
*/

ConfigSetReject
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register char  *sptr;

   int  Comma;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetReject()\n");

   if (!*pptr)
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   for (sptr = pptr; *sptr; *sptr++);
   if (cfptr->RejectHostsPtr == NULL)
      Comma = 0;
   else
      Comma = 1;
   if ((cfptr->RejectHostsPtr =
        realloc (cfptr->RejectHostsPtr,
                 cfptr->RejectHostsLength+sptr-pptr+Comma+1)) == NULL)
      ErrorExitVmsStatus (vaxc$errno, "memory reallocation",
                          __FILE__, __LINE__);
   if (Comma) cfptr->RejectHostsPtr[cfptr->RejectHostsLength++] = ',';
   memcpy (cfptr->RejectHostsPtr+cfptr->RejectHostsLength, pptr, sptr-pptr+1);
   cfptr->RejectHostsLength += sptr-pptr;
   if (Debug)
      fprintf (stdout, "cfptr->RejectHostsPtr |%s|\n", cfptr->RejectHostsPtr);
}

/*****************************************************************************/
/*
Accept or reject the client connection.  Lists of hosts must be a comma-
separated string.  A wildcard '*' in a domain specification matches any string 
that occurs between the dots.
Example: "host1.network1.domain1,*.network2.domain1,*.*.domain2". 
*/ 

boolean ConfigAcceptClientHostName (char *ClientHostName)

{
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ConfigAcceptConnection() |%s|\n", ClientHostName);

   if (Config.AcceptHostsPtr == NULL && Config.RejectHostsPtr == NULL)
      return (true);

   if (!ClientHostName[0] || ClientHostName[0] == '?') return (false);

   /******************/
   /* hosts accepted */
   /******************/

   if (Config.AcceptHostsPtr != NULL)
   {
      sptr = Config.AcceptHostsPtr;
      while (*sptr)
      {
         /** if (Debug) fprintf (stdout, "sptr |%s|\n", sptr); **/
         if (isdigit(*sptr))
            cptr = ClientInternetAddress;
         else
            cptr = ClientHostName;
         while (*cptr)
         {
            if (*sptr == ',') break;
            if (*sptr == '*')
            {
               while (*sptr && *sptr != '.' && *sptr != ',') sptr++;
               while (*cptr && *cptr != '.') cptr++;
            }
            if (tolower(*cptr) != tolower(*sptr)) break;
            if (*cptr) cptr++;
            if (*sptr) sptr++;
         }
         if (!*cptr && (!*sptr || *sptr == ',')) break;
         while (*sptr && *sptr != ',') sptr++;
         if (*sptr) sptr++;
      }
      /* if it was NOT found then reject the connection request */
      if (!(!*cptr && (!*sptr || *sptr == ',')))
         return (false);
   }

   /******************/
   /* hosts rejected */
   /******************/

   if (Config.RejectHostsPtr != NULL)
   {
      sptr = Config.RejectHostsPtr;
      while (*sptr)
      {
         /** if (Debug) fprintf (stdout, "sptr |%s|\n", sptr); **/
         cptr = ClientHostName;
         while (*cptr)
         {
            if (*sptr == ',') break;
            if (*sptr == '*')
            {
               while (*sptr && *sptr != '.' && *sptr != ',') sptr++;
               while (*cptr && *cptr != '.') cptr++;
            }
            if (tolower(*cptr) != tolower(*sptr)) break;
            if (*cptr) cptr++;
            if (*sptr) sptr++;
         }
         if (!*cptr && (!*sptr || *sptr == ',')) break;
         while (*sptr && *sptr != ',') sptr++;
         if (*sptr) sptr++;
      }
      /* if it WAS found then reject the connection request */
      if (!*cptr && (!*sptr || *sptr == ','))
         return (false);
   }

   /* accept the connection */
   return (true);
}

/*****************************************************************************/
/*
Set directory access booleans.  Allowed values OFF, ON, SELECTIVE.
*/

ConfigSetDirAccess
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetDirAccess()\n");

   if (strsame (pptr, "YES", 3) ||
       strsame (pptr, "ON", 2) ||
       strsame (pptr, "ENABLED", 7))
   {
      cfptr->DirAccess = true;
      cfptr->DirAccessSelective = false;
   }
   else
   if (strsame (pptr, "NO", 2) ||
       strsame (pptr, "OFF", 3) ||
       strsame (pptr, "DISABLED", 8))
      cfptr->DirAccess = cfptr->DirAccessSelective = false;
   else
   if (strsame (pptr, "SELECTIVE", 9))
      cfptr->DirAccess = cfptr->DirAccessSelective = true;
   else
      ConfigReportProblem (cfptr, NULL, 0);
}

/*****************************************************************************/
/*
*/

ConfigSetDirLayout
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register char  *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetDirLayout()\n");

   zptr = (sptr = cfptr->DefaultDirLayout) + sizeof(cfptr->DefaultDirLayout);
   while (*pptr && sptr < zptr) *sptr++ = *pptr++;
   if (sptr >= zptr)
   {
      cfptr->DefaultDirLayout[0] = '\0';
      ConfigReportProblem (cfptr, NULL, 0);
   }
}

/*****************************************************************************/
/*
Set directory README file booleans.  Allowed values OFF, TOP, BOTTOM.
*/

ConfigSetDirReadMe
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   if (Debug) fprintf (stdout, "ConfigSetDirReadMe()\n");

   if (strsame (pptr, "NO", 2) ||
       strsame (pptr, "OFF", 3) ||
       strsame (pptr, "DISABLED", 8))
      cfptr->DirReadMeBottom = cfptr->DirReadMeTop = false;
   else
   if (strsame (pptr, "BOTTOM", 6))
   {
      cfptr->DirReadMeTop = false;
      cfptr->DirReadMeBottom = true;
   }
   else
   if (strsame (pptr, "TOP", 3))
   {
      cfptr->DirReadMeTop = true;
      cfptr->DirReadMeBottom = false;
   }
   else
      ConfigReportProblem (cfptr, NULL, 0);
}

/*****************************************************************************/
/*
Add a "read-me" file name.

Format:   "ReadMeFile file.suffix"

Examples: "ReadMeFile README.HTML"
          "ReadMeFile README.TXT"
          "ReadMeFile README."

The directive is stored as an array of pointers to dynamically allocated 
memory containing the "read-me" file name.  Function ConfigReadMeFile() gets 
these by index number.
*/

ConfigSetDirReadMeFile
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   int  status,
        Count;
   char  *NamePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetDirReadMeFile()\n");

   if (cfptr->ReadMeFileCount >= MaxConfigReadMeFiles)
   {
      ConfigReportProblem (cfptr, "maximum read-me files", 0);
      return;
   }

   NamePtr = pptr;
   while (*pptr && !isspace(*pptr))
   {
      *pptr = toupper(*pptr);
      pptr++;
   }
   Count = pptr - NamePtr;
   *pptr = '\0';

   if (!*NamePtr)
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   if (Debug) fprintf (stdout, "malloc()ing %d bytes\n", Count+1);
   if ((cfptr->ReadMeFileArray[cfptr->ReadMeFileCount] =
        malloc (Count+1)) == NULL)
      ErrorExitVmsStatus (vaxc$errno, "malloc()", __FILE__, __LINE__);

   strcpy (cfptr->ReadMeFileArray[cfptr->ReadMeFileCount], NamePtr);
   if (Debug)
      fprintf (stdout, "|%s|\n",
               cfptr->ReadMeFileArray[cfptr->ReadMeFileCount]);
   cfptr->ReadMeFileCount++;
}

/*****************************************************************************/
/*
See Error.c module for information on how these strings are used.
Limitation: due to empty lines being ignored way up there :^) the
'ErrorInfoPtr' text cannot contain any completely empty lines (use "&nbsp;").
*/

boolean ConfigSetError
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register char  *cptr;

   boolean  ErrorInfo;
   int  Index,
        Count;

   /*********/
   /* begin */
   /*********/

   if (ErrorInfo = strsame (cfptr->DirectiveName, "ErrorInfo", -1))
      cptr = cfptr->ErrorInfoPtr;
   else
   {
      if ((cptr = cfptr->ErrorSysAdminPtr) != NULL)
      {
         /* only one per customer :^) */
         free (cptr);
         cptr = NULL;
      }
   }

   Count = strlen(pptr) + 1;
   /* allow for the appended newline */
   if (ErrorInfo) Count++;
   if (Debug) fprintf (stdout, "malloc()ing %d bytes\n", Count);

   if (cptr == NULL)
   {
      if ((cptr = calloc (1, Count)) == NULL)
         ErrorExitVmsStatus (vaxc$errno, "calloc()", __FILE__, __LINE__);
   }
   else
   {
      if ((cptr = realloc (cptr, strlen(cptr) + Count)) == NULL)
         ErrorExitVmsStatus (vaxc$errno, "realloc()", __FILE__, __LINE__);
   }

   if (ErrorInfo)
   {
      strcat (cfptr->ErrorInfoPtr = cptr, pptr);
      strcat (cptr, "\n");
   }
   else
   {
      if (Count > 128)
      {
         ConfigReportProblem (cfptr, "exceeds maximum of 128 characters", 0);
         free (cptr);
      }
      else
         strcpy (cfptr->ErrorSysAdminPtr = cptr, pptr);
   }
}

/*****************************************************************************/
/*
Return a pointer to the "read-me" file name stored against the index number of 
the array.  If none exists return a null string.  The calling function would 
use this by stepping through the index numbers 0 to n, until the null string 
was returned indicating the possible "read-me" file names were exhausted.
*/

char* ConfigReadMeFile (int Number)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigReadMeFile |%d|\n", Number);

   if (Number >= 0 && Number < Config.ReadMeFileCount)
      return (Config.ReadMeFileArray[Number]);
   else
      return ("");
}

/*****************************************************************************/
/*
*/

ConfigSetSearch
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register char  *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetSearch()\n");

   if (*pptr)
   {
      zptr = (sptr = cfptr->KeywordSearch) + sizeof(cfptr->KeywordSearch);
      while (*pptr && !isspace(*pptr) && sptr < zptr) *sptr++ = *pptr++;
      if (sptr >= zptr)
      {
         ConfigReportProblem (cfptr, NULL, 0);
         sptr = cfptr->KeywordSearch;
      }
      *sptr = '\0';
   }
   else
      ConfigReportProblem (cfptr, NULL, 0);
}

/*****************************************************************************/
/*
Add a welcome (home) page file name.

Format:   "Welcome file.suffix"

Examples: "Welcome HOME.HTML"
          "Welcome HOME.HTM"
          "Welcome HOME.MENU"

The directive is stored as an array of pointers to dynamically allocated 
memory containing the home page file name.  Function ConfigHomePage() gets 
these by index number.
*/ 

ConfigSetWelcome
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   int  status,
        Count;
   char  *WelcomePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetWelcome()\n");

   if (cfptr->HomePageCount >= MaxConfigHomePages)
   {
      ConfigReportProblem (cfptr, "maximum welcome files", 0);
      return;
   }

   WelcomePtr = pptr;
   while (*pptr && !isspace(*pptr))
   {
      *pptr = toupper(*pptr);
      pptr++;
   }
   Count = pptr - WelcomePtr;
   *pptr = '\0';

   if (!*WelcomePtr)
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   if (Debug) fprintf (stdout, "malloc()ing %d bytes\n", Count+1);
   if ((cfptr->HomePageArray[cfptr->HomePageCount] = malloc (Count+1)) == NULL)
      ErrorExitVmsStatus (vaxc$errno, "malloc()", __FILE__, __LINE__);

   strcpy (cfptr->HomePageArray[cfptr->HomePageCount], WelcomePtr);
   if (Debug)
      fprintf (stdout, "|%s|\n", cfptr->HomePageArray[cfptr->HomePageCount]);
   cfptr->HomePageCount++;
}

/*****************************************************************************/
/*
Return a pointer to the home page (welcome) file name stored against the index 
number of the array.  If none exists return a null string.  The calling 
function would use this by stepping through the index numbers 0 to n, until 
the null string was returned indicating the possible home page file names were 
exhausted.
*/

char* ConfigHomePage (int Number)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigHomePage |%d|\n", Number);

   if (Number >= 0 && Number < Config.HomePageCount)
      return (Config.HomePageArray[Number]);
   else
      return ("");
}

/*****************************************************************************/
/*
Return a report on the HTTPd server's configuration.  This function blocks
while executing.
*/ 

ConfigReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
boolean  UseServerDatabase
)
{
/* macro to push booleans onto the sys$faol() parameter list */
#define REPBOOL(b) if (b) *vecptr++ = "[yes]"; else *vecptr++ = "[no]";

   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ:!UL ... Server Configuration</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>HTTPd !AZ:!UL</H1>\n\
<H2>Server Configuration</H2>\n\
!AZ\n");

   static $DESCRIPTOR (ProblemReportFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Problem!AZ !AZ</FONT></TH></TR>\n\
<TR><TD><PRE>!AZ</PRE></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (ReportGeneralFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Source: &quot;!AZ&quot;</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=2 NOBORDER>\n\
<TR><TH ALIGN=right>File:</TH>\
<TD ALIGN=left>!AZ</TD>\
<TD>&nbsp;</TD><TH>[<A HREF=\"!AZ\">View</A>]</TH></TR>\n\
<TR><TH ALIGN=right>!AZ:</TH>\
<TD ALIGN=left>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=4>General</TH></TR>\n\
<TR><TH></TH><TH>enabled</TH><TH COLSPAN=2>value</TH></TR>\n\
<TR><TH>Busy</TH><TD ALIGN=center></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>connections</FONT></TR>\n\
<TR><TH>Request History</TH><TD ALIGN=center></TD><TD>!UL</TD></TR>\n\
<TR><TH>Activity History</TH><TD ALIGN=center></TD><TD>!UL</TD>\n\
<TD><FONT SIZE=-1>days</FONT></TD></TR>\n\
<TR><TH>Commments</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>DNS Lookup</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>Logging</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>Monitor</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>IP Port</TH><TD></TD><TD>!UL</TD></TR>\n\
<TR><TH>StreamLF</TH><TD></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>Kbytes</FONT></TD></TR>\n\
<TR><TH>Search Script</TH><TD></TD><TD COLSPAN=2>!AZ</TD></TR>\n\
<TR><TH><FONT SIZE=-1>PUT/POST</FONT> Max.Length</TH><TD></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>Kbytes</FONT></TD></TR>\n\
<TR><TH><FONT SIZE=-1>PUT/POST</FONT> Files</TH><TD></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>version limit</FONT></TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=3>Timeouts</TH></TR>\n\
<TR><TH></TH><TH COLSPAN=2>value</TH></TR>\n\
<TR><TH>Input</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH>Output</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH>Keep-Alive</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1>seconds</FONT></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (ReportWelcomeFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Welcome (Home) Pages</TH></TR>\n");

   static $DESCRIPTOR (ReportHostsFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=2>Hosts</TH></TR>\n\
<TR><TH>Accept</TH><TD>!AZ</TD></TR>\n\
<TR><TH>Reject</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=5>Authentication</TH></TR>\n\
<TR><TH COLSPAN=2></TH><TH>enabled</TH><TH COLSPAN=2>value</TH></TR>\n\
<TR><TH COLSPAN=2>Basic</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH ROWSPAN=3>Digest</TH><TH>enabled</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>GET lifetime</TH><TD></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH>PUT lifetime</TH><TD></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH COLSPAN=2>VMS</TH><TD ALIGN=center>!AZ</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=4>DCL/Scripting</TH></TR>\n\
<TR><TH></TH><TH>enabled</TH><TH COLSPAN=2>value</TH></TR>\n\
<TR><TH>Spawn Authorized Privileges</TH><TD ALIGN=center>!AZ</TD>\
<TD></TD><TD></TD></TR>\n\
<TR><TH>Supply Header and Body</TH><TD ALIGN=center>!AZ</TD>\
<TD></TD><TD></TD></TR>\n\
<TR><TH>Soft Limit</TH><TD></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>subprocesses</FONT></TD></TR>\n\
<TR><TH>Hard Limit</TH><TD></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>subprocesses</FONT></TD></TR>\n\
<TR><TH>Zombie Life-Time</TH><TD></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH>CGIplus Life-Time</TH><TD></TD><TD>!UL</TD>\
<TD><FONT SIZE=-1>minutes</FONT></TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=2>Server-Side Includes</TH></TR>\n\
<TR><TH></TH><TH>enabled</TH></TR>\n\
<TR><TH>SSI</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>Exec</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>Accesses</TH><TD ALIGN=center>!AZ</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=3>Directory</TH></TR>\n\
<TR><TH></TH><TH>enabled</TH><TH>other</TH></TR>\n\
<TR><TH>Access</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>Layout</TH><TD></TD><TD>!AZ</TD></TR>\n\
<TR><TH>Owner</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>Pre-Expired</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>Readme</TH><TD ALIGN=center>!AZ</TD><TD>!AZ</TD></TR>\n\
<TR><TH>Wildcard</TH><TD ALIGN=center>!AZ</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Readme Files</TH></TR>\n");

   static $DESCRIPTOR (ReportIconsFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=4>Directory Icons</TH></TR>\n\
<TR><TH>content-type</TH><TH>icon</TH><TH>path</TH>\
<TH>alternate text</TH></TR>\n");

   static $DESCRIPTOR (ReportContentTypeFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=4>Content-Type</TH></TR>\n\
<TR><TH>suffix</TH><TH>content-type</TH>\
<TH>auto-script</TH><TH>description</TH></TR>\n");

   static $DESCRIPTOR (ContentTypeFaoDsc,
"<TR><TD>!AZ</TD><TD>!AZ</TD><TD>!AZ</TD><TD>!AZ</TD></TR>\n");

   static $DESCRIPTOR (ReportErrorFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=3>Errors</TH></TR>\n\
<TR><TH></TH><TH>enabled</TH><TH>text</TH></TR>\n\
<TR><TH>Recommend</TH><TD ALIGN=center>!AZ</TD></TR>\n\
<TR><TH>Sys.Admin. Path</TH><TD COLSPAN=2>!AZ</TD></TR>\n\
<TR><TH>Information</TH><TD COLSPAN=2>!AZ!AZ!AZ</TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</BODY>\n\
</HTML>\n");

   static char  StringOverflow [] = "*ERROR* string overflow";

   register int  idx;
   register unsigned long  *vecptr;
   register char  *cptr, *sptr;
   register struct ConfigStruct  *cfptr;
   register struct ContentTypeStruct  *clptr;
   register struct IconStruct  *ilptr;

   int  status,
        CommaCount,
        HostsLength;
   unsigned short  Length;
   unsigned long  FaoVector [128];
   char  Buffer [8192],
         HtmlIconAltText [256],
         HtmlErrorInfo [2048],
         HtmlErrorSysAdmin [128],
         IconAltText [256],
         IconPath [256];
   struct ContentTypeStruct  ContentType;
   struct ConfigStruct  LocalConfig;
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigReport()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   if (UseServerDatabase)
      cfptr = &Config;
   else
   {
      cfptr = &LocalConfig;
      cfptr->RequestPtr = rqptr;
      Configure (cfptr);
   }

   if (cfptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      if ((*vecptr++ = cfptr->ProblemCount) == 1)
         *vecptr++ = "";
      else
         *vecptr++ = "s";
      if (UseServerDatabase)
         *vecptr++ = "Reported At Startup";
      else
         *vecptr++ = "Detected During Load";
      *vecptr++ = cfptr->ProblemReportPtr;

      status = sys$faol (&ProblemReportFaoDsc, &Length, &BufferDsc,
                         &FaoVector);

      /* if using a local database free report memory now! */
      if (!UseServerDatabase) free (cfptr->ProblemReportPtr);

      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      Buffer[Length] = '\0';
      NetWrite (rqptr, 0, Buffer, Length);
   }

   vecptr = FaoVector;

   /* source file */
   if (UseServerDatabase)
      *vecptr++ = "Server";
   else
      *vecptr++ = "File";
   *vecptr++ = cfptr->LoadFileName;
   *vecptr++ = MapUrl (NULL, cfptr->LoadFileName, NULL, NULL);
   if (UseServerDatabase)
   {
      *vecptr++ = "Loaded:";
      *vecptr++ = DayDateTime (&cfptr->LoadBinTime, 20);
   }
   else
   {
      *vecptr++ = "Revised:";
      *vecptr++ = DayDateTime (&cfptr->RevBinTime, 20);
   }
   /* general */
   *vecptr++ = cfptr->Busy;
   *vecptr++ = cfptr->RequestHistory;
   *vecptr++ = cfptr->ActivityNumberOfDays;
   REPBOOL (cfptr->IncludeCommentedInfo)
   REPBOOL (cfptr->DnsLookup)
   REPBOOL (cfptr->LoggingEnabled)
   REPBOOL (cfptr->MonitorEnabled)
   *vecptr++ = cfptr->ServerPort;
   *vecptr++ = cfptr->StreamLfConversionMaxKbytes;
   *vecptr++ = cfptr->KeywordSearch;
   *vecptr++ = cfptr->PutMaxKbytes;
   *vecptr++ = cfptr->PutVersionLimit;

   /* timeouts */
   *vecptr++ = cfptr->InputTimeoutMinutes;
   *vecptr++ = cfptr->OutputTimeoutMinutes;
   *vecptr++ = cfptr->KeepAliveTimeoutSeconds;

   status = sys$faol (&ReportGeneralFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   /*****************************/
   /* welcome (home page) files */
   /*****************************/

   NetWrite (rqptr, 0, ReportWelcomeFaoDsc.dsc$a_pointer,
                       ReportWelcomeFaoDsc.dsc$w_length);

   for (idx = 0; idx < cfptr->HomePageCount; idx++)
   {
      memcpy (sptr = Buffer, "<TR><TD>", 8);
      sptr += 8;
      for (cptr = cfptr->HomePageArray[idx]; *cptr; *sptr++ = *cptr++);
      memcpy (sptr, "</TD></TR>\n", 12);  /* includes terminating null */
      sptr += 11;
      NetWrite (rqptr, 0, Buffer, sptr-Buffer);
   }

   NetWrite (rqptr, 0, "</TABLE>\n", 9);

   /*********************/
   /* host access, etc. */
   /*********************/

   vecptr = FaoVector;

   if (cfptr->AcceptHostsPtr == NULL)
      *vecptr++ = "<FONT SIZE=-1><I>(none)</I></FONT>";
   else
   {
      /* introduce spaces to allow the string to be wrapped by the browser */
      CommaCount = 0;
      for (cptr = cfptr->AcceptHostsPtr; *cptr; cptr++)
         if (*cptr == ',') CommaCount++;
      HostsLength = cptr - cfptr->AcceptHostsPtr + CommaCount + 1;
      if ((*vecptr++ = sptr = HeapAlloc (rqptr, HostsLength)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }
      for (cptr = cfptr->AcceptHostsPtr; *cptr; cptr++)
      {
         *sptr++ = *cptr;
         if (*cptr == ',') *sptr++ = ' ';
      }
      *sptr = '\0';
   }

   if (cfptr->RejectHostsPtr == NULL)
      *vecptr++ = "<FONT SIZE=-1><I>(none)</I></FONT>";
   else
   {
      /* introduce spaces to allow the string to be wrapped by the browser */
      CommaCount = 0;
      for (cptr = cfptr->RejectHostsPtr; *cptr; cptr++)
         if (*cptr == ',') CommaCount++;
      HostsLength = cptr - cfptr->RejectHostsPtr + CommaCount + 1;
      if ((*vecptr = sptr = HeapAlloc (rqptr, HostsLength)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }
      for (cptr = cfptr->RejectHostsPtr; *cptr; cptr++)
      {
         *sptr++ = *cptr;
         if (*cptr == ',') *sptr++ = ' ';
      }
      *sptr = '\0';
   }

   /* authentication */
   REPBOOL (cfptr->AuthBasicEnabled)
   REPBOOL (cfptr->AuthDigestEnabled)
   *vecptr++ = cfptr->AuthDigestNonceGetLifeTime;
   *vecptr++ = cfptr->AuthDigestNonceGetLifeTime;
   REPBOOL (cfptr->AuthVmsEnabled)

   /* DCL/Scripting */
   REPBOOL (cfptr->DclSpawnAuthPriv)
   REPBOOL (cfptr->DclFullRequest)
   *vecptr++ = cfptr->DclSubprocessSoftLimit;
   *vecptr++ = cfptr->DclSubprocessHardLimit;
   *vecptr++ = cfptr->DclZombieLifeTime;
   *vecptr++ = cfptr->DclCgiPlusLifeTime;

   /* SSI (.shtml) */
   REPBOOL (cfptr->SsiEnabled)
   REPBOOL (cfptr->SsiExecEnabled)
   REPBOOL (cfptr->SsiAccessesEnabled)

   REPBOOL (cfptr->DirAccess)
   *vecptr++ = cfptr->DefaultDirLayout;
   REPBOOL (cfptr->DirOwnerEnabled)
   REPBOOL (cfptr->DirPreExpired)
   REPBOOL (cfptr->DirReadMeTop | cfptr->DirReadMeBottom)
   if (cfptr->DirReadMeTop)
      *vecptr++ = "top";
   else
   if (cfptr->DirReadMeBottom)
      *vecptr++ = "bottom";
   else
      *vecptr++ = "";
   REPBOOL (cfptr->DirWildcardEnabled)

   status = sys$faol (&ReportHostsFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   /*********************/
   /* directory (contd) */
   /*********************/

   for (idx = 0; idx < cfptr->ReadMeFileCount; idx++)
   {
      memcpy (sptr = Buffer, "<TR><TD>", 8);
      sptr += 8;
      for (cptr = cfptr->ReadMeFileArray[idx]; *cptr; *sptr++ = *cptr++);
      memcpy (sptr, "</TD></TR>\n", 12);  /* includes terminating null */
      sptr += 11;
      NetWrite (rqptr, 0, Buffer, sptr - Buffer);
   }

   NetWrite (rqptr, 0, "</TABLE>\n", 9);

   NetWrite (rqptr, 0, ReportIconsFaoDsc.dsc$a_pointer,
                       ReportIconsFaoDsc.dsc$w_length);

   /*********/
   /* icons */
   /*********/

   ilptr = cfptr->IconListHeadPtr;
   while (ilptr != NULL)
   {
      /* find the ALT="" text from the icon IMG */ 
      IconAltText[0] = IconPath[0] = '\0';
      cptr = (char*)ilptr + cfptr->IconStructOverhead;
      while (*cptr && !isspace(*cptr)) cptr++;
      if (*cptr) cptr++;
      while (*cptr)
      {
         if (*cptr == 'S' && !memcmp (cptr, "SRC=\"", 5))
         {
            cptr += 5;
            sptr = IconPath;
            while (*cptr && *cptr != '\"') *sptr++ = *cptr++;
            *sptr = '\0';
         }
         else
         if (*cptr == 'A' && !memcmp (cptr, "ALT=\"", 5))
         {
            cptr += 5;
            sptr = IconAltText;
            while (*cptr && *cptr != '\"') *sptr++ = *cptr++;
            *sptr = '\0';
         }
         else
            cptr++;
      }
      if (!IconAltText[0])
         strcpy (IconAltText, "<FONT SIZE=-1><I>(none)</I></FONT>");
      CopyTextIntoHtml (HtmlIconAltText, IconAltText, -1);      
      if (!IconPath[0])
         strcpy (IconPath, "<FONT SIZE=-1><I>(none!)</I></FONT>");

      memcpy (sptr = Buffer, "<TR><TD>", 8);
      sptr += 8;
      cptr = (char*)ilptr + cfptr->IconStructOverhead;
      while (*cptr && *cptr != ' ') *sptr++ = tolower(*cptr++);
      if (*cptr) cptr++;
      memcpy (sptr, "</TD><TD ALIGN=center>", 22);
      sptr += 22;
      while (*cptr) *sptr++ = *cptr++;
      memcpy (sptr, "</TD><TD>", 9);
      sptr += 9;
      for (cptr = IconPath; *cptr; *sptr++ = *cptr++);
      memcpy (sptr, "</TD><TD>", 9);
      sptr += 9;
      for (cptr = HtmlIconAltText; *cptr; *sptr++ = *cptr++);
      memcpy (sptr, "</TD></TR>\n", 12);  /* includes null */
      sptr += 11;

      NetWrite (rqptr, 0, Buffer, sptr - Buffer);

      ilptr = ilptr->NextPtr;
   }

   NetWrite (rqptr, 0, "</TABLE>\n", 9);

   /*****************/
   /* content-types */
   /*****************/

   NetWrite (rqptr, 0, ReportContentTypeFaoDsc.dsc$a_pointer,
                       ReportContentTypeFaoDsc.dsc$w_length);

   clptr = cfptr->ContentTypeListHeadPtr;
   for (;;)
   {
      if (clptr == NULL) break;

      if (clptr->AutoScriptNamePtr[0] != '/')
         clptr->AutoScriptNamePtr = "";

      status = sys$fao (&ContentTypeFaoDsc, &Length, &BufferDsc,
                        clptr->SuffixPtr,
                        clptr->ContentTypePtr,
                        clptr->AutoScriptNamePtr,
                        clptr->ContentDescriptionPtr);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$fao()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      Buffer[Length] = '\0';
      NetWrite (rqptr, 0, Buffer, Length);

      clptr = clptr->NextPtr;
   }

   NetWrite (rqptr, 0, "</TABLE>\n", 9);

   /**********/
   /* errors */
   /**********/

   if (CopyToHtml (HtmlErrorSysAdmin, sizeof(HtmlErrorSysAdmin),
                   cfptr->ErrorSysAdminPtr, -1) < 0)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }
   if (CopyToHtml (HtmlErrorInfo, sizeof(HtmlErrorInfo),
                   cfptr->ErrorInfoPtr, -1) < 0)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;
   REPBOOL (cfptr->ErrorRecommend)
   *vecptr++ = HtmlErrorSysAdmin;
   if (cfptr->ErrorInfoPtr[0])
   {
      *vecptr++ = "<PRE>";
      *vecptr++ = HtmlErrorInfo;
      *vecptr++ = "</PRE>";
   }
   else
   {
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "";
   }

   status = sys$faol (&ReportErrorFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   /**************/
   /* end report */
   /**************/

   NetWrite (rqptr, 0, EndPageFaoDsc.dsc$a_pointer,
                       EndPageFaoDsc.dsc$w_length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/
/*
Return a form for the HTTPd server's configuration.  POSTing the form to the
server results in the URL-decoded form as a configuration file.  This function
blocks while executing.
*/ 

ConfigRevise
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
boolean UseServerDatabase
)
{
#define SIZEOF_SPECIAL_ICON 256

#define REPBOOL(b)\
if (b)\
{\
   *vecptr++ = RadioButtonChecked;\
   *vecptr++ = RadioButtonUnchecked;\
}\
else\
{\
   *vecptr++ = RadioButtonUnchecked;\
   *vecptr++ = RadioButtonChecked;\
}

   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ:!UL ... Configuration</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>HTTPd !AZ:!UL</H1>\n\
<H2>Revise Configuration</H2>\n");

   static $DESCRIPTOR (ProblemReportFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Problem!AZ !AZ</FONT></TH></TR>\n\
<TR><TD><PRE>!AZ</PRE></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (ConfigGeneralFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Source: &quot;!AZ&quot;</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=2 NOBORDER>\n\
<TR>\
<TH ALIGN=right>File:</TH>\
<TD ALIGN=left>!AZ</TD>\
<TD>&nbsp;</TD><TH>[<A HREF=\"!AZ\">View</A>]</TH>\
<TH>[<A HREF=\"!AZ!AZ\">Edit</A>]</TH>\
</TR>\n\
<TR><TH ALIGN=right>!AZ</TH>\
<TD ALIGN=left>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<FORM METHOD=POST ACTION=\"!AZ\">\n\
\
<INPUT TYPE=hidden NAME=x VALUE=\"\
\# Configuration:  !AZ:!UL&#10;\
\#                 !AZ&#10;\
\# Last Modified:  !AZ&#10;\
\#                 !AZ@!AZ&#10;\
\">\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>General</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TH>Busy</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Busy]  \">\n\
<INPUT TYPE=text SIZE=2 NAME=Busy VALUE=!UL>\n\
</TD><TD><I>connections</I></TD></TR>\n\
<TR><TH>Request History</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[RequestHistory]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=RequestHistory VALUE=!UL>\n\
</TD></TR>\n\
<TR><TH>Activity History</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ActivityDays]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=ActivityDays VALUE=!UL>\n\
</TD><TD><I>days</I></TD></TR>\n\
<TR><TH>Comments</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CommentedInfo]  \">\n\
<INPUT TYPE=radio NAME=CommentedInfo VALUE=enabled!AZ>yes\n\
<INPUT TYPE=radio NAME=CommentedInfo VALUE=disabled!AZ>no\n\
</TD></TR>\n\
<TR><TH>DNS Lookup</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DNSLookup]  \">\n\
<INPUT TYPE=radio NAME=DNSlookup VALUE=enabled!AZ>yes\n\
<INPUT TYPE=radio NAME=DNSlookup VALUE=disabled!AZ>no\n\
</TD></TR>\n\
<TR><TH>Logging</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Logging]  \">\n\
<INPUT TYPE=radio NAME=Logging VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=Logging VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH>Monitor</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Monitor]  \">\n\
<INPUT TYPE=radio NAME=Monitor VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=Monitor VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH>IP Port</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Port]  \">\n\
<INPUT TYPE=text SIZE=5 NAME=Port VALUE=!UL>\n\
</TD><TD><I>IP port number</I></TD></TR>\n\
<TR><TH>StreamLF</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[StreamLF]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=StreamLF VALUE=!UL>\n\
</TD><TD><I>Kbytes</I></TD></TR>\n\
<TR><TH>Search</TH><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Search]  \">\n\
<INPUT TYPE=text size=40 NAME=Search VALUE=\"!AZ\">\n\
<TR><TH>Put/Post Max.Length</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[PutMaxKbytes]  \">\n\
<INPUT TYPE=text SIZE=5 NAME=MaxContentLength VALUE=!UL>\n\
</TD><TD><I>Kbytes</I></TD></TR>\n\
<TR><TH>Put/Post Files</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[PutVersionLimit]  \">\n\
<INPUT TYPE=text SIZE=5 NAME=PutVersionLimit VALUE=!UL>\n\
</TD><TD><I>version limit</I></TD></TR>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Timeouts</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TH>Input</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[InputTimeout]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=Input VALUE=!UL>\n\
</TD><TD><I>minutes</I></TD></TR>\n\
<TR><TH>Output</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[OutputTimeout]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=Output VALUE=!UL>\n\
</TD><TD><I>minutes</I></TD></TR>\n\
<TR><TH>Keep-Alive</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[KeepAliveTimeout]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=KeepAlive VALUE=!UL>\n\
</TD><TD><I>seconds</I></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Welcome (index/home pages)</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TD COLSPAN=2><I>Format: \"file-name.file-type\" \
(e.g. HOME.HTML)</I></TD></TR>\n\
<TR><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Welcome]&#10;\">\n\
<TEXTAREA NAME=Welcome ROWS=3 COLS=40>");

   static $DESCRIPTOR (ConfigAcceptFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Host Access Control</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TH>Accept</TH></TR>\n\
<TR><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Accept]&#10;\">\n\
<TEXTAREA NAME=Accept ROWS=3 COLS=40>");

   static $DESCRIPTOR (ConfigRejectFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
<TR><TH>Reject</TH></TR>\n\
<TR><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Reject]&#10;\">\n\
<TEXTAREA NAME=Reject ROWS=3 COLS=40>");

   static $DESCRIPTOR (ConfigAuthFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Authorization</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TH>Basic Authorization</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthBasic]  \">\n\
<INPUT TYPE=radio NAME=AuthBasic VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=AuthBasic VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH>Digest Authorization</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthDigest]  \">\n\
<INPUT TYPE=radio NAME=AuthDigest VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=AuthDigest VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TH>GET nonce lifetime</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthDigestGetLife]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=AuthDigestGetLife VALUE=!UL>\n\
</TD><TD><I>minutes</I></TD>\n\
</TD></TR>\n\
<TR><TH>PUT nonce lifetime</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthDigestPutLife]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=AuthDigestPutLife VALUE=!UL>\n\
</TD><TD><I>minutes</I></TD></TR>\n\
<TR><TH>VMS Authentication</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthVMS]  \">\n\
<INPUT TYPE=radio NAME=AuthVMS VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=AuthVMS VALUE=disabled!AZ>disabled\n\
<TR><TH ALIGN=right>Note ...</A></TH>\
<TD>paths must also be configured!!</TD></TR>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>DCL/Scripting</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TH>Spawn Authorized Privileges</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclSpawnAuthPriv]  \">\n\
<INPUT TYPE=radio NAME=DclSpawnAuthPriv VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DclSpawnAuthPriv VALUE=disabled!AZ>disabled\n\
</TD><TD></TD></TR>\n\
<TR><TH>Supply Header and Body</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclFullRequest]  \">\n\
<INPUT TYPE=radio NAME=DclFullRequest VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DclFullRequest VALUE=disabled!AZ>disabled\n\
</TD><TD></TD></TR>\n\
<TR><TH>Soft Limit</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclSoftLimit]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=SoftLimit VALUE=!UL>\n\
</TD><TD><I>subprocesses</I></TD></TR>\n\
<TR><TH>Hard Limit</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclHardLimit]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=HardLimit VALUE=!UL>\n\
</TD><TD><I>subprocesses</I></TD></TR>\n\
<TR><TH>Zombie Life-Time</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclZombieLifeTime]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=ZombieLifeTime VALUE=!UL>\n\
</TD><TD><I>minutes</I></TD></TR>\n\
<TR><TH>CGIplus Life-Time</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclCgiPlusLifeTime]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=CgiPlusLifeTime VALUE=!UL>\n\
</TD><TD><I>minutes</I></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Server-Side Includes</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TH>SSI</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[SSI]  \">\n\
<INPUT TYPE=radio NAME=SSI VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=SSI VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH>Exec</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[SSIexec]  \">\n\
<INPUT TYPE=radio NAME=SSIexec VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=SSIexec VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH>Access Counting</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[SSIaccesses]  \">\n\
<INPUT TYPE=radio NAME=SSIaccesses VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=SSIaccesses VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Directory (Index of...)</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TH>Access</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirAccess]  \">\n\
<INPUT TYPE=radio NAME=DirAccess VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirAccess VALUE=selective!AZ>selective\n\
<INPUT TYPE=radio NAME=DirAccess VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH>Layout</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirLayout]  \">\n\
<INPUT TYPE=text size=40 NAME=DirLayout VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH>Owner</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirOwner]  \">\n\
<INPUT TYPE=radio NAME=DirOwner VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirOwner VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH>Pre-Expired</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirPreExpired]  \">\n\
<INPUT TYPE=radio NAME=DirPreExpired VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirPreExpired VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH>Wildcard</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirWildcard]  \">\n\
<INPUT TYPE=radio NAME=DirWildcard VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirWildcard VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH>Readme</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirReadme]  \">\n\
<INPUT TYPE=radio NAME=DirReadme VALUE=top!AZ>top\n\
<INPUT TYPE=radio NAME=DirReadme VALUE=bottom!AZ>bottom\n\
<INPUT TYPE=radio NAME=DirReadme VALUE=disabled!AZ>none\n\
</TD></TR>\n\
<TR><TH COLSPAN=2>Readme Files</TH></TR>\n\
<TR><TD COLSPAN=2><I>Format: \"file-name.file-type\" \
(e.g. README.HTML)</I></TD></TR>\n\
<TR><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirReadmeFile]&#10;\">\n\
<TEXTAREA NAME=DirReadmeFile ROWS=3 COLS=40>");

   static $DESCRIPTOR (ConfigIconsFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Icons</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TH COLSPAN=2>Icons</TH></TR>\n\
<TR><TD COLSPAN=2><I>Format: \"icon-path [SPACE] alternative-text [SPACE] \
content-type\"</I></TD></TR>\n\
<TR><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddIcon]&#10;\">\n\
<TEXTAREA NAME=AddType ROWS=5 COLS=60>");

   static $DESCRIPTOR (ConfigSpecialIconsFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
<TR><TH COLSPAN=2>Special Icons</TH></TR>\n\
<TR><TD COLSPAN=2><I>Format: \"icon-path [SPACE] \
alternative-text\"</I></TD></TR>\n\
<TR><TH>Blank</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddBlankIcon]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=AddBlankIcon VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH>Directory</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddDirIcon]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=AddDirIcon VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH>Parent Directory</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddParentIcon]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=AddParentIcon VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH>Unknown</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddUnknownIcon]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=AddUnknownIcon VALUE=\"!AZ\">\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (ConfigContentTypesFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Content Types</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TD COLSPAN=2><I>Format: \"file-suffix [SPACE] content-type [SPACE] \
auto-script-path <B>or</B> a hyphen [SPACE] \
description (to-end-of-line)\"</I></TD></TR>\n\
<TR><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddType]&#10;\">\n\
<TEXTAREA NAME=AddType ROWS=5 COLS=60>");

   static $DESCRIPTOR (ConfigErrorFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Errors</TH></TR>\n\
<TR><TD>\n\
<TABLE NOBORDER>\n\
<TR><TH>Recommend</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ErrorRecommend]  \">\n\
<INPUT TYPE=radio NAME=ErrorRecommend VALUE=enabled!AZ>yes\n\
<INPUT TYPE=radio NAME=ErrorRecommend VALUE=disabled!AZ>no\n\
</TD></TR>\n\
<TR><TH>Sys.Admin. Path</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ErrorSysAdmin]  \">\n\
<INPUT TYPE=text NAME=ErrorSysAdmin SIZE=40 VALUE=\"!AZ\">\n\
</TD>\n\
<TR><TH COLSPAN=2>Information</TH></TR>\n\
<TR><TD COLSPAN=2>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ErrorInfo]&#10;\">\n\
<TEXTAREA NAME=ErrorInfo ROWS=5 COLS=80>!AZ</TEXTAREA>\n\
</TD>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;# End!&#10;\">\n\
<P><INPUT TYPE=submit VALUE=\" Commit Changes \">\n\
<INPUT TYPE=reset VALUE=\" Reset \">\n\
</FORM>\n\
</BODY>\n\
</HTML>\n");

   static char  RadioButtonChecked [] = " CHECKED",
                RadioButtonUnchecked [] = "",
                StringOverflow [] = "*ERROR* string overflow\n";

   register int  idx;
   register unsigned long  *vecptr;
   register char  *cptr, *sptr, *zptr;
   register struct ConfigStruct  *cfptr;
   register struct ContentTypeStruct  *clptr;
   register struct IconStruct  *ilptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [128];
   char  *IconAltPtr,
         *IconAltEndPositionPtr,
         *IconAltStartPositionPtr,
         *IconUrlPtr,
         *IconTypePtr,
         *IconTypePositionPtr,
         *SpecialIconPtr;
   char  Buffer [8192],
         HtmlIconUrl [256],
         HtmlErrorInfo [2048],
         HtmlErrorSysAdmin [128],
         Scratch [256],
         SpecialIconBlank [SIZEOF_SPECIAL_ICON],
         SpecialIconDir [SIZEOF_SPECIAL_ICON],
         SpecialIconParent [SIZEOF_SPECIAL_ICON],
         SpecialIconUnknown [SIZEOF_SPECIAL_ICON],
         TimeCurrent [32];
   struct ContentTypeStruct  ContentType;
   struct ConfigStruct  LocalConfig;
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (ScratchDsc, Scratch);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigRevise()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   if (UseServerDatabase)
      cfptr = &Config;
   else
   {
      /* load our own configuration structure from file */
      cfptr = &LocalConfig;
      cfptr->RequestPtr = rqptr;
      Configure (cfptr);
   }

   if (cfptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      if ((*vecptr++ = cfptr->ProblemCount) == 1)
         *vecptr++ = "";
      else
         *vecptr++ = "s";
      if (UseServerDatabase)
         *vecptr++ = "Reported At Startup";
      else
         *vecptr++ = "Detected During Load";
      *vecptr++ = cfptr->ProblemReportPtr;

      status = sys$faol (&ProblemReportFaoDsc, &Length, &BufferDsc,
                         &FaoVector);

      /* if using a local database free report memory now! */
      if (!UseServerDatabase) free (cfptr->ProblemReportPtr);

      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      Buffer[Length] = '\0';
      NetWrite (rqptr, 0, Buffer, Length);
   }

   strcpy (TimeCurrent, DayDateTime (&rqptr->BinaryTime, 20));
   cptr = MapUrl (NULL, cfptr->LoadFileName, NULL, NULL);

   vecptr = FaoVector;

   /* source information (i.e. from executing server or static file) */
   if (UseServerDatabase)
      *vecptr++ = "Server";
   else
      *vecptr++ = "File";
   *vecptr++ = cfptr->LoadFileName;
   *vecptr++ = cptr;
   *vecptr++ = HttpdInternalScriptUpd;
   *vecptr++ = cptr;

   if (UseServerDatabase)
   {
      *vecptr++ = "Loaded:";
      *vecptr++ = DayDateTime (&cfptr->LoadBinTime, 20);
   }
   else
   {
      *vecptr++ = "Revised:";
      *vecptr++ = DayDateTime (&cfptr->RevBinTime, 20);
   }

   /* form action */
   *vecptr++ = cptr;

   /* comments */
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = SoftwareID;
   *vecptr++ = TimeCurrent;
   *vecptr++ = rqptr->RemoteUser;
   *vecptr++ = rqptr->ClientHostName;

   /* general */
   if (cfptr->Busy)
      *vecptr++ = cfptr->Busy;
   else
      *vecptr++ = DefaultMaxConcurrentConnections;
   *vecptr++ = cfptr->RequestHistory;
   *vecptr++ = cfptr->ActivityNumberOfDays;
   REPBOOL (cfptr->IncludeCommentedInfo)
   REPBOOL (cfptr->DnsLookup)
   REPBOOL (cfptr->LoggingEnabled)
   REPBOOL (cfptr->MonitorEnabled)
   if (cfptr->ServerPort)
      *vecptr++ = cfptr->ServerPort;
   else
      *vecptr++ = 80;
   *vecptr++ = cfptr->StreamLfConversionMaxKbytes;
   *vecptr++ = cfptr->KeywordSearch;
   *vecptr++ = cfptr->PutMaxKbytes;
   *vecptr++ = cfptr->PutVersionLimit;

   /* timeouts */
   *vecptr++ = cfptr->InputTimeoutMinutes;
   *vecptr++ = cfptr->OutputTimeoutMinutes;
   *vecptr++ = cfptr->KeepAliveTimeoutSeconds;

   status = sys$faol (&ConfigGeneralFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   /*****************************/
   /* welcome (home page) files */
   /*****************************/

   for (idx = 0; idx < cfptr->HomePageCount; idx++)
   {
      zptr = (sptr = Scratch) + sizeof(Scratch);
      if (!idx) *sptr++ = '\n';
      for (cptr = cfptr->HomePageArray[idx];
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\n';
      if (sptr >= zptr)
         strcpy (Scratch, StringOverflow);
      else
         *sptr = '\0';

      Length = CopyTextIntoHtml (Buffer, Scratch, -1);
      NetWrite (rqptr, 0, Buffer, Length);
   }

   /********/
   /* more */
   /********/

   NetWrite (rqptr, 0, ConfigAcceptFaoDsc.dsc$a_pointer,
                       ConfigAcceptFaoDsc.dsc$w_length);

   /****************/
   /* accept hosts */
   /****************/

   if ((cptr = cfptr->AcceptHostsPtr) == NULL) cptr = "";
   while (*cptr)
   {
      zptr = (sptr = Scratch) + sizeof(Scratch);
      if (cptr == cfptr->AcceptHostsPtr) *sptr++ = '\n';
      while (*cptr && *cptr != ',' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr) cptr++;
      if (sptr < zptr) *sptr++ = '\n';
      if (sptr >= zptr)
         strcpy (Scratch, StringOverflow);
      else
         *sptr = '\0';

      Length = CopyTextIntoHtml (Buffer, Scratch, sizeof(Buffer));
      NetWrite (rqptr, 0, Buffer, Length);
   }

   /********/
   /* more */
   /********/

   NetWrite (rqptr, 0, ConfigRejectFaoDsc.dsc$a_pointer,
                       ConfigRejectFaoDsc.dsc$w_length);

   /****************/
   /* reject hosts */
   /****************/

   if ((cptr = cfptr->RejectHostsPtr) == NULL) cptr = "";
   while (*cptr)
   {
      zptr = (sptr = Scratch) + sizeof(Scratch);
      if (cptr == cfptr->RejectHostsPtr) *sptr++ = '\n';
      while (*cptr && *cptr != ',' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr) cptr++;
      if (sptr < zptr) *sptr++ = '\n';
      if (sptr >= zptr)
         strcpy (Scratch, StringOverflow);
      else
         *sptr = '\0';

      Length = CopyTextIntoHtml (Buffer, Scratch, sizeof(Buffer));
      NetWrite (rqptr, 0, Buffer, Length);
   }

   /********/
   /* more */
   /********/

   vecptr = FaoVector;

   /* authentication */
   REPBOOL (cfptr->AuthBasicEnabled)
   REPBOOL (cfptr->AuthDigestEnabled)
   *vecptr++ = cfptr->AuthDigestNonceGetLifeTime;
   *vecptr++ = cfptr->AuthDigestNonceGetLifeTime;
   REPBOOL (cfptr->AuthVmsEnabled)

   /* DCL/scripting */
   REPBOOL (cfptr->DclSpawnAuthPriv)
   REPBOOL (cfptr->DclFullRequest)
   *vecptr++ = cfptr->DclSubprocessSoftLimit;
   *vecptr++ = cfptr->DclSubprocessHardLimit;
   *vecptr++ = cfptr->DclZombieLifeTime;
   *vecptr++ = cfptr->DclCgiPlusLifeTime;

   /* SSI (.shtml) */
   REPBOOL (cfptr->SsiEnabled)
   REPBOOL (cfptr->SsiExecEnabled)
   REPBOOL (cfptr->SsiAccessesEnabled)

   /* directory */
   if (cfptr->DirAccess)
   {
      *vecptr++ = " CHECKED";
      *vecptr++ = "";
      *vecptr++ = "";
   }
   else
   if (cfptr->DirAccessSelective)
   {
      *vecptr++ = "";
      *vecptr++ = " CHECKED";
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = " CHECKED";
   }
   *vecptr++ = cfptr->DefaultDirLayout;
   REPBOOL (cfptr->DirOwnerEnabled)
   REPBOOL (cfptr->DirPreExpired)
   REPBOOL (cfptr->DirWildcardEnabled)
   if (cfptr->DirReadMeTop)
   {
      *vecptr++ = " CHECKED";
      *vecptr++ = "";
      *vecptr++ = "";
   }
   else
   if (cfptr->DirReadMeBottom)
   {
      *vecptr++ = "";
      *vecptr++ = " CHECKED";
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = " CHECKED";
   }
   REPBOOL (cfptr->DirWildcardEnabled)

   status = sys$faol (&ConfigAuthFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   /****************/
   /* readme files */
   /****************/

   for (idx = 0; idx < cfptr->ReadMeFileCount; idx++)
   {
      if (!idx) NetWrite (rqptr, 0, "\n", 1);
      Length = CopyTextIntoHtml (Buffer, cfptr->ReadMeFileArray[idx],
                                 sizeof(Buffer)-2);
      Buffer[Length++] = '\n';
      Buffer[Length] = '\0';
      NetWrite (rqptr, 0, Buffer, Length);
   }

   /********/
   /* more */
   /********/

   NetWrite (rqptr, 0, ConfigIconsFaoDsc.dsc$a_pointer,
                       ConfigIconsFaoDsc.dsc$w_length);

   /*********/
   /* icons */
   /*********/

   SpecialIconBlank[0] = SpecialIconDir[0] = SpecialIconParent[0] =
      SpecialIconUnknown[0] = '\n';
   SpecialIconBlank[1] = SpecialIconDir[1] = SpecialIconParent[1] =
      SpecialIconUnknown[1] = '\0';

   ilptr = cfptr->IconListHeadPtr;
   while (ilptr != NULL)
   {
      /*
         Convert from 1 back to 2.

         1.  sprintf ((char*)ilptr + cfptr->IconStructOverhead,
                      "%s <IMG ALIGN=top SRC=\"%s\" ALT=\"%s\">",
                      ContentTypePtr, IconUrlPtr, AltTextPtr);

         2.  <URL>[space]<alternative-text>[space]<content-type>
      */

      IconTypePtr = cptr = (char*)ilptr + cfptr->IconStructOverhead;
      while (*cptr && !isspace(*cptr)) cptr++;
      if (*cptr) cptr++;
      while (*cptr)
      {
         if (*cptr == 'S' && !memcmp (cptr, "SRC=\"", 5))
         {
            cptr += 5;
            IconUrlPtr = cptr;
            while (*cptr && *cptr != '\"') cptr++;
            if (*cptr) cptr++;
         }
         else
         if (*cptr == 'A' && !memcmp (cptr, "ALT=\"", 5))
         {
            cptr += 5;
            IconAltPtr = cptr;
            while (*cptr && *cptr != '\"') cptr++;
            if (*cptr) cptr++;
         }
         else
            cptr++;
      }
      zptr = (sptr = Scratch) + sizeof(Scratch);
      if (ilptr == cfptr->IconListHeadPtr) *sptr++ = '\n';
      /* icon URL */
      for (cptr = IconUrlPtr;
           *cptr && *cptr != '\"' && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      /* note the start of alternative text */
      IconAltStartPositionPtr = sptr;
      /* alternative text */
      for (cptr = IconAltPtr;
           *cptr && *cptr != '\"' && sptr < zptr;
           *sptr++ = *cptr++);
      IconAltEndPositionPtr = sptr;
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      /* note position of null-terminated content type in the scratch string */
      IconTypePositionPtr = sptr;
      /* get the content type */
      for (cptr = IconTypePtr;
           *cptr && *cptr != ' ' && sptr < zptr;
           *sptr++ = tolower(*cptr++));
      if (sptr >= zptr)
         Scratch[0] = '\0';
      else
         *sptr = '\0';

      SpecialIconPtr = NULL;
      if (strsame (IconTypePositionPtr, ConfigContentTypeBlank, -1))
      {
         /* replace the blanks with equivalent underscores */
         while (IconAltStartPositionPtr < IconAltEndPositionPtr)
            *IconAltStartPositionPtr++ = '_';
         SpecialIconPtr = SpecialIconBlank;
      }
      else
      if (strsame (IconTypePositionPtr, ConfigContentTypeDir, -1))
         SpecialIconPtr = SpecialIconDir;
      else
      if (strsame (IconTypePositionPtr, ConfigContentTypeParent, -1))
         SpecialIconPtr = SpecialIconParent;
      else
      if (strsame (IconTypePositionPtr, ConfigContentTypeUnknown, -1))
         SpecialIconPtr = SpecialIconUnknown;
      /* do not include content-type in special icon strings */
      if (SpecialIconPtr != NULL) *IconAltEndPositionPtr = '\0';

      if (sptr < zptr) *sptr++ = '\n';
      if (sptr >= zptr)
         strcpy (Scratch, StringOverflow);
      else
         *sptr = '\0';

      Length = CopyTextIntoHtml (Buffer, Scratch, sizeof(Buffer));

      if (SpecialIconPtr == NULL)
         NetWrite (rqptr, 0, Buffer, Length);
      else
      {
         zptr = (sptr = SpecialIconPtr) + SIZEOF_SPECIAL_ICON;
         for (cptr = Buffer; *cptr && sptr < zptr; *sptr++ = *cptr++);
         if (sptr >= zptr)
            strcpy (SpecialIconPtr, StringOverflow);
         else
            *sptr = '\0';
      }

      ilptr = ilptr->NextPtr;
   }

   vecptr = FaoVector;

   *vecptr++ = SpecialIconBlank;
   *vecptr++ = SpecialIconDir;
   *vecptr++ = SpecialIconParent;
   *vecptr++ = SpecialIconUnknown;

   status = sys$faol (&ConfigSpecialIconsFaoDsc, &Length, &BufferDsc,
                      &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   /*****************/
   /* content types */
   /*****************/

   NetWrite (rqptr, 0, ConfigContentTypesFaoDsc.dsc$a_pointer,
                       ConfigContentTypesFaoDsc.dsc$w_length);

   clptr = cfptr->ContentTypeListHeadPtr;
   for (;;)
   {
      if (clptr == NULL) break;

      if (clptr->AutoScriptNamePtr[0] != '/') clptr->AutoScriptNamePtr = "-";

      zptr = (sptr = Scratch) + sizeof(Scratch);
      if (clptr == cfptr->ContentTypeListHeadPtr) *sptr++ = '\n';
      for (cptr = clptr->SuffixPtr;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      for (cptr = clptr->ContentTypePtr;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      for (cptr = clptr->AutoScriptNamePtr;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      for (cptr = clptr->ContentDescriptionPtr;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\n';
      if (sptr >= zptr)
         strcpy (Scratch, StringOverflow);
      else
         *sptr = '\0';

      Length = CopyTextIntoHtml (Buffer, Scratch, sizeof(Buffer));
      NetWrite (rqptr, 0, Buffer, Length);

      clptr = clptr->NextPtr;
   }

   /*********************/
   /* error information */
   /*********************/

   if (CopyToHtml (HtmlErrorSysAdmin, sizeof(HtmlErrorSysAdmin),
                   cfptr->ErrorSysAdminPtr, -1) < 0)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }
   if (CopyToHtml (HtmlErrorInfo, sizeof(HtmlErrorInfo),
                   cfptr->ErrorInfoPtr, -1) < 0)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;
   REPBOOL (cfptr->ErrorRecommend)
   *vecptr++ = HtmlErrorSysAdmin;
   *vecptr++ = HtmlErrorInfo;

   status = sys$faol (&ConfigErrorFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   /************/
   /* end form */
   /************/

   NetWrite (rqptr, 0, EndPageFaoDsc.dsc$a_pointer,
                       EndPageFaoDsc.dsc$w_length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/
