/*****************************************************************************/
/*
                                 DCL.c

*************
** CAUTION **
*************

THIS MODULE IS TASK-ORIENTED, NOT REQUEST-ORIENTED.

That is, most of the functions take a pointer to DCL task rather than a
pointer to request as do other modules. The reason is simple. In this module
the more-permanent data-structures are those associated with the DCL
subprocesses, which persist for multiple requests. With the requests transient
the DCL subprocesses must be managed in their absence. Hence requests are
associated with DCL subprocesses not vice-versa.


OVERVIEW
--------
Provides multi-threaded, concurrent HTTPd subprocesses executing DCL.

This module never returns a valid status and ALWAYS calls the supplied next
task (AST) function.  This should check for a generated error message to
determine is there were any problems.

The DCL can either be in the form of a command, or a procedure or executable 
image file specification.  Both should not be supplied, but file 
specifications have precedence.  If a file specfication is supplied the module 
verifies its existance, and if not qualified with an extension, looks for a 
procedure first (".COM"), then an executable image (".EXE").  Furthermore, the 
DCL can be executed either standalone or as a CGI script (indicated by the 
presence of a script name in the request data structure).  If a CGI script, 
then the output stream is parsed for header information, if not a script then 
the stream is just checked for correct carriage control (terminated by a 
newline).  CGI variables are created for both standalone DCL and scripts, 
although some (e.g. WWW_SCRIPT_NAME) will be empty, and meaningless, for 
standalone DCL. 

The AST-driven nature makes the code a little more difficult to follow, but 
creates a powerful, event-driven, multi-threaded server.  All of the 
necessary functions implementing this module are designed to be non-blocking. 

All of these functions are designed to be, and should be, called at AST 
delivery level, which means they cannot be interrupted at the same level (in 
this case USER mode), and so their actions are essentially atomic at that 
level, meaning no additional synchronization is required for such activities 
as thread disposal, etc.

HTTPD can maintain a number of subprocesses limited only by its process 
quotas, memory is dynamically allocated and there are no fixed data structures 
related to subprocess management. 

The use of byte-streams (effectively "pipes") allows even DCL procedures to
output as HTTP servers, without the need for explicit network I/O. 

Four mailboxes are created for each subprocess' IPC:

  1.  A mailbox connected to its SYS$COMMAND.  This is used to pass DCL
      commands and/or other data to the subprocess.  It effectively allows
      the HTTPd to control the activities of the subprocess this way.

  2.  A mailbox connected to its SYS$OUTPUT.  This recieves records from
      the subprocess, if required appends the HTTP-required carriage-control
      (<CR><LF>), then sends the record to the client via the network socket.
      This allows even DCL procedures to supply a correct output stream to
      the client (see next paragraph).

      If the first line from a script is an HTTP status line (e.g.
      "HTTP/1.0 200 ok") then HTTPD assumes the script will be supplying
      a complete HTTP data stream, including full header and required
      carriage control (<CR><LF> terminating each line).  If the first
      line is not a HTTP status line it assumes CGI script output compliance
      and also ensures each record (line) received has correct HTTP
      carriage-control.

      This stream also attempts to maintain compliance with CGI scripting.
      If the first line output by a script is not an HTTP status line it
      creates and sends one to the client before sending the first line.

  3.  A mailbox defined for the subprocess by the name HTTP$INPUT.
      This may be used to read the request data steam sent by the client.
      Note that from v4.2 this is also the SYS$INPUT <stdin> stream as well,
      which appears to be the more common CGI implementation. As of version
      4.3 the default behaviour is to supply only the request body to the
      script (CGI standard and the more common implmentation).  Using the 
      configuration parameter [DclFullRequest] it is possible to revert
      to the previous behaviour of supplying the header then the body.
      (It's an easy enough modification for most scripts to skip the header by
      reading until the header-separating empty (blank) line is encountered.)

  4.  A mailbox defined by the name CGIPLUSIN.  This allows CGIplus scripts
      to read a stream of CGI variable information.  Each request begins with
      a comment line "!", which can be used for request start synchronisation
      and may always be discarded, and ends with an empty record (blank line),
      with a variable number of records in between.

The script subprocesses can use the basic CGI variables (VMS CERN-like) and 
behave very much like a CGI script.

That is, if a script wants to be CGI-compliant it provides as the first line a 
"Status:", "Content-type:" or a "Location:" then a blank line.   If the first 
line output by a script is a "Content-Type:" header line an HTTP "200" status 
line is prepended.  If the first line output by a script is a "Location:" 
redirection header line the redirection is processed to ensure CERN HTTPD/CGI 
behaviour.  An HTTP "302" status line is prepended if not a local redirection. 
If none of these, HTTPD creates a complete HTTP header comprising status line, 
"Content-Type: text/plain" and blank line (this is an extension of CERN HTTPD 
behaviour). 

If the first characters are "HTTP/1.0 ..." the script will be considered to be 
supplying the raw HTTP stream and record boundaries, carriage-control, etc., 
are of no further concern to the module.  This is the equivalent of a
"no-parse-header" script.  If CGI-compliant each record should represent a 
line of output.   That is lines should not be buffered together and sent as a 
block unless the script is supplying a raw HTTP data stream. 


CGI VARIABLES
-------------
Most of these CGI variable names are those supported by the INTERNET-DRAFT
authored by D.Robinson (drtr@ast.cam.ac.uk), 8 January 1996, plus some
"convenience" variables, breaking the query string into its components (KEY_,
FORM_, etc.) The CGI symbols (CERN/VMS-HTTPd-like DCL symbols instead of Unix
environment variables) are created by the SYS$COMMAND stream of each
subprocess before the script DCL procedure is invoked. By default each
variable name is prefixed by "WWW_" (similar to CERN HTTPd), although this can
be modified at the command line when starting the server. CGI variable
(symbol) values are limited in size to approximately 1000 characters. 


  o  WWW_AUTH_GROUP ........... path authorization group (or empty)    [ext]
  o  WWW_AUTH_REALM ........... authentication realm (or empty)        [ext]
  o  WWW_AUTH_TYPE ............ authentication type (or empty)         [std]

  o  WWW_CONTENT_LENGTH ....... "Content-Length:" from header          [std]
  o  WWW_CONTENT_TYPE ......... "Content-Type:" from header            [std]

  o  WWW_FORM_field ........... query "&" separated form elements      [ext]

  o  WWW_GATEWAY_INTERFACE .... "CGI/1.1"                              [std]

  o  WWW_HTTP_ACCEPT .......... list of browser-accepted content types [std]
  o  WWW_HTTP_AUTHORIZATION ... allows a script to authenticate        [std]
  o  WWW_HTTP_COOKIE .......... any cookie sent by the client          [std]
  o  WWW_HTTP_IF_NOT_MODIFIED . GMT time string                        [std]
  o  WWW_HTTP_PRAGMA .......... any pragma directive of header         [std]
  o  WWW_HTTP_REFERER ......... source document URL for this request   [std]
  o  WWW_HTTP_USER_AGENT ...... client/browser identification string   [std]

  o  WWW_KEY_n ................ query string "+" separated elements    [ext]
  o  WWW_KEY_COUNT ............ number of "+" separated elements       [ext]

  o  WWW_PATH_INFO ............ virtual path of data requested in URL  [std]
  o  WWW_PATH_TRANSLATED ...... VMS file path of data requested in URL [std]

  o  WWW_QUERY_STRING ......... string following "?" in URL            [std]

  o  WWW_REMOTE_ADDR .......... IP host address of HTTP client         [std]
  o  WWW_REMOTE_HOST .......... IP host name of HTTP client            [std]
  o  WWW_REMOTE_USER .......... authenticated username (or empty)      [std]
  o  WWW_REQUEST_METHOD ....... "GET", "PUT", etc.                     [std]
  o  WWW_REQUEST_TIME_GMT ..... request GMT time                       [ext]
  o  WWW_REQUEST_TIME_LOCAL ... request local time                     [ext]

  o  WWW_SCRIPT_NAME .......... name of script (e.g. "/query")         [std]
  o  WWW_SERVER_NAME .......... IP host name of server system          [std]
  o  WWW_SERVER_PROTOCOL ...... HTTP protocol version ("HTTP/1.0")     [std]
  o  WWW_SERVER_PORT .......... IP port request was received on        [std]
  o  WWW_SERVER_SOFTWARE ...... software ID of the HTTPD daemon        [std]


CGI-PLUS
--------
CGI plus lower latency, plus greater efficiency, plus less system impact!

CGIplus attempts to eliminate the overhead associated with creating the
subprocess and then executing the image of a CGI script.  It does this by
allowing the subprocess and associated image to continue executing in-between
uses, eliminating the startup overhead.  This both reduces the load on the
system and the request latency where one of these scripts is involved.
In this sense these advantages parallel those offered by commercial HTTP
server-integration APIs, such as Netscape NSAPI and Microsoft ISAPI.

The script interface is still largely CGI, which means a new API does not need
to be learned and that existing CGI scripts are simple to modify. The script
must read the CGI variables from CGIPLUSIN. They are supplied as a series of
records (lines) containing the CGI variable name, an equate symbol and then
the variable value. This format may be easily parsed and as the value contains
no encoded characters may be directly used. An empty record (blank link)
indicates the end of the request information. The request may be processed at
that stage.

After processing the CGIplus script can loop, waiting to read the details of
the next request from CGIPLUSIN. The first line read may ALWAYS be discarded
allowing a read to be used as a simple synchronization mechanism.

HTTP output (to the client) is written to SYS$OUTPUT (stdout). End of output
MUST be indicated by writing a special EOF string to the output stream. This
is a KLUDGE, and the least elegant part of CGIplus desgin, BUT it is also the
simplest implementation and should work in all but the most exceptional
circumstances. The special EOF string has been chosen to be most unlikely to
occur in normal output (a hopefully unique 280 bit sequence), but there is
still a very, very small chance! The CGIplus EOF string is obtained from the
logical name CGIPLUSEOF, defined in the subprocess's process table, using the
language's equivalent of F$TRNLNM(), SYS$TRNLNM(), or a getenv() call in the C
Language. This string will always contain less than 64 characters and comprise
only common, printable characters. It must be written at the conclusion of a
request's output to the output stream as a single record (line) but may also
contain a <CR><LF> or just <LF> trailing carriage-control (to allow for
programming language constraints). See examples in HT_ROOT:[SRC.CGIPLUS]

HTTP input (raw request stream, header and any body) is still available to the
CGIplus script.

Multiple CGIplus scripts may be executing in subprocesses at any one time.
This includes multiple instances of any particular script.  It is the server's
task to track these, distributing appropriate requests to idle subprocesses,
monitoring those currently processing requests, creating new instances if and
when necessary, and deleting the least-used, idle CGIplus subprocesses when
configurable thresholds are reached.

A CGIplus script can be terminated by the server at any time (the subprocess
SYS$DELPRC()ed) so resources should be largely quiescent when not actually
processing. CGIplus subprocesses may also be terminated from the command line
using STOP/ID=. The server administration menu provides a simple mechansim to
purge (stop) all CGIplus processes, allowing the server to be flushed of all
subprocesses. This can be usefull if some new compilation of a CGIplus script
executable needs to made available.

CGIplus scripts are differentiated from "normal" CGI scripts in the mapping
rule configuration file using the "script+" and "exec+" directives.  Of course
it would be possible to design a script to simply modify it's behaviour so it
was possible to execute in both environments.  Simply detecting the presence
or absence of one of the "normal" CGI variables (i.e. DCL symbols,
e.g. WWW_PATH_INFO, WWW_CGI_GATEWAY_INTERFACE, etc.) would be sufficient
indication.

See examples in HT_ROOT:[SRC.CGIPLUS]


ZOMBIES
-------
The reuse of DCL subprocesses for CGI scripting provides very significant
performance gains with very little _real_ possibility of undesirable
interaction between uses (where scripts are "well-behaved", which should be
all enviornments). When a non-zero zombie lifetime is specified DCL
subprocesses implicitly persist between uses for standard CGI and DCL (SSI)
commands as well as explicitly with CGIplus scripts. When not being used to
process a request these subprocesses are termed "zombies" ;^) they are neither
"alive" (executing a script and processing a request) nor are they "dead"
(subprocess deleted and task structure free). Instead the subprocess is in a
LEF state waiting for more input via CGIPLUSIN. A great deal of care is
taken to ensure there is no interaction between uses (all symbols are deleted,
output mailbox is emptied), but there can be no "iron-clad" guarantee. Use of
zombies (persistant DCL processes) is disabled by setting the appropriate
configuration parameter to zero. The same DCL EOF mechansism is used to signal
end-of-output in all process-persistant environments.


LOGICAL NAMES
-------------
HTTPD$VERIFY    if defined, turns on script/DCL-level verify


VERSION HISTORY
---------------
01-AUG-97  MGD  allow supplying request header AS WELL AS body, or only body,
                added AUTH_REALM, AUTH_GROUP, HTTP_ACCEPT, and 
                REQUEST_TIME_GMT/LOCAL CGI variables
01-JUN-97  MGD  Persistant DCL processes, CGIplus, new for HTTPd v4.2,
                DCL.C completely re-designed!
26-APR-97  MGD  bugfix; serious flaw POST content handling since modifications
                to REQUEST.C, PUT.C in version 4.0 (rewrite of HTTP$INPUT)
01-FEB-97  MGD  HTTPd version 4
14-NOV-96  MGD  bugfix; no status was being returned after "DELETE/SYMBOL X"
                in DclSysCommandStringSymbol()
06-APR-96  MGD  miscellaneous refinements
26-MAR-96  MGD  added WWW_HTTP_AUTHENTICATION, allowing scripts to authenticate
01-DEC-95  MGD  HTTPd version 3
19-SEP-95  MGD  changed carriage-control on records from <CR><LF> (the strict
                HTTP requirement) to single newline (<LF>, de facto standard)
                This will be slightly more efficient, and "more compliant"!
21-APR-95  MGD  bugfix to DclSysOutputAst()
03-APR-95  MGD  added remote user authentication and CGI symbol
20-MAR-95  MGD  bugfix to DclQioHttpInput()
20-DEC-94  MGD  initial development as a module for multi-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>

/* VMS related header files */

/* cmbdef.h is not defined for VAXC 3.n */
#define CMB$M_READONLY 0x01
#define CMB$M_WRITEONLY 0x02

#include <descrip.h>
#include <dvidef.h>

#include <iodef.h>
/* these should be, but are not defined for VAXC 3.n */
#define IO$M_READERCHECK 0x100
#define IO$M_WRITERCHECK 0x200
#define IO$M_READERWAIT 0x400
#define IO$M_WRITERWAIT 0x800

#include <jpidef.h>
#include <prvdef.h>
#include <rms.h>

#include <ssdef.h>
/* these should be, but are not defined for VAXC 3.n */
#define SS$_NOREADER 9412
#define SS$_NOWRITER 9420

#include <stsdef.h>
#include <syidef.h>

/* application header files */
#include "wasd.h"
#include "dcl.h"
#include "error.h"
#include "httpd.h"
#include "net.h"
#include "support.h"

/**********************/
/* CGIPLUSEOF string! */
/**********************/

/*
   A hopefully unique sequence of 280 bits (which is a fair bit :^)
   The string is generated using the subprocess PID (first !8XL) and
   and the ones-complement of the quadword binary time (subsequent
   !8XLs) each time a subprocess is uniquely used.  This provides a
   continuous, non-repeating series of unlikely bit combinations with
   a one in 2^280 chance (I think!) of presence in an output stream.
*/
$DESCRIPTOR (DclEofFaoDsc, "^z<!8XL>~EoF[!8XL!8XL]!!");

/******************/
/* global storage */
/******************/

boolean  DclUseZombies;

int  DclCgiPlusLifeTimePurgeCount,
     DclHitHardLimitCount,
     DclPurgeCount,
     DclPurgeAllSubprocessesCount,
     DclSoftLimitPurgeCount,
     DclSubprocessCount,
     DclSubprocessHardLimit,
     DclSubprocessSoftLimit,
     DclZombieLifeTimePurgeCount;

struct GenericListStruct  DclTaskList;

#define DefaultCgiVariablePrefix "WWW_"
char DclCgiVariablePrefix [32] = DefaultCgiVariablePrefix;
int DclCgiVariablePrefixLength = sizeof(DefaultCgiVariablePrefix)-1;

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern int  DclSysOutputSize;
extern int  NetReadBufferSize;
extern int  ServerPort;
extern char  ErrorStringSize[];
extern char  ErrorSanityCheck[];
extern char  HtmlSgmlDoctype[];
extern char  HttpProtocol[];
extern char  SoftwareID[];
extern char  ServerHostName[];
extern struct AccountingStruct Accounting;
extern struct ConfigStruct Config;

/*****************************************************************************/
/*
Set and ensure limits are reasonable at server startup.
*/ 

DclInit ()

{
   /*********/
   /* begin */
   /*********/

   DclSubprocessSoftLimit = Config.DclSubprocessSoftLimit;
   if (DclSubprocessSoftLimit <= 0 || DclSubprocessSoftLimit > 99)
      DclSubprocessSoftLimit = Config.Busy;

   DclSubprocessHardLimit = Config.DclSubprocessHardLimit;
   if (DclSubprocessHardLimit <= DclSubprocessSoftLimit ||
       DclSubprocessHardLimit > 99)
      DclSubprocessHardLimit = DclSubprocessSoftLimit +
                               (DclSubprocessSoftLimit / 4);

   if (Config.DclZombieLifeTime) DclUseZombies = true;
}

/*****************************************************************************/
/*
This function does not return a status value.  If an error occurs the
'NextTaskFunction()' is executed.  The calling routine may assume that this
module will always execute the 'NextTaskFunction()' at some stage.
*/ 

DclBegin
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
char *DclCommand,
char *CgiFileName
)
{
   register struct DclTaskStruct  *tkptr;

   boolean  IsDclProc;
   int  status,
        Count,
        Length;
   char  *ContentPtr;

   /*********/
   /* begin */
   /*********/

   if (CgiFileName == NULL) CgiFileName = "";
   if (DclCommand == NULL) DclCommand = "";

   if (Debug)
   {
      fprintf (stdout,
"DclBegin()\n\
DclCommand     |%s|\n\
CgiFileName    |%s|\n\
ScriptName     |%s|%d|\n\
PathInfoPtr    |%s|\n\
QueryStringPtr |%s|\n",
      DclCommand, CgiFileName, rqptr->ScriptName, rqptr->IsCgiPlusScript,
      rqptr->PathInfoPtr, rqptr->QueryStringPtr);
   }

   /*****************/
   /* begin request */
   /*****************/

   if (rqptr->ErrorMessagePtr != NULL)
   {
      /* previous error, cause threaded processing to unravel */
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   /************************************/
   /* check DCL procedure/image exists */
   /************************************/

   if (CgiFileName[0])
   {
      if (VMSnok (status =
          DclFindScript (CgiFileName, &IsDclProc)))
      {
         if (status == RMS$_FNF && rqptr->ScriptName[0])
         {
            rqptr->ResponseStatusCode = 404;
            ErrorGeneral (rqptr, "Script not found.", __FILE__, __LINE__);
         }
         else
         {
            rqptr->ErrorTextPtr = "finding file";
            rqptr->ErrorHiddenTextPtr = CgiFileName;
            ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         }

         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }
   }

   /***************************/
   /* allocate task structure */
   /***************************/

   if (VMSnok (status = DclAllocateTask (rqptr)))
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   /*********************************/
   /* initialize the task structure */
   /*********************************/

   /* get a local pointer to the newly allocated DCL task structure */
   tkptr = rqptr->DclTaskPtr;

   tkptr->NextTaskFunction = NextTaskFunction;

   if (rqptr->ScriptName[0])
      strcpy (tkptr->ScriptName, rqptr->ScriptName);
   else
      tkptr->ScriptName[0] = '\0';

   if (DclCommand[0])
   {
      strncpy (tkptr->DclCommand, DclCommand, sizeof(tkptr->DclCommand));
      tkptr->DclCommand[sizeof(tkptr->DclCommand)-1] = '\0';
   }
   else
      tkptr->DclCommand[0] = '\0';

   memcpy (&tkptr->LastUsedBinaryTime, &rqptr->BinaryTime, 8);

   tkptr->CheckSubprocessResponseHeader =
      tkptr->AddNewline = tkptr->BufferRecords = false;
   if (tkptr->IsCgiScript || tkptr->IsCgiPlusScript)
      tkptr->CheckSubprocessResponseHeader = true;
   else
      tkptr->AddNewline = tkptr->BufferRecords = true;

   if (tkptr->IsDclCommand)
      Accounting.DoDclCommandCount++;

   if (!rqptr->AccountingDone)
   {
      rqptr->AccountingDone = true;
      if (tkptr->IsCgiScript)
         Accounting.DoScriptCount++;
      else
      if (tkptr->IsCgiPlusScript)
      {
         Accounting.DoCgiPlusScriptCount++;
         if (tkptr->CgiPlusUsageCount > 1)
            Accounting.DclCgiPlusReusedCount++;
      }
   }

   /*********************************************/
   /* provide data to subprocess' input streams */
   /*********************************************/

   if (tkptr->IsCgiScript || tkptr->IsCgiPlusScript)
   {
      /* provide request stream (HTTP$INPUT) to subprocess */
      if (Debug) 
         fprintf (stdout, "|%d|%s|\n",
                  rqptr->ContentLength, rqptr->ContentBufferPtr);

      /* initialize content information (POST, PUT, etc., only) */
      tkptr->ContentPtr = rqptr->ContentBufferPtr;
      tkptr->ContentLength = rqptr->ContentLength;

      if (Config.DclFullRequest)
      {
         /* supply request header and body (pre-4.3 behaviour) */
         if (VMSnok (status =
             sys$qio (0, tkptr->HttpInputChannel,
                      IO$_WRITELBLK, &tkptr->HttpInputIOsb,
                      &DclHttpInputAst, tkptr,
                      rqptr->RequestHeaderPtr, rqptr->RequestHeaderLength,
                      0, 0, 0, 0)))
         {
            rqptr->ErrorTextPtr = "sys$qio()";
            ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
            DclConcludeTask (tkptr, true);
            return;
         }
      }
      else
      if (tkptr->ContentLength)
      {
         /* supply only the body of the request (standard CGI behaviour) */
         if (tkptr->ContentLength > NetReadBufferSize)
            Length = NetReadBufferSize;
         else
            Length = tkptr->ContentLength;
         ContentPtr = tkptr->ContentPtr;
         tkptr->ContentPtr += Length;
         tkptr->ContentLength -= Length;

         if (VMSnok (status =
             sys$qio (0, tkptr->HttpInputChannel,
                      IO$_WRITELBLK, &tkptr->HttpInputIOsb,
                      &DclHttpInputAst, tkptr,
                      ContentPtr, Length,
                      0, 0, 0, 0)))
         {
            rqptr->ErrorTextPtr = "sys$qio()";
            ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
            DclConcludeTask (tkptr, true);
            return;
         }
      }
      else
      {
         /* the request has no body */
         if (VMSnok (status =
             sys$qio (0, tkptr->HttpInputChannel,
                      IO$_WRITEOF, &tkptr->HttpInputIOsb,
                      &DclHttpInputAst, tkptr,
                      0, 0, 0, 0, 0, 0)))
         {
            rqptr->ErrorTextPtr = "sys$qio()";
            ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
            DclConcludeTask (tkptr, true);
            return;
         }
      }
      tkptr->QueuedHttpInput++;
   }

   if (tkptr->IsCgiScript)
   {
      tkptr->QueuedSysCommandAllowed = 0;
      rqptr->ErrorTextPtr = "DCL environment";
      status = DclCgiScriptSysCommand (tkptr, "", CgiFileName, IsDclProc);
   }
   else
   if (tkptr->IsCgiPlusScript)
   {
      /* allow for the outstanding queued "STOP/ID=0" and EOF */
      tkptr->QueuedSysCommandAllowed = 2;
      if (tkptr->QueuedSysCommand < tkptr->QueuedSysCommandAllowed)
      {
         /* only need to supply the SYS$COMMAND stream once per subprocess */
         rqptr->ErrorTextPtr = "DCL environment";
         status = DclCgiPlusScriptSysCommand (tkptr, CgiFileName, IsDclProc);
      }
      else
         status = SS$_NORMAL;
      if (VMSok (status))
      {
         rqptr->ErrorTextPtr = "CGIplus variables";
         status = DclCgiPlusScriptCgiPlusIn (tkptr);
      }
   }
   else
   {
      tkptr->QueuedSysCommandAllowed = 0;
      rqptr->ErrorTextPtr = "DCL command";
      status = DclCgiScriptSysCommand (tkptr, DclCommand, "", false);
   }

   /* remove default error message */
   rqptr->ErrorTextPtr = NULL;

   if (VMSnok (status))
   {
      if (rqptr->ErrorMessagePtr == NULL)
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      DclConcludeTask (tkptr, true);
      return;
   }

   /*
      Return to the calling routine now.  Subsequent processing is
      event-driven.  Routine completion ASTs drives the script output.
   */
}

/*****************************************************************************/
/*
Allocate a DCL task structure to the request. All task structures are linked
together in a single list, function and state indicated by the various flags
and counters associated with each. If a CGIplus task script is to be executed
then check for an existing, idle CGIplus task structure executing that
particular script. If none found, or CGIplus script not required, and zombies
in use (persistant-subprocesses) then look through the list for an idle zombie
subprocess. If no zombie available (or not enabled) then check if we have
reached the subprocess creation hard-limit. If not reached look through the
list for an existing but no-subprocess-executing DCL task structure. If none
found create an additional DCL task structure and add it to the list.
Initialize the task structure (no matter from which scan it originated) and if
necessary create a subprocess for it. (An obvious improvement to processing
would be to have multiple lists, but that will have to wait for another time
:^) If an error is encountered an error message is generated and the error
status returned. It is up to the calling routine to abort the processing.
*/

DclAllocateTask (struct RequestStruct *rqptr)

{
   register struct DclTaskStruct  *tkptr;
   register struct GenericListEntryStruct  *leptr;

   int  status;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclAllocateTask()\n");

   tkptr = NULL;

   if (rqptr->IsCgiPlusScript)
   {
      /******************************************************/
      /* look for an unused instance of this CGIplus script */
      /******************************************************/

      for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
      {
         tkptr = (struct DclTaskStruct*)(leptr->DataPtr);

/*
         if (Debug)
         {
            fprintf (stdout, "leptr %d %d %d (%d)\n",
                     leptr->PrevPtr, leptr, leptr->NextPtr, tkptr);

            fprintf (stdout, "%08.08X %d %d %d %d %d %d %d |%s|\n",
                     tkptr->SubprocessPid, tkptr->rqptr,
                     tkptr->CgiPlusUsageCount,
                     tkptr->TotalUsageCount, tkptr->QueuedSysCommand,
                     tkptr->QueuedSysOutput, tkptr->QueuedCgiPlusIn,
                     tkptr->QueuedHttpInput, tkptr->ScriptName);
         }
*/
 
         if (!tkptr->SubprocessPid ||
             !tkptr->IsCgiPlusScript ||
             tkptr->QueuedSysCommand > tkptr->QueuedSysCommandAllowed ||
             tkptr->QueuedSysOutput ||
             tkptr->QueuedCgiPlusIn ||
             tkptr->QueuedHttpInput ||
             tkptr->rqptr ||
             !strsame (rqptr->ScriptName, tkptr->ScriptName, -1))
         {
            tkptr = NULL;
            continue;
         }

         break;
      }

      if (Debug) fprintf (stdout, "CGIplus tkptr: %d\n", tkptr);
   }

   if (tkptr == NULL && DclUseZombies)
   {
      /************************/
      /* look for idle zombie */
      /************************/

      for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
      {
         tkptr = (struct DclTaskStruct*)(leptr->DataPtr);

/*
         if (Debug)
         {
            fprintf (stdout, "leptr %d %d %d (%d)\n",
                     leptr->PrevPtr, leptr, leptr->NextPtr, tkptr);

            fprintf (stdout, "%08.08X %d %d %d %d %d %d %d |%s|\n",
                     tkptr->SubprocessPid, tkptr->rqptr,
                     tkptr->ZombieCount,
                     tkptr->TotalUsageCount, tkptr->QueuedSysCommand,
                     tkptr->QueuedSysOutput, tkptr->QueuedCgiPlusIn,
                     tkptr->QueuedHttpInput, tkptr->ScriptName);
         }
*/

         if (!tkptr->SubprocessPid ||
             tkptr->IsCgiPlusScript ||
             tkptr->QueuedSysCommand ||
             tkptr->QueuedSysOutput ||
             tkptr->QueuedCgiPlusIn ||
             tkptr->QueuedHttpInput ||
             tkptr->rqptr != NULL)
         {
            tkptr = NULL;
            continue;
         }
         break;
      }

      if (Debug) fprintf (stdout, "idle zombie tkptr: %d\n", tkptr);
   }

   if (tkptr == NULL && DclSubprocessCount >= DclSubprocessHardLimit)
   {
      /* let's see if we can do something about it! */
      DclSubprocessPurge ();

      DclHitHardLimitCount++;
      ErrorGeneral (rqptr, "Exceeded DCL task hard-limit!",
                    __FILE__, __LINE__);
      return (STS$K_ERROR);
   }

   if (tkptr == NULL)
   {
      /********************************/
      /* look for free task structure */
      /********************************/

      for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
      {
         tkptr = (struct DclTaskStruct*)(leptr->DataPtr);

/*
         if (Debug)
         {
            fprintf (stdout, "leptr %d %d %d (%d)\n",
                     leptr->PrevPtr, leptr, leptr->NextPtr, tkptr);

            fprintf (stdout, "%08.08X %d %d %d %d %d %d %d |%s|\n",
                     tkptr->SubprocessPid, tkptr->rqptr,
                     tkptr->ZombieCount,
                     tkptr->TotalUsageCount, tkptr->QueuedSysCommand,
                     tkptr->QueuedSysOutput, tkptr->QueuedCgiPlusIn,
                     tkptr->QueuedHttpInput, tkptr->ScriptName);
         }
*/

         if (tkptr->SubprocessPid ||
             tkptr->QueuedSysCommand ||
             tkptr->QueuedSysOutput ||
             tkptr->QueuedCgiPlusIn ||
             tkptr->QueuedHttpInput ||
             tkptr->rqptr != NULL)
         {
            tkptr = NULL;
            continue;
         }
         break;
      }

      if (Debug) fprintf (stdout, "free tkptr: %d\n", tkptr);
   }

   if (tkptr == NULL)
   {
      /* if we're getting short of subprocesses then start purging */
      if (DclSubprocessCount >= DclSubprocessSoftLimit) DclSubprocessPurge ();
   }

   if (tkptr == NULL)
   {
      /*********************/
      /* create a new task */
      /*********************/

      if ((tkptr = calloc (1, sizeof (struct DclTaskStruct))) == NULL)
         ErrorExitVmsStatus (status, "calloc()", __FILE__, __LINE__);

/*
      if (Debug) fprintf (stdout, "new tkptr: %d\n", tkptr);
*/

      /*
         Allocate memory in the DCL task for SYS$OUTPUT buffer.
         Allow two bytes for carriage control and terminating null.
      */
      if ((tkptr->SysOutputPtr = calloc (1, DclSysOutputSize+2)) == NULL)
      {
         free (tkptr);
         ErrorGeneral (rqptr, "calloc()", __FILE__, __LINE__);
         return (STS$K_ERROR);
      }
      tkptr->SysOutputSize = DclSysOutputSize;

      if (VMSnok (status = DclCreateMailboxes (tkptr)))
      {
         free (tkptr->SysOutputPtr);
         free (tkptr);
         rqptr->ErrorTextPtr = "creating mailbox IPC";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         return (status);
      }

      ListAddTail (&DclTaskList, &tkptr->DclListEntry, tkptr);
   }

   /*******************/
   /* initialize task */
   /*******************/

   /* associate the DCL task and the request */
   rqptr->DclTaskPtr = tkptr;
   tkptr->rqptr = rqptr;          

   tkptr->TotalUsageCount++;
   tkptr->IsMarkedForDelete = false;

   if (rqptr->ScriptName[0])
   {
      /* script is begin executed */
      if (rqptr->IsCgiPlusScript)
      {
         /* CGIplus script is being executed */
         tkptr->IsCgiPlusScript = true;
         tkptr->IsDclCommand = tkptr->IsCgiScript = false;
         tkptr->CgiPlusUsageCount++;

         /* give it three-score years and ten if life-time is specified */
         if (Config.DclCgiPlusLifeTime)
            status = DclSetCgiPlusLifeTime (tkptr);

         /* reset the zombie count, CGIplus scripts never become zombies */
         tkptr->ZombieCount = 0;

         /* CGIplus MUST retain the original EOF until subprocess deletion */
      }
      else
      {
         /* standard CGI script is being executed */
         tkptr->IsCgiScript = true;
         tkptr->IsDclCommand = tkptr->IsCgiPlusScript = false;
         tkptr->CgiPlusUsageCount = 0;

         /* limited life in the twilight zone */
         if (DclUseZombies && Config.DclZombieLifeTime)
            status = DclSetZombieLifeTime (tkptr);

         /* always generate a new EOF string for standard CGI scripts */
         tkptr->DclEof[0] = '\0';
         tkptr->DclEofLength = 0;
      }
   }
   else
   {
      /* DCL command is begin executed (from SSI) */
      tkptr->IsDclCommand = true;
      tkptr->IsCgiScript = tkptr->IsCgiPlusScript = false;
      tkptr->CgiPlusUsageCount = 0;

      /* limited life in the twilight zone */
      if (DclUseZombies && Config.DclZombieLifeTime)
         status = DclSetZombieLifeTime (tkptr);

      /* always generate a new EOF string for DCL commands */
      tkptr->DclEof[0] = '\0';
      tkptr->DclEofLength = 0;
   }

   if (tkptr->SubprocessPid)
   {
      /*********************/
      /* subprocess exists */
      /*********************/

      /* queue the initial read of the subprocess' SYS$OUTPUT */
      DclQioSysOutput (tkptr);

      return (SS$_NORMAL);
   }

   /*********************/
   /* create subprocess */
   /*********************/

   if (VMSok (status = DclCreateSubprocess (tkptr)))
   {
      /* if a CGIplus script and a life-time is specified then set it */
      if (tkptr->IsCgiPlusScript && Config.DclCgiPlusLifeTime)
         DclSetCgiPlusLifeTime (tkptr);

      tkptr->ZombieCount = 0;
      DclSubprocessCount++;
      return (status);
   }
   else
   {
      /* disassociate the DCL task and request structures */
      rqptr->DclTaskPtr = tkptr->rqptr = NULL;
      return (status);
   }
}

/*****************************************************************************/
/*
*/

int DclSetZombieLifeTime (struct DclTaskStruct *tkptr)

{
   static unsigned long  LifeTimeDeltaBinTime [2] = { 0, 0 };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclSetZombieLifeTime() %d\n", tkptr);

   if (!Config.DclZombieLifeTime) return (SS$_NORMAL);

   if (!LifeTimeDeltaBinTime[1])
   {
      int  Days, Hours, Minutes;
      unsigned short  Length;
      char  LifeTime [32];
      $DESCRIPTOR (LifeTimeDsc, LifeTime);
      $DESCRIPTOR (LifeTimeFaoDsc, "!UL !UL:!UL:00.00");
   
      Minutes = Config.DclZombieLifeTime;
      Days = Minutes / 1440;
      Minutes = Minutes - Days * 1440;
      Hours = Minutes / 60;
      Minutes = Minutes - Hours * 60;
      if (VMSnok (status =
          sys$fao (&LifeTimeFaoDsc, &Length, &LifeTimeDsc,
                   Days, Hours, Minutes)))
         ErrorExitVmsStatus (status, "Zombie life-time", __FILE__, __LINE__);
      LifeTime[LifeTimeDsc.dsc$w_length = Length] = '\0';
      if (Debug) fprintf (stdout, "LifeTime |%s|\n", LifeTime);
      if (VMSnok (status =
          sys$bintim (&LifeTimeDsc, &LifeTimeDeltaBinTime)))
         ErrorExitVmsStatus (status, "Zombie life-time", __FILE__, __LINE__);
   }

   sys$cantim (tkptr, 0);
   return (sys$setimr (0, &LifeTimeDeltaBinTime,
                       &DclZombieLifeTimeAst, tkptr, 0));
}

/*****************************************************************************/
/*
This AST is called when the sys$setimr() of DclSetZombieLifeTime() expires.
*/ 

DclZombieLifeTimeAst (struct DclTaskStruct *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclZombieLifeTimeAst() %d\n", tkptr);

   if (tkptr->SubprocessPid &&
       !tkptr->IsCgiPlusScript &&
       !tkptr->QueuedSysCommand &&
       !tkptr->QueuedSysOutput &&
       !tkptr->QueuedCgiPlusIn &&
       !tkptr->QueuedHttpInput &&
       tkptr->rqptr == NULL)
   {
      /* zombie subprocess exists, not active, abort marked for delete */
      tkptr->IsMarkedForDelete = true;
      DclConcludeTask (tkptr, true);
      DclZombieLifeTimePurgeCount++;
      return;
   }

   /* looks like task became active after the timer expired, ignore */
}

/*****************************************************************************/
/*
*/

int DclSetCgiPlusLifeTime (struct DclTaskStruct *tkptr)

{
   static unsigned long  LifeTimeDeltaBinTime [2] = { 0, 0 };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclSetCgiPlusLifeTime() %d\n", tkptr);

   if (!Config.DclCgiPlusLifeTime) return (SS$_NORMAL);

   if (!LifeTimeDeltaBinTime[1])
   {
      int  Days, Hours, Minutes;
      unsigned short  Length;
      char  LifeTime [32];
      $DESCRIPTOR (LifeTimeDsc, LifeTime);
      $DESCRIPTOR (LifeTimeFaoDsc, "!UL !UL:!UL:00.00");
   
      Minutes = Config.DclCgiPlusLifeTime;
      Days = Minutes / 1440;
      Minutes = Minutes - Days * 1440;
      Hours = Minutes / 60;
      Minutes = Minutes - Hours * 60;
      if (VMSnok (status =
          sys$fao (&LifeTimeFaoDsc, &Length, &LifeTimeDsc,
                   Days, Hours, Minutes)))
         ErrorExitVmsStatus (status, "CgiPlus life-time", __FILE__, __LINE__);
      LifeTime[LifeTimeDsc.dsc$w_length = Length] = '\0';
      if (Debug) fprintf (stdout, "LifeTime |%s|\n", LifeTime);
      if (VMSnok (status =
          sys$bintim (&LifeTimeDsc, &LifeTimeDeltaBinTime)))
         ErrorExitVmsStatus (status, "CgiPlus life-time", __FILE__, __LINE__);
   }

   sys$cantim (tkptr, 0);
   return (sys$setimr (0, &LifeTimeDeltaBinTime,
                       &DclCgiPlusLifeTimeAst, tkptr, 0));
}

/*****************************************************************************/
/*
This AST is called when the sys$setimr() of DclSetCgiPlusLifeTime() expires.
*/ 

DclCgiPlusLifeTimeAst (struct DclTaskStruct *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclCgiPlusLifeTimeAst() %d\n", tkptr);

   if (tkptr->SubprocessPid &&
       tkptr->IsCgiPlusScript &&
       !tkptr->QueuedSysOutput &&
       !tkptr->QueuedCgiPlusIn &&
       !tkptr->QueuedHttpInput &&
       tkptr->rqptr == NULL)
   {
      /* CGIplus subprocess exists, not active, abort marked for delete */
      tkptr->IsMarkedForDelete = true;
      DclConcludeTask (tkptr, true);
      DclCgiPlusLifeTimePurgeCount++;
      return;
   }

   /* looks like task became active after the timer expired, ignore */
}

/*****************************************************************************/
/*
Called if the HTTPd server is being restarted (image run down and then
restarted).  Does not need to be done if process is exiting.  This function
should be included in the image's onexit() function.
*/

DclExit ()

{
   register struct DclTaskStruct  *tkptr;
   register struct GenericListEntryStruct  *leptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclExit()\n");

   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)(leptr->DataPtr);
      if (!tkptr->SubprocessPid) continue;
      status = sys$delprc (&tkptr->SubprocessPid, 0);
      if (Debug)
         fprintf (stdout, "sys$delprc() %d %08.08X %%X%08.08X\n",
                  tkptr, tkptr->SubprocessPid, status);
   }
}

/*****************************************************************************/
/*
The maximum number of concurrent subprocesses has been reached. Look through
the DCL task structure for a CGIplus subprocess not currently in use. Find the
least used subprocess and delete it.
*/

DclSubprocessPurge ()

{
   register struct DclTaskStruct  *tkptr;
   register struct GenericListEntryStruct  *leptr;

   int  status,
        MinUsageCount;
   struct DclTaskStruct  *MinTkPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclSubprocessPurge() %d %d\n",
               DclSubprocessCount, DclSubprocessSoftLimit);

   DclPurgeCount++;
   MinUsageCount = 999999999;
   MinTkPtr = NULL;

   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)(leptr->DataPtr);

      if (!tkptr->SubprocessPid ||
          !tkptr->IsCgiPlusScript ||
          tkptr->QueuedSysOutput ||
          tkptr->QueuedCgiPlusIn ||
          tkptr->QueuedHttpInput ||
          tkptr->rqptr != NULL ||
          tkptr->IsMarkedForDelete)
         continue;

      if (tkptr->CgiPlusUsageCount < MinUsageCount)
         MinUsageCount = (MinTkPtr = tkptr)->CgiPlusUsageCount;
   }

   if (MinTkPtr == NULL) return;

   MinTkPtr->IsMarkedForDelete = true;
   DclConcludeTask (MinTkPtr, true);
   DclSoftLimitPurgeCount++;
}

/*****************************************************************************/
/*
This function can be called at any stage to rundown or abort a DCL task.
If there is still outstanding IO this is cancelled as appropriate to task
rundown or abort.  If no outstanding IO then if there is an associated request
that request's next task function is called.
*/

DclConcludeTask
(
struct DclTaskStruct *tkptr,
boolean AbortTask
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout, "DclConcludeTask(%d) %d %d %d %d %d %d %08.08X\n",
         AbortTask, tkptr, tkptr->rqptr, tkptr->QueuedSysCommand,
         tkptr->QueuedSysOutput, tkptr->QueuedCgiPlusIn,
         tkptr->QueuedHttpInput, tkptr->SubprocessPid);
   }

   if (AbortTask && tkptr->SubprocessPid)
   {
      tkptr->IsMarkedForDelete = true;
      tkptr->QueuedSysCommandAllowed = 0;
   }

   if (tkptr->QueuedSysCommand > tkptr->QueuedSysCommandAllowed ||
       tkptr->QueuedSysOutput ||
       tkptr->QueuedCgiPlusIn ||
       tkptr->QueuedHttpInput)
   {
      /****************/
      /* rundown task */
      /****************/

      if (AbortTask)
      {
         /* cancel everything everywhere! */
         sys$cancel (tkptr->SysCommandChannel);
         sys$cancel (tkptr->SysOutputChannel);
         sys$cancel (tkptr->CgiPlusInChannel);
         sys$cancel (tkptr->HttpInputChannel);
      }
      else
      {
         /* cancel any outstanding I/O on input channels */
         if (tkptr->QueuedCgiPlusIn) sys$cancel (tkptr->CgiPlusInChannel);
         if (tkptr->QueuedHttpInput) sys$cancel (tkptr->HttpInputChannel);
         if (tkptr->QueuedSysCommand > tkptr->QueuedSysCommandAllowed)
            sys$cancel (tkptr->SysCommandChannel);
      }
   }
   else
   {
      /*********************/
      /* task has finished */
      /*********************/

      if (tkptr->rqptr != NULL)
      {
         /* still has an associated request, declare the next task */
         if (VMSnok (status =
             sys$dclast (tkptr->NextTaskFunction, tkptr->rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

         /* disassociate the DCL task and request structures */
         tkptr->rqptr->DclTaskPtr = NULL;
         tkptr->rqptr = tkptr->NextTaskFunction = NULL;
      }

      if (DclUseZombies || tkptr->IsCgiPlusScript)
      {
         /* empty any remains in sys$output! */
         DclEmptySysOutput (tkptr);
      }

      /* if marked for process deletion */
      if (tkptr->IsMarkedForDelete)
      {
         tkptr->IsMarkedForDelete = false;
         if (tkptr->SubprocessPid)
         {
            status = sys$delprc (&tkptr->SubprocessPid, 0);
            if (Debug)
               fprintf (stdout, "sys$delprc() %08.08X %%X%08.08X\n",
                        tkptr->SubprocessPid, status);
         }
      }

      if (DclUseZombies && !tkptr->IsCgiPlusScript) tkptr->ZombieCount++;
   }
}

/*****************************************************************************/
/*
If an error is encountered an error message is generated and the error status
returned.  It is up to the calling routine to abort the processing.  Queue a
writer-wait I/O to the SYS$OUTPUT channel to stall I/O until the subprocess
has started.
*/ 

DclCreateSubprocess (struct DclTaskStruct *tkptr)

{
   /* spawn DCL subprocess with account's authorized privileges */
   static unsigned long  /* authpriv,trusted,nowait,noclisym,nolognam,nokey */
                         AuthPrivSpawnFlags = 0xcf;

   /* spawn DCL subprocess with current privileges (none at all!) */
   static unsigned long  /* nowait,noclisym,nolognam,nokey */
                         CurrPrivSpawnFlags = 0x0f;

   static int  SubprocessNumber;
   static char  SubprocessName [16];
   static $DESCRIPTOR (SubprocessNameFaoDsc, "!15<HTTPd:!UL-!UL!>");
   static $DESCRIPTOR (SubprocessNameDsc, SubprocessName);

   register struct RequestStruct  *rqptr;

   int  status,
        Count;
   unsigned short  Length;
   unsigned long  SpawnFlags;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclCreateSubprocess()\n");

   /* get a pointer to the request structure */
   rqptr = tkptr->rqptr;

   /*
      Queue up an asynchronous wait for a writer channel to be assigned
      to the mailbox as the subprocess' SYS$OUTPUT.  When the channel is
      assigned the AST function will queue the first read from the mailbox.
   */
   if (VMSnok (status =
       sys$qio (0, tkptr->SysOutputChannel,
                IO$_SETMODE | IO$M_WRITERWAIT,
                &tkptr->SysOutputIOsb,
                &DclSysOutputWaitAst, tkptr,
                0, 0, 0, 0, 0, 0)))
   {
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      return (status);
   }
   if (Debug) fprintf (stdout, "sys$qio() READERWAIT %%X%08.08X\n", status);
   tkptr->QueuedSysOutput++;

   if (Config.DclSpawnAuthPriv)
      SpawnFlags = AuthPrivSpawnFlags;
   else
      SpawnFlags = CurrPrivSpawnFlags;

   for (Count = 100; Count; Count--)
   {
      /* subprocesses are consecutively numbered from 1..999 */
      SubprocessNameDsc.dsc$w_length = sizeof(SubprocessName)-1;
      sys$fao (&SubprocessNameFaoDsc, &Length, &SubprocessNameDsc, 
               ServerPort, SubprocessNumber++ % 1000 + 1);
      SubprocessName[SubprocessNameDsc.dsc$w_length = Length] = '\0';
      if (Debug) fprintf (stdout, "SubprocessName |%s|\n", SubprocessName);

      status = lib$spawn (0,
                          &tkptr->SysCommandDevNameDsc,
                          &tkptr->SysOutputDevNameDsc,
                          &SpawnFlags,
                          &SubprocessNameDsc,
                          &tkptr->SubprocessPid,
                          &tkptr->SubprocessCompletionStatus,
                          0,
                          &DclSubprocessCompletionAST, tkptr,
                          0, 0, 0);

      if (Debug)
         fprintf (stdout, "lib$spawn() %%X%08.08X PID: %08.08X\n",
                  status, tkptr->SubprocessPid);

      if (status != SS$_DUPLNAM) break;
   }

   if (VMSnok (status))
   {
      /* cancel the writer-wait IO */
      sys$cancel (tkptr->SysOutputChannel);

      rqptr->ErrorTextPtr = "creating subprocess";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      return (status);
   }

   Accounting.DclSpawnCount++;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This AST is called when the subprocesses exits.
*/

DclSubprocessCompletionAST (struct DclTaskStruct *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclSubprocessCompletionAST() %08.08X %%X%08.08X\n",
         tkptr->SubprocessPid, tkptr->SubprocessCompletionStatus);
   }

   if (tkptr->IsCgiPlusScript)
   {
      /* ensure SYS$COMMAND gets emptied! */
      tkptr->QueuedSysCommandAllowed = 0;

      /* may be called due to life-time expiry, but this won't hurt anyway! */
      if (Config.DclCgiPlusLifeTime) sys$cantim (tkptr, 0);
   }
   else
   if (DclUseZombies)
   {
      /* may be called due to life-time expiry, but this won't hurt anyway! */
      sys$cantim (tkptr, 0);
   }

   /* set the PID to zero to indicate the subprocess no longer exists */
   tkptr->SubprocessPid = 0;

   /* keep track of how many subprocesses are executing */
   if (DclSubprocessCount) DclSubprocessCount--;

   /* ensure any old EOF string is not reused */
   tkptr->DclEof[0] = '\0';
   tkptr->DclEofLength = 0;

   DclConcludeTask (tkptr, false);
}

/*****************************************************************************/
/*
If a script application writes any output before it was/is terminated that
output will still be laying unread in the mailbox.  Clear any such noise lest
it be read by the next subprocess to use the mailbox.  Does this synchronously
(i.e. with QIO waits).
*/

DclEmptySysOutput (struct DclTaskStruct *tkptr)

{
   struct MbxSenseIOsb  SenseIOsb;
   int  LastMessageCount;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclEmptySysOutput()\n");

   LastMessageCount = 999999999;
   SenseIOsb.MessageCount = 0;

   for (;;)
   {
      sys$qiow (0, tkptr->SysOutputChannel,
                IO$_SENSEMODE, &SenseIOsb,
                0, 0, 0, 0, 0, 0, 0, 0);

      if (VMSnok (SenseIOsb.Status))
         ErrorExitVmsStatus (SenseIOsb.Status, "sys$qiow()",
                             __FILE__, __LINE__);

      if (Debug)
         fprintf (stdout, "sys$qiow() IO$_SENSEMODE %%X%08.08X %d %d\n", 
                  SenseIOsb.Status, SenseIOsb.MessageCount,
                  SenseIOsb.MessageBytes);

      if (!SenseIOsb.MessageCount) break;
      /* potential infinite loop, check message count is decreasing! */
      if (LastMessageCount <= SenseIOsb.MessageCount) break;
      LastMessageCount = SenseIOsb.MessageCount;

      sys$qiow (0, tkptr->SysOutputChannel,
                IO$_READLBLK | IO$M_WRITERCHECK,
                &tkptr->SysOutputIOsb, 0, 0,
                tkptr->SysOutputPtr, tkptr->SysOutputSize,
                0, 0, 0, 0);

      if (Debug)
      {
         tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count] = '\0';
         fprintf (stdout, "sys$qiow() IO$_READLBLK %%X%08.08X %d\n|%s|\n", 
                  tkptr->SysOutputIOsb.Status, tkptr->SysOutputIOsb.Count,
                  tkptr->SysOutputPtr);
      }
   }
}

/*****************************************************************************/
/*
If an error is encountered an error message is generated and the error status
returned.  It is up to the calling routine to abort the processing.  Create
four mailboxes that will be associated with the subprocess I/O streams.  If
an error occurs any mailbox created up to that point is deleted and the
channel set back to zero.
*/ 

DclCreateMailboxes (struct DclTaskStruct *tkptr)

{
   static unsigned long  DevNamItem = DVI$_DEVNAM;

   int  status;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclCreateMailboxes()\n");

   /**********************************/
   /* create the SYS$COMMAND mailbox */
   /**********************************/

   if (VMSnok (status =
       sys$crembx (0,
                   &tkptr->SysCommandChannel,
                   DclSysOutputSize/2, DclSysOutputSize/2,
                   SubprocessMbxProtectionMask,
                   0, 0, CMB$M_WRITEONLY)))
   {
      if (Debug) fprintf (stdout, "sys$crembx() %%X%08.08X\n", status);
      return (status);
   }

   tkptr->SysCommandDevNameDsc.dsc$w_length = sizeof(tkptr->SysCommandDevName);
   tkptr->SysCommandDevNameDsc.dsc$a_pointer = tkptr->SysCommandDevName;
   tkptr->SysCommandDevNameDsc.dsc$b_class = DSC$K_CLASS_S;
   tkptr->SysCommandDevNameDsc.dsc$b_dtype = DSC$K_DTYPE_T;

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &tkptr->SysCommandChannel,
                   0, 0, &tkptr->SysCommandDevNameDsc, &Length)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      return (status);
   }
   tkptr->SysCommandDevName
      [tkptr->SysCommandDevNameDsc.dsc$w_length = Length] = '\0';
   if (Debug)
      fprintf (stdout, "SysCommandDevName |%s|\n", tkptr->SysCommandDevName);

   /*********************************/
   /* create the SYS$OUTPUT mailbox */
   /*********************************/

   if (VMSnok (status =
       sys$crembx (0,
                   &tkptr->SysOutputChannel,
                   DclSysOutputSize, DclSysOutputSize,
                   SubprocessMbxProtectionMask,
                   0, 0, CMB$M_READONLY)))
   {
      if (Debug) fprintf (stdout, "sys$crembx() %%X%08.08X\n", status);
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      return (status);
   }

   tkptr->SysOutputDevNameDsc.dsc$w_length = sizeof(tkptr->SysOutputDevName);
   tkptr->SysOutputDevNameDsc.dsc$a_pointer = tkptr->SysOutputDevName;
   tkptr->SysOutputDevNameDsc.dsc$b_class = DSC$K_CLASS_S;
   tkptr->SysOutputDevNameDsc.dsc$b_dtype = DSC$K_DTYPE_T;

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &tkptr->SysOutputChannel,
                   0, 0, &tkptr->SysOutputDevNameDsc, &Length)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      tkptr->HttpInputChannel = tkptr->SysCommandChannel = 0;
      return (status);
   }
   tkptr->SysOutputDevName
      [tkptr->SysOutputDevNameDsc.dsc$w_length = Length] = '\0';
   if (Debug)
      fprintf (stdout, "SysOutputDevName |%s|\n", tkptr->SysOutputDevName);

   /********************************/
   /* create the CGIPLUSIN mailbox */
   /********************************/

   if (VMSnok (status =
       sys$crembx (0,
                   &tkptr->CgiPlusInChannel,
                   DclSysOutputSize/2, DclSysOutputSize/2,
                   SubprocessMbxProtectionMask,
                   0, 0, CMB$M_WRITEONLY)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      if (Debug) fprintf (stdout, "sys$crembx() %%X%08.08X\n", status);
      return (status);
   }

   tkptr->CgiPlusInDevNameDsc.dsc$w_length = sizeof(tkptr->CgiPlusInDevName);
   tkptr->CgiPlusInDevNameDsc.dsc$a_pointer = tkptr->CgiPlusInDevName;
   tkptr->CgiPlusInDevNameDsc.dsc$b_class = DSC$K_CLASS_S;
   tkptr->CgiPlusInDevNameDsc.dsc$b_dtype = DSC$K_DTYPE_T;

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &tkptr->CgiPlusInChannel,
                   0, 0, &tkptr->CgiPlusInDevNameDsc, &Length)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      sys$dassgn (tkptr->CgiPlusInChannel);
      tkptr->CgiPlusInChannel = 0;
      return (status);
   }
   tkptr->CgiPlusInDevName
     [tkptr->CgiPlusInDevNameDsc.dsc$w_length = Length] = '\0';
   if (Debug)
      fprintf (stdout, "CgiPlusInDevName |%s|\n", tkptr->CgiPlusInDevName);

   /*********************************/
   /* create the HTTP$INPUT mailbox */
   /*********************************/

   if (VMSnok (status =
       sys$crembx (0,
                   &tkptr->HttpInputChannel,
                   NetReadBufferSize, NetReadBufferSize,
                   SubprocessMbxProtectionMask,
                   0, 0, CMB$M_WRITEONLY)))
   {
      if (Debug) fprintf (stdout, "sys$crembx() %%X%08.08X\n", status);
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      sys$dassgn (tkptr->CgiPlusInChannel);
      tkptr->CgiPlusInChannel = 0;
      return (status);
   }

   tkptr->HttpInputDevNameDsc.dsc$w_length = sizeof(tkptr->HttpInputDevName);
   tkptr->HttpInputDevNameDsc.dsc$a_pointer = tkptr->HttpInputDevName;
   tkptr->HttpInputDevNameDsc.dsc$b_class = DSC$K_CLASS_S;
   tkptr->HttpInputDevNameDsc.dsc$b_dtype = DSC$K_DTYPE_T;

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &tkptr->HttpInputChannel,
                   0, 0, &tkptr->HttpInputDevNameDsc, &Length)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      sys$dassgn (tkptr->CgiPlusInChannel);
      tkptr->CgiPlusInChannel = 0;
      sys$dassgn (tkptr->HttpInputChannel);
      tkptr->HttpInputChannel = 0;
      return (status);
   }
   tkptr->HttpInputDevName
     [tkptr->HttpInputDevNameDsc.dsc$w_length = Length] = '\0';
   if (Debug)
      fprintf (stdout, "HttpInputDevName |%s|\n", tkptr->HttpInputDevName);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
A channel has been assigned to the subprocess' SYS$OUTPUT mailbox.  If it is
still associated with an actual subprocess (PID is not empty) queue the first
asynchronous read of the subprocess' output.
*/ 

DclSysOutputWaitAst (struct DclTaskStruct *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclSysOutputWaitAst() %d %d %d %d %d %d %%X%08.08X\n",
         tkptr, tkptr->rqptr, tkptr->QueuedSysCommand,
         tkptr->QueuedSysOutput, tkptr->QueuedCgiPlusIn,
         tkptr->QueuedHttpInput, tkptr->SysOutputIOsb.Status);
   }

   if (tkptr->QueuedSysOutput) tkptr->QueuedSysOutput--;

   /* abort if a request or subprocess is not associated with the task */
   if (tkptr->rqptr == NULL || !tkptr->SubprocessPid)
   {
      DclConcludeTask (tkptr, true);
      return;
   }

   /* if the subprocess is in the process or aborting/being deleted */
   if (tkptr->IsMarkedForDelete)
   {
      DclConcludeTask (tkptr, false);
      return;
   }

   if (VMSnok (tkptr->SysOutputIOsb.Status))
   {
      if (tkptr->SysOutputIOsb.Status == SS$_ABORT ||
          tkptr->SysOutputIOsb.Status == SS$_CANCEL)
      {
         /* looks like the I/O has been cancelled before it started! */
         DclConcludeTask (tkptr, true);
         return;
      }

      ErrorVmsStatus (tkptr->rqptr, tkptr->SysOutputIOsb.Status,
                      __FILE__, __LINE__);
      DclConcludeTask (tkptr, true);
      return;
   }

   /* queue the initial read of the subprocess' SYS$OUTPUT */
   DclQioSysOutput (tkptr);
}

/*****************************************************************************/
/*
Queue up a read from the subprocess "SYS$OUTPUT" mailbox.  When the read 
completes call function DclSysOutputAst(), do any post-processing 
required and write the data to the client over the network.  The next read 
from the subprocess via the mailbox will be queued by the network write 
completion AST function.
*/ 

DclQioSysOutput (struct DclTaskStruct *tkptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclSysOutput() %d %d\n", tkptr, tkptr->rqptr);

   if (VMSok (status =
       sys$qio (0, tkptr->SysOutputChannel,
                IO$_READLBLK | IO$M_WRITERCHECK,
                &tkptr->SysOutputIOsb,
                &DclSysOutputAst, tkptr,
                tkptr->SysOutputPtr, tkptr->SysOutputSize,
                0, 0, 0, 0)))
   {
      tkptr->QueuedSysOutput++;
      return (status);
   }
   return (status);
}

/*****************************************************************************/
/*
A queued asynchronous read from the subprocess "SYS$OUTPUT" mailbox has 
completed.  If this record is part of the HTTP header do some specific 
processing.  If required append carriage-control (newline) to this record to 
make it HTTP compliant.  Queue a write of this data to the client over the 
network.  The next read from the subprocess via the mailbox will be queued by 
the network write completion AST function, DclQioSysOutput().
*/ 

DclSysOutputAst (struct DclTaskStruct *tkptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclSysOutputAst() %d %d %08.08X %d %d %d %d %d %%X%08.08X\n",
         tkptr, tkptr->rqptr, tkptr->SubprocessPid,
         tkptr->QueuedSysCommand, tkptr->QueuedSysOutput,
         tkptr->QueuedCgiPlusIn, tkptr->QueuedHttpInput,
         tkptr->SysOutputIOsb.Count, tkptr->SysOutputIOsb.Status);

/*
      tkptr->SysOutputPtr [tkptr->SysOutputIOsb.Count] = '\0';
      fprintf (stdout, "|%s|\n", tkptr->SysOutputPtr);
*/
   }

   if (tkptr->QueuedSysOutput) tkptr->QueuedSysOutput--;

   if (tkptr->SysOutputIOsb.Status == SS$_ENDOFFILE)
   {
      /* 
          If a script spawns multiple subprocesses each will
          terminate by queueing an end-of-file.  Ignore these.
          Queue the next read of the subprocess' SYS$OUTPUT, a
          SS$_NOWRITER or DCL EOF signals completion of output!
      */
      DclQioSysOutput (tkptr);
      return;
   }

   if (VMSnok (tkptr->SysOutputIOsb.Status))
   {
      if (tkptr->SysOutputIOsb.Status == SS$_NOWRITER)
      {
         /* subprocess has terminated */
         DclConcludeTask (tkptr, false);
         return;
      }

      if (tkptr->SysOutputIOsb.Status == SS$_ABORT ||
          tkptr->SysOutputIOsb.Status == SS$_CANCEL)
      {
         DclConcludeTask (tkptr, true);
         return;
      }

      ErrorVmsStatus (tkptr->rqptr, tkptr->SysOutputIOsb.Status,
                      __FILE__, __LINE__);
      DclConcludeTask (tkptr, true);
      return;
   }

   if (tkptr->DclEofLength)
   {
      if (tkptr->SysOutputPtr[0] == tkptr->DclEof[0] &&
          tkptr->SysOutputIOsb.Count >= tkptr->DclEofLength &&
          tkptr->SysOutputIOsb.Count <= tkptr->DclEofLength+2 &&
          !memcmp (tkptr->SysOutputPtr, tkptr->DclEof, tkptr->DclEofLength))
      {
         /******************************/
         /* end of output from script! */
         /******************************/

         if (Debug) fprintf (stdout, "DCL EOF!\n");
         DclConcludeTask (tkptr, false);
         return;
      }
   }

   if (tkptr->rqptr->LocationPtr != NULL)
   {
      /*
         Absorb any output from script until the request processing ends.
         The script has output a "Location:" redirection header for a
         partial URL that can be processed locally.  This request will be
         resubmitted to the server as soon as the current script subprocess
         terminates.
      */
      DclQioSysOutput (tkptr);
      return;
   }

   if (tkptr->IsCgiScript || tkptr->IsCgiPlusScript)
   {
      /********************/
      /* script execution */
      /********************/

      if (tkptr->CheckSubprocessResponseHeader)
      {
         if (DclSysOutputHeader (tkptr))
            return;
      }

      if (tkptr->rqptr->LocationPtr != NULL)
      {
         /* redirection, absorb any output */
         DclQioSysOutput (tkptr);
         return;
      }
   }

   if (tkptr->AddNewline)
   {
      /*************************/
      /* HTTP carriage-control */
      /*************************/

      if (tkptr->SysOutputIOsb.Count)
      {
         if (tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count-1] != '\n')
            tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count++] = '\n';
      }
      else
      {
         /* must be a blank line (empty record), add a newline */
         tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count++] = '\n';
      }
      if (Debug) tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count] = '\0';

      if (tkptr->BufferRecords)
      {
         /*********************/
         /* buffer the record */
         /*********************/

         /*
            The AST routine address supplied ensures that at the completion
            of the network write to the client DclQioSysOutput() function
            is called to queue another read from the subprocess.
         */

         if (VMSok (status =
            NetBufferOutput (tkptr->rqptr, &DclSysOutputToClientAst,
                             tkptr->SysOutputPtr,
                             tkptr->SysOutputIOsb.Count)))
         {
            tkptr->QueuedSysOutput++;
            return;
         }

         DclConcludeTask (tkptr, true);
         return;
      }
   }

   /********************/
   /* write the record */
   /********************/

   if (Debug)
   {
      tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count] = '\0';
      fprintf (stdout, "|%s|\n", tkptr->SysOutputPtr);
   }

   /*
      The AST routine address supplied ensures that at the completion
      of the network write to the client DclQioSysOutput() function
      is called to queue another read from the subprocess.
   */
   if (VMSok (status =
      NetWrite (tkptr->rqptr, &DclSysOutputToClientAst,
                tkptr->SysOutputPtr, tkptr->SysOutputIOsb.Count)))
   {
      tkptr->QueuedSysOutput++;
      return;
   }

   ErrorVmsStatus (tkptr->rqptr, status, __FILE__, __LINE__);
   DclConcludeTask (tkptr, true);
}

/*****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************  Due to is being an AST from a general network write function.

A queued asynchronous write of subprocess SYS$OUTPUT (mailbox) to the client
over the network has completed.
*/ 

DclSysOutputToClientAst (struct RequestStruct *rqptr)

{
   register struct DclTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   /* get a pointer to the DCL task from the request structure */
   tkptr = rqptr->DclTaskPtr;

   if (Debug)
   {
      fprintf (stdout,
         "DclSysOutputToClientAst() %d %d %d %d %d %d %%X%08.08X %%X%08.08X\n",
         tkptr, tkptr->rqptr, tkptr->QueuedSysCommand,
         tkptr->QueuedSysOutput, tkptr->QueuedCgiPlusIn,
         tkptr->QueuedHttpInput, tkptr->SysOutputIOsb.Status,
         rqptr->NetWriteIOsb.Status);
   }

   if (tkptr->QueuedSysOutput) tkptr->QueuedSysOutput--;

   /* if I/O cancelled then just return */
   if (rqptr->NetWriteIOsb.Status == SS$_ABORT ||
       rqptr->NetWriteIOsb.Status == SS$_CANCEL)
      return;

   /* abort task if NETWORK ERROR when writing TO CLIENT */
   if (VMSnok (rqptr->NetWriteIOsb.Status))
   {
      DclConcludeTask (tkptr, true);
      return;
   }

   /* queue the next read of the subprocess' SYS$OUTPUT */
   DclQioSysOutput (tkptr);
}

/*****************************************************************************/
/*
Queue up a write of data to the subprocess "SYS$COMMAND" mailbox.  This is the 
subprocesses' "SYS$COMMAND", supplying the DCL commands to execute, CGI
information, etc.
*/ 

DclQioSysCommand 
(
struct DclTaskStruct *tkptr,
char *DataPtr,
int DataLength
)
{
   register int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclQioSysCommand() %d %d |%s|\n",
               tkptr, tkptr->rqptr, DataPtr);

   if (DataPtr == NULL)
   {
      /* NULL pointer means write an end-of-file to the channel */
      if (VMSok (status =
          sys$qio (0, tkptr->SysCommandChannel,
                   IO$_WRITEOF, &tkptr->SysCommandIOsb,
                   &DclSysCommandAst, tkptr,
                   0, 0, 0, 0, 0, 0)))
         tkptr->QueuedSysCommand++;
      return (status);
   }
   else
   {
      if (VMSok (status =
          sys$qio (0, tkptr->SysCommandChannel,
                   IO$_WRITELBLK, &tkptr->SysCommandIOsb,
                   &DclSysCommandAst, tkptr,
                   DataPtr, DataLength,
                   0, 0, 0, 0)))
         tkptr->QueuedSysCommand++;
      return (status);
   }
}

/*****************************************************************************/
/*
A queued write to the subprocess "SYS$COMMAND" mailbox has completed.  This is 
the subprocesses' "SYS$COMMAND", supplying the DCL commands to execute.
*/ 

DclSysCommandAst (struct DclTaskStruct *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclSysCommandAst() %d %d %08.08X %d %d %d %d %%X%08.08X\n",
         tkptr, tkptr->rqptr, tkptr->SubprocessPid,
         tkptr->QueuedSysCommand, tkptr->QueuedSysOutput,
         tkptr->QueuedCgiPlusIn, tkptr->QueuedHttpInput,
         tkptr->SysCommandIOsb.Status);
   }

   if (tkptr->QueuedSysCommand) tkptr->QueuedSysCommand--;

   /* if (effectively) no outstanding I/O then conclude the DCL task */
   if (tkptr->QueuedSysCommand <= tkptr->QueuedSysCommandAllowed &&
       !tkptr->QueuedSysOutput &&
       !tkptr->QueuedCgiPlusIn &&
       !tkptr->QueuedHttpInput)
   {
      DclConcludeTask (tkptr, false);
      return;
   }

   /* if I/O cancelled then just return */
   if (tkptr->SysCommandIOsb.Status == SS$_ABORT ||
       tkptr->SysCommandIOsb.Status == SS$_CANCEL)
      return;

   /* don't know why, first AST is delivered %x00000000, make %X00000001 */
   if (!tkptr->SysCommandIOsb.Status) tkptr->SysCommandIOsb.Status = 1;

   /* abort if an error writing SYS$COMMAND stream to subprocess */
   if (VMSnok (tkptr->SysCommandIOsb.Status))
   {
      ErrorVmsStatus (tkptr->rqptr, tkptr->SysCommandIOsb.Status,
                      __FILE__, __LINE__);
      DclConcludeTask (tkptr, true);
      return;
   }
}

/*****************************************************************************/
/*
Queue up a write of data to the subprocess "CGIPLUSIN" mailbox.  This is a
CGIplus subprocesses' CGI variable stream mailbox.             
*/ 

DclQioCgiPlusIn 
(
struct DclTaskStruct *tkptr,
char *DataPtr,
int DataLength
)
{
   register int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclQioCgiPlusIn() %d %d |%s|\n",
               tkptr, tkptr->rqptr, DataPtr);

   if (DataPtr == NULL)
   {
      /* NULL pointer means write an end-of-file to the channel */
      if (VMSok (status =
          sys$qio (0, tkptr->CgiPlusInChannel,
                   IO$_WRITEOF, &tkptr->CgiPlusInIOsb,
                   &DclCgiPlusInAst, tkptr,
                   0, 0, 0, 0, 0, 0)))
         tkptr->QueuedCgiPlusIn++;
      return (status);
   }
   else
   {
      if (VMSok (status =
          sys$qio (0, tkptr->CgiPlusInChannel,
                   IO$_WRITELBLK, &tkptr->CgiPlusInIOsb,
                   &DclCgiPlusInAst, tkptr,
                   DataPtr, DataLength,
                   0, 0, 0, 0)))
         tkptr->QueuedCgiPlusIn++;
      return (status);
   }
}

/*****************************************************************************/
/*
A queued write to the subprocess "CGIPLUSIN" mailbox has completed.  This is a
CGIplus subprocesses' CGI variable stream mailbox.             
*/ 

DclCgiPlusInAst (struct DclTaskStruct *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclCgiPlusInAst() %d %d %08.08X %d %d %d %%X%08.08X\n",
         tkptr, tkptr->rqptr, tkptr->SubprocessPid, tkptr->QueuedCgiPlusIn,
         tkptr->QueuedSysOutput, tkptr->QueuedHttpInput,
         tkptr->CgiPlusInIOsb.Status);
   }

   if (tkptr->QueuedCgiPlusIn) tkptr->QueuedCgiPlusIn--;

   /* if (effectively) no outstanding I/O then conclude the DCL task */
   if (tkptr->QueuedSysCommand <= tkptr->QueuedSysCommandAllowed &&
       !tkptr->QueuedSysOutput &&
       !tkptr->QueuedCgiPlusIn &&
       !tkptr->QueuedHttpInput)
   {
      DclConcludeTask (tkptr, false);
      return;
   }

   /* if I/O cancelled then just return */
   if (tkptr->CgiPlusInIOsb.Status == SS$_ABORT ||
       tkptr->CgiPlusInIOsb.Status == SS$_CANCEL)
      return;

   /* don't know why, first AST is delivered %x00000000, make %X00000001 */
   if (!tkptr->CgiPlusInIOsb.Status) tkptr->CgiPlusInIOsb.Status = 1;

   /* abort if an error writing CGIPLUSIN stream to subprocess */
   if (VMSnok (tkptr->CgiPlusInIOsb.Status))
   {
      ErrorVmsStatus (tkptr->rqptr, tkptr->CgiPlusInIOsb.Status,
                      __FILE__, __LINE__);
      DclConcludeTask (tkptr, true);
      return;
   }
}

/*****************************************************************************/
/*
A queued write to the subprocess "HTTP$INPUT" mailbox has completed.
Provide more (possibly first) of the request body, or EOF.
*/ 

DclHttpInputAst (struct DclTaskStruct *tkptr)

{
   int  status,
        Length;
   char  *ContentPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclHttpInputAst() %d %d %08.08X %d %d %d %d %d %%X%08.08X\n",
         tkptr, tkptr->rqptr, tkptr->SubprocessPid,
         tkptr->QueuedSysCommand, tkptr->QueuedSysOutput,
         tkptr->QueuedCgiPlusIn, tkptr->QueuedHttpInput,
         tkptr->HttpInputIOsb.Count, tkptr->HttpInputIOsb.Status);
   }

   if (tkptr->QueuedHttpInput) tkptr->QueuedHttpInput--;

   /* if (effectively) no outstanding I/O then conclude the DCL task */
   if (tkptr->QueuedSysCommand <= tkptr->QueuedSysCommandAllowed &&
       !tkptr->QueuedSysOutput &&
       !tkptr->QueuedCgiPlusIn &&
       !tkptr->QueuedHttpInput)
   {
      DclConcludeTask (tkptr, false);
      return;
   }

   /* if I/O cancelled then just return */
   if (tkptr->HttpInputIOsb.Status == SS$_ABORT ||
       tkptr->HttpInputIOsb.Status == SS$_CANCEL)
      return;

   /* abort if an error writing HTTP stream to subprocess */
   if (VMSnok (tkptr->HttpInputIOsb.Status))
   {
      ErrorVmsStatus (tkptr->rqptr, tkptr->HttpInputIOsb.Status,
                      __FILE__, __LINE__);
      DclConcludeTask (tkptr, true);
      return;
   }

   if (Debug)
      fprintf (stdout, "%d %d\n", tkptr->ContentLength, tkptr->ContentPtr);

   if (tkptr->ContentLength)
   {
      if (tkptr->ContentLength > NetReadBufferSize)
         Length = NetReadBufferSize;
      else
         Length = tkptr->ContentLength;
      ContentPtr = tkptr->ContentPtr;
      tkptr->ContentPtr += Length;
      tkptr->ContentLength -= Length;

      if (VMSok (status =
          sys$qio (0, tkptr->HttpInputChannel,
                   IO$_WRITELBLK, &tkptr->HttpInputIOsb,
                   &DclHttpInputAst, tkptr,
                   ContentPtr, Length,
                   0, 0, 0, 0)))
      {
          tkptr->QueuedHttpInput++;
          return;
      }
   }
   else
   {
      if (VMSok (status =
          sys$qio (0, tkptr->HttpInputChannel,
                   IO$_WRITEOF, &tkptr->HttpInputIOsb,
                   &DclHttpInputAst, tkptr,
                   0, 0, 0, 0, 0, 0)))
      {
          tkptr->QueuedHttpInput++;
          return;
      }
   }

   ErrorVmsStatus (tkptr->rqptr, status, __FILE__, __LINE__);
   DclConcludeTask (tkptr, true);
}

/*****************************************************************************/
/*
For CGI-compliance the first line of the header should be a "Content-Type:", 
"Location:" or "Status:".  If a "Content-Type:" prepend a "200" header status 
line.  If a "Status:" use this to create a unique header status line.  If a 
"Location:" redirection header line call a function to establish the level of 
redirection, and do any associated processing.  For non-parse-header style CGI 
scripts check if the first line is a full HTTP status line.  If it is then do 
nothing, the script will supply the raw HTTP stream.  If none of the above 
then consider the script will supply nothing but plain text, none of the HTTP 
header, etc., and prepend a "200" status line and a "Content-Type: text/plain" 
line to create a complete HTTP header.  This is an extension to CGI behaviour, 
but makes simple, plain-text DCL scripts very easy!  Return true any action
required has been taken by this function, false if further processing is
required by the calling function.
*/ 

DclSysOutputHeader (struct DclTaskStruct *tkptr)

{
   register char  *cptr, *sptr;
   register struct RequestStruct  *rqptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      tkptr->SysOutputPtr [tkptr->SysOutputIOsb.Count] = '\0';
      fprintf (stdout, "DclSysOutputHeader()\n|%s|\n", tkptr->SysOutputPtr);
   }

   /* get a pointer to the request structure */
   rqptr = tkptr->rqptr;

   tkptr->CheckSubprocessResponseHeader = false;

   /************************************/
   /* first line(s) output from script */
   /************************************/

   tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count] = '\0';
   cptr = tkptr->SysOutputPtr;

   if (toupper(*cptr) == 'H' && strsame (cptr, "HTTP/", 5))
   {
      /*************************************************/
      /* script is supplying the full HTTP data stream */
      /*************************************************/

      if (Debug) fprintf (stdout, "(HTTP data stream)\n");
      tkptr->AddNewline = false;
      tkptr->BufferRecords = true;

      /* get the response status code for logging purposes */
      while (*cptr && !isspace(*cptr)) cptr++;
      while (*cptr && !isdigit(*cptr)) cptr++;
      if (isdigit(*cptr)) rqptr->ResponseStatusCode = atoi(cptr);
      return (false);
   }
   else
   if (toupper(*cptr) == 'L' && strsame (cptr, "Location:", 9))
   {
      /*******************/
      /* redirection URL */
      /*******************/

      cptr += 9;
      /* locate the first character of the "Location:" URL */
      while (isspace(*cptr)) cptr++;
      sptr = cptr;
      while (*cptr && !isspace(*cptr) && *cptr != '\r' && *cptr != '\n') cptr++;
      if ((rqptr->LocationPtr = HeapAlloc (rqptr, cptr-sptr+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         DclConcludeTask (tkptr, true);
         return (true);
      }
      memcpy (rqptr->LocationPtr, sptr, cptr-sptr);
      rqptr->LocationPtr[cptr-sptr] = '\0';
      if (Debug) fprintf (stdout, "LocationPtr |%s|\n", rqptr->LocationPtr);
      return (false);
   }
   else
   if (toupper(*cptr) == 'C' && strsame (cptr, "Content-Type:", 13))
   {

      /**************************/
      /* content-type specified */
      /**************************/

      /* make sure each record received has HTTP carriage control */
      if (Debug) fprintf (stdout, "(RMS record stream)\n");
      tkptr->AddNewline = true;
      rqptr->ResponseStatusCode = 200;

      if ((rqptr->ResponseHeaderPtr =
           HttpHeader (rqptr, rqptr->ResponseStatusCode,
                       NULL, -1, NULL, NULL)) == NULL)
      {
         DclConcludeTask (tkptr, true);
         return (true);
      }
      /* remove generated empty header line to allow the script to supply it */
      if (rqptr->ResponseHeaderLength > 2)
      {
         rqptr->ResponseHeaderPtr[rqptr->ResponseHeaderLength-2] = '\0';
         rqptr->ResponseHeaderLength -= 2;
      }
      return (false);
   }
   else
   if (toupper(*cptr) == 'S' && strsame (cptr, "Status:", 7))
   {
      /********************/
      /* status specified */
      /********************/

      /* make sure each record received has HTTP carriage control */
      if (Debug) fprintf (stdout, "(RMS record stream)\n");
      tkptr->AddNewline = true;

      /* create HTTP header status line using supplied status */
      cptr += 7;
      while (*cptr++ && !isdigit(*cptr) && *cptr != '\n') cptr++;
      /* get the response status code for header and logging purposes */
      if (isdigit(*cptr)) rqptr->ResponseStatusCode = atoi(cptr);

      if ((rqptr->ResponseHeaderPtr =
           HttpHeader (rqptr, rqptr->ResponseStatusCode,
                       "text/plain", -1, NULL, NULL)) == NULL)
      {
         DclConcludeTask (tkptr, true);
         return (true);
      }
      /* remove generated empty header line, allow the script to supply it */
      if (rqptr->ResponseHeaderLength > 2)
      {
         rqptr->ResponseHeaderPtr[rqptr->ResponseHeaderLength-2] = '\0';
         rqptr->ResponseHeaderLength -= 2;
      }
      return (false);

   }
   else
   {
      /******************/
      /* non-CGI output */
      /******************/

      /* make sure each record received has HTTP carriage control */
      if (Debug) fprintf (stdout, "(RMS record stream)\n");
      tkptr->AddNewline = true;
      rqptr->ResponseStatusCode = 200;

      if ((rqptr->ResponseHeaderPtr =
           HttpHeader (rqptr, rqptr->ResponseStatusCode,
                       "text/plain", -1, NULL, NULL)) == NULL)
      {
         DclConcludeTask (tkptr, true);
         return (true);
      }
      return (false);
   }
}

/*****************************************************************************/
/*
Check if the script procedure or executable file exists.
*/ 

int DclFindScript
(
char *FileName,
boolean *IsDclProcPtr
)
{
   int  status;
   char  ExpandedFileName [256];
   struct FAB  SearchFab;
   struct NAM  SearchNam;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclFindScript() |%s|\n", FileName);

   *IsDclProcPtr = false;

   SearchFab = cc$rms_fab;
   SearchFab.fab$l_fna = FileName;
   SearchFab.fab$b_fns = strlen(FileName);
   SearchFab.fab$l_fop = FAB$M_NAM;
   SearchFab.fab$l_nam = &SearchNam;
   SearchNam = cc$rms_nam;
   SearchNam.nam$l_esa = ExpandedFileName;
   SearchNam.nam$b_ess = sizeof(ExpandedFileName)-1;

   /* first, default to a DCL procedure */
   SearchFab.fab$l_dna = ".COM;";
   SearchFab.fab$b_dns = 5;

   if (VMSok (status = sys$parse (&SearchFab, 0, 0)))
   {
      if ((status = sys$search (&SearchFab, 0, 0)) == RMS$_FNF)
      {
         /* next, default to an executable */
         SearchFab.fab$l_dna = ".EXE;";
         SearchFab.fab$b_dns = 5;

         if (VMSok (status = sys$parse (&SearchFab, 0, 0)))
            status = sys$search (&SearchFab, 0, 0);
      }
      else
         *IsDclProcPtr = true;
   }

   /* release parse and search internal data structures */
   SearchFab.fab$l_fna = "a:[b]c.d;";
   SearchFab.fab$b_fns = 9;
   SearchFab.fab$b_dns = 0;
   SearchNam.nam$b_nop = NAM$M_SYNCHK;
   sys$parse (&SearchFab, 0, 0);

   return (status);
}

/*****************************************************************************/
/*
ANY CHANGE TO CGI VARIABLES IN THIS FUNCTION MUST ALSO BE MADE TO
-----------------------------------------------------------------
DclCgiPlusScriptCgiPlusIn()

Send DCL commands to a standard CGI script or DCL command subprocess'
SYS$COMMAND. This sets up the DCL environment (defines logical names, assigns
symbols) then executes the specified DCL command, procedure or image.

Some CGI variables have the potential to include double-quotation marks
characters which can cause complications on the DCL command line.  Ensure
these have quotes escaped.

Some CGI variables have a potential to be larger than DCL symbols can be 
created on the DCL command line (which is limited to 255 characters).  For 
example, PATH_INFO, QUERY_STRING, etc.  These variables are handled by a 
function that allows variables (DCL symbols) to have up to 1000 characters via 
a symbol assignment kludge.
*/ 

DclCgiScriptSysCommand
(
struct DclTaskStruct *tkptr,
char *DclCommand,
char *CgiFileName,
boolean IsDclProc
)
{
   static char  DclHttpdVerify [] = "!'F$VERIFY(F$TRNLNM(\"HTTPD$VERIFY\"))";
   static char  DeleteIsDelete [] = "DELETE=\"DELETE\"";
   static char  DeleteAllLocalSymbol[] = "DELETE/SYMBOL/LOCAL/ALL";
   static char  DeleteAllGlobalSymbol[] = "DELETE/SYMBOL/GLOBAL/ALL";
   static char  NoVerify [] = "!'F$VERIFY(0)";
   static char  SetNoOn[] = "SET NOON";
   static char  StopId [] = "STOP/ID=0";
   static char  WriteIsWrite [] = "WRITE=\"WRITE\"";

   static $DESCRIPTOR (WriteDclEofFaoDsc, "WRITE SYS$OUTPUT \"!AZ\"");

   static $DESCRIPTOR (DefineHttpInputFaoDsc,
                       "DEFINE/NOLOG/EXEC HTTP$INPUT !AZ");
   static $DESCRIPTOR (DefineSysInputFaoDsc,
                       "DEFINE/NOLOG/EXEC SYS$INPUT !AZ");

   static $DESCRIPTOR (AuthGroupFaoDsc, "!AZAUTH_GROUP==\"!AZ\"");
   static $DESCRIPTOR (AuthRealmFaoDsc, "!AZAUTH_REALM==\"!AZ\"");
   static $DESCRIPTOR (AuthTypeFaoDsc, "!AZAUTH_TYPE==\"!AZ\"");

   static $DESCRIPTOR (ContentLengthFaoDsc, "!AZCONTENT_LENGTH==\"!UL\"");
   static $DESCRIPTOR (ContentTypeFaoDsc, "!AZCONTENT_TYPE==\"!AZ\"");

   static $DESCRIPTOR (GatewayFaoDsc, "!AZGATEWAY_INTERFACE==\"CGI/1.1\"");

   static $DESCRIPTOR (HttpAcceptFaoDsc, "!AZHTTP_ACCEPT==\"!AZ\"");

   static $DESCRIPTOR (IfModifiedSinceFaoDsc,
                       "!AZHTTP_IF_MODIFIED_SINCE==\"!AZ\"");
   /* HTTP_REFERER is created using DclSysCommandStringSymbol() */
   static $DESCRIPTOR (UserAgentFaoDsc, "!AZHTTP_USER_AGENT==\"!AZ\"");
   /* HTTP_AUTHORIZATION is created using DclSysCommandStringSymbol() */

   /* PATH_INFO is created using DclSysCommandStringSymbol() */
   static $DESCRIPTOR (PathTranslatedFaoDsc, "!AZPATH_TRANSLATED==\"!AZ\"");

   /* QUERY_STRING is created using DclSysCommandStringSymbol() */

   static $DESCRIPTOR (RemoteAddressFaoDsc, "!AZREMOTE_ADDR==\"!AZ\"");
   static $DESCRIPTOR (RemoteHostNameFaoDsc, "!AZREMOTE_HOST==\"!AZ\"");
   static $DESCRIPTOR (RemoteUserFaoDsc, "!AZREMOTE_USER==\"!AZ\"");
   static $DESCRIPTOR (RequestMethodFaoDsc, "!AZREQUEST_METHOD==\"!AZ\"");
   static $DESCRIPTOR (RequestTimeGmtFaoDsc, "!AZREQUEST_TIME_GMT==\"!AZ\"");
   static $DESCRIPTOR (RequestTimeLocalFaoDsc, "!AZREQUEST_TIME_LOCAL==\"!AZ\"");

   static $DESCRIPTOR (ScriptNameFaoDsc, "!AZSCRIPT_NAME==\"!AZ\"");
   static $DESCRIPTOR (ServerNameFaoDsc, "!AZSERVER_NAME==\"!AZ\"");
   static $DESCRIPTOR (ServerSoftwareFaoDsc, "!AZSERVER_SOFTWARE==\"!AZ\"");
   static $DESCRIPTOR (ServerProtocolFaoDsc, "!AZSERVER_PROTOCOL==\"!AZ\"");
   static $DESCRIPTOR (ServerPortFaoDsc, "!AZSERVER_PORT==\"!UL\"");

   static $DESCRIPTOR (KeyWordFieldNameFaoDsc, "KEY_!UL");
   static $DESCRIPTOR (KeyWordCountFaoDsc, "!AZKEY_COUNT==\"!UL\"");

   register struct RequestStruct *rqptr;
   register char  c;
   register char  *cptr, *sptr, *zptr;

   int  status,
        Count;
   unsigned short  Length;
   char  FormFieldName [256],
         KeyWordFieldName [16],
         LocalDateTime [48],
         Scratch [1024],
         DclLine [256];
   $DESCRIPTOR (KeyWordFieldNameDsc, KeyWordFieldName);
   $DESCRIPTOR (DclLineDsc, DclLine);
   $DESCRIPTOR (DclEofDsc, tkptr->DclEof);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclCgiScriptSysCommand() |%s|%s|%d|\n",
               DclCommand, CgiFileName, IsDclProc);

   /* get the pointer to the request structure */
   rqptr = tkptr->rqptr;

   HttpLocalTimeString (LocalDateTime, &rqptr->BinaryTime);

   if (DclUseZombies)
   {
      /* create a pseudo-unique DCL EOF string */
      sys$fao (&DclEofFaoDsc, &Length, &DclEofDsc,
               tkptr->SubprocessPid, ~rqptr->BinaryTime[1],
               ~rqptr->BinaryTime[0]);
      tkptr->DclEof[tkptr->DclEofLength = Length] = '\0';
   }

   if (VMSnok (status = DclQioSysCommand (tkptr, SetNoOn, sizeof(SetNoOn)-1)))
      return (status);

   if (VMSnok (status =
       DclQioSysCommand (tkptr, DeleteIsDelete, sizeof(DeleteIsDelete)-1)))
      return (status);

   if (VMSnok (status =
       DclQioSysCommand (tkptr, DeleteAllLocalSymbol,
                         sizeof(DeleteAllLocalSymbol )-1)))
      return (status);

   if (VMSnok (status =
       DclQioSysCommand (tkptr, DeleteAllGlobalSymbol,
                         sizeof(DeleteAllGlobalSymbol )-1)))
      return (status);

   if (VMSnok (status =
       DclQioSysCommand (tkptr, DclHttpdVerify, sizeof(DclHttpdVerify)-1)))
      return (status);

   if (tkptr->IsCgiScript)
   {
      sys$fao (&DefineSysInputFaoDsc, &Length, &DclLineDsc,
               tkptr->HttpInputDevName);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
         return (status);

      /* for backward compatibility */
      sys$fao (&DefineHttpInputFaoDsc, &Length, &DclLineDsc,
               tkptr->HttpInputDevName);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
         return (status);
   }

   /***************************/ 
   /* CGI compliant variables */
   /***************************/ 

   sys$fao (&AuthGroupFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->AuthGroupPtr);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&AuthRealmFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->AuthRealmPtr);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&AuthTypeFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->AuthType);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   if (rqptr->HttpContentTypePtr != NULL)
      sys$fao (&ContentTypeFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpContentTypePtr);
   else
   if (rqptr->ContentType.ContentTypePtr != NULL)
      sys$fao (&ContentTypeFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->ContentType.ContentTypePtr);
   else
      sys$fao (&ContentTypeFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, "");
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ContentLengthFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->ContentLength);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&GatewayFaoDsc, &Length, &DclLineDsc, DclCgiVariablePrefix);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   if (cptr = rqptr->HttpAcceptPtr != NULL)
   {
      sys$fao (&HttpAcceptFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpAcceptPtr);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
         return (status);
   }

   if (rqptr->HttpAuthorizationPtr != NULL)
   {
      cptr = rqptr->HttpAuthorizationPtr;
      zptr = (sptr = Scratch) + sizeof(Scratch);
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '\"' && sptr < zptr) *sptr++ = '\"';
         if (sptr < zptr) *sptr++ = *cptr++;
      }
      if (sptr >= zptr)
      {
         ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
         return (STS$K_ERROR);
      }
      *sptr = '\0';
      if (VMSnok (status =
          DclSysCommandStringSymbol (tkptr, "HTTP_AUTHORIZATION", Scratch)))
      {
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         return (status);
      }
   }

   if (rqptr->HttpIfModifiedSincePtr != NULL)
   {
      sys$fao (&IfModifiedSinceFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpIfModifiedSincePtr);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
         return (status);
   }

   if (rqptr->HttpRefererPtr != NULL)
   {
      cptr = rqptr->HttpRefererPtr;
      zptr = (sptr = Scratch) + sizeof(Scratch);
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '\"' && sptr < zptr) *sptr++ = '\"';
         if (sptr < zptr) *sptr++ = *cptr++;
      }
      if (sptr >= zptr)
      {
         ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
         return (STS$K_ERROR);
      }
      *sptr = '\0';
      if (VMSnok (status =
          DclSysCommandStringSymbol (tkptr, "HTTP_REFERER", Scratch)))
      {
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         return (status);
      }
   }

   if (rqptr->HttpUserAgentPtr != NULL)
   {
      cptr = rqptr->HttpUserAgentPtr;
      zptr = (sptr = Scratch) + sizeof(Scratch);
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '\"' && sptr < zptr) *sptr++ = '\"';
         if (sptr < zptr) *sptr++ = *cptr++;
      }
      if (sptr >= zptr)
      {
         ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
         return (STS$K_ERROR);
      }
      *sptr = '\0';
      sys$fao (&UserAgentFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, Scratch);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
         return (status);
   }

   cptr = rqptr->PathInfoPtr;
   zptr = (sptr = Scratch) + sizeof(Scratch);
   while (*cptr && sptr < zptr)
   {
      if (*cptr == '\"' && sptr < zptr) *sptr++ = '\"';
      if (sptr < zptr) *sptr++ = *cptr++;
   }
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      return (STS$K_ERROR);
   }
   *sptr = '\0';
   if (VMSnok (status =
       DclSysCommandStringSymbol (tkptr, "PATH_INFO", Scratch)))
   {
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      return (status);
   }

   sys$fao (&PathTranslatedFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->RequestFileName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   cptr = rqptr->QueryStringPtr;
   zptr = (sptr = Scratch) + sizeof(Scratch);
   while (*cptr && sptr < zptr)
   {
      if (*cptr == '\"' && sptr < zptr) *sptr++ = '\"';
      if (sptr < zptr) *sptr++ = *cptr++;
   }
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      return (STS$K_ERROR);
   }
   *sptr = '\0';
   if (VMSnok (status =
       DclSysCommandStringSymbol (tkptr, "QUERY_STRING", Scratch)))
   {
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      return (status);
   }

   sys$fao (&RemoteHostNameFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->ClientHostName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RemoteAddressFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->ClientInternetAddress);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RemoteUserFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->RemoteUser);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RequestMethodFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->HttpMethodName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RequestTimeGmtFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->GmDateTime);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RequestTimeLocalFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, LocalDateTime);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ScriptNameFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->ScriptName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ServerNameFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, ServerHostName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ServerProtocolFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, HttpProtocol);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ServerSoftwareFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, SoftwareID);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ServerPortFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, ServerPort);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   /***************/
   /* any pragma? */
   /***************/

   if (rqptr->HttpPragmaPtr != NULL)
   {
      zptr = (sptr = Scratch) + sizeof(Scratch);
      cptr = rqptr->HttpPragmaPtr;
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '\"')
         {
            /* escape quotes for DCL (not sure if they'll ever be sent :^) */
            *sptr++ = '\"';
         }
         if (sptr < zptr) *sptr++ = *cptr++;
      }
      if (sptr >= zptr)
      {
         ErrorGeneral (rqptr, "Pragma too big!", __FILE__, __LINE__);
         return (STS$K_ERROR);
      }
      *sptr = '\0';

      if (VMSnok (status =
          DclSysCommandStringSymbol (tkptr, "HTTP_PRAGMA", Scratch)))
      {
          ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
          return (status);
      }
   }

   /****************/
   /* any cookies? */
   /****************/

   if (rqptr->HttpCookiePtr != NULL)
   {
      zptr = (sptr = Scratch) + sizeof(Scratch);
      cptr = rqptr->HttpCookiePtr;
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '\"')
         {
            /* escape quotes for DCL (not sure if they'll ever be sent :^) */
            *sptr++ = '\"';
         }
         else
         if (*cptr == '\'')   
         {
            ErrorGeneral (rqptr, "Disallowed cookie character!",
                          __FILE__, __LINE__);
            return (STS$K_ERROR);
         }
         if (sptr < zptr) *sptr++ = *cptr++;
      }
      if (sptr >= zptr)
      {
         ErrorGeneral (rqptr, "Cookie too big!", __FILE__, __LINE__);
         return (STS$K_ERROR);
      }
      *sptr = '\0';

      if (VMSnok (status =
          DclSysCommandStringSymbol (tkptr, "HTTP_COOKIE", Scratch)))
      {
          ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
          return (status);
      }
   }

   /***************************/
   /* query string components */
   /***************************/

   if (rqptr->QueryStringPtr[0])
   {
      Count = 0;

      cptr = rqptr->QueryStringPtr;
      while (*cptr && *cptr != '=') cptr++;
      /* if an equal symbol was found then its a form not a keyword search */
      if (*cptr)
      {
         /***************/
         /* form fields */
         /***************/

         memcpy (FormFieldName, "FORM_", 5);
         cptr = rqptr->QueryStringPtr;
         while (*cptr)
         {
            sptr = FormFieldName + 5;
            zptr = FormFieldName + sizeof(FormFieldName);
            while (*cptr && *cptr != '=' && *cptr != '&' && sptr < zptr)
            {
               if (isalnum(*cptr))
                  *sptr++ = toupper(*cptr++);
               else
               {
                  *sptr++ = '_';
                  cptr++;
               }
            }
            if (sptr >= zptr)
            {
               ErrorVmsStatus (rqptr, SS$_BUFFEROVF, __FILE__, __LINE__);
               return (status);
            }
            *sptr = '\0';

            if (!FormFieldName[5] || *cptr != '=')
            {
               /* error; back-to-back '&' and/or '=', or no equate */
               ErrorGeneral (rqptr, "URL-encoded form error.",
                            __FILE__, __LINE__);
               return (STS$K_ERROR);
            }

            /* must have encountered an '=' */
            if (*cptr) cptr++;
            zptr = (sptr = Scratch) + sizeof(Scratch);
            while (*cptr && *cptr != '&' && sptr < zptr)
            {
               if (*cptr == '+')
               {
                  *sptr++ = ' ';
                  cptr++;
               }
               else
               if (*cptr == '%')
               {
                  /* an escaped character ("%xx" where xx is a hex number) */
                  cptr++;
                  c = 0;
                  if (*cptr >= '0' && *cptr <= '9')
                     { c = (*cptr - (int)'0') << 4; cptr++; }
                  else
                  if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
                     { c = (toupper(*cptr) - (int)'A' + 10) << 4; cptr++; }
                  else
                  {
                     ErrorGeneral (rqptr, "URL hexadecimal error.",
                                   __FILE__, __LINE__);
                     return (STS$K_ERROR);
                  }
                  if (*cptr >= '0' && *cptr <= '9')
                     { c += (*cptr - (int)'0'); cptr++; }
                  else
                  if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
                     { c += (toupper(*cptr) - (int)'A' + 10); cptr++; }
                  else
                  {
                     ErrorGeneral (rqptr, "URL hexadecimal error.",
                                   __FILE__, __LINE__);
                     return (STS$K_ERROR);
                  }
                  if (c == '\"' && sptr < zptr)
                  {
                     /* ensure enough quotes are passed for a DCL procedure */
                     *sptr++ = '\"';
                  }
                  if (sptr < zptr) *sptr++ = c;
               }
               else
                  *sptr++ = *cptr++;
            }
            if (sptr >= zptr)
            {
               ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
               return (STS$K_ERROR);
            }
            *sptr = '\0';
            if (*cptr) cptr++;

            if (VMSnok (status =
                DclSysCommandStringSymbol (tkptr, FormFieldName, Scratch)))
            {
                ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
                return (status);
            }
         }

         sys$fao (&KeyWordCountFaoDsc, &Length, &DclLineDsc,
                  DclCgiVariablePrefix, 0);
         DclLine[Length] = '\0';
         if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
            return (status);
      }
      else
      {
         /******************/
         /* query keywords */
         /******************/

         memcpy (FormFieldName, "KEY_", 4);
         cptr = rqptr->QueryStringPtr;
         while (*cptr)
         {
            zptr = (sptr = Scratch) + sizeof(Scratch);
            while (*cptr && *cptr != '+' && sptr < zptr)
            {
               if (*cptr == '%')
               {
                  /* an escaped character ("%xx" where xx is a hex number) */
                  cptr++;
                  c = 0;
                  if (*cptr >= '0' && *cptr <= '9')
                     { c = (*cptr - (int)'0') << 4; cptr++; }
                  else
                  if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
                     { c = (toupper(*cptr) - (int)'A' + 10) << 4; cptr++; }
                  else
                  {
                     ErrorGeneral (rqptr, "URL hexadecimal error.",
                                   __FILE__, __LINE__);
                     return (STS$K_ERROR);
                  }
                  if (*cptr >= '0' && *cptr <= '9')
                     { c += (*cptr - (int)'0'); cptr++; }
                  else
                  if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
                     { c += (toupper(*cptr) - (int)'A' + 10); cptr++; }
                  else
                  {
                     ErrorGeneral (rqptr, "URL hexadecimal error.",
                                   __FILE__, __LINE__);
                     return (STS$K_ERROR);
                  }
                  if (c == '\"' && sptr < zptr)
                  {
                     /* ensure enough quotes are passed for a DCL procedure */
                     *sptr++ = '\"';
                  }
                  if (sptr < zptr) *sptr++ = c;
               }
               else
                  *sptr++ = *cptr++;
            }
            if (sptr >= zptr)
            {
               ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
               return (STS$K_ERROR);
            }
            *sptr = '\0';
            if (*cptr) cptr++;

            /* just use 'DclLine' to contain the name of the symbol here! */
            sys$fao (&KeyWordFieldNameFaoDsc, &Length, &KeyWordFieldNameDsc,
                     ++Count);
            KeyWordFieldName[Length] = '\0';
            if (VMSnok (status =
                DclSysCommandStringSymbol (tkptr, KeyWordFieldName,
                                         Scratch)))
            {
                ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
                return (status);
            }
         }

         sys$fao (&KeyWordCountFaoDsc, &Length, &DclLineDsc,
                  DclCgiVariablePrefix, Count);
         DclLine[Length] = '\0';
         if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
            return (status);
      }
   }
   else
   {
      sys$fao (&KeyWordCountFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, 0);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
         return (status);
   }

   /*******************************/
   /* DCL command/procedure/image */
   /*******************************/

   if (CgiFileName[0])
   {
      sptr = DclLine;
      if (IsDclProc)
         *sptr++ = '@';
      else
      {
         cptr = "RUN ";
         while (*cptr) *sptr++ = *cptr++;
      }
      cptr = CgiFileName;
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
      Length = sptr - (cptr = DclLine);
   }
   else
   {
      for (cptr = sptr = DclCommand; *sptr; sptr++);
      Length = sptr - cptr;
   }
   if (VMSnok (status = DclQioSysCommand (tkptr, cptr, Length)))
      return (status);

   /*********************/
   /* after script runs */
   /*********************/

   if (DclUseZombies)
   {
      if (VMSnok (status =
          DclQioSysCommand (tkptr, NoVerify, sizeof(NoVerify)-1)))
         return (status);

      if (VMSnok (status =
          DclQioSysCommand (tkptr, WriteIsWrite, sizeof(WriteIsWrite)-1)))
         return (status);

      sys$fao (&WriteDclEofFaoDsc, &Length, &DclLineDsc, tkptr->DclEof);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
         return (status);

      /* do not send and end-of-file! */
      return (status);
   }
   else
   {
      /* ensure subprocess terminates! */
      if (VMSnok (status = DclQioSysCommand (tkptr, StopId, sizeof(StopId))))
         return (status);

      /* send end-of-file */
      return (DclQioSysCommand (tkptr, NULL, 0));
   }
}

/*****************************************************************************/
/*
DCL symbol creation at the command line is limited by the CLI command line 
length (255 characters).  Symbol values however can be up to approximately 
1000 characters, probably enough for any CGI variable value.  If a CGI value 
is too large for for a single command-line assignment then build it up using 
multiple assignments, a symbol assignment kludge!
*/ 

DclSysCommandStringSymbol
(
struct DclTaskStruct *tkptr,
char *SymbolName,
char *SymbolValue
)
{
#  define MaxDclLineSize 255

   static char  DclLine [MaxDclLineSize+1];
   static $DESCRIPTOR (DclLineDsc, DclLine);
   static $DESCRIPTOR (AssignFaoDsc, "!AZ!AZ==X");
   static $DESCRIPTOR (SymbolFaoDsc, "!AZ!AZ==\"!AZ\"");

   register struct RequestStruct *rqptr;

   int  status;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclSysCommandStringSymbol() |%s|\n", SymbolName);

   /* get the pointer to the request structure */
   rqptr = tkptr->rqptr;

   status = sys$fao (&SymbolFaoDsc, &Length, &DclLineDsc,
                     DclCgiVariablePrefix, SymbolName, SymbolValue);
   if (VMSok (status) && status != SS$_BUFFEROVF)
   {
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
         return (status);
   }
   if (status != SS$_BUFFEROVF) return (status);

   /* buffer overflowed!  create it using multiple symbol assignments */
   {
      /* register storage */
      register char  *cptr, *sptr, *zptr;

      /* create, by assigning the temporary symbol with an empty string */
      if (VMSnok (status = DclQioSysCommand (tkptr, "X=\"\"", 4)))
         return (status);

      memcpy (DclLine, "X=X+\"", 5);
      /* minus two, one each for the closing quote and the terminating null */
      zptr = DclLine + sizeof(DclLine) - 2;
      cptr = SymbolValue;
      while (*cptr)
      {
         sptr = DclLine + 5;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         *sptr++ = '\"';
         *sptr = '\0';
         if (VMSnok (status =
             DclQioSysCommand (tkptr, DclLine, sptr-DclLine)))
            return (status);
      }

      /* assign the temporary symbol value to the CGI symbol */
      if (VMSnok (status =
          sys$fao (&AssignFaoDsc, &Length, &DclLineDsc,
                   DclCgiVariablePrefix, SymbolName)))
         return (status);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
         return (status);

      /* not really necessary, but let's be tidy */
      return (DclQioSysCommand (tkptr, "DELETE/SYMBOL X", 15));
   }
}

/*****************************************************************************/
/*
Send DCL commands to the CGIplus subprocess' SYS$COMMAND. This sets up the DCL
environment (defines logical names, assigns symbols) executes the procedure or
image.
*/ 

DclCgiPlusScriptSysCommand
(
struct DclTaskStruct *tkptr,
char *CgiFileName,
boolean IsDclProc
)
{
   static char  DclHttpdVerify [] = "!'F$VERIFY(F$TRNLNM(\"HTTPD$VERIFY\"))";
   static char  DeleteIsDelete [] = "DELETE=\"DELETE\"";
   static char  DeleteAllLocalSymbol[] = "DELETE/SYMBOL/LOCAL/ALL";
   static char  DeleteAllGlobalSymbol[] = "DELETE/SYMBOL/GLOBAL/ALL";
   static char  SetNoOn [] = "SET NOON";
   static char  StopId [] = "STOP/ID=0";

   static $DESCRIPTOR (DefineCgiPlusInFaoDsc,
                       "DEFINE/NOLOG/EXEC CGIPLUSIN !AZ");
   static $DESCRIPTOR (DefineCgiPlusEofFaoDsc,
                       "DEFINE/NOLOG/EXEC CGIPLUSEOF \"!AZ\"");
   static $DESCRIPTOR (DefineHttpInputFaoDsc,
                       "DEFINE/NOLOG/EXEC HTTP$INPUT !AZ");
   static $DESCRIPTOR (DefineSysInputFaoDsc,
                       "DEFINE/NOLOG/EXEC SYS$INPUT !AZ");

   register char  *cptr, *sptr;
   register struct RequestStruct *rqptr;

   int  status;
   unsigned short  Length;
   char  DclLine [256];
   $DESCRIPTOR (DclLineDsc, DclLine);
   $DESCRIPTOR (DclEofDsc, tkptr->DclEof);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclCgiPlusScriptSysCommand()\n");

   /* get the pointer to the request structure */
   rqptr = tkptr->rqptr;

   /* multiple requests to the same subprocess use the same EOF string! */
   if (!tkptr->DclEofLength)
   {
      /* only generate if a new one is necessary */
      sys$fao (&DclEofFaoDsc, &Length, &DclEofDsc,
               tkptr->SubprocessPid, ~rqptr->BinaryTime[1],
               ~rqptr->BinaryTime[0]);
      tkptr->DclEof[tkptr->DclEofLength = Length] = '\0';
   }

   if (VMSnok (status = DclQioSysCommand (tkptr, SetNoOn, sizeof(SetNoOn)-1)))
      return (status);

   if (VMSnok (status =
       DclQioSysCommand (tkptr, DeleteIsDelete, sizeof(DeleteIsDelete)-1)))
      return (status);

   if (VMSnok (status =
       DclQioSysCommand (tkptr, DeleteAllLocalSymbol,
                         sizeof(DeleteAllLocalSymbol )-1)))
      return (status);

   if (VMSnok (status =
       DclQioSysCommand (tkptr, DeleteAllGlobalSymbol,
                         sizeof(DeleteAllGlobalSymbol )-1)))
      return (status);

   if (VMSnok (status =
       DclQioSysCommand (tkptr, DclHttpdVerify, sizeof(DclHttpdVerify)-1)))
      return (status);

   sys$fao (&DefineSysInputFaoDsc, &Length, &DclLineDsc,
            tkptr->HttpInputDevName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   /* for backward compatibility */
   sys$fao (&DefineHttpInputFaoDsc, &Length, &DclLineDsc,
            tkptr->HttpInputDevName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&DefineCgiPlusInFaoDsc, &Length, &DclLineDsc,
            tkptr->CgiPlusInDevName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&DefineCgiPlusEofFaoDsc, &Length, &DclLineDsc, tkptr->DclEof);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioSysCommand (tkptr, DclLine, Length)))
      return (status);

   sptr = DclLine;
   if (IsDclProc)
      *sptr++ = '@';
   else
   {
      cptr = "RUN ";
      while (*cptr) *sptr++ = *cptr++;
   }
   cptr = CgiFileName;
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';
   Length = sptr - (cptr = DclLine);
   if (VMSnok (status = DclQioSysCommand (tkptr, cptr, Length)))
      return (status);

   /* ensure subprocess terminates! */
   if (VMSnok (status = DclQioSysCommand (tkptr, StopId, sizeof(StopId))))
      return (status);

   /* send end-of-file */
   return (DclQioSysCommand (tkptr, NULL, 0));
}

/*****************************************************************************/
/*
ANY CHANGE TO CGI VARIABLES IN THIS FUNCTION MUST ALSO BE MADE TO
-----------------------------------------------------------------
DclCgiScriptSysCommand()

Send CGI vraiables to the subprocess' CGIPLUSIN input stream.
*/ 

DclCgiPlusScriptCgiPlusIn (struct DclTaskStruct *tkptr)

{
   static $DESCRIPTOR (AuthorizationFaoDsc, "!AZHTTP_AUTHORIZATION=!AZ");
   static $DESCRIPTOR (AuthGroupFaoDsc, "!AZAUTH_GROUP=!AZ");
   static $DESCRIPTOR (AuthRealmFaoDsc, "!AZAUTH_REALM=!AZ");
   static $DESCRIPTOR (AuthTypeFaoDsc, "!AZAUTH_TYPE=!AZ");

   static $DESCRIPTOR (ContentLengthFaoDsc, "!AZCONTENT_LENGTH=!UL");
   static $DESCRIPTOR (ContentTypeFaoDsc, "!AZCONTENT_TYPE=!AZ");
   static $DESCRIPTOR (CookieFaoDsc, "!AZHTTP_COOKIE=!AZ");

   static $DESCRIPTOR (GatewayFaoDsc, "!AZGATEWAY_INTERFACE=CGI/1.1");

   static $DESCRIPTOR (HttpAcceptFaoDsc, "!AZHTTP_ACCEPT=!AZ");

   static $DESCRIPTOR (IfModifiedSinceFaoDsc,
                       "!AZHTTP_IF_MODIFIED_SINCE=!AZ");

   static $DESCRIPTOR (RefererFaoDsc, "!AZHTTP_REFERER=!AZ");

   static $DESCRIPTOR (PathInfoFaoDsc, "!AZPATH_INFO=!AZ");
   static $DESCRIPTOR (PathTranslatedFaoDsc, "!AZPATH_TRANSLATED=!AZ");
   static $DESCRIPTOR (PragmaFaoDsc, "!AZHTTP_PRAGMA=!AZ");

   static $DESCRIPTOR (QueryStringFaoDsc, "!AZQUERY_STRING=!AZ");

   static $DESCRIPTOR (RemoteAddressFaoDsc, "!AZREMOTE_ADDR=!AZ");
   static $DESCRIPTOR (RemoteHostNameFaoDsc, "!AZREMOTE_HOST=!AZ");
   static $DESCRIPTOR (RemoteUserFaoDsc, "!AZREMOTE_USER=!AZ");
   static $DESCRIPTOR (RequestMethodFaoDsc, "!AZREQUEST_METHOD=!AZ");
   static $DESCRIPTOR (RequestTimeGmtFaoDsc, "!AZREQUEST_TIME_GMT=!AZ");
   static $DESCRIPTOR (RequestTimeLocalFaoDsc, "!AZREQUEST_TIME_LOCAL=!AZ");

   static $DESCRIPTOR (ScriptNameFaoDsc, "!AZSCRIPT_NAME=!AZ");
   static $DESCRIPTOR (ServerNameFaoDsc, "!AZSERVER_NAME=!AZ");
   static $DESCRIPTOR (ServerSoftwareFaoDsc, "!AZSERVER_SOFTWARE=!AZ");
   static $DESCRIPTOR (ServerProtocolFaoDsc, "!AZSERVER_PROTOCOL=!AZ");
   static $DESCRIPTOR (ServerPortFaoDsc, "!AZSERVER_PORT=!UL");

   static $DESCRIPTOR (UserAgentFaoDsc, "!AZHTTP_USER_AGENT=!AZ");

   static $DESCRIPTOR (FormFieldFaoDsc, "!AZ!AZ=!AZ");
   static $DESCRIPTOR (KeyWordFaoDsc, "!AZKEY_!UL=!AZ");
   static $DESCRIPTOR (KeyWordCountFaoDsc, "!AZKEY_COUNT=!UL");

   register struct RequestStruct *rqptr;
   register char  c;
   register char  *cptr, *sptr, *zptr;

   int  status,
        Count;
   unsigned short  Length;
   char  FormFieldName [256],
         LocalDateTime [48],
         Scratch [1024],
         DclLine [1096];
   $DESCRIPTOR (DclLineDsc, DclLine);
   $DESCRIPTOR (DclEofDsc, tkptr->DclEof);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclCgiPlusScriptCgiPlusIn()\n");

   /* get the pointer to the request structure */
   rqptr = tkptr->rqptr;

   HttpLocalTimeString (LocalDateTime, &rqptr->BinaryTime);

   /* just a line for "start-of-request" that can always be discarded */
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, "!", 1)))
      return (status);

   sys$fao (&AuthGroupFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->AuthGroupPtr);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&AuthRealmFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->AuthRealmPtr);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&AuthTypeFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->AuthType);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   if (rqptr->HttpContentTypePtr != NULL)
      sys$fao (&ContentTypeFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpContentTypePtr);
   else
   if (rqptr->ContentType.ContentTypePtr != NULL)
      sys$fao (&ContentTypeFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->ContentType.ContentTypePtr);
   else
      sys$fao (&ContentTypeFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, "");
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ContentLengthFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->ContentLength);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&GatewayFaoDsc, &Length, &DclLineDsc, DclCgiVariablePrefix);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   if (cptr = rqptr->HttpAcceptPtr != NULL)
   {
      sys$fao (&HttpAcceptFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpAcceptPtr);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
        return (status);
   }

   if (rqptr->HttpAuthorizationPtr != NULL)
   {
      sys$fao (&AuthorizationFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpAuthorizationPtr);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
         return (status);
   }

   if (rqptr->HttpCookiePtr != NULL)
   {
      sys$fao (&CookieFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpCookiePtr);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
         return (status);
   }

   if (rqptr->HttpIfModifiedSincePtr != NULL)
   {
      sys$fao (&IfModifiedSinceFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpIfModifiedSincePtr);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
         return (status);
   }

   if (rqptr->HttpPragmaPtr != NULL)
   {
      sys$fao (&PragmaFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpPragmaPtr);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
         return (status);
   }

   if (rqptr->HttpRefererPtr != NULL)
   {
      sys$fao (&RefererFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpRefererPtr);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
         return (status);
   }

   if (rqptr->HttpUserAgentPtr != NULL)
   {
      sys$fao (&UserAgentFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, rqptr->HttpUserAgentPtr);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
         return (status);
   }

   sys$fao (&PathInfoFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->PathInfoPtr);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&PathTranslatedFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->RequestFileName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&QueryStringFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->QueryStringPtr);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RemoteHostNameFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->ClientHostName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RemoteAddressFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->ClientInternetAddress);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RemoteUserFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->RemoteUser);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RequestMethodFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->HttpMethodName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RequestTimeGmtFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->GmDateTime);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&RequestTimeLocalFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, LocalDateTime);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ScriptNameFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, rqptr->ScriptName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ServerNameFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, ServerHostName);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ServerProtocolFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, HttpProtocol);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ServerSoftwareFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, SoftwareID);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   sys$fao (&ServerPortFaoDsc, &Length, &DclLineDsc,
            DclCgiVariablePrefix, ServerPort);
   DclLine[Length] = '\0';
   if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
      return (status);

   /***************************/
   /* query string components */
   /***************************/

   if (rqptr->QueryStringPtr[0])
   {
      Count = 0;

      cptr = rqptr->QueryStringPtr;
      while (*cptr && *cptr != '=') cptr++;
      /* if an equal symbol was found then its a form not a keyword search */
      if (*cptr)
      {
         /***************/
         /* form fields */
         /***************/

         memcpy (FormFieldName, "FORM_", 5);
         cptr = rqptr->QueryStringPtr;
         while (*cptr)
         {
            sptr = FormFieldName + 5;
            zptr = FormFieldName + sizeof(FormFieldName);
            while (*cptr && *cptr != '=' && *cptr != '&' && sptr < zptr)
            {
               if (isalnum(*cptr))
                  *sptr++ = toupper(*cptr++);
               else
               {
                  *sptr++ = '_';
                  cptr++;
               }
            }
            if (sptr >= zptr)
            {
               ErrorVmsStatus (rqptr, SS$_BUFFEROVF, __FILE__, __LINE__);
               return (status);
            }
            *sptr = '\0';

            if (!FormFieldName[5] || *cptr != '=')
            {
               /* error; back-to-back '&' and/or '=', or no equate */
               ErrorGeneral (rqptr, "URL-encoded form error.",
                            __FILE__, __LINE__);
               return (STS$K_ERROR);
            }

            /* must have encountered an '=' */
            if (*cptr) cptr++;
            zptr = (sptr = Scratch) + sizeof(Scratch);
            while (*cptr && *cptr != '&' && sptr < zptr)
            {
               if (*cptr == '+')
               {
                  *sptr++ = ' ';
                  cptr++;
               }
               else
               if (*cptr == '%')
               {
                  /* an escaped character ("%xx" where xx is a hex number) */
                  cptr++;
                  c = 0;
                  if (*cptr >= '0' && *cptr <= '9')
                     { c = (*cptr - (int)'0') << 4; cptr++; }
                  else
                  if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
                     { c = (toupper(*cptr) - (int)'A' + 10) << 4; cptr++; }
                  else
                  {
                     ErrorGeneral (rqptr, "URL hexadecimal error.",
                                   __FILE__, __LINE__);
                     return (STS$K_ERROR);
                  }
                  if (*cptr >= '0' && *cptr <= '9')
                     { c += (*cptr - (int)'0'); cptr++; }
                  else
                  if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
                     { c += (toupper(*cptr) - (int)'A' + 10); cptr++; }
                  else
                  {
                     ErrorGeneral (rqptr, "URL hexadecimal error.",
                                   __FILE__, __LINE__);
                     return (STS$K_ERROR);
                  }
                  if (c == '\"' && sptr < zptr)
                  {
                     /* ensure enough quotes are passed for a DCL procedure */
                     *sptr++ = '\"';
                  }
                  if (sptr < zptr) *sptr++ = c;
               }
               else
                  *sptr++ = *cptr++;
            }
            if (sptr >= zptr)
            {
               ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
               return (STS$K_ERROR);
            }
            *sptr = '\0';
            if (*cptr) cptr++;

            sys$fao (&FormFieldFaoDsc, &Length, &DclLineDsc,
                     DclCgiVariablePrefix, FormFieldName, Scratch);
            DclLine[Length] = '\0';
            if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
               return (status);
         }

         sys$fao (&KeyWordCountFaoDsc, &Length, &DclLineDsc,
                  DclCgiVariablePrefix, 0);
         DclLine[Length] = '\0';
         if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
            return (status);
      }
      else
      {
         /******************/
         /* query keywords */
         /******************/

         memcpy (FormFieldName, "KEY_", 4);
         cptr = rqptr->QueryStringPtr;
         while (*cptr)
         {
            zptr = (sptr = Scratch) + sizeof(Scratch);
            while (*cptr && *cptr != '+' && sptr < zptr)
            {
               if (*cptr == '%')
               {
                  /* an escaped character ("%xx" where xx is a hex number) */
                  cptr++;
                  c = 0;
                  if (*cptr >= '0' && *cptr <= '9')
                     { c = (*cptr - (int)'0') << 4; cptr++; }
                  else
                  if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
                     { c = (toupper(*cptr) - (int)'A' + 10) << 4; cptr++; }
                  else
                  {
                     ErrorGeneral (rqptr, "URL hexadecimal error.",
                                   __FILE__, __LINE__);
                     return (STS$K_ERROR);
                  }
                  if (*cptr >= '0' && *cptr <= '9')
                     { c += (*cptr - (int)'0'); cptr++; }
                  else
                  if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
                     { c += (toupper(*cptr) - (int)'A' + 10); cptr++; }
                  else
                  {
                     ErrorGeneral (rqptr, "URL hexadecimal error.",
                                   __FILE__, __LINE__);
                     return (STS$K_ERROR);
                  }
                  if (sptr < zptr) *sptr++ = c;
               }
               else
                  *sptr++ = *cptr++;
            }
            if (sptr >= zptr)
            {
               ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
               return (STS$K_ERROR);
            }
            *sptr = '\0';
            if (*cptr) cptr++;

            sys$fao (&KeyWordFaoDsc, &Length, &DclLineDsc,
                     DclCgiVariablePrefix, ++Count, Scratch);
            DclLine[Length] = '\0';
            if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
               return (status);
         }

         sys$fao (&KeyWordCountFaoDsc, &Length, &DclLineDsc,
                  DclCgiVariablePrefix, Count);
         DclLine[Length] = '\0';
         if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
            return (status);
      }
   }
   else
   {
      sys$fao (&KeyWordCountFaoDsc, &Length, &DclLineDsc,
               DclCgiVariablePrefix, 0);
      DclLine[Length] = '\0';
      if (VMSnok (status = DclQioCgiPlusIn (tkptr, DclLine, Length)))
         return (status);
   }

   /***************************************/
   /* empty line terminates CGI variables */
   /***************************************/

   return (DclQioCgiPlusIn (tkptr, "", 0));
}

/*****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************  Due to it being a general report processing function.

Return a report on the DCL task structure.  This function blocks while
executing.
*/ 

DclScriptingReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ:!UL ... CGI/DCL Scripting</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>HTTPd !AZ:!UL</H1>\n\
<H2>CGI/DCL Scripting</H2>\n\
!AZ\n\
\
<P><TABLE CELLPADDING=0 BORDER=0>\n\
<TR><TD VALIGN=top>\n\
\
<TABLE CELLPADDING=2 BORDER=1>\n\
<TR><TH COLSPAN=2>Statistics</TH></TR>\n\
<TH ALIGN=left>CGI <I>script</I></TH><TD>!UL</TD></TR>\n\
<TR ALIGN=left><TH>CGIplus <I>script</I></TH>\
<TD>!UL &nbsp;<I>(!UL multiple)</I></TD></TR>\n\
<TR ALIGN=left><TH>DCL command</TH><TD>!UL</TD></TR>\n\
<TR ALIGN=left><TH>Spawned</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
\
</TD><TD>&nbsp;&nbsp;&nbsp;&nbsp;</TD><TD VALIGN=top>\n\
\
<TABLE CELLPADDING=2 BORDER=1>\n\
<TR><TH COLSPAN=3>Subprocesses</TH></TR>\n\
<TR><TH COLSPAN=2 ALIGN=right>Current</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ROWSPAN=2>Soft-Limit</TH>\
<TH ALIGN=right>Value</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TD ALIGN=right>Purged At</TD><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ROWSPAN=2>Hard-Limit</TH>\
<TH ALIGN=right>Value</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TD ALIGN=right>Hit Against</TD><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ROWSPAN=2>Purge</TH>\
<TD ALIGN=right>At Soft-Limit</TD><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TD ALIGN=right>Explicit</TD><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ROWSPAN=2>Zombie</TH>\
<TH ALIGN=right>Life-Time</TH><TD ALIGN=right>!AZ</TD></TR>\n\
<TR><TD ALIGN=right>Purged at Life-Time</TD>\
<TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ROWSPAN=3>CGIplus</TH>\
<TH ALIGN=right>Life-Time</TH><TD ALIGN=right>!AZ</TD></TR>\n\
<TR><TD ALIGN=right>Purged at Life-Time</TD><TD ALIGN=right>!UL</TD></TR>\n\
</TABLE>\n\
\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER=1>\n\
<TR><TH></TH><TH>Script/DCL</TH><TH>PID</TH>\
<TH>Zombie</TH><TH>CGIplus</TH><TH>Total</TH><TH>Last</TH></TR>\n\
<TR><TH></TH><TH COLSPAN=6>Client</TH></TR>\n\
<TR><TH COLSPAN=6></TH></TR>\n");

   static $DESCRIPTOR (TaskFaoDsc,
"<TR><TD ALIGN=right>!UL</TD>\
<TD ALIGN=left>!AZ</TD><TD>!AZ</TD>\
<TD ALIGN=right>!AZ</TD><TD ALIGN=right>!AZ</TD><TD ALIGN=right>!UL</TD>\
<TD><FONT SIZE=-2>!AZ</FONT></TD></TR>\n\
<TR><TD></TD><TD ALIGN=left COLSPAN=6>!AZ!AZ!AZ</TD></TR>\n");

   static $DESCRIPTOR (ButtonsFaoDsc,
"</TABLE>\n\
<P><TABLE BORDER=0>\n\
<TR><TD>\n\
<FORM METHOD=GET ACTION=\"!AZ!AZ\">\n\
<INPUT TYPE=submit VALUE=\" Purge \">\n\
</FORM>\n\
</TD><TD>&nbsp;</TD><TD>\n\
<FORM METHOD=GET ACTION=\"!AZ!AZ\">\n\
<INPUT TYPE=submit VALUE=\" Force Delete \">\n\
</FORM>\n\
</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n");

   static $DESCRIPTOR (PidFaoDsc, "<B><TT>!8XL</TT></B>\0");
   static $DESCRIPTOR (CgiPlusCountFaoDsc, "!UL\0");
   static $DESCRIPTOR (LifeTimeFaoDsc, "!UL\0");
   static $DESCRIPTOR (ZombieCountFaoDsc, "!UL\0");

   register unsigned long  *vecptr;
   register struct DclTaskStruct  *tkptr;
   register struct GenericListEntryStruct  *leptr;

   int  status,
        Count;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   char  Buffer [4096],
         CgiPlusCount [32],
         CgiPlusLifeTime [32],
         Pid [32],
         ZombieCount [32],
         ZombieLifeTime [32];
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (CgiPlusCountDsc, CgiPlusCount);
   $DESCRIPTOR (CgiPlusLifeTimeDsc, CgiPlusLifeTime);
   $DESCRIPTOR (PidDsc, Pid);
   $DESCRIPTOR (ZombieCountDsc, ZombieCount);
   $DESCRIPTOR (ZombieLifeTimeDsc, ZombieLifeTime);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclScriptingReport()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   *vecptr++ = Accounting.DoScriptCount;
   *vecptr++ = Accounting.DoCgiPlusScriptCount;
   *vecptr++ = Accounting.DclCgiPlusReusedCount;
   *vecptr++ = Accounting.DoDclCommandCount;
   *vecptr++ = Accounting.DclSpawnCount;

   *vecptr++ = DclSubprocessCount;
   *vecptr++ = DclSubprocessSoftLimit;
   *vecptr++ = DclSoftLimitPurgeCount;
   *vecptr++ = DclSubprocessHardLimit;
   *vecptr++ = DclHitHardLimitCount;
   *vecptr++ = DclPurgeCount;
   *vecptr++ = DclPurgeAllSubprocessesCount;
   if (DclUseZombies && Config.DclZombieLifeTime)
   {
      sys$fao (&LifeTimeFaoDsc, 0, &ZombieLifeTimeDsc,
               Config.DclZombieLifeTime);
      *vecptr++ = ZombieLifeTime;
   }
   else
      *vecptr++ = "<I>disabled</I>";
   *vecptr++ = DclZombieLifeTimePurgeCount;
   if (Config.DclCgiPlusLifeTime)
   {
      sys$fao (&LifeTimeFaoDsc, 0, &CgiPlusLifeTimeDsc,
               Config.DclCgiPlusLifeTime);
      *vecptr++ = CgiPlusLifeTime;
   }
   else
      *vecptr++ = "<I>none</I>";
   *vecptr++ = DclCgiPlusLifeTimePurgeCount;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   Count = 0;

   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)(leptr->DataPtr);

      vecptr = FaoVector;
      *vecptr++ = ++Count;
      if (tkptr->IsDclCommand)
         *vecptr++ = tkptr->DclCommand;
      else
         *vecptr++ = tkptr->ScriptName;
      if (tkptr->SubprocessPid)
      {
         sys$fao (&PidFaoDsc, 0, &PidDsc, tkptr->SubprocessPid);
         *vecptr++ = Pid;
      }
      else
         *vecptr++ = "";
      if (tkptr->ZombieCount)
      {
         sys$fao (&ZombieCountFaoDsc, 0, &ZombieCountDsc,
                  tkptr->ZombieCount);
         *vecptr++ = ZombieCount;
      }
      else
         *vecptr++ = "";
      if (tkptr->IsCgiPlusScript)
      {
         sys$fao (&CgiPlusCountFaoDsc, 0, &CgiPlusCountDsc,
                  tkptr->CgiPlusUsageCount);
         *vecptr++ = CgiPlusCount;
      }
      else
         *vecptr++ = "";
      *vecptr++ = tkptr->TotalUsageCount;
      *vecptr++ = DayDateTime (&tkptr->LastUsedBinaryTime, 23);
      if (tkptr->rqptr != NULL)
      {
         if (tkptr->rqptr->RemoteUser[0])
         {
            *vecptr++ = tkptr->rqptr->RemoteUser;
            *vecptr++ = "@";
         }
         else
         {
            *vecptr++ = "";
            *vecptr++ = "";
         }
         *vecptr++ = tkptr->rqptr->ClientHostName;
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      status = sys$faol (&TaskFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      Buffer[Length] = '\0';
      NetWrite (rqptr, 0, Buffer, Length);
   }

   vecptr = FaoVector;
   *vecptr++ = HttpdInternalControl;
   *vecptr++ = "subprocess_purge";
   *vecptr++ = HttpdInternalControl;
   *vecptr++ = "subprocess_delete";

   status = sys$faol (&ButtonsFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************  Due to it being a general control processing function.

Control function to delete and mark-for-delete all subprocesses. This is a
quick and effective means for purging all zombie and CGIplus subprocesses
forcing the next activation to reload the (possibly new) image.  The
'WithExtremePrejudice' option deletes subprocesses unconditionally, processing
or not.  It can be used to clear errant scripts without restarting the server. 
Called from Admin.c module.
*/

DclPurgeAllSubprocesses
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
boolean WithExtremePrejudice
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>Success 200</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>SUCCESS!!</H1>\n\
<P>Reported by server.\n\
<P>Server !AZ:!UL had !UL subprocess!AZ deleted, \
!UL marked for deletion.\n\
</BODY>\n\
</HTML>\n");

   register struct  GenericListEntryStruct  *leptr;
   register struct DclTaskStruct  *tkptr;
   register unsigned long  *vecptr;

   int  status,
        DeletedCount,
        MarkedCount;
   unsigned long  FaoVector [16];
   unsigned short  Length;
   char  Buffer [2048];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclPurgeAllSubprocesses()\n");

   DclPurgeAllSubprocessesCount++;

   DeletedCount = MarkedCount = 0;

   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)(leptr->DataPtr);

      if (!tkptr->SubprocessPid) continue;

      tkptr->IsMarkedForDelete = true;

      if (WithExtremePrejudice ||
          (tkptr->QueuedSysCommand <= tkptr->QueuedSysCommandAllowed &&
           !tkptr->QueuedSysOutput &&
           !tkptr->QueuedCgiPlusIn &&
           !tkptr->QueuedHttpInput &&
           tkptr->rqptr == NULL))
      {
         /* forced delete or subprocess not currently active, abort task */
         DclConcludeTask (tkptr, true);
         DeletedCount++;
      }
      else
      {
         /* subprocess is currently active, just leave marked for delete */
         MarkedCount++;
      }
   }

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;
   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = DeletedCount;
   if (DeletedCount == 1)
      *vecptr++ = "";
   else
      *vecptr++ = "es";
   *vecptr++ = MarkedCount;

   sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   Buffer[Length] = '\0';
   NetWrite (rqptr, 0, Buffer, Length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/

