/*****************************************************************************/
/*
                                  Dir.c

This module implements a full multi-threaded, AST-driven, asynchronous 
directory listing.  The AST-driven nature makes the code a little more 
difficult to follow, but creates a powerful, event-driven, multi-threaded 
server.  All of the necessary functions implementing this module are designed 
to be non-blocking. 

This module never returns a valid status and ALWAYS calls the supplied next
task (AST) function.  This should check for a generated error message to
determine is there were any problems.

This module uses the NetBufferOutput() function to buffer any output it can
into larger chunks before sending it to the client.

A directory listing is recognised by the server whenever a wildcard is present 
in a specification and there is no query string directing another activity 
(e.g. a query/search).  Compliant with other HTTPD implementations, a 
directory listing is also generated if a URL specifies a directory only and 
the directory contains no home page. 

The directory listing is designed to look very much like the default layout of 
the CERN and NCSA HTTPD implementations, with the exception that all 
directories are grouped at the top, not dispersed throughout the file names.  
This looks and functions better than the above (at least in the experience of 
the author). 

Access Control
--------------

If the server configuration specifies "DirAccessSelective" then the directory
must contain the file ".WWW_BROWSABLE" to be listed by this module.

If a directory contains the file ".WWW_HIDDEN" it cannot be listed with this
module, but individual files are accessable for server access.  If a
subdirectory contains this file it does not appear in a listing.

If a directory contains the file ".WWW_NOP" it does not show a parent.
If a directory contains the file ".WWW_NOS" it does not show subdirectories.
If a directory contains the file ".WWW_NOPS" both the above apply.

Listing Layout
--------------

The layout of the listing can be controlled from the configuration file or via
a server directive.  A decimal number may optionally, immediately precede any
directive, this becomes the width of that particular field.  If a width is not
specified an appropriate default is used.  Information wider than the field
width is truncated without indication.  The following layout directives
provide:

  _     (underscore) each underscore provides one space between fields
  C     creation date
  D     content-type description (BEST be the last field specified)
  I     icon
  L     file anchor (including hot-spot name)
  N     file name (without anchor, why I can't imagine :^)
  O     file owner (optional, configuration file enable/disable)
  R     revision date
  S     file size

These may placed in any order ("D" is best last because it does not use a 
field width) and with any (reasonable) field-width.  For example see 
'DefaultDirLayout' below. 


VERSION HISTORY
---------------
20-APR-97  MGD  changed file sizes from 1024 to 1000 kilos, megas, etc.
01-FEB-97  MGD  HTTPd version 4
24-APR-96  MGD  chagrin ... just realized I don't need full URLs in HREFs
01-DEC-95  MGD  HTTPd version 3
27-SEP-95  MGD  major revision of file and directory anchor generation
                (far fewer MapUrl()s yielding greater efficiency);
                added query string capability;
                added automatic script capability
07-AUG-95  MGD  include VMS-style directory layout and file sizes
16-JUN-95  MGD  file contents description for non-HTML files (see config.c)
25-MAR-95  MGD  bugfix for error handling with DirReadMe() and DirFileExists()
20-DEC-94  MGD  multi-threaded daemon (it suddenly got very complex!)
20-JUN-94  MGD  single-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <string.h>

/* VMS related header files */
#include <atrdef.h>
#include <descrip.h>
#include <dvidef.h>
#include <fibdef.h>
#include <iodef.h>
#include <rmsdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>

/* application related header file */
#include "wasd.h"
#include "config.h"
#include "descr.h"
#include "dir.h"
#include "error.h"
#include "file.h"
#include "httpd.h"
#include "net.h"
#include "support.h"

/**********/
/* macros */
/**********/

#define DefaultFieldWidthInterField 1
#define DefaultFieldWidthCdt 15
#define DefaultFieldWidthName 25
#define DefaultFieldWidthOwner 20
#define DefaultFieldWidthRdt 15
#define DefaultFieldWidthSize 5

#define DirBlankContentTypeIcon "x-internal/blank"
#define DirParentContentTypeIcon "x-internal/parent"
#define DirUnknownContentTypeIcon "x-internal/unknown"

/******************/
/* global storage */
/******************/

char  DirBrowsableFileName [] = ".WWW_BROWSABLE",
      DirHiddenFileName [] = ".WWW_HIDDEN",
      DirLayoutProblem [] = "Directory layout problem.",
      DirNopFileName [] = ".WWW_NOP",
      DirNosFileName [] = ".WWW_NOS",
      DirNopsFileName [] = ".WWW_NOPS",
      DirParentDirDescription [] = "parent directory",
      DirSubDirDescription [] = "subdirectory";

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern int  FileBufferSize;
extern char  ErrorFacilityDisabled[];
extern char  ErrorStringSize[];
extern char  HtmlSgmlDoctype[];
extern char  SoftwareID[];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;

/*****************************************************************************/
/*
This function never returns a valid status and ALWAYS calls the supplied next
task (AST) function.  This should check for a generated error message to
determine is there were any problems.
*/ 
 
DirBegin
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
char *DirSpec,
char *DirQueryString
)
{
   static $DESCRIPTOR (IndexOfFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>Index of !AZ</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>Index of <TT>!AZ</TT></NOBR></H1>\n");

   register char  *cptr, *sptr, *zptr;
   register unsigned long  *vecptr;
   register struct DirTaskStruct  *tkptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [8];
   char  Scratch [256],
         String [1024];
   void  *AstFunctionPtr;
   $DESCRIPTOR (StringDsc, String);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DirBegin() %d |%s|\n",
               rqptr->ErrorMessagePtr, DirSpec);

   if (!rqptr->AccountingDone)
      rqptr->AccountingDone = ++Accounting.DoDirectoryCount;

   if (rqptr->ErrorMessagePtr != NULL)
   {
      /* previous error, cause threaded processing to unravel */
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   if (!Config.DirAccess)
   {
      Accounting.RequestForbiddenCount++;
      rqptr->ResponseStatusCode = 501;
      ErrorGeneral (rqptr, ErrorFacilityDisabled, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   if (rqptr->DirTaskPoolPtr == NULL)
   {
      if (Debug) fprintf (stdout, "tkptr from HEAP\n");
      /* allocate heap memory for the file task structure */
      if ((tkptr = (struct DirTaskStruct*)
           HeapAlloc (rqptr, sizeof(struct DirTaskStruct))) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }
   }
   else
   {
      if (Debug) fprintf (stdout, "tkptr from POOL\n");
      /* get the top entry from the pool of previously used structures */
      tkptr = rqptr->DirTaskPoolPtr;
      /* set the top of pool to the next (if any) */
      rqptr->DirTaskPoolPtr = tkptr->DirTaskPtr;
      /* clear out any previous content */
      memset (tkptr, 0, sizeof(struct DirTaskStruct));
   }

   /* buffer the previous file task (if any) */
   tkptr->DirTaskPtr = rqptr->DirTaskPtr;
   /* set the current file task */
   rqptr->DirTaskPtr = tkptr;

   /* buffer the pointer to the next task */
   tkptr->NextTaskFunction = NextTaskFunction;

   cptr = MapUrl (NULL, DirSpec, NULL, NULL);
   if (!cptr[0])
   {
      ErrorGeneral (rqptr, cptr+1, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }
   zptr = (sptr = tkptr->DirPath) + sizeof(tkptr->DirPath);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }
   *sptr = '\0';

   cptr = DirSpec;
   zptr = (sptr = tkptr->DirSpec) + sizeof(tkptr->DirSpec);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }
   *sptr = '\0';
   tkptr->DirSpecLength = sptr - tkptr->DirSpec;

   if (sptr > tkptr->DirSpec && *(sptr-1) == ']')
      tkptr->DirSpecIncludedFilePart = false;
   else
      tkptr->DirSpecIncludedFilePart = true;

   cptr = DirQueryString;
   zptr = (sptr = tkptr->QueryString) + sizeof(tkptr->QueryString);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }
   *sptr = '\0';

   /********************************************/
   /* check if there are any server directives */
   /********************************************/

   if (Debug) fprintf (stdout, "QueryString |%s|\n", tkptr->QueryString);

   rqptr->ResponsePreExpired = Config.DirPreExpired;
   tkptr->AutoScriptEnabled = true;

   if (tkptr->QueryString[0])
   {
      cptr = tkptr->QueryString;
      while (*cptr)
      {
         if (strsame (cptr, "autoscript=", 11))
         {
            cptr += 11;
            /* if "false", "no" or "0" then turn off auto-scripting */
            if (tolower(*cptr) == 'f' || tolower(*cptr) == 'n' || *cptr == '0')
               tkptr->AutoScriptEnabled = false;
            else
               tkptr->AutoScriptEnabled = true;
         }
         else
         /* experience shows both make it easier! */
         if (strsame (cptr, "expired=", 8) || strsame (cptr, "expire=", 7))
         {
            if (cptr[7] == '=')
               cptr += 8;
            else
               cptr += 7;
            /* "true", "yes", "1", "false", "no" or "0" */
            if (tolower(*cptr) == 't' || tolower(*cptr) == 'y' || *cptr == '1')
               rqptr->ResponsePreExpired = true;
            else
            if (tolower(*cptr) == 'f' || tolower(*cptr) == 'n' || *cptr == '0')
               rqptr->ResponsePreExpired = false;
         }
         else
         if (strsame (cptr, "layout=", 7))
         {
            tkptr->LayoutPtr = cptr+7;
            while (*cptr && *cptr != '&') cptr++;
            if (*cptr) cptr++;
         }
         else
         if (strsame (cptr, "script=", 7))
         {
            /* local storage */
            register char  *zptr;

            cptr += 7;
            zptr = (sptr = tkptr->ScriptName) + sizeof(tkptr->ScriptName);
            if (*cptr != '/') *sptr++ = '/';
            while (*cptr && *cptr != '&' && sptr < zptr) *sptr++ = *cptr++;
            if (sptr >= zptr)
            {
               ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
               DirEnd (rqptr);
               return;
            }
            *sptr = '\0';
         }
         while (*cptr && *cptr != '&') cptr++;
         if (*cptr) cptr++;
      }
   }

   /*************************************/
   /* parse and get required components */
   /*************************************/

   tkptr->ParseInUse = true;

   tkptr->SearchFab = cc$rms_fab;
   tkptr->SearchFab.fab$l_dna = "*.*";
   tkptr->SearchFab.fab$b_dns = 3;
   tkptr->SearchFab.fab$l_fna = tkptr->DirSpec;
   tkptr->SearchFab.fab$b_fns = tkptr->DirSpecLength;
   tkptr->SearchFab.fab$l_fop = FAB$M_NAM;
   tkptr->SearchFab.fab$l_nam = &tkptr->SearchNam;

   tkptr->SearchNam = cc$rms_nam;
   tkptr->SearchNam.nam$l_esa = tkptr->ExpandedFileName;
   tkptr->SearchNam.nam$b_ess = sizeof(tkptr->ExpandedFileName)-1;

   if (VMSnok (status = sys$parse (&tkptr->SearchFab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$parse() %%X%08.08X\n", status);
      rqptr->ErrorTextPtr = tkptr->DirPath;
      rqptr->ErrorHiddenTextPtr = tkptr->DirSpec;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }

   if (tkptr->SearchNam.nam$l_fnb & NAM$M_WILD_DIR)
   {
      ErrorGeneral (rqptr, "Wildcard not permitted in directory component.",
                    __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }

   memcpy (tkptr->DirectoryPart,
           tkptr->SearchNam.nam$l_dev,
           Length = tkptr->SearchNam.nam$l_name - tkptr->SearchNam.nam$l_dev);
   tkptr->DirectoryPart[Length] = '\0';

   /* if a wildcard or explicit version number supplied the VMS format */
   if (tkptr->SearchNam.nam$l_fnb & NAM$M_WILD_VER ||
       tkptr->SearchNam.nam$l_fnb & NAM$M_EXP_VER)
   {
      memcpy (tkptr->FilePart,
              tkptr->SearchNam.nam$l_name,
              Length = tkptr->SearchNam.nam$l_ver -
                       tkptr->SearchNam.nam$l_name +
                       tkptr->SearchNam.nam$b_ver);
      tkptr->FormatLikeVms = true;
   }
   else
      memcpy (tkptr->FilePart,
              tkptr->SearchNam.nam$l_name,
              Length = tkptr->SearchNam.nam$l_ver -
                       tkptr->SearchNam.nam$l_name);
   tkptr->FilePart[Length] = '\0';

   Scratch[0] = '\0';
   sptr = MapUrl (Scratch, tkptr->DirectoryPart, NULL, NULL);
   if (!sptr[0] && sptr[1])
   {
      /* mapping report */
      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, sptr+1, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }
   Length = strlen(Scratch) + 1;
   if ((tkptr->DirectoryPathPtr = HeapAlloc (rqptr, Length)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }
   memcpy (tkptr->DirectoryPathPtr, Scratch, Length);

   if (Debug)
      fprintf (stdout,
      "DirectoryPart |%s|\nFilePart |%s|\nDirDirectoryUrl |%s|\n",
      tkptr->DirectoryPart, tkptr->FilePart,
      tkptr->DirectoryPathPtr);

   /************************/
   /* directory access ok? */
   /************************/

   if (Config.DirAccessSelective)
      status = FileExists (tkptr->DirectoryPart, DirBrowsableFileName);
   else
   {
      if (VMSok (status =
          FileExists (tkptr->DirectoryPart, DirHiddenFileName)))
         status = RMS$_FNF;
      else
      if (status == RMS$_FNF)
         status = SS$_NORMAL;
   }
   if (VMSnok (status))
   {
      /* give some "disinformation"  ;^)  */
      if (status == RMS$_FNF) status = RMS$_DNF;
      rqptr->ErrorTextPtr = tkptr->DirPath;
      rqptr->ErrorHiddenTextPtr = tkptr->DirSpec;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }

   /********************/
   /* begin processing */
   /********************/

   tkptr->FileCount = tkptr->DirectoryCount = 0;

   if (VMSnok (DirFormatLayout (rqptr)))
   {
      DirEnd (rqptr);
      return;
   }

   if (rqptr->ResponseHeaderPtr == NULL)
   {
      /* "index of"s can have pre-expiry controlled from the query string */
      rqptr->ResponseStatusCode = 200;
      if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
      {
         DirEnd (rqptr);
         return;
      }

      rqptr->RequestHeaderLength = strlen(rqptr->ResponseHeaderPtr);

      if (rqptr->HttpMethod == HTTP_METHOD_HEAD)
      {
         DirEnd (rqptr);
         return;
      }

      vecptr = FaoVector;

      *vecptr++ = HtmlSgmlDoctype;
      *vecptr++ = HtmlMetaInfo (rqptr, tkptr->DirSpec);

      if (tkptr->FormatLikeVms)
      {
         *vecptr++ = tkptr->DirSpec;
         *vecptr++ = tkptr->DirSpec;
      }
      else
      {
         *vecptr++ = tkptr->DirPath;
         *vecptr++ = tkptr->DirPath;
      }

      status = sys$faol (&IndexOfFaoDsc, &Length, &StringDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         DirEnd (rqptr);
         return;
      }
      String[Length] = '\0';

      if (Config.DirReadMeTop)
         AstFunctionPtr = &DirReadMeTop;
      else
         AstFunctionPtr = &DirHeading;

      if (VMSnok (NetBufferOutput (rqptr, AstFunctionPtr, String, Length)))
      {
         DirEnd (rqptr);
         return;
      }
   }
   else
      DirHeading (rqptr);
}

/*****************************************************************************/
/*
Release any dynamic memory allocated for parse structures.
*/ 

DirEnd (struct RequestStruct *rqptr)

{
   register struct DirTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirEnd()\n");

   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   if (tkptr->ParseInUse)
   {
      /* ensure parse internal data structures are released */
      tkptr->SearchFab.fab$l_fna = "a:[b]c.d;";
      tkptr->SearchFab.fab$b_fns = 9;
      tkptr->SearchFab.fab$b_dns = 0;
      tkptr->SearchNam.nam$b_nop = NAM$M_SYNCHK;
      sys$parse (&tkptr->SearchFab, 0, 0);
   }

   /* restore any directory task (if any) */
   rqptr->DirTaskPtr = tkptr->DirTaskPtr;
   /* new top points to the old top (if any) */
   tkptr->DirTaskPtr = rqptr->DirTaskPoolPtr;
   /* top of the pool now points at the one just finished with */
   rqptr->DirTaskPoolPtr = tkptr;

   /* declare the next task */
   if (VMSnok (status = sys$dclast (tkptr->NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
}

/*****************************************************************************/
/*
AST-driven output of column headings (these have already generated by 
DirFormatLayout()).
*/ 

DirHeading (struct RequestStruct *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirHeading()\n");

   if (VMSnok (NetBufferOutput (rqptr, &DirBeginDirectories,
                                rqptr->DirTaskPtr->LayoutHeadingPtr,
                                rqptr->DirTaskPtr->LayoutHeadingLength)))
      DirEnd (rqptr);
}

/*****************************************************************************/
/*
Begin a listing with directories if not expressly forbidden by configuration 
or local directory-specific configuration files.  If required check if there 
is an accessable parent directory and if so generate a formatted entry for it. 
*/ 
 
DirBeginDirectories (struct RequestStruct *rqptr)

{
   register char  *cptr, *sptr;
   register struct DirTaskStruct  *tkptr;

   boolean  ParentDirectory,
            SubDirectories;
   int  status;
   char  ParentDirPath [256],
         ParentDirVms [256];
   void  *AstFunctionPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirBeginDirectories()\n");

   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   /*
      Determine if any parent directory or subdirectories will be
      displayed by checking for the presence of directory listing
      control files (and a kludge for user directory mapping).
   */
   ParentDirectory = SubDirectories = true;
   cptr = tkptr->DirectoryPart;
   if (FileExists (cptr, DirNopsFileName) != RMS$_FNF)
      ParentDirectory = SubDirectories = false;
   else
   {
      if (FileExists (cptr, DirNosFileName) != RMS$_FNF)
         SubDirectories = false;

      if (FileExists (cptr, DirNopFileName) != RMS$_FNF)
         ParentDirectory = false;
      else
      if (*(unsigned short*)(cptr = tkptr->DirPath) == 0x7e2f)
      {
         /*
            This is a kludge to better support user directories ("/~username/").
            It prevents a "top-level" user directory from displaying a parent
            directory even if it is not a "top-level" VMS directory.  It does
            this by looking to see if there is any subdirectory to the "top-
            level" user directory (e.g. "/~username/subdirectory/").
         */

         /* step over the leading "/~" */
         cptr += 2;
         /* scan to the end of the username */
         while (*cptr && *cptr != '/') cptr++;
         if (*cptr)
         {
            /* check to see if there is a directory beyond the username */
            cptr++;
            while (*cptr && *cptr != '/') cptr++;
         }
         if (!*cptr) ParentDirectory = false;
      }
   }

   if (SubDirectories)
   {
      /* next function to be executed will be the search for directories */
      AstFunctionPtr = &DirSearchDirectories;

      tkptr->ParseInUse = true;

      /* set up ready to search for all files ending in ".DIR" (directories) */
      tkptr->SearchFab = cc$rms_fab;
      /* set the FAB user context to the client thread pointer */
      tkptr->SearchFab.fab$l_ctx = rqptr;
      tkptr->SearchFab.fab$l_dna = "*.DIR;";
      tkptr->SearchFab.fab$b_dns = 6;
      tkptr->SearchFab.fab$l_fna = tkptr->DirectoryPart;
      tkptr->SearchFab.fab$b_fns = strlen(tkptr->DirectoryPart);
      tkptr->SearchFab.fab$l_fop = FAB$M_NAM;
      tkptr->SearchFab.fab$l_nam = &tkptr->SearchNam;

      tkptr->SearchNam = cc$rms_nam;
      tkptr->SearchNam.nam$l_esa = tkptr->ExpandedFileName;
      tkptr->SearchNam.nam$b_ess = sizeof(tkptr->ExpandedFileName)-1;
      tkptr->SearchNam.nam$l_rsa = tkptr->ResultantFileName;
      tkptr->SearchNam.nam$b_rss = sizeof(tkptr->ResultantFileName)-1;

      if (VMSnok (status = sys$parse (&tkptr->SearchFab, 0, 0)))
      {
         if (Debug) fprintf (stdout, "sys$parse() %%X%08.08X\n", status);
         rqptr->ErrorTextPtr = tkptr->DirPath;
         rqptr->ErrorHiddenTextPtr = tkptr->DirectoryPart;
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         DirEnd (rqptr);
         return;
      }
   }
   else
   {
      /* next function to be executed will be to start the search for files */
      AstFunctionPtr = &DirBeginFiles;
   }

   if (ParentDirectory)
   {
      cptr = tkptr->DirectoryPart;
      sptr = ParentDirVms;
      while (*cptr && *cptr != '[') *sptr++ = *cptr++;
      if (!memcmp (cptr, "[000000]", 8))
      {
         /* in Master-File-Directory, therefore no parent */
         ParentDirectory = false;
      }
      else
      {
         /*
            Not in a Master-File-Directory, create a parent directory
            (e.g. "DEVICE:[DIR1.DIR2]" from "DEVICE:[DIR1.DIR2.DIR3]",
            and "DEVICE:[000000]" from "DEVICE:[DIR1]", etc.)
         */
         if (!memcmp (cptr, "[000000.", 8))
         {
            /* skip over the Master-File-Directory in the specification */
            *sptr++ = *cptr++;
            cptr += 7;
         }
         while (*cptr) *sptr++ = *cptr++;
         *sptr = '\0';
         while (*sptr != ']') sptr--;
         while (*sptr != '.' && *sptr != '[') sptr--;
         if (*sptr == '.')
         {
            *sptr++ = ']';
            *sptr = '\0';
         }
         else
            memcpy (sptr, "[000000]", 9);
         if (Debug) fprintf (stdout, "ParentDirVms |%s|\n", ParentDirVms);

         /*
            If access is allowed display the details.
            Otherwise, its as if the directory just didn't exist!
         */
         if ((Config.DirAccessSelective &&
              VMSnok (FileExists (ParentDirVms, DirBrowsableFileName))) ||
             FileExists (ParentDirVms, DirHiddenFileName) != RMS$_FNF)
            ParentDirectory = false;
      }

      if (ParentDirectory)
      {
         ParentDirPath[0] = '\0';
         sptr = MapUrl (ParentDirPath, ParentDirVms, NULL, NULL);
         if (sptr[0])
         {
            DirFormat (rqptr, AstFunctionPtr, ParentDirPath, false);
            return;
         }
         /*
            Mapping error, most probably access forbidden, we'll assume that.
            This can happen where a subdirectory is mapable but the parent is
            not (e.g. "pass /parent/subd/* /device/parent/subd/*"), making it
            forbidden to request "/parent/file" or "/parent/*.*".
            Hence, for directory listings, it's as if parent does not exist!
         */
         ParentDirectory = false;
      }
   }

   if (!ParentDirectory)
   {
      /* parent directory was not output, explicitly drive the search */
      if (VMSnok (status = sys$dclast (AstFunctionPtr, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
   }
}

/*****************************************************************************/
/*
AST function to invoke another sys$search() call when listing directories.
*/ 

DirSearchDirectories (struct RequestStruct *rqptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirSearchDirectories()\n");

   /* call RMS directory search routine, AST completion to DirDirectories() */
   sys$search (&rqptr->DirTaskPtr->SearchFab, &DirDirectories, &DirDirectories);
}

/*****************************************************************************/
/*
AST completion routine called each time sys$search() completes.  It will 
either point to another file name found or have "no more files found" status 
(or an error!).  If a file (directory) has been found check if its a "hidden" 
directory and if not format its details for the listing.
*/ 

DirDirectories (struct FAB *FabPtr)

{
   register char  *cptr, *sptr;
   register struct RequestStruct  *rqptr;
   register struct DirTaskStruct  *tkptr;

   int  status;
   char  LocalDirectory [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
      "DirDirectories() sts: %%X%08.08X stv: %%X%08.08X\n",
      FabPtr->fab$l_sts, FabPtr->fab$l_stv);

   /* retrieve the pointer to the client thread from the FAB user context */
   rqptr = FabPtr->fab$l_ctx;
   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   if (VMSnok (status = tkptr->SearchFab.fab$l_sts))
   {
      /* if its a search list treat directory not found as if file not found */
      if ((tkptr->SearchNam.nam$l_fnb & NAM$M_SEARCH_LIST) &&
           status == RMS$_DNF)
         status = RMS$_FNF;

      if (status == RMS$_FNF || status == RMS$_NMF)
      {
         /***************************/
         /* end of directory search */
         /***************************/

         tkptr->ParseInUse = false;
         DirBeginFiles (rqptr);
         return;
      }

      /**********************/
      /* sys$search() error */
      /**********************/

      rqptr->ErrorTextPtr = tkptr->DirPath;
      rqptr->ErrorHiddenTextPtr = tkptr->SearchNam.nam$l_dev;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }

   /* terminate following the last character in the version number */
   tkptr->SearchNam.nam$l_ver[tkptr->SearchNam.nam$b_ver] = '\0';
   if (Debug) fprintf (stdout, "Directory |%s|\n", tkptr->ResultantFileName);

   /************************/
   /* directory access ok? */
   /************************/

   /*
      Create a directory specification from a directory file name.
      (e.g. "DEVICE:[DIR1.DIR2]" from "DEVICE:[DIR1]DIR2.DIR")
   */
   sptr = LocalDirectory;
   cptr = tkptr->SearchNam.nam$l_dev;
   *tkptr->SearchNam.nam$l_type = '\0';
   while (*cptr) *sptr++ = *cptr++;
   *tkptr->SearchNam.nam$l_type = '.';
   *sptr++ = ']';
   *sptr = '\0';
   sptr -= 2;
   while (*sptr != ']') sptr--;
   *sptr = '.';

   /*
      If access is allowed display the details.
      Otherwise, its as if the directory just didn't exist!
   */
   if (Config.DirAccessSelective)
      status = FileExists (LocalDirectory, DirBrowsableFileName);
   else
   {
      if (VMSok (status = FileExists (LocalDirectory, DirHiddenFileName)))
         status = STS$K_ERROR;
      else
         status = SS$_NORMAL;
   }
   if (VMSnok (status))
   {
      /* directory shouldn't be accessed */
      DirSearchDirectories (rqptr);
      return;
   }

   tkptr->Description[0] = '\0';
   ConfigContentType (&tkptr->ContentType, tkptr->SearchNam.nam$l_type);
   DirFormat (rqptr, &DirSearchDirectories, "", false);
}

/*****************************************************************************/
/*
Initiate the listing of non-directory files.
*/ 

DirBeginFiles (struct RequestStruct *rqptr)

{
   register struct DirTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirBeginFiles()\n");

   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   tkptr->ParseInUse = true;

   tkptr->SearchFab = cc$rms_fab;
   /* set the FAB user context to the client thread pointer */
   tkptr->SearchFab.fab$l_ctx = rqptr;
   tkptr->SearchFab.fab$l_dna = tkptr->DirectoryPart;
   tkptr->SearchFab.fab$b_dns = strlen(tkptr->DirectoryPart);
   tkptr->SearchFab.fab$l_fna = tkptr->FilePart;
   tkptr->SearchFab.fab$b_fns = strlen(tkptr->FilePart);
   tkptr->SearchFab.fab$l_fop = FAB$M_NAM;
   tkptr->SearchFab.fab$l_nam = &tkptr->SearchNam;
   tkptr->SearchNam = cc$rms_nam;
   tkptr->SearchNam.nam$l_esa = tkptr->ExpandedFileName;
   tkptr->SearchNam.nam$b_ess = sizeof(tkptr->ExpandedFileName)-1;
   tkptr->SearchNam.nam$l_rsa = tkptr->ResultantFileName;
   tkptr->SearchNam.nam$b_rss = sizeof(tkptr->ResultantFileName)-1;

   if (VMSnok (status = sys$parse (&tkptr->SearchFab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$parse() %%X%08.08X\n", status);
      rqptr->ErrorTextPtr = tkptr->DirPath;
      rqptr->ErrorHiddenTextPtr = tkptr->DirectoryPart;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }

   /***************/
   /* file search */
   /***************/

   /* call RMS directory search routine, AST completion to DirFiles() */
   sys$search (&tkptr->SearchFab, &DirFiles, &DirFiles);
}

/*****************************************************************************/
/*
AST function to invoke another sys$search() call when listing files.
*/ 

DirSearchFiles (struct RequestStruct *rqptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirSearchFiles()\n");

   /* call RMS directory search routine, AST completion to DirFiles() */
   sys$search (&rqptr->DirTaskPtr->SearchFab, &DirFiles, &DirFiles);
}

/*****************************************************************************/
/*
This is an AST completion routine called each time sys$search() completes.  It 
will either point to another file name found or have "no more files found" 
status (or an error!).  If a file call a function to determine if the file 
will contain a "description".  When that function returns just return to the 
AST interrupted routine.
*/ 

DirFiles (struct FAB *FabPtr)

{
   register struct RequestStruct  *rqptr;
   register struct DirTaskStruct  *tkptr;

   int  status;
   void *AstFunctionPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
               "DirFiles() sts: %%X%08.08X stv: %%X%08.08X\n",
               FabPtr->fab$l_sts, FabPtr->fab$l_stv);

   /* retrieve the pointer to the client thread from the FAB user context */
   rqptr = FabPtr->fab$l_ctx;
   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   if (VMSnok (status = tkptr->SearchFab.fab$l_sts))
   {
      /* if its a search list treat directory not found as if file not found */
      if ((tkptr->SearchNam.nam$l_fnb & NAM$M_SEARCH_LIST) &&
          status == RMS$_DNF)
         status = RMS$_FNF;

      if (status == RMS$_FNF || status == RMS$_NMF)
      {
         /**********************/
         /* end of file search */
         /**********************/

         tkptr->ParseInUse = false;
         if (tkptr->FormatLikeVms)
            DirFormat (rqptr, &DirEndOutput, "", true);
         else
            DirEndOutput (rqptr);

         return;
      }

      /**********************/
      /* sys$search() error */
      /**********************/

      rqptr->ErrorTextPtr = tkptr->DirPath;
      rqptr->ErrorHiddenTextPtr = tkptr->SearchNam.nam$l_dev;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }

   /* terminate following the last character in the version number */
   tkptr->SearchNam.nam$l_ver[tkptr->SearchNam.nam$b_ver] = '\0';

   /* we've already output the directories, ignore them */
   if (!memcmp (tkptr->SearchNam.nam$l_type, ".DIR;", 5))
   {
      sys$search (&tkptr->SearchFab, &DirFiles, &DirFiles);
      return;
   }

   if (Debug) fprintf (stdout, "File |%s|\n", tkptr->SearchNam.nam$l_dev);

   /* in true Unix-style "hidden" files do not appear (those beginning ".") */
   if (*(char*)tkptr->SearchNam.nam$l_name == '.')
   {
      sys$search (&tkptr->SearchFab, &DirFiles, &DirFiles);
      return;
   }

   ConfigContentType (&tkptr->ContentType, tkptr->SearchNam.nam$l_type);

   if (VMSnok (status =
       Description (rqptr,
                    &DirEndDescription, 
                    tkptr->SearchNam.nam$l_rsa,
                    tkptr->SearchNam.nam$l_rsa,
                    tkptr->Description,
                    sizeof(tkptr->Description),
                    DESCRIPTION_TEXT_HTML)))
      DirFormat (rqptr, &DirSearchFiles, "", false);
}

/*****************************************************************************/
/*
*/ 

boolean DirEndDescription (struct RequestStruct *rqptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirEndDescription()\n");

   DirFormat (rqptr, &DirSearchFiles, "", false);
}

/*****************************************************************************/
/*
Directories have been listed, files have been listed.  If appropriate provide 
a readme at the bottom, then conclude the HTML and the directory listing.
*/ 

boolean DirEndOutput (struct RequestStruct *rqptr)

{
   register struct DirTaskStruct  *tkptr;

   int  status;
   void *AstFunctionPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirEndOutput()\n");

   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   if (Config.DirReadMeBottom)
      AstFunctionPtr = &DirReadMeBottom;
   else
      AstFunctionPtr = &DirEnd;
 
   if (tkptr->DirectoryCount || tkptr->FileCount)
   {
      if (VMSnok (NetBufferOutput (rqptr, AstFunctionPtr,
                  "</PRE><HR SIZE=3>\n</BODY>\n</HTML>\n", 35)))
         DirEnd (rqptr);
   }
   else
   {
      if (VMSnok (NetBufferOutput (rqptr, AstFunctionPtr,
                  "</PRE><HR SIZE=3>\n</BODY>\n</HTML>\n", 35)))
         DirEnd (rqptr);
   }
}

/*****************************************************************************/
/*
Using the directive string pointed to by 'tkptr->LayoutPtr' format the 
directory listing column headings and sys$fao() directive string used to format
the information for each directory/file.  Return normal or error status.  This
function buffers the details of the default generic and VMS directory layout
the first time each is required.  There-after the buffered layout information
is used, reducing overhead.  Only if a user-supplied layout is required does
any format processing occur again.
*/ 
 
int DirFormatLayout (struct RequestStruct *rqptr)

{
#  define LayoutCapacity 256

   static int  FieldWidthCdt,
               FieldWidthName,
               FieldWidthOwner,
               FieldWidthRdt,
               FieldWidthSize,
               LayoutFaoLength,
               LayoutHeadingLength,
               VmsFieldWidthCdt,
               VmsFieldWidthName,
               VmsFieldWidthOwner,
               VmsFieldWidthRdt,
               VmsFieldWidthSize,
               VmsLayoutFaoLength,
               VmsLayoutHeadingLength;
   static char  LayoutFao [LayoutCapacity/2],
                LayoutHeading [LayoutCapacity],
                VmsLayoutFao [LayoutCapacity/2],
                VmsLayoutHeading [LayoutCapacity];

   register int  cnt;
   register char  *cptr, *lptr, *sptr, *wptr, *zptr;
   register struct DirTaskStruct  *tkptr;

   int  size;
   char  String [LayoutCapacity];
   char  *LayoutFaoPtr,
         *LayoutHeadingPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirFormatLayout()\n");

   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   if (tkptr->LayoutPtr == NULL)
   {
      /******************/
      /* default layout */
      /******************/

      if (Config.DefaultDirLayout[0])
         tkptr->LayoutPtr = Config.DefaultDirLayout;
      else
         tkptr->LayoutPtr = DefaultDirListingLayout;

      if (!tkptr->FormatLikeVms && LayoutFaoLength)
      {
         /* default generic heading and fao have already been generated */
         tkptr->LayoutFaoPtr = LayoutFao;
         tkptr->LayoutFaoLength = LayoutFaoLength;
         tkptr->LayoutHeadingPtr = LayoutHeading;
         tkptr->LayoutHeadingLength = LayoutHeadingLength;
         tkptr->FieldWidthCdt = FieldWidthCdt;
         tkptr->FieldWidthName = FieldWidthName;
         tkptr->FieldWidthOwner = FieldWidthOwner;
         tkptr->FieldWidthRdt = FieldWidthRdt;
         tkptr->FieldWidthSize = FieldWidthSize;

         if (!Debug) return (SS$_NORMAL);

         fprintf (stdout,"|%d|%s|\n|%d|%s|\n",
                  LayoutHeadingLength, LayoutHeading,
                  LayoutFaoLength, LayoutFao);
         return (SS$_NORMAL);
      }

      if (tkptr->FormatLikeVms && VmsLayoutFaoLength)
      {
         /* default VMS heading and fao have already been generated */
         tkptr->LayoutFaoPtr = VmsLayoutFao;
         tkptr->LayoutFaoLength = VmsLayoutFaoLength;
         tkptr->LayoutHeadingPtr = VmsLayoutHeading;
         tkptr->LayoutHeadingLength = VmsLayoutHeadingLength;
         tkptr->FieldWidthCdt = VmsFieldWidthCdt;
         tkptr->FieldWidthName = VmsFieldWidthName;
         tkptr->FieldWidthOwner = VmsFieldWidthOwner;
         tkptr->FieldWidthRdt = VmsFieldWidthRdt;
         tkptr->FieldWidthSize = VmsFieldWidthSize;

         if (!Debug) return (SS$_NORMAL);

         fprintf (stdout,"|%d|%s|\n|%d|%s|\n",
                  VmsLayoutHeadingLength, VmsLayoutHeading,
                  VmsLayoutFaoLength, VmsLayoutFao);
         return (SS$_NORMAL);
      }

      /* default headings and layouts have not been generated ... do it! */
      if (tkptr->FormatLikeVms)
      {
         /* use the function-internal VMS buffers */
         LayoutHeadingPtr = VmsLayoutHeading;
         LayoutFaoPtr = VmsLayoutFao;
      }
      else
      {
         /* use the function-internal generic buffers */
         LayoutHeadingPtr = LayoutHeading;
         LayoutFaoPtr = LayoutFao;
      }
   }
   else
   {
      /**********************/
      /* non-default layout */
      /**********************/

      /* use scratch space to generate the non-default heading and fao */
      LayoutHeadingPtr = LayoutFaoPtr = String;
   }

   /******************************/
   /* first, the column headings */
   /******************************/

   zptr = (sptr = LayoutHeadingPtr) + LayoutCapacity;
   lptr = tkptr->LayoutPtr;
   wptr = "";

   cptr = "<PRE>";
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;

   while (*lptr && *lptr != '&' && sptr < zptr)
   {
      /** if (Debug) fprintf (stdout, "lptr |%s|\n", lptr); **/

      if (isdigit(*lptr))
      {
         wptr = lptr;
         while (*lptr && isdigit(*lptr)) lptr++;
         continue;
      }

      if (wptr[0])
      {
         cnt = atoi(wptr);
         if (cnt < 0) cnt = 0;
         wptr = "";
      }
      else
         cnt = 0;

      switch (toupper(*lptr))
      {
         case ' ' :
         case '_' :
            if (!cnt) cnt = DefaultFieldWidthInterField;
            break;

         case 'C' :
            cptr = "Created";
            if (!cnt) cnt = DefaultFieldWidthCdt;
            if (cnt < (size = 7)) cnt = size;
            if (tkptr->FormatLikeVms) cnt += 5;
            tkptr->FieldWidthCdt = cnt;
            break;

         case 'D' :
            /* no flush either way, must always be the last field */
            cptr = "Description";
            break;

         case 'I' :  /* icon */
            break;

         case 'L' :
         case 'N' :
            cptr = "Name";
            if (!cnt) cnt = DefaultFieldWidthName;
            if (cnt < (size = 4)) cnt = size;
            tkptr->FieldWidthName = cnt;
            break;

         case 'O' :
            cptr = "Owner";
            if (!cnt) cnt = DefaultFieldWidthOwner;
            if (cnt < (size = 5)) cnt = size;
            tkptr->FieldWidthOwner = cnt;
            break;

         case 'R' :
            cptr = "Revised";
            if (!cnt) cnt = DefaultFieldWidthRdt;
            if (cnt < (size = 7)) cnt = size;
            if (tkptr->FormatLikeVms) cnt += 5;
            tkptr->FieldWidthRdt = cnt;
            break;

         case 'S' :
            cptr = "Size";
            if (!cnt) cnt = DefaultFieldWidthSize;
            if (cnt < (size = 4)) cnt = size;
            /* expand size field width by 7 for blocks allocated/used */
            if (tkptr->FormatLikeVms) cnt += 7;
            tkptr->FieldWidthSize = cnt;
            break;

         default :
            /* unknown layout directive character */
            rqptr->ResponseStatusCode = 501;
            ErrorGeneral (rqptr, DirLayoutProblem, __FILE__, __LINE__);
            return (STS$K_ERROR);
      }

      switch (toupper(*lptr))
      {
         case ' ' :
         case '_' :
            while (cnt-- && sptr < zptr) *sptr++ = ' ';
            break;

         case 'C' :
         case 'R' :
         case 'S' :
            /* flush the column heading right using spaces */
            cnt -= size;
            while (cnt-- && sptr < zptr) *sptr++ = ' ';
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            break;

         case 'L' :
         case 'N' :
         case 'O' :
            /* flush the column heading left using spaces */
            cnt -= size;
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            while (cnt-- && sptr < zptr) *sptr++ = ' ';
            break;

         case 'D' :
            /* no flush either way, must always be the last field */
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            break;

         case 'I' :
            cptr = ConfigIconFor (&tkptr->IconUsed, DirBlankContentTypeIcon);
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            break;
      }
      lptr++;
   }

   cptr = "\n<HR SIZE=3>\n";
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;

   if (sptr >= zptr)
   {
      /* should be enough for all but a deliberate formatting hack! */
      rqptr->ResponseStatusCode = 501;
      ErrorGeneral (rqptr, DirLayoutProblem, __FILE__, __LINE__);
      return (STS$K_ERROR);
   }
   *sptr = '\0';
   cnt = sptr - LayoutHeadingPtr;

   if (Debug)
      fprintf (stdout, "LayoutHeadingPtr |%d|%s|\n", cnt, LayoutHeadingPtr);

   if (LayoutHeadingPtr == LayoutHeading)
   {
      /* first time a generic layout has been required, buffer details */
      tkptr->LayoutHeadingPtr = LayoutHeading;
      tkptr->LayoutHeadingLength = LayoutHeadingLength = cnt;
      FieldWidthCdt = tkptr->FieldWidthCdt;
      FieldWidthName = tkptr->FieldWidthName;
      FieldWidthOwner = tkptr->FieldWidthOwner;
      FieldWidthRdt = tkptr->FieldWidthRdt;
      FieldWidthSize = tkptr->FieldWidthSize;
   }
   else
   if (LayoutHeadingPtr == VmsLayoutHeading)
   {
      /* first time a VMS layout has been required, buffer details */
      tkptr->LayoutHeadingPtr = VmsLayoutHeading;
      tkptr->LayoutHeadingLength = VmsLayoutHeadingLength = cnt;
      VmsFieldWidthCdt = tkptr->FieldWidthCdt;
      VmsFieldWidthName = tkptr->FieldWidthName;
      VmsFieldWidthOwner = tkptr->FieldWidthOwner;
      VmsFieldWidthRdt = tkptr->FieldWidthRdt;
      VmsFieldWidthSize = tkptr->FieldWidthSize;
   }
   else
   {
      /* this is a user-supplied layout, place it in the thread's heap */
      if ((tkptr->LayoutHeadingPtr = lptr =
          HeapAlloc (rqptr, cnt+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         return (SS$_INSFMEM);
      }
      memcpy (lptr, LayoutHeadingPtr, cnt+1);
      tkptr->LayoutHeadingLength = cnt;
   }

   /******************************************/
   /* second, the sys$fao() directive string */
   /******************************************/

   zptr = (sptr = LayoutFaoPtr) + LayoutCapacity / 2;
   lptr = tkptr->LayoutPtr;
   wptr = "";

   /* accomodate an optional leading new-line (for creating a blank line) */
   cptr = "!AZ";
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;

   while (*lptr && *lptr != '&' && sptr < zptr)
   {
      /** if (Debug) fprintf (stdout, "lptr |%s|\n", lptr); **/

      if (isdigit(*lptr))
      {
         wptr = lptr;
         while (*lptr && isdigit(*lptr)) lptr++;
         continue;
      }

      if (wptr[0])
      {
         cnt = atoi(wptr);
         if (cnt < 0) cnt = 0;
         wptr = "";
      }
      else
         cnt = 0;

      switch (toupper(*lptr))
      {
         case ' ' :
         case '_' :
            if (!cnt) cnt = DefaultFieldWidthInterField;
            while (cnt-- && sptr < zptr) *sptr++ = ' ';
            break;

         case 'C' :
         case 'R' :
         case 'S' :
            /* right justify, then value */
            cptr = "!#<!#AZ!AZ!>";
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            break;

         case 'D' :
            /* nothing fancy (allow for leading and trailing quotes) */
            cptr = "!AZ!AZ!AZ";
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            break;

         case 'I' :
            /* nothing fancy */
            cptr = "!AZ";
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            break;

         case 'L' :
            /* left justify, then link, name, and name overflow */
            cptr = "!AZ!AZ!AZ!#<!AZ!AZ!AZ!>";
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            break;

         case 'N' :
            /* left justify, string, plus overflow */
            cptr = "!#<!AZ!AZ!>";
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            break;

         case 'O' :
            /* '!#!AZ' will left justify value with spaces */
            cptr = "!#AZ";
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            break;
      }
      lptr++;
   }
   if (sptr < zptr) *sptr++ = '\n';

   if (sptr >= zptr)
   {
      /* should be enough for all but a deliberate formatting hack! */
      rqptr->ResponseStatusCode = 501;
      ErrorGeneral (rqptr, DirLayoutProblem, __FILE__, __LINE__);
      return (STS$K_ERROR);
   }
   *sptr = '\0';
   cnt = sptr - LayoutFaoPtr;

   if (Debug) fprintf (stdout, "LayoutFaoPtr |%d|%s|\n", cnt, LayoutFaoPtr);

   if (LayoutFaoPtr == LayoutFao)
   {
      /* first time a generic layout has been required, buffer details */
      tkptr->LayoutFaoPtr = LayoutFao;
      tkptr->LayoutFaoLength = LayoutFaoLength = cnt;
   }
   else
   if (LayoutFaoPtr == VmsLayoutFao)
   {
      /* first time a VMS layout has been required, buffer details */
      tkptr->LayoutFaoPtr = VmsLayoutFao;
      tkptr->LayoutFaoLength = VmsLayoutFaoLength = cnt;
   }
   else
   {
      /* this is a user-supplied layout, place it in the thread's heap */
      if ((tkptr->LayoutFaoPtr = lptr =
          HeapAlloc (rqptr, cnt+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         return (SS$_INSFMEM);
      }
      memcpy (lptr, LayoutFaoPtr, cnt+1);
      tkptr->LayoutFaoLength = cnt;
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This overly-long (sorry) function formats and outputs a line of information 
for a parent directory, directory file, non-directory file, or for the final 
line giving used/allocated totals for a VMS format listing. 

For directory and ordinary files retrieve and display the specified file's 
statistics (name, size, modification  time, etc.)  This function completes 
synchronously, that is it waits for the single QIO on the file details to 
complete before assembling the details for the listing, because of this the 
function introduces slight granularity.  It is a candidate for future 
improvement. 

This function uses the ACP-QIO interface detailed in the "OpenVMS I/O User's 
Reference Manual".
*/ 

int DirFormat
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *ParentDirPath,
boolean BlockTotals
)
{
   static char  *MonthName [] =
      { "", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

   static $DESCRIPTOR (BlocksFaoDsc, "!UL/!UL");
   static $DESCRIPTOR (BytesFaoDsc, "!UL");
   static $DESCRIPTOR (CdtDsc, "");
   static $DESCRIPTOR (DateTimeFaoDsc, "!2ZW-!AZ-!2ZW !2ZW:!2ZW:!2ZW");
   static $DESCRIPTOR (DeviceDsc, "");
   static $DESCRIPTOR (LayoutFaoDsc, "");
   static $DESCRIPTOR (KBytesFaoDsc, "!ULK");
   static $DESCRIPTOR (MBytesFaoDsc, "!ULM");
   static $DESCRIPTOR (OwnerFaoDsc, "!%I");
   static $DESCRIPTOR (OwnerDsc, "");
   static $DESCRIPTOR (RdtDsc, "");
   static $DESCRIPTOR (SizeDsc, "");
   static $DESCRIPTOR (SizeFaoDsc, "!#* !AZ");
   static $DESCRIPTOR (StringDsc, "");
   static $DESCRIPTOR (TotalFilesDsc, "");
   static $DESCRIPTOR (TotalFilesFaoDsc, "!UL files");
   static $DESCRIPTOR (VmsDateTimeFaoDsc, "!%D");

   register int  idx, MaxIdx;
   register char  *cptr, *sptr, *zptr;
   register struct DirTaskStruct  *tkptr;

   boolean  ThisIsADirectory;
   int  status;
   unsigned short  AcpChannel,
                   Length;
   unsigned short  NumTime [7];
   unsigned long  Bytes,
                  AllocatedVbn,
                  EndOfFileVbn;
   unsigned long  FaoVector [64];
   char  *IconPtr,
         *AnchorNameOverflowPtr;
   char  AnchorLink [1024],
         AnchorName [128],
         Cdt [64],
         Description [256],
         Owner [128],
         Rdt [64],
         Scratch [256],
         Size [32],
         String [2048],
         TotalFiles [32];

   unsigned long  AtrUic;
   unsigned long  AtrCdt [2],
                  AtrRdt [2];
   struct {
      unsigned long  OfNoInterest1;
      unsigned short  AllocatedVbnHi;
      unsigned short  AllocatedVbnLo;
      unsigned short  EndOfFileVbnHi;
      unsigned short  EndOfFileVbnLo;
      unsigned short  FirstFreeByte;
      unsigned short  OfNoInterest2;
      unsigned long  OfNoInterestLots [4];
   } AtrRecord;

   struct fibdef  SearchFib;
   struct atrdef  SearchAtr [] =
   {
      { sizeof(AtrRecord), ATR$C_RECATTR, &AtrRecord },
      { sizeof(AtrCdt), ATR$C_CREDATE, &AtrCdt },
      { sizeof(AtrRdt), ATR$C_REVDATE, &AtrRdt },
      { sizeof(AtrUic), ATR$C_UIC, &AtrUic },
      { 0, 0, 0 }
   };

   struct {
      unsigned short  Length;
      unsigned short  Unused;
      unsigned long  Address;
   } FileNameAcpDsc,
     SearchFibAcpDsc,
     SearchAtrAcpDsc;

   struct {
      unsigned short  Status;
      unsigned short  Unused1;
      unsigned long  Unused2;
   } AcpIOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DirFormat() |%s|\n",
               rqptr->DirTaskPtr->SearchNam.nam$l_dev);

   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   if (!(ParentDirPath[0] || BlockTotals))
   {
      /*******************************************/
      /* queue an ACP I/O to get file attributes */
      /*******************************************/

      /* assign a channel to the disk device containing the file */
      DeviceDsc.dsc$w_length = tkptr->SearchNam.nam$b_dev;
      DeviceDsc.dsc$a_pointer = tkptr->SearchNam.nam$l_dev;
      status = sys$assign (&DeviceDsc, &AcpChannel, 0, 0, 0);
      if (Debug) fprintf (stdout, "sys$assign() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         rqptr->ErrorTextPtr = tkptr->DirPath;
         rqptr->ErrorHiddenTextPtr = tkptr->SearchNam.nam$l_dev;
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         DirEnd (rqptr);
         return;
      }

      /* set up the File Information Block for the ACP interface */
      memset (&SearchFib, 0, sizeof(struct fibdef));
      SearchFibAcpDsc.Length = sizeof(SearchFib);
      SearchFibAcpDsc.Address = &SearchFib;
      /* quick work around, different syntax for this structure with DEC C? */
#ifdef __DECC
      memcpy (&SearchFib.fib$w_did, &tkptr->SearchNam.nam$w_did, 6);
#else
      memcpy (&SearchFib.fib$r_did_overlay.fib$w_did,
              &tkptr->SearchNam.nam$w_did, 6);
#endif

      FileNameAcpDsc.Length =
         tkptr->SearchNam.nam$b_name +
         tkptr->SearchNam.nam$b_type +
         tkptr->SearchNam.nam$b_ver;
      FileNameAcpDsc.Address = tkptr->SearchNam.nam$l_name;

      status = sys$qiow (0, AcpChannel, IO$_ACCESS, &AcpIOsb, 0, 0, 
                         &SearchFibAcpDsc, &FileNameAcpDsc, 0, 0,
                         &SearchAtr, 0);

      /* immediately deassign the channel in case we return on an error */
      sys$dassgn (AcpChannel);

      if (Debug)
         fprintf (stdout, "sys$qio() %%X%08.08X IOsb: %%X%08.08X\n",
                  status, AcpIOsb.Status);

      if (VMSok (status)) status = AcpIOsb.Status;
      if (VMSnok (status)) 
      {
         rqptr->ErrorTextPtr = tkptr->DirPath;
         rqptr->ErrorHiddenTextPtr = tkptr->SearchNam.nam$l_dev;
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         DirEnd (rqptr);
         return;
      }

      /* test for numeric equivalent of "DIR;" */
      if (ThisIsADirectory = !memcmp(tkptr->SearchNam.nam$l_type, ".DIR;", 5))
      {
         /* no need to display the MFD */
         if (tkptr->SearchNam.nam$l_name[0] == '0' &&
             !memcmp (tkptr->SearchNam.nam$l_name, "000000.DIR;", 11))
         {
            /* can't return without queueing the next bit of processing! */
            if (VMSnok (status = sys$dclast (AstFunctionPtr, rqptr, 0)))
               ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
            return;
         }
      }
   }

   /********************************/
   /* process file name components */
   /********************************/

   MaxIdx = sizeof(FaoVector);
   idx = 0;

   if (BlockTotals)
   {
      /****************/
      /* block totals */
      /****************/

      /* must be block totals, which requires a leading newline */
      if (idx < MaxIdx) FaoVector[idx++] = "\n";
   }
   else
   if (ParentDirPath[0])
   {
      /********************/
      /* parent directory */
      /********************/

      /* no leading newline required for the parent directory */
      if (idx < MaxIdx) FaoVector[idx++] = "";
      tkptr->DirectoryCount = -1;

      zptr = (sptr = AnchorLink) + sizeof(AnchorLink);
      /* copy in the directory portion of the URL */
      cptr = ParentDirPath;
      while (*cptr && sptr < zptr) *sptr++ = tolower(*cptr++);
      /* append any file name/type/version from the original spec */
      if (tkptr->DirSpecIncludedFilePart)
      {
         cptr = tkptr->FilePart;
         while (*cptr && sptr < zptr) *sptr++ = tolower(*cptr++);
      }
      /* propagate any query string */
      if (tkptr->QueryString[0])
      {
         if (sptr < zptr) *sptr++ = '?';
         for (cptr = tkptr->QueryString;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
      }
      if (sptr >= zptr)
      {
         ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
         DirEnd (rqptr);
         return;
      }
      *sptr = '\0';
      /** if (Debug) fprintf (stdout, "AnchorLink |%s|\n", AnchorLink); **/

      if (tkptr->FormatLikeVms)
         memcpy (AnchorName, "[-]", 4);
      else
         memcpy (AnchorName, "../", 4);
      sptr = AnchorName + 4;
      /** if (Debug) fprintf (stdout, "AnchorName |%s|\n", AnchorName); **/
   }
   else
   if (ThisIsADirectory)
   {
      /***************/
      /* a directory */
      /***************/

      /* no leading newline required for directories */
      if (idx < MaxIdx) FaoVector[idx++] = "";
      if (tkptr->DirectoryCount <= 0)
         tkptr->DirectoryCount = 1;
      else
         tkptr->DirectoryCount++;

      /* terminate to remove the file type (".DIR") */
      *tkptr->SearchNam.nam$l_type = '\0';

      zptr = (sptr = AnchorLink) + sizeof(AnchorLink);

      /* copy in the directory (file) name */
      cptr = tkptr->SearchNam.nam$l_name;
      while (*cptr && sptr < zptr) *sptr++ = tolower(*cptr++);
      /* end directory */
      if (sptr < zptr) *sptr++ = '/';
      /* append any file name/type/version from the original spec */
      if (tkptr->DirSpecIncludedFilePart)
      {
         cptr = tkptr->FilePart;
         while (*cptr && sptr < zptr) *sptr++ = tolower(*cptr++);
      }
      /* propagate any query string */
      if (tkptr->QueryString[0])
      {
         if (sptr < zptr) *sptr++ = '?';
         for (cptr = tkptr->QueryString;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
      }
      if (sptr >= zptr)
      {
         ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
         DirEnd (rqptr);
         return;
      }
      *sptr = '\0';
      /** if (Debug) fprintf (stdout, "AnchorLink |%s|\n", AnchorLink); **/

      sptr = AnchorName;
      if (tkptr->FormatLikeVms)
      {
         *sptr++ = '[';
         *sptr++ = '.';
      }
      cptr = tkptr->SearchNam.nam$l_name;
      while (*cptr) *sptr++ = *cptr++;
      if (tkptr->FormatLikeVms)
         *sptr++ = ']';
      else
         *sptr++ = '/';
      *sptr = '\0';
      /** if (Debug) fprintf (stdout, "AnchorName |%s|\n", AnchorName); **/

      /* restore the type delimitter */
      *tkptr->SearchNam.nam$l_type = '.';
   }
   else
   {
      /**********/
      /* a file */
      /**********/

      /* leading newline required for first file after directories */
      if (tkptr->DirectoryCount && !tkptr->FileCount)
         { if (idx < MaxIdx) FaoVector[idx++] = "\n"; }
      else
         if (idx < MaxIdx) FaoVector[idx++] = "";
      tkptr->FileCount++;

      /* if appropriate, terminate to remove the version number */
      if (!tkptr->FormatLikeVms) *tkptr->SearchNam.nam$l_ver = '\0';

      /* don't need string limit checking, no user supplied component here! */
      sptr = AnchorLink;

      if (tkptr->ScriptName[0])
      {
         /* prepend the (optional) script component of the path */
         cptr = tkptr->ScriptName;
         while (*cptr) *sptr++ = *cptr++;

         /* need the directory portion of the URL for scripting */
         cptr = tkptr->DirectoryPathPtr;
         while (*cptr) *sptr++ = tolower(*cptr++);
      }
      /* copy in the file name and type */
      cptr = tkptr->SearchNam.nam$l_name;
      while (*cptr) *sptr++ = tolower(*cptr++);
      /* disable auto-scripting by appending a (most-recent) version number */
      if (!tkptr->FormatLikeVms && !tkptr->AutoScriptEnabled)
      {
         *sptr++ = ';';
         *sptr++ = '0';
      }
      *sptr = '\0';
      /** if (Debug) fprintf (stdout, "AnchorLink |%s|\n", AnchorLink); **/

      sptr = AnchorName;
      cptr = tkptr->SearchNam.nam$l_name;
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
      /** if (Debug) fprintf (stdout, "AnchorName |%s|\n", AnchorName); **/

      /* if appropriate, restore the version delimitter */
      if (!tkptr->FormatLikeVms) *tkptr->SearchNam.nam$l_ver = ';';
   }

   if (!BlockTotals)
   {
      /********************************/
      /* check if file name overflows */
      /********************************/

      /* assumes that the end of 'AnchorName' is still pointed to by 'sptr' */
      if (sptr - AnchorName > tkptr->FieldWidthName)
      {
         AnchorNameOverflowPtr = "+";
         AnchorName[tkptr->FieldWidthName-1] = '\0';
      }
      else
         AnchorNameOverflowPtr = "";
   }

   /***********************************/
   /* build the sys$faol() paremeters */
   /***********************************/

   for (cptr = tkptr->LayoutPtr; *cptr; cptr++)
   {
      if (!isalpha(*cptr)) continue;
      /** if (Debug) fprintf (stdout, "idx: %d cptr |%s|\n", idx, cptr); **/

      switch (toupper(*cptr))
      {
         case 'C' :

            /*****************/
            /* creation date */
            /*****************/

            if (ParentDirPath[0] || BlockTotals)
            {
               /* make it an empty field */
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthCdt;
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthCdt;
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = "";
               break;
            }

            /* format the creation date/time */
            CdtDsc.dsc$a_pointer = Cdt;
            CdtDsc.dsc$w_length = sizeof(Cdt)-1;
            if (tkptr->FormatLikeVms)
               sys$fao (&VmsDateTimeFaoDsc, &Length, &CdtDsc, &AtrCdt);
            else
            {
               sys$numtim (&NumTime, &AtrCdt);
               sys$fao (&DateTimeFaoDsc, &Length, &CdtDsc,
                        NumTime[2], MonthName[NumTime[1]], NumTime[0]%100,
                        NumTime[3], NumTime[4], NumTime[5]);
            }
            Cdt[Length] = '\0';
            /** if (Debug) fprintf (stdout, "Cdt |%s|\n", Cdt); **/
            if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthCdt;
            /* next parameter right justifies the field */
            if (tkptr->FieldWidthCdt - Length > 0)
            {
               if (idx < MaxIdx)
                  FaoVector[idx++] = tkptr->FieldWidthCdt - Length;
            }
            else
               if (idx < MaxIdx) FaoVector[idx++] = 0;
            if (idx < MaxIdx) FaoVector[idx++] = "";
            if (idx < MaxIdx) FaoVector[idx++] = Cdt;
            break;

         case 'D' :

            /***************/
            /* description */
            /***************/

            if (ParentDirPath[0])
            {
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = DirParentDirDescription;
               if (idx < MaxIdx) FaoVector[idx++] = "";
            }
            else
            if (BlockTotals)
            {
               if (tkptr->DirectoryCount < 0)
                  tkptr->DirectoryCount = 0;
               TotalFilesDsc.dsc$a_pointer = TotalFiles;
               TotalFilesDsc.dsc$w_length = sizeof(TotalFiles)-1;
               sys$fao (&TotalFilesFaoDsc, &Length, &TotalFilesDsc,
                        tkptr->DirectoryCount + tkptr->FileCount);
               TotalFiles[Length] = '\0';
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = TotalFiles;
               if (idx < MaxIdx) FaoVector[idx++] = "";
            }
            else
            if (ThisIsADirectory)
            {
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = DirSubDirDescription;
               if (idx < MaxIdx) FaoVector[idx++] = "";
            }
            else
            if (tkptr->Description[0] &&
                tkptr->Description[0] != DESCRIPTION_IMPOSSIBLE)
            {
               if (idx < MaxIdx) FaoVector[idx++] = "\"";
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->Description;
               if (idx < MaxIdx) FaoVector[idx++] = "\"";
            }
            else
            {
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx)
                  FaoVector[idx++] = tkptr->ContentType.ContentDescriptionPtr;
               if (idx < MaxIdx) FaoVector[idx++] = "";
            }
            break;

         case 'I' :

            /********/
            /* icon */
            /********/

            if (ParentDirPath[0])
               IconPtr = ConfigIconFor (&tkptr->IconUsed,
                                        DirParentContentTypeIcon);
            else
            if (BlockTotals)
               IconPtr = ConfigIconFor (&tkptr->IconUsed,
                                        DirBlankContentTypeIcon);
            else
               IconPtr = ConfigIconFor (&tkptr->IconUsed,
                                        tkptr->ContentType.ContentTypePtr);
            if (!IconPtr[0])
               IconPtr = ConfigIconFor (&tkptr->IconUsed,
                                        DirUnknownContentTypeIcon);
            if (idx < MaxIdx) FaoVector[idx++] = IconPtr;
            break;

         case 'L' :

            /***************/
            /* link/anchor */
            /***************/

            if (BlockTotals)
            {
               /* make it an empty field */
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthName;
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = "";
               break;
            }

            if (idx < MaxIdx) FaoVector[idx++] = "<A HREF=\"";
            if (idx < MaxIdx) FaoVector[idx++] = AnchorLink;
            if (idx < MaxIdx) FaoVector[idx++] = "\">";
            if (idx < MaxIdx)
               FaoVector[idx++] = tkptr->FieldWidthName + sizeof("</A>")-1;
            if (idx < MaxIdx) FaoVector[idx++] = AnchorName;
            if (idx < MaxIdx) FaoVector[idx++] = "</A>";
            if (idx < MaxIdx) FaoVector[idx++] = AnchorNameOverflowPtr;
            break;

         case 'N' :

            /********/
            /* name */
            /********/

            if (BlockTotals)
            {
               /* make it an empty field */
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthName;
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = "";
               break;
            }

            if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthName;
            if (idx < MaxIdx) FaoVector[idx++] = AnchorName;
            if (idx < MaxIdx) FaoVector[idx++] = AnchorNameOverflowPtr;
            break;

         case 'O' :

            /*********/
            /* owner */
            /*********/

            if (ParentDirPath[0] || BlockTotals)
            {
               /* make it an empty field */
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthOwner;
               if (idx < MaxIdx) FaoVector[idx++] = "";
               break;
            }

            if (Config.DirOwnerEnabled)
            {
               OwnerDsc.dsc$a_pointer = Owner;
               OwnerDsc.dsc$w_length = sizeof(Owner)-1;
               sys$fao (&OwnerFaoDsc, &Length, &OwnerDsc, AtrUic);
               Owner[Length] = '\0';
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthOwner;
               if (idx < MaxIdx) FaoVector[idx++] = Owner;
               break;
            }

            memcpy (Owner, "<I>(disabled)</I>", 18);
            if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthOwner + 7;
            if (idx < MaxIdx) FaoVector[idx++] = Owner;
            break;

         case 'R' :

            /*****************/
            /* revision date */
            /*****************/

            if (ParentDirPath[0] || BlockTotals)
            {
               /* make it an empty field */
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthRdt;
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthRdt;
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = "";
               break;
            }

            /* format the revision date/time */
            RdtDsc.dsc$a_pointer = Rdt;
            RdtDsc.dsc$w_length = sizeof(Rdt)-1;
            if (tkptr->FormatLikeVms)
               sys$fao (&VmsDateTimeFaoDsc, &Length, &RdtDsc, &AtrRdt);
            else
            {
               sys$numtim (&NumTime, &AtrRdt);
               sys$fao (&DateTimeFaoDsc, &Length, &RdtDsc,
                        NumTime[2], MonthName[NumTime[1]], NumTime[0]%100,
                        NumTime[3], NumTime[4], NumTime[5]);
            }
            Rdt[Length] = '\0';
            /** if (Debug) fprintf (stdout, "Rdt |%s|\n", Rdt); **/
            if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthRdt;
            /* next parameter right justifies the field */
            if (tkptr->FieldWidthRdt - Length > 0)
            {
               if (idx < MaxIdx)
                  FaoVector[idx++] = tkptr->FieldWidthRdt - Length;
            }
            else
               if (idx < MaxIdx) FaoVector[idx++] = 0;
            if (idx < MaxIdx) FaoVector[idx++] = "";
            if (idx < MaxIdx) FaoVector[idx++] = Rdt;
            break;

         case 'S' :

            /********/
            /* size */
            /********/

            if (ParentDirPath[0])
            {                                   
               /* make it an empty field */
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthSize;
               if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthSize;
               if (idx < MaxIdx) FaoVector[idx++] = "";
               if (idx < MaxIdx) FaoVector[idx++] = "";
               break;
            }

            SizeDsc.dsc$a_pointer = Size;
            SizeDsc.dsc$w_length = sizeof(Size)-1;

            if (BlockTotals)
            {
               sys$fao (&BlocksFaoDsc, &Length, &SizeDsc,
                        tkptr->TotalUsedBlocks,
                        tkptr->TotalAllocatedBlocks);
            }
            else
            {
               AllocatedVbn = AtrRecord.AllocatedVbnLo +
                              (AtrRecord.AllocatedVbnHi << 16);
               EndOfFileVbn = AtrRecord.EndOfFileVbnLo +
                              (AtrRecord.EndOfFileVbnHi << 16);

/**
               if (Debug)
                  fprintf (stdout,
                  "AllocatedVbn: %d EndOfFileVbn: %d FirstFreeByte %d\n",
                  AllocatedVbn, EndOfFileVbn, AtrRecord.FirstFreeByte);
**/

               if (ThisIsADirectory && tkptr->FormatLikeVms)
                  EndOfFileVbn--;
               else
               if (!ThisIsADirectory &&
                   EndOfFileVbn && !AtrRecord.FirstFreeByte)
                  EndOfFileVbn--;

               tkptr->TotalAllocatedBlocks += AllocatedVbn;
               tkptr->TotalUsedBlocks += EndOfFileVbn;

               if (tkptr->FormatLikeVms)
               {
                  sys$fao (&BlocksFaoDsc, &Length, &SizeDsc,
                           EndOfFileVbn, AllocatedVbn);
               }
               else
               {
                  /* the "<< 9" is more efficient than equivalent "* 512" */
                  if (EndOfFileVbn <= 1)
                     Bytes = AtrRecord.FirstFreeByte;
                  else
                     Bytes = ((EndOfFileVbn - 1) << 9) +
                             AtrRecord.FirstFreeByte;
                  /** if (Debug) fprintf (stdout, "Bytes %d\n", Bytes); **/
                  if (Bytes >= 1000000)
                     sys$fao (&MBytesFaoDsc, &Length, &SizeDsc, Bytes/1000000);
                  else
                  if (Bytes >= 1000)
                     sys$fao (&KBytesFaoDsc, &Length, &SizeDsc, Bytes/1000);
                  else
                     sys$fao (&BytesFaoDsc, &Length, &SizeDsc, Bytes);
               }
            }

            Size[Length] = '\0';
            /** if (Debug) fprintf (stdout, "Size |%s|\n", Size); **/

            if (idx < MaxIdx) FaoVector[idx++] = tkptr->FieldWidthSize;
            /* next parameter right justifies the field */
            if (tkptr->FieldWidthSize - Length > 0)
            {
               if (idx < MaxIdx)
                  FaoVector[idx++] = tkptr->FieldWidthSize - Length;
            }
            else
               if (idx < MaxIdx) FaoVector[idx++] = 0;
            if (idx < MaxIdx) FaoVector[idx++] = "";
            if (idx < MaxIdx) FaoVector[idx++] = Size;
            break;
      }

      if (idx >= MaxIdx)
      {
         /* should be enough for all but a deliberate formatting hack! */
         rqptr->ResponseStatusCode = 501;
         ErrorGeneral (rqptr, DirLayoutProblem, __FILE__, __LINE__);
         DirEnd (rqptr);
         return;
      }
   }

   /** if (Debug) fprintf (stdout, "idx: %d\n", idx); **/

   /*********************/
   /* format and output */
   /*********************/

/**
   if (Debug)
      fprintf (stdout, "LayoutFaoPtr |%s|\n", tkptr->LayoutFaoPtr);
**/
   LayoutFaoDsc.dsc$a_pointer = tkptr->LayoutFaoPtr;
   LayoutFaoDsc.dsc$w_length = tkptr->LayoutFaoLength;
   StringDsc.dsc$a_pointer = String;
   StringDsc.dsc$w_length = sizeof(String)-1;

   status = sys$faol (&LayoutFaoDsc, &Length, &StringDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = tkptr->DirPath;
      rqptr->ErrorHiddenTextPtr = tkptr->LayoutFaoPtr;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      DirEnd (rqptr);
      return;
   }
   String[Length] = '\0';
   /** if (Debug) fprintf (stdout, "String: %d |%s|\n", Length, String); **/

   /* empty any file-internal description previously generated */
   tkptr->Description[0] = '\0';

   if (VMSnok (NetBufferOutput (rqptr, AstFunctionPtr, String, Length)))
      DirEnd (rqptr);
}

/*****************************************************************************/
/*
These comments apply equally to DirReadMeBottom() below.

Attempt to send one of possible multiple files ("README.", et. al.) in the 
specified directory.  Then, provided there was not serious error with the 
send, execute the function specified by the address in 'NextTaskFunction'.  
The file can be HTML or plain-text.  With plain-text the contents will be 
encapsulated.
*/ 
 
int DirReadMeTop (struct RequestStruct *rqptr)

{
   register char  *cptr, *rptr, *sptr;
   register struct DirTaskStruct  *tkptr;

   boolean  PreformatFile;
   int  status;
   char  *TypePtr;
   char  FileName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirReadMeTop()\n");

   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   if (*(rptr = ConfigReadMeFile(tkptr->ReadMeFileIndex++)))
   {
      sptr = FileName;
      for (cptr = tkptr->DirectoryPart; *cptr; *sptr++ = *cptr++);
      TypePtr = NULL;
      for (/* 'rptr' initialized above! */; *rptr; *sptr++ = *rptr++)
         if (*rptr == '.') TypePtr = rptr;
      *sptr = '\0';
      if (TypePtr == NULL) TypePtr = rptr;

      cptr = ConfigContentType (&tkptr->ContentType, TypePtr);
      if (strsame (cptr, Config.PlainTextContentType, -1))
         PreformatFile = true;
      else
         PreformatFile = false;

       FileBegin (rqptr, &DirHeading, &DirReadMeTop,
                  FileName, cptr, PreformatFile, false);
       return;
   }

   DirHeading (rqptr);
}

/*****************************************************************************/
/*
See comment in DirReadMeTop() above.
*/ 
 
int DirReadMeBottom (struct RequestStruct *rqptr)

{
   register char  *cptr, *rptr, *sptr;
   register struct DirTaskStruct  *tkptr;

   boolean  PreformatFile;
   int  status;
   char  *TypePtr;
   char  FileName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DirReadMeBottom()\n");

   /* get the pointer to the task structure */
   tkptr = rqptr->DirTaskPtr;

   if (*(rptr = ConfigReadMeFile(tkptr->ReadMeFileIndex++)))
   {
      sptr = FileName;
      for (cptr = tkptr->DirectoryPart; *cptr; *sptr++ = *cptr++);
      TypePtr = NULL;
      for (/* 'rptr' initialized above! */; *rptr; *sptr++ = *rptr++)
         if (*rptr == '.') TypePtr = rptr;
      *sptr = '\0';
      if (TypePtr == NULL) TypePtr = rptr;

      cptr = ConfigContentType (&tkptr->ContentType, TypePtr);
      if (strsame (cptr, Config.PlainTextContentType, -1))
         PreformatFile = true;
      else
         PreformatFile = false;

       FileBegin (rqptr, &DirEnd, &DirReadMeBottom,
                  FileName, cptr, PreformatFile, false);
       return;
   }

   DirEnd (rqptr);
}

/*****************************************************************************/
