/*****************************************************************************/
/*
                                Request.c

Get, process and execute the HTTP request from the client.

Some server directives are detected in this module.  Server directives take
the form of a special path or query string and are CASE SENSISTIVE.

  ?httpd=index&...            used with a wildcard specification, generates
                              a directory listing with specified format

  /httpd/-/admin/             generates a server administration menu
  /httpd/-/admin/graphic/...  generates graphics
  /httpd/-/admin/report/...   generates various reports
  /httpd/-/admin/revise/...   allows form-based configuration
  /httpd/-/admin/control/...  allows server to be stopped, restarted, etc.

A request for an HTTP/1.0 persistent connection is looked for in the request
header and if found a boolean set to indicate the request.

AUTHORIZATION MAY BE PERFORMED TWICE!

The first, to authorize the full path provided with the request (after it
has been mapped!)  The second, if the request path contained a script any
path specification following the script component is also checked.  In this
way access to controlled information is not inadvertantly allowed because
the script itself is not controlled.  In general, script specifications
should not be part of an authorized path if that script is used to access
other paths (i.e. does not return data generated by itself), apply
authorization to data paths.

Error counting within this module attempts to track the reason that any 
connection accepted fails before being passsed to another module for 
processing (i.e. request format error, path forbidden by rule).


VERSION HISTORY
---------------
27-JUL-97  MGD  modified "Accept:" header lines processing
08-JUN-97  MGD  added "Pragma:" header field detection
27-MAR-97  MGD  added "temporary" file detection (for UPD/PUT preview)
01-FEB-97  MGD  HTTPd version 4
01-OCT-96  MGD  added more reports
25-JUL-96  MGD  use optional "length=" within "If-Modified-Since:" header
12-APR-96  MGD  RMS parse structures moved to thread data;
                persistent connections ("keep-alive");
                changed internal directive from query string to path;
                observed Multinet disconnection/zero-byte behaviour
                (request now aborts if Multinet returns zero bytes)
01-DEC-95  MGD  HTTPd version 3
27-SEP-95  MGD  extensive rework of some functions;
                added 'Referer:', 'User-Agent:', 'If-Modified-Since:'
01-APR-95  MGD  initial development for addition to multi-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <iodef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"
#include "admin.h"
#include "config.h"
#include "control.h"
#include "dcl.h"
#include "dir.h"
#include "error.h"
#include "file.h"
#include "httpd.h"
#include "ismap.h"
#include "menu.h"
#include "net.h"
#include "request.h"
#include "ssi.h"
#include "support.h"

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern boolean  ControlExitRequested;
extern boolean  ControlRestartRequested;
extern int  NetReadBufferSize;
extern int  ExitStatus;
extern char  ErrorAuthorizationRequired[];
extern char  ErrorFacilityDisabled[];
extern char  ErrorStringSize[];
extern char  ErrorRequestBodyRead[];
extern char  ErrorRequestFormat[];
extern char  ErrorRequestHttpMethod[];
extern char  ErrorUrlEncoding[];
extern char  SoftwareID[];
extern char  Utility[];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;

/*****************************************************************************/
/*
Process first packet read from the client.  In most cases this will contain 
the complete HTTP header and it can be processed without building up a 
buffered header.  If it does not contain the full header allocate heap memory
to contain the current packet and queue subsequent read(s), expanding the 
request header heap memory, until all is available (or it becomes completely 
rediculous!).
*/ 
 
RequestGet (struct RequestStruct *rqptr)

{
   register int  cnlcnt, cnt;
   register char  *cptr, *sptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
"RequestGet() Status: %%X%08.08X Count: %d BufferSize: %d HeaderLength: %d\n",
               rqptr->NetReadIOsb.Status,
               rqptr->NetReadIOsb.Count,
               rqptr->NetReadBufferSize,
               rqptr->RequestHeaderLength);

   /* zero bytes with a normal status is a definite no-no */
   if (VMSok (rqptr->NetReadIOsb.Status) && !rqptr->NetReadIOsb.Count)
      rqptr->NetReadIOsb.Status = SS$_ABORT;

   if (VMSnok (rqptr->NetReadIOsb.Status))
   {
      /*
         Most common cause will be broken connection.  Most common reason
         for the break ... "keep-alive" timeout on persistent connection.
         Don't bother to report!
      */
      if (rqptr->KeepAliveCount) rqptr->KeepAliveTimeout = true;
      RequestEnd (rqptr);
      return;
   }

   if (rqptr->KeepAliveCount)
   {
      /* request is underway, reinitialize the (longer) request timeout */
      if (VMSnok (status = TimeoutSet (rqptr, TIMEOUT_INPUT)))
      {
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         RequestEnd (rqptr);
         return;
      }
   }

   rqptr->BytesRx += (cnt = rqptr->NetReadIOsb.Count);

   if (rqptr->NetReadBufferSize >= 8192)
   {
      /* not likely to be a legitimate HTTP request! */
      RequestEnd (rqptr);
      return;
   }

   if (rqptr->RequestHeaderPtr == NULL)
   {
      rqptr->RequestHeaderPtr = rqptr->NetReadBufferPtr;
      rqptr->RequestHeaderLength = 0;
   }

   rqptr->RequestHeaderPtr[cnt] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", rqptr->RequestHeaderPtr); 

   /* look for the end of header blank line */
   cnlcnt = rqptr->HeaderConsecutiveNewLineCount;
   cptr = rqptr->RequestHeaderPtr;
   while (*cptr && cnlcnt < 2)
   {
      if (*cptr == '\n')
      {
         cptr++;
         cnlcnt++;
         continue;
      }
      if (*cptr == '\r' && cptr[1] == '\n')
      {
         cptr += 2;
         cnlcnt++;
         continue;
      }
      cptr++;
      cnlcnt = 0;
   }

   if (cnlcnt >= 2)
   {
      /* point (possibly back) to the start of the header */
      rqptr->RequestHeaderPtr = rqptr->NetReadBufferPtr;
      /* make the header length the actual length of the header! */
      rqptr->RequestHeaderLength = cptr - (char*)rqptr->NetReadBufferPtr;

      RequestParseAndExecute (rqptr);

      return;
   }

   rqptr->HeaderConsecutiveNewLineCount = cnlcnt;

   /**********************************************/
   /* request header not completely received yet */
   /**********************************************/

   rqptr->RequestHeaderLength += cnt;

   if (rqptr->RequestHeaderLength >= rqptr->NetReadBufferSize)
   {
      /* need more buffer space, allow one for termination of the buffer */
      rqptr->NetReadBufferSize += NetReadBufferSize;
      if ((rqptr->NetReadBufferPtr =
           HeapRealloc (rqptr, rqptr->NetReadBufferPtr,
                        rqptr->NetReadBufferSize + 1))
         == NULL)
      {
         Accounting.RequestErrorCount++;
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         RequestEnd (rqptr);
         return;
      }
   }

   /* because a realloc() may move the memory recalculate the pointer */
   rqptr->RequestHeaderPtr = rqptr->NetReadBufferPtr +
                             rqptr->RequestHeaderLength;

   /* 
      Queue an asynchronous read to get more of the header from the client.
      When the read completes call RequestSubsequent() AST completion
      function again to further process the request.
   */
   if (VMSnok (status =
       NetRead (rqptr, &RequestGet, rqptr->RequestHeaderPtr,
                rqptr->NetReadBufferSize - rqptr->RequestHeaderLength)))
   {
      /* IO failed, abandon the request */
      Accounting.RequestErrorCount++;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return;
   }
   /* further processing of the request will be AST-driven */
}

/****************************************************************************/
/*
Get the method, path and query string, and some header field line values, then
execute the request.  This function can be called from two points.  First, from
the function RequestGet(), after it has received a complete HTTP header from
the client.  Second, from the function HttpdRedirect(), where an HTTP header
has been reconstructed in the 'NetReadBufferPtr' in order to effect a local
redirection.
*/

RequestParseAndExecute (struct RequestStruct *rqptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout, "RequestParseAndExecute() size: %d\n",
               rqptr->RequestHeaderLength);
/**
      fprintf (stdout, "|%.*s|\n",
               rqptr->RequestHeaderLength, rqptr->RequestHeaderPtr);
**/
   }

   Accounting.RequestTotalCount++;
   Accounting.RequestParseCount++;
   if (rqptr->KeepAliveCount)
      Accounting.RequestKeepAliveCount++;
   if (rqptr->KeepAliveCount > Accounting.RequestKeepAliveMax)
      Accounting.RequestKeepAliveMax = rqptr->KeepAliveCount;  

   if (!RequestMethodPathQuery (rqptr)) return;
   if (!RequestFields (rqptr)) return;

   /* evaluated in order of probable occurance */
   if (!strcmp (rqptr->HttpMethodName, "GET"))
   {
      Accounting.MethodGetCount++;
      rqptr->HttpMethod = HTTP_METHOD_GET;
      RequestExecute (rqptr);
      return;
   }
   if (!strcmp (rqptr->HttpMethodName, "PUT"))
   {
      Accounting.MethodPutCount++;
      rqptr->HttpMethod = HTTP_METHOD_PUT;
      rqptr->KeepAliveRequest = false;
      RequestBodyReadBegin (rqptr);
      return;
   }
   if (!strcmp (rqptr->HttpMethodName, "POST"))
   {
      Accounting.MethodPostCount++;
      rqptr->HttpMethod = HTTP_METHOD_POST;
      rqptr->KeepAliveRequest = false;
      RequestBodyReadBegin (rqptr);
      return;
   }
   if (!strcmp (rqptr->HttpMethodName, "DELETE"))
   {
      Accounting.MethodDeleteCount++;
      rqptr->HttpMethod = HTTP_METHOD_DELETE;
      rqptr->KeepAliveRequest = false;
      RequestExecute (rqptr);
      return;
   }
   if (!strcmp (rqptr->HttpMethodName, "HEAD"))
   {
      Accounting.MethodHeadCount++;
      rqptr->HttpMethod = HTTP_METHOD_HEAD;
      RequestExecute (rqptr);
      return;
   }
/*
   if (!strcmp (rqptr->HttpMethodName, "TRACE"))
   {
      Accounting.MethodTraceCount++;
      rqptr->HttpMethod = HTTP_METHOD_TRACE;
      RequestExecute (rqptr);
      return;
   }
*/

   Accounting.RequestErrorCount++;
   rqptr->ResponseStatusCode = 501;
   ErrorGeneral (rqptr, ErrorRequestHttpMethod, __FILE__, __LINE__);
   RequestEnd (rqptr);
}

/*****************************************************************************/
/*
Read the body of a request (POST, PUT, etc., methods) into a buffer of heap
memory pointed to by 'rqptr->ContentBufferPtr'.
*/ 
 
RequestBodyReadBegin (struct RequestStruct *rqptr)

{
   int  status,
        ReadSize;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestBodyReadBegin() %d bytes\n",
               rqptr->ContentLength);

   if (rqptr->ContentBufferPtr != NULL)
   {
      /********************************/
      /* body exists from redirection */
      /********************************/

      RequestExecute (rqptr);
      return;
   }

   /****************************************/
   /* request body allowed to be this big? */
   /****************************************/

   if (Config.PutMaxKbytes &&
       rqptr->ContentLength / 1000 > Config.PutMaxKbytes)
   {
      /* local storage */
      static $DESCRIPTOR (MaxLengthFaoDsc,
      "Attempted !AZ of !UL Kbytes exceeds allowed maximum of !UL Kbytes.\0");
      char  String [128];
      $DESCRIPTOR (StringDsc, String);

      sys$fao (&MaxLengthFaoDsc, 0, &StringDsc,
               rqptr->HttpMethodName, rqptr->ContentLength / 1000,
               Config.PutMaxKbytes);
      rqptr->ResponseStatusCode = 400;
      ErrorGeneral (rqptr, String, __FILE__, __LINE__);

      /*
         Navigator 3 and MSIE 3 (all browsers?) spit if a POST is not read
         before returning an error, hence continue on if not an unreasonable
         number of bytes!  Disable this by defining NO_UPLOAD_GRACE.
         10 Mbytes doesn't seem an unreasonable limit!
      */

#ifndef NO_UPLOAD_GRACE
      if (rqptr->ContentLength < 10000000)
      {
         /* fake these for the discard AST routine checking */
         rqptr->NetReadIOsb.Status = SS$_NORMAL;
         rqptr->NetReadIOsb.Count = 1;

         if (VMSnok (status = ErrorSendToClient (rqptr, &RequestDiscardAst)))
            RequestEnd (rqptr);
      }
      else
#endif /* NO_UPLOAD_GRACE */
         if (VMSnok (status = ErrorSendToClient (rqptr, &ConcludeProcessing)))
            RequestEnd (rqptr);

      return;
   }

   /****************************/
   /* allocate buffer for body */
   /****************************/

   /*
      We're going to stick an extra null char on the end
      so we can parse it as if it's a string if we want to!
   */
   if ((rqptr->ContentBufferPtr =
        HeapAlloc (rqptr, rqptr->ContentLength+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return;
   }
   /* might as well do the above now! */
   rqptr->ContentBufferPtr[rqptr->ContentLength] = '\0';

   /*  allow for any content received along with the header */
   rqptr->ContentCount = rqptr->BytesRx - rqptr->RequestHeaderLength;
   /* ensure this does not exceed the specified content length */
   if (rqptr->ContentCount > rqptr->ContentLength)
      rqptr->ContentCount = rqptr->ContentLength;

   if (Debug)
      fprintf (stdout,
"ContentLength: %d BytesRx: %d RequestHeaderLength: %d ContentCount: %d\n",
          rqptr->ContentLength, rqptr->BytesRx,
          rqptr->RequestHeaderLength, rqptr->ContentCount);

   if (rqptr->ContentCount)
   {
      /* copy into the buffer any content received along with the header */
      memcpy (rqptr->ContentBufferPtr,
              rqptr->NetReadBufferPtr + rqptr->RequestHeaderLength,
              rqptr->ContentCount);
   }

   if (rqptr->ContentCount >= rqptr->ContentLength)
   {
      /******************************************************/
      /* looks like it all arrived with the request header! */
      /******************************************************/

      RequestExecute (rqptr);
      return;
   }

   /*******************************/
   /* read (more) from the client */
   /*******************************/

   rqptr->ContentPtr = rqptr->ContentBufferPtr + rqptr->ContentCount;
   if (rqptr->ContentLength - rqptr->ContentCount > 32767)
      ReadSize = 32767;
   else
      ReadSize = rqptr->ContentLength - rqptr->ContentCount;

   if (VMSnok (status =
       NetRead (rqptr, &RequestBodyReadAst, rqptr->ContentPtr, ReadSize)))
   {
      /* IO failed, abandon the request */
      rqptr->ErrorTextPtr = ErrorRequestBodyRead;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return;
   }
}

/*****************************************************************************/
/*
A read from the client directly into the content buffer has completed.  Check
the status of that read.  If ok check whether all of the content body has been
read.  If not read more of the content body, otherwise declare an AST to
execute the next task.
*/ 

RequestBodyReadAst (struct RequestStruct *rqptr)

{
   int  status,
        ReadSize;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
         "RequestBodyReadAst() NetReadIOsb.Status %%X%08.08X .Count %d\n",
         rqptr->NetReadIOsb.Status, rqptr->NetReadIOsb.Count);

   /* zero bytes with a normal status is a definite no-no (TGV-Multinet) */
   if (VMSok (rqptr->NetReadIOsb.Status) && !rqptr->NetReadIOsb.Count)
      rqptr->NetReadIOsb.Status = SS$_ABORT;

   if (VMSnok (rqptr->NetReadIOsb.Status))
   {
      rqptr->ErrorTextPtr = ErrorRequestBodyRead;
      ErrorVmsStatus (rqptr, rqptr->NetReadIOsb.Status, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return;
   }

   rqptr->BytesRx += rqptr->NetReadIOsb.Count;
   rqptr->ContentCount += rqptr->NetReadIOsb.Count;

   if (Debug)
   {
      fprintf (stdout, "ContentLength: %d ContentCount: %d\n",
               rqptr->ContentLength, rqptr->ContentCount);
      /** if (Debug) fprintf (stdout, "|%s|\n", rqptr->ContentPtr); **/
   }

   if (rqptr->ContentCount >= rqptr->ContentLength)
   {
      /*******************************/
      /* looks like it's all arrived */
      /*******************************/

      RequestExecute (rqptr);
      return;
   }

   /*****************************/
   /* read more from the client */
   /*****************************/

   rqptr->ContentPtr = rqptr->ContentBufferPtr + rqptr->ContentCount;
   if (rqptr->ContentLength - rqptr->ContentCount > 32767)
      ReadSize = 32767;
   else
      ReadSize = rqptr->ContentLength - rqptr->ContentCount;

   if (VMSnok (status =
       NetRead (rqptr, &RequestBodyReadAst, rqptr->ContentPtr, ReadSize)))
   {
      /* IO failed, abandon the request */
      rqptr->ErrorTextPtr = ErrorRequestBodyRead;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return;
   }
}

/*****************************************************************************/
/*
Throw-away anything read from the client.
*/ 

RequestDiscardAst (struct RequestStruct *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
         "RequestDiscardAst() NetReadIOsb.Status %%X%08.08X .Count %d\n",
         rqptr->NetReadIOsb.Status, rqptr->NetReadIOsb.Count);

   /* zero bytes with a normal status is a definite no-no (TGV-Multinet) */
   if (VMSok (rqptr->NetReadIOsb.Status) && !rqptr->NetReadIOsb.Count)
      rqptr->NetReadIOsb.Status = SS$_ABORT;

   if (VMSnok (rqptr->NetReadIOsb.Status))
   {
      RequestEnd (rqptr);
      return;
   }

   if (VMSnok (status =
       NetRead (rqptr, &RequestDiscardAst,
                rqptr->RequestHeaderPtr, rqptr->NetReadBufferSize)))
   {
      /* IO failed, abandon the request */
      RequestEnd (rqptr);
      return;
   }
}

/****************************************************************************/
/*
Parse the first <CR><LF> (strict HTTP) or newline (de facto HTTP) terminated
string from the client's request packet into HTTP method, path information and
query string.  Return true to continue processing the request, false to stop
immediately.  If false this function or any it called must have reported the
problem to the client and/or disposed of the thread.
*/ 
 
RequestMethodPathQuery (struct RequestStruct *rqptr)

{
   register unsigned char  *cptr, *rptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestMethodPathQuery()\n");

   cptr = rqptr->RequestHeaderPtr;

   zptr = (sptr = rqptr->HttpMethodName) + sizeof(rqptr->HttpMethodName);
   while (*cptr && !isspace(*cptr) && *cptr != '\r' && *cptr != '\n' &&
          sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr) sptr = rqptr->HttpMethodName;
   *sptr = '\0';

   if (!rqptr->HttpMethodName[0])
   {
      Accounting.RequestErrorCount++;
      rqptr->ResponseStatusCode = 400;
      ErrorGeneral (rqptr, ErrorRequestFormat, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return (false);
   }

   /* skip across white-space between method and URI */
   while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n') cptr++;

   /* get the path information, noting the start of the requested resource */
   rptr = zptr = cptr;
   while (*zptr && !isspace(*zptr) && *zptr != '?' &&
          *zptr != '\r' && *zptr != '\n') zptr++;

   if (cptr == zptr)
   {
      /* no path supplied */
      Accounting.RequestErrorCount++;
      rqptr->ResponseStatusCode = 400;
      ErrorGeneral (rqptr, ErrorRequestFormat, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return (false);
   }

   if ((rqptr->PathInfoPtr = sptr = HeapAlloc (rqptr, zptr-cptr+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return (false);
   }
   memcpy (sptr, cptr, rqptr->PathInfoLength = zptr-cptr);
   sptr[zptr-cptr] = '\0';
   cptr = zptr;

   /* get any query string, or create an empty one */
   if (*cptr == '?') cptr++;
   zptr = cptr;
   while (*zptr && !isspace(*zptr) && *zptr != '\r' && *zptr != '\n') zptr++;
   if ((rqptr->QueryStringPtr = sptr =
       HeapAlloc (rqptr, zptr-cptr+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return (false);
   }
   if (rqptr->QueryStringLength = zptr-cptr)
      memcpy (sptr, cptr, rqptr->QueryStringLength);
   sptr[rqptr->QueryStringLength] = '\0';

   /* store the requested resource */
   if ((rqptr->ResourcePtr = sptr = HeapAlloc (rqptr, zptr-rptr+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return (false);
   }
   memcpy (sptr, rptr, rqptr->ResourceLength = zptr-rptr);
   sptr[zptr-rptr] = '\0';

   return (true);
}

/*****************************************************************************/
/*
Scan the header buffer looking for relevant fields. Return true to continue
processing the request, false to stop immediately.  If false this function or
any it called must have reported the problem to the client and/or disposed of
the thread.
*/ 
 
boolean RequestFields (struct RequestStruct *rqptr)

{
   register char  *cptr, *hptr, *sptr, *zptr;

   int  Count,
        Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestFields()\n");

   /* ensure any old field values are ignored (in the case of redirection) */
   rqptr->HttpAcceptPtr = rqptr->HttpAuthorizationPtr =
       rqptr->HttpContentTypePtr = rqptr->HttpIfModifiedSincePtr =
      rqptr->HttpRefererPtr = rqptr->HttpUserAgentPtr = NULL;
   rqptr->ContentLength = rqptr->HttpAcceptLength =
      rqptr->IfModifiedSinceBinaryTime[0] =
      rqptr->IfModifiedSinceBinaryTime[1] = 0;

   /* the header is already null-terminated */
   hptr = rqptr->RequestHeaderPtr;
   /* look for the request-terminating blank line */
   while (*hptr)
   {
      /** if (Debug) fprintf (stdout, "hptr |%s|\n", hptr); **/

      cptr = hptr;
      while (*hptr && *hptr != '\r' && *hptr != '\n') hptr++;
      if (hptr[0] == '\n' || (hptr[0] == '\r' && hptr[1] == '\n'))
      {
         /* break if blank line (i.e. end-of-header) */
         if (cptr == hptr) break;
      }
      /* step over the carriage control at the end of the line */
      if (*hptr == '\r') hptr++;
      if (*hptr == '\n') hptr++;
/*
      if (Debug)
         fprintf (stdout, "field |%*.*s|\n", hptr-cptr, hptr-cptr, cptr);
*/

      /***********/
      /* Accept: */
      /***********/

      /* reduce function call overhead by explicitly checking the first two */
      if (toupper(cptr[0]) == 'A' && toupper(cptr[1]) == 'C' &&
          strsame (cptr, "Accept:", 7))
      {
         cptr += 7;
         while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n')
                cptr++;
         sptr = cptr;
         Count = 1;
         while (*sptr && *sptr != '\r' && *sptr != '\n')
            if (*sptr++ == ',') Count++;
         Length = sptr - cptr;
         if (rqptr->HttpAcceptPtr == NULL)
         {
            if ((rqptr->HttpAcceptPtr = sptr =
                HeapAlloc (rqptr, Length+Count*2+1)) == NULL)
            {
               ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
               RequestEnd (rqptr);
               return (false);
            }
         }
         else
         {
            if ((rqptr->HttpAcceptPtr =
                HeapRealloc (rqptr, rqptr->HttpAcceptPtr, 
                             rqptr->HttpAcceptLength+Length+Count*2+2))
                == NULL)
            {
               ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
               RequestEnd (rqptr);
               return (false);
            }
            sptr = rqptr->HttpAcceptPtr + rqptr->HttpAcceptLength;
            *sptr++ = ',';
            *sptr++ = ' ';
         }
         while (*cptr && *cptr != '\r' && *cptr != '\n')
         {
            if (!isspace(*cptr))
               *sptr++ = *cptr++;
            else
            {
               *sptr++ = *cptr++;
               while (isspace(*cptr)) cptr++;
            }
         }
         *sptr = '\0';
         rqptr->HttpAcceptLength = sptr - (char*)rqptr->HttpAcceptPtr;
         if (Debug)
            fprintf (stdout, "HttpAcceptPtr %d |%s|\n",
                     rqptr->HttpAcceptLength, rqptr->HttpAcceptPtr);
         continue;
      }

      /******************/
      /* Authorization: */
      /******************/

      /* reduce function call overhead by explicitly checking the first two */
      if (toupper(cptr[0]) == 'A' && toupper(cptr[1]) == 'U' &&
          strsame (cptr, "Authorization:", 14))
      {
         cptr += 14;
         while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n')
            cptr++;
         zptr = cptr;
         while (*zptr && *zptr != '\r' && *zptr != '\n') zptr++;
         if ((rqptr->HttpAuthorizationPtr = sptr =
             HeapAlloc (rqptr, zptr-cptr+1)) == NULL)
         {
            ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
            RequestEnd (rqptr);
            return (false);
         }
         memcpy (sptr, cptr, rqptr->HttpAuthorizationLength = zptr-cptr);
         sptr[zptr-cptr] = '\0';
         /* overwrite the authorization string in the buffer with spaces */
/*
         memset (cptr, 0x20, zptr-cptr);
*/
         if (Debug) fprintf (stdout, "Authorization |%s|\n", sptr);

         continue;
      }

      /***************/
      /* Connection: */
      /***************/

      /*
         Only check for persistent connection request if it's configured.
         Persistent connection only allowed for object retrievals.
      */
      if (Config.KeepAliveTimeoutSeconds)
      {
         /* reduce function call overhead by explicitly checking the first two */
         if (toupper(cptr[0]) == 'C' && toupper(cptr[1]) == 'O' &&
             strsame (cptr, "Connection:", 11))
         {
            cptr += 11;
            while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n')
                cptr++;
            if (toupper(cptr[0]) == 'K' && strsame (cptr, "Keep-Alive", 10))
            {
               rqptr->KeepAliveRequest = true;
               if (Debug) fprintf (stdout, "KEEP-ALIVE!\n");
            }
            continue;
         }
      }

      /*******************/
      /* Content-Length: */
      /*******************/

      /* reduce function call overhead by explicitly checking the first two */
      if (toupper(cptr[0]) == 'C' && toupper(cptr[1]) == 'O' &&
          strsame (cptr, "Content-Length:", 15))
      {
         cptr += 15;
         while (*cptr && !isdigit(*cptr) && *cptr != '\r' && *cptr != '\n')
             cptr++;
         if (!isdigit(*cptr))
         {
            Accounting.RequestErrorCount++;
            rqptr->ResponseStatusCode = 400;
            ErrorGeneral (rqptr, ErrorRequestFormat, __FILE__, __LINE__);
            RequestEnd (rqptr);
            return (false);
         }
         rqptr->ContentLength = atoi (cptr);
         if (Debug)
            fprintf (stdout, "ContentLength |%d|\n", rqptr->ContentLength);
         continue;
      }

      /*****************/
      /* Content-Type: */
      /*****************/

      /* reduce function call overhead by explicitly checking the first */
      if (toupper(cptr[0]) == 'C' && toupper(cptr[2]) == 'N' &&
          strsame (cptr, "Content-Type:", 13))
      {
         cptr += 13;
         while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n')
             cptr++;
         zptr = cptr;
         while (*zptr && *zptr != '\r' && *zptr != '\n') zptr++;
         if ((rqptr->HttpContentTypePtr = sptr =
              HeapAlloc (rqptr, zptr-cptr+1)) == NULL)
         {
            ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
            RequestEnd (rqptr);
            return (false);
         }
         memcpy (sptr, cptr, rqptr->ContentTypeLength = zptr-cptr);
         sptr[zptr-cptr] = '\0';
         if (Debug)
            fprintf (stdout, "HttpContentType |%s|\n", sptr);
         continue;
      }

      /***********/
      /* Cookie: */
      /***********/

      /* reduce function call overhead by explicitly checking the first */
      if (toupper(cptr[0]) == 'C' && toupper(cptr[2]) == 'O' &&
          strsame (cptr, "Cookie:", 7))
      {
         cptr += 7;
         while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n')
             cptr++;
         zptr = cptr;
         while (*zptr && *zptr != '\r' && *zptr != '\n') zptr++;
         if ((rqptr->HttpCookiePtr = sptr =
              HeapAlloc (rqptr, zptr-cptr+1)) == NULL)
         {
            ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
            RequestEnd (rqptr);
            return (false);
         }
         memcpy (sptr, cptr, zptr-cptr);
         sptr[zptr-cptr] = '\0';
         if (Debug) fprintf (stdout, "HttpCookie |%s|\n", sptr);
         continue;
      }

      /**********************/
      /* If-Modified-Since: */
      /**********************/

      /* reduce function call overhead by explicitly checking the first two */
      if (toupper(cptr[0]) == 'I' && toupper(cptr[1]) == 'F' &&
          strsame (cptr, "If-Modified-Since:", 18))
      {
         cptr += 18;
         while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n')
             cptr++;
         zptr = cptr;
         while (*zptr && *zptr != '\r' && *zptr != '\n') zptr++;
         if ((rqptr->HttpIfModifiedSincePtr = sptr =
              HeapAlloc (rqptr, zptr-cptr+1)) == NULL)
         {
            ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
            RequestEnd (rqptr);
            return (false);
         }
         memcpy (sptr, cptr, zptr-cptr);
         sptr[zptr-cptr] = '\0';
         if (Debug) fprintf (stdout, "HttpIfModifiedSince |%s|\n", sptr);

         if (VMSnok (HttpGmTime (sptr, &rqptr->IfModifiedSinceBinaryTime)))
         {
            if (Debug) fprintf (stdout, "If-Modified-Since: NBG!\n");
            rqptr->HttpIfModifiedSincePtr = NULL;
            rqptr->IfModifiedSinceBinaryTime[0] =
            rqptr->IfModifiedSinceBinaryTime[1] = 0;

            continue;
         }

         /* indicate that no "length=" was found (if none is found :^) */
         rqptr->IfModifiedSinceLength = -1;

         /* look for a "length=" parameter to this header field */
         while (*sptr)
         {
            while (*sptr && *sptr != ';') sptr++; 
            if (!*sptr) break;
            sptr++;
            while (*sptr && isspace(*sptr)) sptr++; 
            if (!*sptr) break;
            if (!strsame (sptr, "length", 6)) continue;
            sptr += 6;
            while (*sptr && *sptr != '=') sptr++; 
            if (!*sptr) break;
            sptr++;
            while (*sptr && isspace(*sptr)) sptr++; 
            if (!isdigit (*sptr)) continue;
            rqptr->IfModifiedSinceLength = atol(sptr);
            if (Debug)
               fprintf (stdout, "length=%d\n",
                        rqptr->IfModifiedSinceLength);
         }

         continue;
      }

      /***********/
      /* Pragma: */
      /***********/

      /* reduce function call overhead by explicitly checking the first */
      if (toupper(cptr[0]) == 'P' && toupper(cptr[1]) == 'R' &&
          strsame (cptr, "Pragma:", 7))
      {
         cptr += 7;
         while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n')
             cptr++;
         zptr = cptr;
         while (*zptr && *zptr != '\r' && *zptr != '\n') zptr++;
         if ((rqptr->HttpPragmaPtr = sptr =
              HeapAlloc (rqptr, zptr-cptr+1)) == NULL)
         {
            ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
            RequestEnd (rqptr);
            return (false);
         }
         memcpy (sptr, cptr, zptr-cptr);
         sptr[zptr-cptr] = '\0';
         if (strsame (sptr, "no-cache", -1))
         {
            rqptr->PragmaNoCache = true;
            Accounting.RequestPragmaNoCacheCount++;
         }
         if (Debug)
            fprintf (stdout, "HttpPragma |%s| no-cache: %d\n",
                     sptr, rqptr->PragmaNoCache);
         continue;
      }

      /************/
      /* Referer: */
      /************/

      /* reduce function call overhead by explicitly checking the first two */
      if (toupper(cptr[0]) == 'R' && toupper(cptr[1]) == 'E' &&
          strsame (cptr, "Referer:", 8))
      {
         cptr += 8;
         while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n')
             cptr++;
         zptr = cptr;
         while (*zptr && *zptr != '\r' && *zptr != '\n') zptr++;
         if ((rqptr->HttpRefererPtr = sptr =
             HeapAlloc (rqptr, zptr-cptr+1)) == NULL)
         {
            ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
            RequestEnd (rqptr);
            return (false);
         }
         memcpy (sptr, cptr, rqptr->HttpRefererLength = zptr-cptr);
         sptr[zptr-cptr] = '\0';
         if (Debug) fprintf (stdout, "Referer |%s|\n", sptr);
         continue;
      }

      /***************/
      /* User-Agent: */
      /***************/

      /* reduce function call overhead by explicitly checking the first two */
      if (toupper(cptr[0]) == 'U' && toupper(cptr[1]) == 'S' &&
          strsame (cptr, "User-Agent:", 11))
      {
         cptr += 11;
         while (*cptr && isspace(*cptr) && *cptr != '\r' && *cptr != '\n')
             cptr++;
         zptr = cptr;
         while (*zptr && *zptr != '\r' && *zptr != '\n') zptr++;
         if ((rqptr->HttpUserAgentPtr = sptr =
             HeapAlloc (rqptr, zptr-cptr+1)) == NULL)
         {
            ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
            RequestEnd (rqptr);
            return (false);
         }
         memcpy (sptr, cptr, rqptr->HttpUserAgentLength = zptr-cptr);
         sptr[zptr-cptr] = '\0';
         if (Debug) fprintf (stdout, "UserAgent |%s|\n", sptr);
         continue;
      }
   }

   return (true);
}

/*****************************************************************************/
/*
If a file specification does not contain a file name the function looks for a
home page in the directory.  If no home page found it generates a directory
listing ("Index of" documents).

If the file specification contains a wildcard and there is no query string a
directory listing (index) is generated.  If there is a query string it must
begin with the literal "httpd=index" for a directory listing (index) to be
generated (this allows directives in the query string to be passed to the
directory listing functions).  If it does not begin with this literal the
default query (search) script is invoked.

Detects file types that have an associated script.  If a script is returned by
ConfigContentType(), and the file specification does not include a specific
version number (even such as ";0"), an automatic redirection is generated so
that the specified document is processed by the script, and the output from
that returned to the client.  If it does contain a version number the file is
always directly returned to the client.
*/ 

int RequestExecute (struct RequestStruct *rqptr)

{
   register char  *cptr, *sptr;

   int  status,
        Count;
   char  FileSpec [1024],
         ScriptSpec [256],
         String [512];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestExecute()\n");

   /** fprintf(stdout,"|%s|\n",rqptr->RequestHeaderPtr); **/

   /* ensure any old values are ignored, in case its a redirection */
   rqptr->LocationPtr = NULL;

   /* request is underway, initialize the output (response) timeout */
   if (VMSnok (status = TimeoutSet (rqptr, TIMEOUT_OUTPUT)))
   {
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return;
   }

   /****************/
   /* map the path */
   /****************/

   FileSpec[0] = rqptr->ScriptName[0] = ScriptSpec[0] = '\0';

   /* map the path, converting it to a VMS file path (or script file name) */
   sptr = MapUrl (rqptr->PathInfoPtr, FileSpec, rqptr->ScriptName, ScriptSpec);
   if (!sptr[0] && sptr[1])
   {
      Accounting.RequestForbiddenCount++;
      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, sptr+1, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return;
   }

   if (rqptr->ScriptName[0] == '+')
   {
      rqptr->IsCgiPlusScript = true;
      rqptr->ScriptName[0] = '/';
   }
   else
      rqptr->IsCgiPlusScript = false;

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%d|%s|%s|\n",
               rqptr->PathInfoPtr, FileSpec,
               rqptr->ScriptName, rqptr->IsCgiPlusScript, ScriptSpec, sptr);

   /********************************************/
   /* check for redirection (from the mapping) */
   /********************************************/

   /* find full redirection URL (i.e. the "://" out of "http://host/path") */
   for (cptr = sptr; *cptr && *cptr != '/'; cptr++)
      if (!memcmp (cptr, "://", 3)) break;
   if (*cptr == ':')
   {
      if ((rqptr->LocationPtr = HeapAlloc (rqptr, strlen(sptr)+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         RequestEnd (rqptr);
         return;
      }
      strcpy (rqptr->LocationPtr, sptr);
      RequestEnd (rqptr);
      return;
   }

   /*********************************/
   /* authorization of request path */
   /*********************************/

   if (VMSnok (Authorize (rqptr)))
   {
      RequestEnd (rqptr);
      return;
   }

   /************************************/
   /* handled internally by the server */
   /************************************/

   if (strsame (rqptr->ResourcePtr, HttpdInternalAdmin,
                sizeof(HttpdInternalAdmin)-1))
   {
      AdminBegin (rqptr, &RequestEnd);
      return;
   }

   if (strsame (rqptr->ResourcePtr, HttpdInternalPasswordChange,
                sizeof(HttpdInternalPasswordChange)-1))
   {
      HTAdminBegin (rqptr, &RequestEnd);
      return;
   }

   /******************************/
   /* check for script execution */
   /******************************/

   if (rqptr->ScriptName[0])
   {
      /* get the path derived from the script specification */
      if ((rqptr->PathInfoPtr =
          HeapRealloc (rqptr, rqptr->PathInfoPtr, 
                       ((rqptr->PathInfoLength=strlen(sptr))+1))) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         RequestEnd (rqptr);
         return (false);
      }
      memcpy (rqptr->PathInfoPtr, sptr, rqptr->PathInfoLength+1);

      if (rqptr->PathInfoLength > 1)
      {
         /******************************************/
         /* authorization of script path parameter */
         /******************************************/

         if (VMSnok (Authorize (rqptr)))
         {
            RequestEnd (rqptr);
            return;
         }
      }

      RequestScript (rqptr, ScriptSpec, FileSpec);
      return;
   }

   /********************************/
   /* parse the file specification */
   /********************************/

   rqptr->RequestParseFab = cc$rms_fab;
   rqptr->RequestParseFab.fab$l_fna = FileSpec;
   rqptr->RequestParseFab.fab$b_fns = strlen(FileSpec);
   rqptr->RequestParseFab.fab$l_fop = FAB$M_NAM;
   rqptr->RequestParseFab.fab$l_nam = &rqptr->RequestParseNam;
   rqptr->RequestParseNam = cc$rms_nam;
   rqptr->RequestParseNam.nam$l_esa = rqptr->RequestFileName;
   rqptr->RequestParseNam.nam$b_ess = sizeof(rqptr->RequestFileName)-1;
   /* parse without disk I/O, syntax check only! */
   rqptr->RequestParseNam.nam$b_nop = NAM$M_SYNCHK;

   if (VMSnok (status = sys$parse (&rqptr->RequestParseFab, 0, 0)))
   {
      Accounting.RequestRmsErrorCount++;
      rqptr->ErrorHiddenTextPtr = FileSpec;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return;
   }

   /* 
       If a file version was explicitly specified or a wildcard
       specified in any version component then terminate after that version.
       If there is no file name or type (i.e. a directory specification)
       then terminate after the ']'.
       Otherwise terminate after the file type.
   */
   if ((rqptr->RequestParseNam.nam$l_fnb & NAM$M_EXP_VER) ||
       (rqptr->RequestParseNam.nam$l_fnb & NAM$M_WILD_VER))
      rqptr->RequestParseNam.nam$l_ver[rqptr->RequestParseNam.nam$b_ver] = '\0';
   else
   if (rqptr->RequestParseNam.nam$l_name ==  rqptr->RequestParseNam.nam$l_type)
      rqptr->RequestParseNam.nam$l_name[0] = '\0';
   else
      rqptr->RequestParseNam.nam$l_ver[0] = '\0';
   if (Debug) fprintf (stdout, "FileName |%s|\n", rqptr->RequestFileName);

   /*
      If the file name component comprises a leading hyphen followed by
      16 digits and a trailing hyphen  ("-yyyymmddhhmmsshh-") then consider
      it a temporary file and delete it on close.
   */
   if (*rqptr->RequestParseNam.nam$l_name == '-' &&
       rqptr->RequestParseNam.nam$b_name == 18)
   {
      for (cptr = rqptr->RequestParseNam.nam$l_name+1;
           *cptr && isdigit(*cptr);
           cptr++);
      if (*cptr == '-')
      {
         rqptr->DeleteOnClose = true;
         rqptr->ResponsePreExpired = PRE_EXPIRE_DELETE_ON_CLOSE;
      }
   }

   /* all these "file write" methods are handled by the one set of functions */
   if (rqptr->HttpMethod == HTTP_METHOD_PUT ||
       rqptr->HttpMethod == HTTP_METHOD_POST ||
       rqptr->HttpMethod == HTTP_METHOD_DELETE)
   {
      PutBegin (rqptr, &RequestEnd);
      return;
   }

   if ((tolower(rqptr->QueryStringPtr[0]) == 'h' &&
        strsame (rqptr->QueryStringPtr, "httpd=index", 11)) ||
       ((rqptr->RequestParseNam.nam$l_fnb & NAM$M_WILDCARD) &&
        !rqptr->QueryStringPtr[0]))
   {
      /******************************/
      /* generate directory listing */
      /******************************/

      if (Config.DirWildcardEnabled)
      {
         DirBegin (rqptr, &RequestEnd, FileSpec, rqptr->QueryStringPtr);
         return;
      }
      else
      {
         Accounting.RequestForbiddenCount++;
         rqptr->ResponseStatusCode = 501;
         ErrorGeneral (rqptr, ErrorFacilityDisabled, __FILE__, __LINE__);
         RequestEnd (rqptr);
         return;
      }
   }

   if (!rqptr->RequestParseNam.nam$b_name &&
       rqptr->RequestParseNam.nam$b_type == 1 &&
       !rqptr->QueryStringPtr[0])
   {
      /*********************************************/
      /* no file name supplied, look for home page */
      /*********************************************/

      RequestHomePage (rqptr);
      return;
   }

   /********************************/
   /* get the content-type details */
   /********************************/

   ConfigContentType (&rqptr->ContentType, rqptr->RequestParseNam.nam$l_type);

   if (!isdigit(rqptr->RequestParseNam.nam$l_ver[1]))
   {
      /* 
         If a file specification does not include a specific version
         number (which indicates send-me-this-file-regardless) and
         ConfigContentType() has returned a script name, indicating
         an automatic script handling for this file type, then redirect.
      */
      if (rqptr->ContentType.AutoScriptNamePtr[0] == '/')
      {
         /********************************/
         /* automatic script redirection */
         /********************************/

         Accounting.DoAutoScriptCount++;

         Count = strlen(rqptr->ContentType.AutoScriptNamePtr);
         if (rqptr->PathInfoPtr != NULL) Count += rqptr->PathInfoLength;
         if (rqptr->QueryStringPtr[0])
         {
             /* allow one extra character for the query-string leading '?' */
             Count++;
             Count += rqptr->QueryStringLength;
         }
         /* allow for the terminating null */
         Count++;

         if ((rqptr->LocationPtr = sptr = HeapAlloc (rqptr, Count)) == NULL)
         {
            ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
            RequestEnd (rqptr);
            return;
         }

         cptr = rqptr->ContentType.AutoScriptNamePtr;
         while (*cptr) *sptr++ = *cptr++;
         if (rqptr->PathInfoPtr != NULL)
         {
            cptr = rqptr->PathInfoPtr;
            while (*cptr) *sptr++ = *cptr++;
         }
         if (rqptr->QueryStringPtr[0])
         {
            *sptr++ = '?';
            cptr = rqptr->QueryStringPtr;
            while (*cptr) *sptr++ = *cptr++;
         }
         *sptr = '\0';
         if (Debug)
            fprintf (stdout, "LocationPtr |%s|\n", rqptr->LocationPtr);

         /* dispose of thread function does the actual redirection */
         RequestEnd (rqptr);
         return;
      }
   }

   /**************************************/
   /* internally processed content-types */
   /**************************************/

   if (strsame (rqptr->ContentType.ContentTypePtr, Config.IsMapContentType, -1))
   {
      /* image mapping configuration file (has a query string!) */
      IsMapBegin (rqptr, &RequestEnd, rqptr->RequestFileName);
      return;
   }

   /******************************************************************/
   /* query string (without script name), an implicit keyword search */
   /******************************************************************/

   if (rqptr->QueryStringPtr[0])
   {
      /*
         No need to authorize the path here.  No script component was
         supplied in the HTTP request, but a query string was, making
         this an ISINDEX search on the already authorized path.
      */
      strcpy (rqptr->ScriptName, Config.KeywordSearch);
      MapUrl_UrlToVms (rqptr->ScriptName, ScriptSpec);
      RequestScript (rqptr, ScriptSpec, FileSpec);
      return;
   }

   /****************************************/ 
   /* otherwise ... send the file contents */
   /****************************************/ 

   RequestFile (rqptr, &RequestEnd, NULL);
}

/*****************************************************************************/
/*
This function can be called one or more time per request.  It steps through all
possible home page files until either one is found or a default directory
listing is generated.
*/ 

int RequestHomePage (struct RequestStruct *rqptr)

{
   register char  *cptr;

   char  *HomePageNamePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestHomePage() %d %d\n",
               rqptr->RequestHomePageIndex, rqptr->ResponseHeaderPtr);

   if (rqptr->ResponseHeaderPtr != NULL || rqptr->ErrorMessagePtr != NULL)
   {
      RequestEnd (rqptr);
      return;
   }

   if (!*(HomePageNamePtr = ConfigHomePage(rqptr->RequestHomePageIndex++)))
   {
      /********************************************/
      /* no home page, generate directory listing */
      /********************************************/

      if (!Config.DirAccess)
      {
         /* directory listing is disabled */
         rqptr->ErrorTextPtr = rqptr->PathInfoPtr;
         rqptr->ErrorHiddenTextPtr = rqptr->RequestParseNam.nam$l_dev;
         ErrorVmsStatus (rqptr, RMS$_FNF, __FILE__, __LINE__);
         RequestEnd (rqptr);
         return;
      }

      rqptr->RequestParseNam.nam$l_name[0] = '\0';
      DirBegin (rqptr, &RequestEnd,
                rqptr->RequestParseNam.nam$l_dev, rqptr->QueryStringPtr);
      return;
   }

   /********************************/
   /* generate home page file name */
   /********************************/

   /* overwrite any existing file name in the NAM block */
   strcpy (rqptr->RequestParseNam.nam$l_name, HomePageNamePtr);
   /* locate the start of the file type */
   for (cptr = rqptr->RequestParseNam.nam$l_name;
        *cptr && *cptr != '.';
        cptr++);
   /* set the mime content type for this file type */
   ConfigContentType (&rqptr->ContentType, cptr);

   RequestFile (rqptr, &RequestEnd, &RequestHomePage);
}

/*****************************************************************************/
/*
Send a file to the client.  Special cases are menu and pre-processed HTML.  
*/ 

int RequestFile
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
void *FileNotFoundFunction
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestFile() |%s|%s|\n",
               rqptr->RequestFileName, rqptr->ContentType.ContentTypePtr);

   if (strsame (rqptr->ContentType.ContentTypePtr, Config.MenuContentType, -1))
   {
      MenuBegin (rqptr, NextTaskFunction, FileNotFoundFunction,
                 rqptr->RequestFileName);
      return;
   }

   if (strsame (rqptr->ContentType.ContentTypePtr, Config.SsiContentType, -1))
   {
      SsiBegin (rqptr, NextTaskFunction, FileNotFoundFunction,
                rqptr->RequestFileName);
      return;
   }

   FileBegin (rqptr, NextTaskFunction, FileNotFoundFunction,
              rqptr->RequestFileName, rqptr->ContentType.ContentTypePtr,
              false, false);
}

/*****************************************************************************/
/*
Process request by executing a script.  The 'FileSpec' was the non-script-name 
portion of the path, returned by the original MapUrl(), attempt to parse this
as a VMS file specification, but if that fails (and it well might not be a
file specification) then do not report it.
*/ 

RequestScript
(
struct RequestStruct *rqptr,
char *ScriptSpec,
char *FileSpec
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestScript() |%s|%s|%s|\n",
               rqptr->ScriptName, FileSpec, rqptr->PathInfoPtr);

   if (FileSpec[0])
   {
      /* parse the file specification */
      rqptr->RequestParseFab = cc$rms_fab;
      rqptr->RequestParseFab.fab$l_fna = FileSpec;
      rqptr->RequestParseFab.fab$b_fns = strlen(FileSpec);
      rqptr->RequestParseFab.fab$l_fop = FAB$M_NAM;
      rqptr->RequestParseFab.fab$l_nam = &rqptr->RequestParseNam;

      rqptr->RequestParseNam = cc$rms_nam;
      rqptr->RequestParseNam.nam$l_esa = rqptr->RequestFileName;
      rqptr->RequestParseNam.nam$b_ess = sizeof(rqptr->RequestFileName)-1;
      /* parse without disk I/O, syntax check only! */
      rqptr->RequestParseNam.nam$b_nop = NAM$M_SYNCHK;

      status = sys$parse (&rqptr->RequestParseFab, 0, 0);
      if (VMSok (status))
      {
         /* if its a directory specification then terminate after the ']' */
         if (rqptr->RequestParseNam.nam$l_name == 
             rqptr->RequestParseNam.nam$l_type)
            *rqptr->RequestParseNam.nam$l_name = '\0';
         else
            rqptr->RequestParseNam.nam$l_ver
               [rqptr->RequestParseNam.nam$b_ver] = '\0';
         ConfigContentType (&rqptr->ContentType,
                            rqptr->RequestParseNam.nam$l_type);
      }
      else
      {
         /* local storage */
         register char  *cptr, *sptr, *zptr;

         zptr = (sptr = rqptr->RequestFileName) +
                sizeof(rqptr->RequestFileName);
         for (cptr = FileSpec; *cptr && sptr < zptr; *sptr++ = *cptr++);
         if (sptr >= zptr) sptr = rqptr->RequestFileName;
         *sptr = '\0';
         while (sptr > rqptr->RequestFileName && *sptr != '.') sptr--;
         if (*sptr == '.') ConfigContentType (&rqptr->ContentType, sptr);
      }
   }
   else
      rqptr->RequestFileName[0] = '\0';

   if (Debug)
      fprintf (stdout, "rqptr->RequestFileName |%s|\n", rqptr->RequestFileName);

   /* check for internal server "scripts" */
   if (strsame (rqptr->ScriptName, HttpdInternalScriptUpd,
                sizeof(HttpdInternalScriptUpd)-1) ||
       strsame (rqptr->ScriptName, HttpdInternalScriptTree,
                sizeof(HttpdInternalScriptTree)-1))      
   {
      UpdBegin (rqptr, &RequestEnd);
      return;
   }
   else
   {
      DclBegin (rqptr, &RequestEnd, NULL, ScriptSpec);
      return;
   }
}

/*****************************************************************************/
/*
*/ 

int RequestEnd (struct RequestStruct *rqptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestEnd()\n");

   ConcludeProcessing (rqptr);
}

/*****************************************************************************/

