/*****************************************************************************/
/*
                                   SSI.c

This module implements a full multi-threaded, AST-driven, asynchronous HTML
Server Side Includes pre-processor.  The AST-driven nature makes the code a
little more difficult to follow, but creates a powerful, event-driven,
multi-threaded server.  All of the necessary functions implementing this module
are designed to be non-blocking. 

This module uses the NetBufferOutput() function to buffer any output it can
into larger chunks before sending it to the client.

An SSI directive CAN BE SPLIT OVER MULTIPLE LINES, up to a maximum dictated by
the size of the directive buffer (currently approx. 255 characters).

Included files are automatically encapsulated in <PRE></PRE> tags and HTML-
escaped if not "text/html" content-type (i.e. if they are "text/plain" 
content-type).  They can be forced to be directly included by including a 
'par="text/html"' parameter.  Conversly, "text/html" files can be forced to be 
included as plain-text using the 'par="text/plain"' parameter. 

Including another SSI file (".shtml") will only have its contents included, it
WILL NOT be preprocessed. 

Format ('fmt=""') specifications for time values follow those allowed by 
strftime().  If none is specified it defaults to a fairly standard looking 
VMS-style time.

*PRIVILEGED* directives are only allowed in documents owned by SYSTEM ([1,4])
and that are NOT world writable!


DIRECTIVES
----------

<!--#accesses [since=""] [timefmt="[]"] -->   number of accesses of document

<!--#config errmsg="" -->               (not implemented, ignored)
<!--#config timefmt="" -->              set default time format
<!--#config sizefmt="" -->              set file size output format

NOTE: the "dcl" and "exec" directives are identical

<!--#dir file="" [par=""] -->           directory (index of) file spec
<!--#dir virtual="" [par=""] -->        directory (index of) virtual spec
<!--#index file="" [par=""] -->         synonym for the above
<!--#index virtual="" [par=""] -->      synonym for the above

<!--#echo created[=fmt] -->             current document creation date/time
<!--#echo date_local[=fmt] -->          current local date/time
<!--#echo date_gmt[=fmt] -->            current GMT date/time
<!--#echo document_name -->             current document URL path
<!--#echo file_name -->                 current document VMS file path
<!--#echo last_modified[=fmt] -->       current document last modified date

<!--#exec dir="" [par=""] -->           DIRECTORY file-spec [qualifiers]
<!--#exec vdir="" [par=""] -->          DIRECTORY virtual-file-spec [qualifiers]
<!--#exec show="" -->                   SHOW command
<!--#exec say="" -->                    WRITE SYS$OUTPUT 'anything'

<!--#exec exec="" -->                   *PRIVILEGED* execute any DCL command
<!--#exec file="" [par=""] -->          *PRIVILEGED* execute command procedure
<!--#exec run="" [par=""] -->           *PRIVILEGED* execute image
<!--#exec virtual="" [par=""] -->       *PRIVILEGED* execute command procedure
<!--#exec vrun="" [par=""] -->          *PRIVILEGED* execute virtual image

<!--#fcreated file="" [fmt=""] -->      specified file creation date/time
<!--#fcreated virtual="" [fmt=""] -->   specified URL document creation date
<!--#flastmod file="" [fmt=""] -->      specified file last modified date/time
<!--#flastmod virtual="" [fmt=""] -->   specified URL document last modified 
<!--#fsize file="" -->                  specified file size (bytes, Kb, Mb)
<!--#fsize virtual="" -->               specified URL document size

<!--#include file="" [type=""] -->      include the file's contents
<!--#include virtual="" [type=""] -->   include the URL document contents


VERSION HISTORY
---------------
27-FEB-97  MGD  delete on close for "temporary" files
01-FEB-97  MGD  HTTPd version 4 (also changed module name to SSI.c);
                extensive rework of some functions;
                directives now allowed to span multiple lines
04-JUN-96  MGD  bugfix; SsiFileDetails() error reporting
01-DEC-95  MGD  new for HTTPd version 3
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <time.h>

/* VMS related header files */
#include <atrdef.h>
#include <descrip.h>
#include <fibdef.h>
#include <iodef.h>
#include <prvdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>

/* application-related header files */
#include "wasd.h"
#include "dcl.h"
#include "dir.h"
#include "error.h"
#include "file.h"
#include "httpd.h"
#include "mapurl.h"
#include "net.h"
#include "Ssi.h"
#include "support.h"

#define DefaultTimeFormat "%Od-%b-%Y %T"  /* a little VMSish! */
#define DefaultSizeFormat "abbrev"

#define FILE_FCREATED 1
#define FILE_FLASTMOD 2
#define FILE_FSIZE 3

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern char  ErrorFacilityDisabled[];
extern char  ErrorStringSize[];
extern char  SoftwareID[];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;

/*****************************************************************************/
/*
Open the specified file.  If there is a problem opening it then just return 
the status, do not report an error or anything.  This is used to determine 
whether the file exists, as when attempting to open one of multiple, possible 
home pages.

As pre-processed HTML files are by definition dynamic, no check of any
"If-Modified-Since:" request field date/time is made, and no "Last-Modified:" 
field is included in any response header. 

When successfully opened and connected generate an HTTP header, if required.
Once open and connected the pre-processing becomes AST-driven.
*/

SsiBegin
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
void *FileOpenErrorFunction,
char *FileName
)
{
   static $DESCRIPTOR (CommentedInfoFaoDsc,
"<!!-- Server=\"!AZ\" -->\n\
<!!-- VMS=\"!AZ\" -->\n");
   static $DESCRIPTOR (StringDsc, "");

   register char  *cptr, *sptr, *zptr;
   register struct SsiTaskStruct  *tkptr;

   int  status;
   unsigned short  CommentLength;
   char  CommentedInfo [256],
         ExpandedFileName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SsiBegin() |%s|\n", FileName);

   if (rqptr->ErrorMessagePtr != NULL)
   {
      /* previous error, cause threaded processing to unravel */
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   if (!Config.SsiEnabled)
   {
      Accounting.RequestForbiddenCount++;
      rqptr->ResponseStatusCode = 501;
      ErrorGeneral (rqptr, ErrorFacilityDisabled, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   if (rqptr->SsiTaskPoolPtr == NULL)
   {
      if (Debug) fprintf (stdout, "tkptr from HEAP\n");
      /* allocate heap memory for the SSI task structure */
      if ((tkptr = (struct SsiTaskStruct*)
           HeapAlloc (rqptr, sizeof(struct SsiTaskStruct))) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }
   }
   else
   {
      if (Debug) fprintf (stdout, "tkptr from POOL\n");
      /* get the top entry from the pool of previously used structures */
      tkptr = rqptr->SsiTaskPoolPtr;
      /* set the top of pool to the next (if any) */
      rqptr->SsiTaskPoolPtr = tkptr->SsiTaskPtr;
      /* clear out any previous content */
      memset (tkptr, 0, sizeof(struct SsiTaskStruct));
   }

   /* buffer the previous SSI task (if any) */
   tkptr->SsiTaskPtr = rqptr->SsiTaskPtr;
   /* set the current SSI task */
   rqptr->SsiTaskPtr = tkptr;

   /* buffer the pointer to the next task */
   tkptr->NextTaskFunction = NextTaskFunction;

   /* buffer the escape-HTML flag */
   tkptr->NetBufferOutputEscapeHtml = rqptr->NetBufferOutputEscapeHtml;

   cptr = FileName;
   zptr = (sptr = tkptr->FileName) + sizeof(tkptr->FileName);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }
   *sptr = '\0';
   tkptr->FileNameLength = sptr - tkptr->FileName;

   tkptr->FileFab = cc$rms_fab;
   tkptr->FileFab.fab$b_fac = FAB$M_GET;
   tkptr->FileFab.fab$l_fna = tkptr->FileName;  
   tkptr->FileFab.fab$b_fns = tkptr->FileNameLength;
   tkptr->FileFab.fab$b_shr = FAB$M_SHRGET;

   /* initialize the NAM block */
   tkptr->FileNam = cc$rms_nam;
   tkptr->FileFab.fab$l_nam = &tkptr->FileNam;
   tkptr->FileNam.nam$l_esa = ExpandedFileName;
   tkptr->FileNam.nam$b_ess = sizeof(ExpandedFileName)-1;

   /* initialize the date extended attribute block */
   tkptr->FileXabDat = cc$rms_xabdat;
   tkptr->FileFab.fab$l_xab = &tkptr->FileXabDat;

   /* initialize and link in the protection extended attribute block */
   tkptr->FileXabPro = cc$rms_xabpro;
   tkptr->FileXabDat.xab$l_nxt = &tkptr->FileXabPro;

   /* "temporary" file, automatic delete on closing it */
   if (rqptr->DeleteOnClose) tkptr->FileFab.fab$l_fop |= FAB$M_DLT;

   if (VMSnok (status = sys$open (&tkptr->FileFab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$open() %%X%08.08X\n", status);

      /* if its a search list treat directory not found as if file not found */
      if ((tkptr->FileNam.nam$l_fnb & NAM$M_SEARCH_LIST) && status == RMS$_DNF)
         status = RMS$_FNF;

      if (FileOpenErrorFunction != NULL)
      {
         /* do not report the error, the alternate function will handle it */
         tkptr->NextTaskFunction = FileOpenErrorFunction;
      }
      else
      {
         if (!rqptr->AccountingDone)
            rqptr->AccountingDone = ++Accounting.DoSsiCount;
         rqptr->ErrorTextPtr = MapUrl (NULL, tkptr->FileName, NULL, NULL);
         rqptr->ErrorHiddenTextPtr = tkptr->FileName;
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      }

      SsiEnd (rqptr);
      return;
   }

   if (!rqptr->AccountingDone)
      rqptr->AccountingDone = ++Accounting.DoSsiCount;

   /***************/
   /* file exists */
   /***************/

   tkptr->FileRab = cc$rms_rab;
   tkptr->FileRab.rab$l_fab = &tkptr->FileFab;
   /* 2 buffers and read ahead performance option */
   tkptr->FileRab.rab$b_mbf = 2;
   tkptr->FileRab.rab$l_rop = RAB$M_RAH;
   tkptr->FileRab.rab$l_ubf = tkptr->ReadBuffer;
   /* allow two bytes for carriage control and terminating null */
   tkptr->FileRab.rab$w_usz = sizeof(tkptr->ReadBuffer)-2;

   if (VMSnok (status = sys$connect (&tkptr->FileRab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
      rqptr->ErrorTextPtr = MapUrl (NULL, tkptr->FileName, NULL, NULL);
      rqptr->ErrorHiddenTextPtr = tkptr->FileName;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   /********************/
   /* begin processing */
   /********************/

   if (rqptr->ResponseHeaderPtr == NULL)
   {
      if (Config.IncludeCommentedInfo &&
          rqptr->HttpMethod == HTTP_METHOD_GET)
      {
         /*************************/
         /* commented information */
         /*************************/

         StringDsc.dsc$a_pointer = CommentedInfo;
         StringDsc.dsc$w_length = sizeof(CommentedInfo)-1;
         sys$fao (&CommentedInfoFaoDsc, &CommentLength, &StringDsc,
                  SoftwareID, tkptr->FileName);
         CommentedInfo[CommentLength] = '\0';
      }
      else
         CommentLength = 0;
      if (Debug) fprintf (stdout, "CommentLength: %d\n", CommentLength);

      rqptr->ResponsePreExpired = PRE_EXPIRE_SSI;
      rqptr->ResponseStatusCode = 200;
      if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
      {
         SsiEnd (rqptr);
         return;
      }

      if (CommentLength && rqptr->HttpMethod != HTTP_METHOD_HEAD)
      {
         /* sneakily append the commented info to the response header :^) */
         if ((rqptr->ResponseHeaderPtr =
              HeapStringAppend (rqptr, rqptr->ResponseHeaderPtr,
                                CommentedInfo, CommentLength)) == NULL)
         {
            SsiEnd (rqptr);
            return;
         }
      }
      rqptr->ResponseHeaderLength += CommentLength;

      if (rqptr->HttpMethod == HTTP_METHOD_HEAD)
      {
         SsiEnd (rqptr);
         return;
      }
   }

   tkptr->TimeFmtPtr = DefaultTimeFormat;
   tkptr->SizeFmtPtr = DefaultSizeFormat;

   /* set then RAB user context storage to the client thread pointer */
   tkptr->FileRab.rab$l_ctx = rqptr;

   tkptr->LineNumber = 0;

   SsiNextRecord (rqptr);
}

/*****************************************************************************/
/*
*/ 

SsiEnd (struct RequestStruct *rqptr)

{
   register struct SsiTaskStruct  *tkptr;

   int  status,
        SetPrvStatus;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SsiEnd()\n");

   tkptr = rqptr->SsiTaskPtr;

   if (tkptr->FileFab.fab$w_ifi)
   {
      if (rqptr->DeleteOnClose)
      {
         /* use SYSPRV to ensure deletion-on-close of file */
         EnableSysPrv ();
         sys$close (&tkptr->FileFab, 0, 0);
         DisableSysPrv ();
      }
      else
         sys$close (&tkptr->FileFab, 0, 0);
   }

   /* restore the escape-HTML flag */
   rqptr->NetBufferOutputEscapeHtml = tkptr->NetBufferOutputEscapeHtml;

   /* restore SSI task (if any) */
   rqptr->SsiTaskPtr = tkptr->SsiTaskPtr;
   /* new top points to the old top (if any) */
   tkptr->SsiTaskPtr = rqptr->SsiTaskPoolPtr;
   /* top of the pool now points at the one just finished with */
   rqptr->SsiTaskPoolPtr = tkptr;

   /* declare the next task */
   if (VMSnok (status = sys$dclast (tkptr->NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
}

/*****************************************************************************/
/*
Queue an asynchronous read of the next pre-processed file record.
*/ 
 
int SsiNextRecord (struct RequestStruct *rqptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SsiNextRecord()\n");

   /* queue the next record read */
   sys$get (&rqptr->SsiTaskPtr->FileRab, &SsiNextRecordAST, &SsiNextRecordAST);
}

/*****************************************************************************/
/*
The asynchronous read of the next pre-processed file record has completed.
*/ 

SsiNextRecordAST (struct RAB *RabPtr)

{
   register char  *cptr, *sptr;
   register struct SsiTaskStruct  *tkptr;

   int  status;
   struct RequestStruct  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
      "SsiNextRecordAST() sts: %%X%08.08X stv: %%X%08.08X\n",
      RabPtr->rab$l_sts, RabPtr->rab$l_stv);

   /* get the pointer to the thread from the RAB user context storage */
   rqptr = RabPtr->rab$l_ctx;

   tkptr = rqptr->SsiTaskPtr;

   if (VMSnok (tkptr->FileRab.rab$l_sts))
   {
      if (tkptr->FileRab.rab$l_sts == RMS$_EOF)
      {
         /***************/
         /* end of file */
         /***************/

         SsiEnd (rqptr);
         return;
      }

      /**********************/
      /* error reading file */
      /**********************/

      rqptr->ErrorTextPtr = MapUrl (NULL, tkptr->FileName, NULL, NULL);
      rqptr->ErrorHiddenTextPtr = tkptr->FileName;
      ErrorVmsStatus (rqptr, tkptr->FileRab.rab$l_sts, __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   /*****************************/
   /* process the record (line) */
   /*****************************/

   /* first check if any error occured during any previous processing */
   if (rqptr->ErrorMessagePtr != NULL)
   {
      SsiEnd (rqptr);
      return;
   }

   tkptr->LineNumber++;

   /* terminate the line */
   tkptr->FileRab.rab$l_ubf[tkptr->FileRab.rab$w_rsz] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", tkptr->FileRab.rab$l_ubf);

   /* this pointer keeps track of the current parse position in the line */
   tkptr->LineParsePtr = tkptr->FileRab.rab$l_ubf;

   /* begin parsing the line looking for commented SSI directives */
   SsiParseLine (rqptr);

   /*
      Return to the AST-interrupted routine now.  Continued processing is
      event-driven.  Routine completion ASTs drive the pre-processing.
   */
}

/*****************************************************************************/
/*
'tkptr->LineParsePtr' points to the currently parsed-up-to position on 
the line.  Continue parsing that line looking for preprocessor directives.
*/ 

SsiParseLine (struct RequestStruct *rqptr)

{
   register char  *cptr, *sptr;
   register struct SsiTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiParseLine()\n|%s|\n",
               rqptr->SsiTaskPtr->LineParsePtr);

   tkptr = rqptr->SsiTaskPtr;

   /* first check if any error occured during any previous processing */
   if (rqptr->ErrorMessagePtr != NULL)
   {
      SsiEnd (rqptr);
      return;
   }

   if (tkptr->DirectivePtr == NULL)
   {
      /* ensure the escape-HTML flag is off from any previous processing */
      rqptr->NetBufferOutputEscapeHtml = false;

      /* find the start of any directive */
      for (cptr = sptr = tkptr->LineParsePtr; *sptr; sptr++)
      {
         if (*sptr != '<') continue;
         if (memcmp (sptr, "<!--#", 5)) continue;
         break;
      }

      if (*sptr)
      {
         /*****************/
         /* SSI directive */
         /*****************/
   
         tkptr->DirectiveLineNumber = tkptr->LineNumber;
         tkptr->LineParsePtr = sptr;
         if (sptr - cptr)
         {
            /* buffer the part line */
            if (VMSnok (NetBufferOutput (rqptr, &SsiDirective,
                                         cptr, sptr-cptr)))
               SsiEnd (rqptr);
            return;
         }
         SsiDirective (rqptr);
         return;
      }
      else
      {
         /*******************************/
         /* send line (or part thereof) */
         /*******************************/

         /* end-of-line, ensure it has correct carriage control */
         if (sptr > tkptr->FileRab.rab$l_ubf)
         {
            if (sptr[-1] != '\n')
            {
               *sptr++ = '\n';
               *sptr = '\0';
            }
         }
         else
         {
            /* blank (empty) line */
            *sptr++ = '\n';
            *sptr = '\0';
         }

         if (VMSnok (NetBufferOutput (rqptr, &SsiNextRecord,
                                      cptr, sptr-cptr)))
            SsiEnd (rqptr);
         return;
      }
   }
   else
   {
      /* continue to parse directive */
      SsiDirective (rqptr);
   }
}

/*****************************************************************************/
/*
'tkptr->LineParsePtr' points to the start of a pre-processor directive.  
Parse that directive and execute it, or provide an error message.
*/ 

SsiDirective (struct RequestStruct *rqptr)

{
   register char  *cptr, *sptr, *zptr;
   register struct SsiTaskStruct  *tkptr;

   boolean  ok;
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiDirective()\n|%s|\n",
               rqptr->SsiTaskPtr->LineParsePtr);

   tkptr = rqptr->SsiTaskPtr;

   if (tkptr->DirectivePtr == NULL)
   {
      tkptr->DirectivePtr = tkptr->Directive;
      cptr = tkptr->LineParsePtr + 4;
   }
   else
      cptr = tkptr->LineParsePtr;

   /* point at the directive buffer and set a pointer to the end of buffer */
   sptr = tkptr->DirectivePtr;
   zptr = tkptr->Directive + sizeof(tkptr->Directive);

   if (tkptr->DirectivePtr > tkptr->Directive)
   {
      /* continued lines introduce an implied space */
      if (sptr < zptr) *sptr++ = ' ';
   }

   while (*cptr && sptr < zptr)
   {
      while (*cptr && *cptr != '-' && sptr < zptr) *sptr++ = *cptr++;
      if (cptr[1] == '-' && cptr[2] == '>') break;
      if (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }
   if (sptr >= zptr)
   {
      SsiExplainError (rqptr, "Directive not terminated, or too large",
                       __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }
   *(tkptr->DirectivePtr = sptr) = '\0';
   if (Debug) fprintf (stdout, "tkptr->Directive |%s|\n", tkptr->Directive);

   if (!*cptr)
   {
      /* got to the end of the line before seeing a directive terminator */
      SsiNextRecord (rqptr);
      return;
   }

   /* note the character position immediately following the terminator */
   tkptr->LineParsePtr = cptr + 3;

   cptr = tkptr->Directive;

   if (strsame (cptr, "#ACCESSES ", 10))
      SsiAccesses (rqptr);
   else
   if (strsame (cptr, "#CONFIG ", 8))
      SsiConfig (rqptr);
   else
   if (strsame (cptr, "#DCL ", 5) || strsame (cptr, "#EXEC ", 6))
      SsiDcl (rqptr);
   else
   if (strsame (cptr, "#DIR ", 5) || strsame (cptr, "#INDEX ", 7))
      SsiDir (rqptr);
   else
   if (strsame (cptr, "#ECHO ", 6))
      SsiEcho (rqptr);
   else
   if (strsame (cptr, "#FCREATED ", 10))
      SsiFCreated (rqptr);
   else
   if (strsame (cptr, "#FLASTMOD ", 10))
      SsiFLastMod (rqptr);
   else
   if (strsame (cptr, "#FSIZE ", 7))
      SsiFSize (rqptr);
   else
   if (strsame (cptr, "#INCLUDE ", 9))
      SsiIncludeFile (rqptr);
   else
   {
      SsiExplainError (rqptr, "Unknown directive", __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   /* indicate that there is no current directive being buffered */
   tkptr->DirectivePtr = NULL;
}

/*****************************************************************************/
/*
Number of times this document has been accessed.

This operation is fairly expensive in terms of I/O, and because it is atomic 
introduces a fair degree of granularity.  Its expense means the "#accesses" 
functionality should be used sparingly.  This functionality can be disabled. 
*/ 

SsiAccesses (struct RequestStruct *rqptr)

{
   static $DESCRIPTOR (AccessesFaoDsc, "!UL!AZ!AZ");
   static $DESCRIPTOR (StringDsc, "");

   register char  *dptr;

   boolean  SupplySinceTime;
   int  status;
   unsigned short  Length;
   unsigned long  AccessCount;
   unsigned long  SinceBinTime [2];
   char  SinceText [256],
         String [1024],
         TimeFormat [256],
         TimeString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiAccesses() |%s|\n", rqptr->SsiTaskPtr->Directive);

   if (!Config.SsiAccessesEnabled)
   {
      SsiExplainError (rqptr, "<I>#accesses</I> directive disabled",
                       __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   SinceText[0] = TimeFormat[0] = TimeString[0] = '\0';
   SupplySinceTime = false;

   dptr = rqptr->SsiTaskPtr->Directive;
   while (*dptr && !isspace(*dptr)) dptr++;

   while (*dptr)
   {
      while (*dptr && isspace(*dptr)) dptr++;
      if (!*dptr) break;

      if (strsame (dptr, "SINCE=", 6))
      {
         dptr += SsiGetTagValue (rqptr, dptr, SinceText);
         SupplySinceTime = true;
      }
      else
      if (strsame (dptr, "TIMEFMT=", 8))
         dptr += SsiGetTagValue (rqptr, dptr, TimeFormat);
      else
      {
         SsiExplainError (rqptr, "Unknown tag", __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
   }

   if (VMSnok (SsiAccessCount (rqptr, &AccessCount, &SinceBinTime)))
   {
      SsiEnd (rqptr);
      return;
   }

   if (SupplySinceTime)
      if (!SsiTimeString (rqptr, &SinceBinTime, TimeFormat,
                            TimeString, sizeof(TimeString)))
      {
         SsiEnd (rqptr);
         return;
      }

   StringDsc.dsc$a_pointer = String;
   StringDsc.dsc$w_length = sizeof(String)-1;
   status = sys$fao (&AccessesFaoDsc, &Length, &StringDsc,
                     AccessCount, SinceText, TimeString);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$fao()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }
   String[Length] = '\0';

   if (VMSnok (NetBufferOutput (rqptr, &SsiParseLine, String, Length)))
      SsiEnd (rqptr);
}

/*****************************************************************************/
/*
Config sets the default behaviour for a specific action for the rest of the 
document.
*/ 

SsiConfig (struct RequestStruct *rqptr)

{
   register char  *cptr, *dptr;
   register struct SsiTaskStruct  *tkptr;

   int  status;
   char  TagValue [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiConfig() |%s|\n", rqptr->SsiTaskPtr->Directive);

   tkptr = rqptr->SsiTaskPtr;

   dptr = tkptr->Directive;
   while (*dptr && !isspace(*dptr)) dptr++;
   while (*dptr && isspace(*dptr)) dptr++;

   if (strsame (dptr, "ERRMSG", 6))
   {
      /* this directive is completely ignored :^) */
   }
   else
   if (strsame (dptr, "TIMEFMT", 7))
   {
      dptr += SsiGetTagValue (rqptr, dptr, TagValue);
      if ((cptr = HeapAlloc (rqptr, strlen(TagValue)+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
      strcpy (tkptr->TimeFmtPtr = cptr, TagValue);
   }
   else
   if (strsame (dptr, "SIZEFMT", 7))
   {
      dptr += SsiGetTagValue (rqptr, dptr, TagValue);
      if (!(strsame (TagValue, "abbrev", -1) ||
            strsame (TagValue, "bytes", -1) ||
            strsame (TagValue, "blocks", -1)))
      {
         SsiExplainError (rqptr, "Invalid tag value", __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
      if ((cptr = HeapAlloc (rqptr, strlen(TagValue)+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
      strcpy (tkptr->SizeFmtPtr = cptr, TagValue);
   }
   else
   {
      SsiExplainError (rqptr, "Unknown tag", __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   /* declare an AST to execute the required function */
   if (VMSnok (status = sys$dclast (&SsiParseLine, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
}

/*****************************************************************************/
/*
Process a DCL/EXEC directive.  The array at the beginning of this function
provides  the allowed DCL directives and their DCL command equivalents.  Only
innocuous  DCL commands are allowed (hopefully! e.g. SHOW, WRITE SYS$OUTPUT)
for the  average document.  For documents owned by SYSTEM and NOT WORLD
WRITEABLE (for obvious reasons) the execution of any DCL command or command
procedure is allowed allowing maximum flexibility for the privileged document
author.
*/ 

SsiDcl (struct RequestStruct *rqptr)

{
   /*
      First element is allowed DCL command.
      Second is actual DCL verb/command/syntax.
      Third, if non-empty, idicates a file specification is required.
      Fourth, if non-empty, indicates it is for privileged documents only!
   */
   static char  *SupportedDcl [] =
   {
      "cmd", "", "", "P",                    /* privileged only! */
      "dir", "directory ", "", "",           /* any document */
      "exec", "", "", "P",                   /* privileged only! */
      "file", "@", "Y", "P",                 /* privileged only! */
      "run", "run ", "Y", "P",               /* privileged only! */
      "say", "write sys$output", "", "",     /* any document */
      "show", "show", "", "",                /* any document */
      "vdir", "directory ", "", "",          /* any document */
      "virtual", "@", "Y", "P",              /* privileged only! */
      "vrun", "run ", "Y", "P",              /* privileged only! */
      "", "", "", "P"  /* must be terminated by empty/privileged command! */
   };

   register int  idx;
   register char  *cptr, *dptr, *sptr, *zptr;
   register struct SsiTaskStruct  *tkptr;

   int  status;
   char  DclCommand [1024],
         TagValue [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiDcl() |%s|\n", rqptr->SsiTaskPtr->Directive);

   tkptr = rqptr->SsiTaskPtr;

   dptr = rqptr->SsiTaskPtr->Directive;
   while (*dptr && !isspace(*dptr)) dptr++;
   while (*dptr && isspace(*dptr)) dptr++;

   for (idx = 0; *SupportedDcl[idx]; idx += 4)
   {
      if (Debug)
         fprintf (stdout, "|%s|%s|\n", SupportedDcl[idx], SupportedDcl[idx+1]);
      /* compare the user-supplied DCL command to the list of allowed */
      cptr = SupportedDcl[idx];
      sptr = dptr;
      while (*cptr && *sptr != '=' && toupper(*cptr) == toupper(*sptr))
         { cptr++; sptr++; }
      if (!*cptr && *sptr == '=') break;
   }

   if (!SupportedDcl[idx][0] &&
       !strsame (dptr = rqptr->SsiTaskPtr->Directive, "#exec ", 6))
   {
      /************************/
      /* unsupported command! */
      /************************/

      SsiExplainError (rqptr, "Unsupported DCL", __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   /*************************/
   /* supported DCL command */
   /*************************/

   if (SupportedDcl[idx+3][0])
   {
      /******************************/
      /* "privileged" DCL requested */
      /******************************/

      if (!Config.SsiExecEnabled)
      {
         SsiExplainError (rqptr, "DCL execution disabled", __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }

      /* SYSTEM is UIC [1,4] */
      if (tkptr->FileXabPro.xab$w_grp != 1 ||
          tkptr->FileXabPro.xab$w_mbm != 4)
      {
         SsiExplainError (rqptr, "Document must be owned by SYSTEM",
                          __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }

      /* protect word: wwggooss, protect bits: dewr, 1 denies access */
      if (!(tkptr->FileXabPro.xab$w_pro & 0x2000))
      {
         SsiExplainError (rqptr, "Document cannot be WORLD writable",
                          __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
   }

   /****************************************************/
   /* create the DCL command from array and parameters */
   /****************************************************/

   cptr = SupportedDcl[idx+1];
   sptr = DclCommand;
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';

   if (SupportedDcl[idx+2][0])
   {
      /* file specification involved */
      dptr += SsiGetFileSpec (rqptr, dptr, TagValue);
      for (cptr = TagValue; *cptr; *sptr++ = *cptr++);
      *sptr = '\0';
   }
   else
   {
      /* some DCL verb/parameters or another */
      dptr += SsiGetTagValue (rqptr, dptr, TagValue);
      if (TagValue[0] && sptr > DclCommand) *sptr++ = ' ';
      for (cptr = TagValue; *cptr; *sptr++ = *cptr++);
      *sptr = '\0';
   }

   if (Debug) fprintf (stdout, "DclCommand |%s|\n", DclCommand);

   /* by default HTML-forbidden characters in DCL output are escaped */
   rqptr->NetBufferOutputEscapeHtml = true;

   while (*dptr)
   {
      while (*dptr && isspace(*dptr)) dptr++;
      if (!*dptr) break;

      if (strsame (dptr, "PAR=", 4))
      {
         /* parameters to the command procedure, directory, etc. */
         dptr += SsiGetTagValue (rqptr, dptr, TagValue);
         if (TagValue[0]) *sptr++ = ' ';
         for (cptr = TagValue; *cptr; *sptr++ = *cptr++);
         *sptr = '\0';
      }
      else
      if (strsame (dptr, "TYPE=", 5))
      {
         dptr += SsiGetTagValue (rqptr, dptr, TagValue);
         if (TagValue[0] && strsame (TagValue, "text/html", -1))
            rqptr->NetBufferOutputEscapeHtml = false;
         else
            rqptr->NetBufferOutputEscapeHtml = true;
      }
      else
      if (*dptr)
      {
         SsiExplainError (rqptr, "Unknown tag", __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
   }

   DclBegin (rqptr, &SsiParseLine, DclCommand, NULL);
}

/*****************************************************************************/
/*
Generate an "Index of" directory listing by calling DirBegin() task.
*/

SsiDir (struct RequestStruct *rqptr)

{
   register char  *cptr, *dptr;

   int  status;
   char  DirSpec [256],
         TagValue [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiDir() |%s|\n", rqptr->SsiTaskPtr->Directive);

   TagValue[0] = '\0';

   dptr = rqptr->SsiTaskPtr->Directive;
   while (*dptr && !isspace(*dptr)) dptr++;

   while (*dptr)
   {
      while (*dptr && isspace(*dptr)) dptr++;
      if (!*dptr) break;

      if (strsame (dptr, "FILE=", 5) || strsame (dptr, "VIRTUAL=", 8))
         dptr += SsiGetFileSpec (rqptr, dptr, DirSpec);
      else
      if (strsame (dptr, "PAR=", 4))
         dptr += SsiGetTagValue (rqptr, dptr, TagValue);
      else
      {
         SsiExplainError (rqptr, "Unknown tag", __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
   }

   if (Debug) fprintf (stdout, "|%s|%s|\n", DirSpec, TagValue);
   if (TagValue[0])
   {
      if ((cptr = HeapAlloc (rqptr, strlen(TagValue)+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
      strcpy (cptr, TagValue);
   }
   else
      cptr = "";

   DirBegin (rqptr, &SsiParseLine, DirSpec, cptr);
}

/*****************************************************************************/
/*
Output the special variable.
*/

boolean SsiEcho (struct RequestStruct *rqptr)

{
   register char  *cptr, *dptr, *sptr, *zptr;
   register struct SsiTaskStruct  *tkptr;

   int  status,
        Length;
   unsigned long  BinTime [2];
   char  *StringPtr;
   char  TimeFormat [256],
         TimeString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiEcho() |%s|\n", rqptr->SsiTaskPtr->Directive);

   tkptr = rqptr->SsiTaskPtr;

   dptr = rqptr->SsiTaskPtr->Directive;
   while (*dptr && !isspace(*dptr)) dptr++;
   while (*dptr && isspace(*dptr)) dptr++;

   if (strsame (dptr, "DATE_LOCAL", 10))
   {
      SsiGetTagValue (rqptr, dptr+10, TimeFormat);
      if (!(Length = SsiTimeString (rqptr, NULL, TimeFormat,
                                    TimeString, sizeof(TimeString))))
      {
         SsiEnd (rqptr);
         return;
      }
      StringPtr = TimeString;
   }
   else
   if (strsame (dptr, "DATE_GMT", 8))
   {
      SsiGetTagValue (rqptr, dptr+10, TimeFormat);
      sys$gettim (&BinTime);
      TimeAdjustGMT (true, &BinTime);
      if (!(Length = SsiTimeString (rqptr, &BinTime, TimeFormat,
                                    TimeString, sizeof(TimeString))))
      {
         SsiEnd (rqptr);
         return;
      }
      memcpy (TimeString+Length, " GMT", 5);
      Length += 4;
      StringPtr = TimeString;
   }
   else
   if (strsame (dptr, "DOCUMENT_NAME", 13))
      Length = strlen(StringPtr = rqptr->PathInfoPtr);
   else
   if (strsame (dptr, "CREATED", 7))
   {
      SsiGetTagValue (rqptr, dptr+7, TimeFormat);
      if (!(Length = SsiTimeString (rqptr, &tkptr->FileXabDat.xab$q_cdt,
                        TimeFormat, TimeString, sizeof(TimeString))))
      {
         SsiEnd (rqptr);
         return;
      }
      StringPtr = TimeString;
   }
   else
   if (strsame (dptr, "FILE_NAME", 9))
   {
      StringPtr = tkptr->FileName;
      Length = tkptr->FileNameLength;
   }
   else
   if (strsame (dptr, "LAST_MODIFIED", 13))
   {
      SsiGetTagValue (rqptr, dptr+13, TimeFormat);
      if (!(Length = SsiTimeString (rqptr, &tkptr->FileXabDat.xab$q_rdt,
                        TimeFormat, TimeString, sizeof(TimeString))))
      {
         SsiEnd (rqptr);
         return;
      }
      StringPtr = TimeString;
   }
   else
   {
      SsiExplainError (rqptr, "Unknown tag", __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   if (VMSnok (NetBufferOutput (rqptr, &SsiParseLine, StringPtr, Length)))
      SsiEnd (rqptr);
}

/*****************************************************************************/
/*
Output the specified file's creation date and time.
*/

boolean SsiFCreated (struct RequestStruct *rqptr)

{
   register char  *dptr, *sptr, *zptr;

   char  FileName [256],
         TimeFormat [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiFCreated() |%s|\n", rqptr->SsiTaskPtr->Directive);

   FileName[0] = TimeFormat[0] = '\0';

   dptr = rqptr->SsiTaskPtr->Directive;
   while (*dptr && !isspace(*dptr)) dptr++;

   while (*dptr)
   {
      while (*dptr && isspace(*dptr)) dptr++;
      if (!*dptr) break;

      if (strsame (dptr, "FILE=", 5) || strsame (dptr, "VIRTUAL=", 8))
         dptr += SsiGetFileSpec (rqptr, dptr, FileName);
      else
      if (strsame (dptr, "FMT=", 4))
         dptr += SsiGetTagValue (rqptr, dptr, TimeFormat);
      else
      {
         SsiExplainError (rqptr, "Unknown tag", __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
   }

   SsiFileDetails (rqptr, &SsiParseLine, FileName, FILE_FCREATED, TimeFormat);
}

/*****************************************************************************/
/*
Output the specified file's last modification date and time.
*/

SsiFLastMod (struct RequestStruct *rqptr)

{
   register char  *dptr, *sptr, *zptr;

   char  FileName [256],
         TimeFormat [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiFLastMod() |%s|\n", rqptr->SsiTaskPtr->Directive);

   FileName[0] = TimeFormat[0] = '\0';

   dptr = rqptr->SsiTaskPtr->Directive;
   while (*dptr && !isspace(*dptr)) dptr++;

   while (*dptr)
   {
      while (*dptr && isspace(*dptr)) dptr++;
      if (!*dptr) break;

      if (strsame (dptr, "FILE=", 5) || strsame (dptr, "VIRTUAL=", 8))
         dptr += SsiGetFileSpec (rqptr, dptr, FileName);
      else
      if (strsame (dptr, "FMT=", 4))
         dptr += SsiGetTagValue (rqptr, dptr, TimeFormat);
      else
      {
         SsiExplainError (rqptr, "Unknown tag", __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
   }

   SsiFileDetails (rqptr, &SsiParseLine, FileName, FILE_FLASTMOD, TimeFormat);
}

/*****************************************************************************/
/*
Output the specified file's size.
*/

SsiFSize (struct RequestStruct *rqptr)

{
   register char  *dptr;

   char  FileSize [32],
         FileName [256],
         SizeFormat [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiFSize() |%s|\n", rqptr->SsiTaskPtr->Directive);

   FileName[0] = SizeFormat[0] = '\0';

   dptr = rqptr->SsiTaskPtr->Directive;
   while (*dptr && !isspace(*dptr)) dptr++;

   while (*dptr)
   {
      while (*dptr && isspace(*dptr)) dptr++;
      if (!*dptr) break;

      if (strsame (dptr, "FILE=", 5) || strsame (dptr, "VIRTUAL=", 8))
         dptr += SsiGetFileSpec (rqptr, dptr, FileName);
      else
      if (strsame (dptr, "FMT=", 4))
         dptr += SsiGetTagValue (rqptr, dptr, SizeFormat);
      else
      {
         SsiExplainError (rqptr, "Unknown tag", __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }
   }

   SsiFileDetails (rqptr, &SsiParseLine, FileName, FILE_FSIZE, SizeFormat);
}

/*****************************************************************************/
/*
Include the contents of the specified file by calling FileBegin() task.
*/

SsiIncludeFile (struct RequestStruct *rqptr)

{
   register char  *cptr, *dptr;

   boolean  PreTagFileContents;
   int  status;
   char  FileFormat [256],
         FileName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiIncludeFile() |%s|\n", rqptr->SsiTaskPtr->Directive);

   FileName[0] = FileFormat[0] = '\0';

   dptr = rqptr->SsiTaskPtr->Directive;
   while (*dptr && !isspace(*dptr)) dptr++;

   while (*dptr)
   {
      while (*dptr && isspace(*dptr)) dptr++;
      if (!*dptr) break;

      if (strsame (dptr, "FILE=", 5) || strsame (dptr, "VIRTUAL=", 8))
         dptr += SsiGetFileSpec (rqptr, dptr, FileName);
      else
      if (strsame (dptr, "TYPE=", 5))
         dptr += SsiGetTagValue (rqptr, dptr, FileFormat);
      else
      {
         SsiExplainError (rqptr, "Unknown tag", __FILE__, __LINE__);
         SsiEnd (rqptr);
         return;
      }               
   }

   if (Debug) fprintf (stdout, "|%s|%s|\n", FileName, FileFormat);

   /* find the file extension and set the content type */
   for (cptr = FileName; *cptr && *cptr != ']'; cptr++);
   while (*cptr && *cptr != '.') cptr++;
   cptr = ConfigContentType (NULL, cptr);

   if (!strsame (cptr, "text/", 5))
   {
      SsiExplainError (rqptr, "File content type is not text",
                       __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   if (strsame (FileFormat, "text/plain", -1))
      PreTagFileContents = true;
   else
      PreTagFileContents = false;

   if (VMSnok (status =
       FileBegin (rqptr, &SsiParseLine, &SsiParseLine, FileName,
                  cptr, PreTagFileContents, false)))
   {
      while (*dptr)
      {
         while (*dptr && *dptr != '-') dptr++;
         if (dptr[1] == '-' && dptr[2] == '>') break;
         dptr++;
      }
      *dptr = '\0';
      rqptr->ErrorTextPtr = rqptr->SsiTaskPtr->Directive;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   return (true);
}

/*****************************************************************************/
/*
Using the locale formatting capabilities of function strftime(), output the 
time represented by the specified VMS quadword, binary time.  If 
'BinaryTimePtr' is NULL then default to the current time.  Returns number of 
characters placed into 'TimeString', or zero if an error.
*/ 

int SsiTimeString
(
struct RequestStruct *rqptr,
unsigned long *BinaryTimePtr,
char *TimeFmtPtr,
char *TimeString,
int SizeOfTimeString
)
{
   int  Length;
   unsigned long  BinaryTime [2];
   struct tm  UnixTime;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SsiTimeString() |%s|\n", TimeFmtPtr);

   if (!TimeFmtPtr[0]) TimeFmtPtr = rqptr->SsiTaskPtr->TimeFmtPtr;
   if (BinaryTimePtr == NULL) sys$gettim (BinaryTimePtr = &BinaryTime);
   TimeVmsToUnix (BinaryTimePtr, &UnixTime);

   if (!(Length =
         strftime (TimeString, SizeOfTimeString, TimeFmtPtr, &UnixTime)))
   {
      SsiExplainError (rqptr, "Problem with date/time format string",
                       __FILE__, __LINE__);
      SsiEnd (rqptr);
      return (0);
   }
   return (Length);
}

/*****************************************************************************/
/*
Get the value of a tag parameter (e.g. tag_name="value").  Maximum number of
characters allowed in value is 256.   Returns the number of characters scanned
to get the value.
*/ 

int SsiGetTagValue
(
struct RequestStruct *rqptr,
char *String,
char *Value
)
{
   register char  *sptr, *vptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SsiGetTagValue() |%s|\n", String);

   sptr = String;

   while (*sptr && *sptr != '=') sptr++;
   if (*sptr) sptr++;
   if (*sptr == '\"') sptr++;
   zptr = (vptr = Value) + 256;
   while (*sptr && *sptr != '\"' && vptr < zptr)
   {
      if (*sptr == '\\') sptr++;
      *vptr++ = *sptr++;
   }
   *vptr = '\0';
   if (*sptr == '\"') sptr++;

   if (Debug) fprintf (stdout, "Value |%s|\n", Value);
   return (sptr - String);
}

/*****************************************************************************/
/*
Get a 'FILE="file_name"' or a 'VIRTUAL="file_name"'.  Maximum number of 
characters allowed in value is 256.   Returns the number of characters scanned 
to get the value.
*/ 

int SsiGetFileSpec
(
struct RequestStruct *rqptr,
char *String,
char *FileName
)
{
   register char  *fptr, *sptr, *zptr;

   char  VirtualPath [256],
         VirtualFileName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SsiGetFileSpec() |%s|\n", String);

   sptr = String;

   if (toupper(*sptr) == 'V')
   {
      /* virtual path to file */
      while (*sptr && *sptr != '=') sptr++;
      if (*sptr) sptr++;
      if (*sptr == '\"') sptr++;
      zptr = (fptr = VirtualPath) + sizeof(VirtualPath);
      while (*sptr && *sptr != ' ' && *sptr != '\"' && fptr < zptr)
      {
         if (*sptr == '\\') sptr++;
         *fptr++ = *sptr++;
      }
      *fptr = '\0';
      if (*sptr == '\"') sptr++;
      if (Debug) fprintf (stdout, "VirtualPath |%s|\n", VirtualPath);

      MapUrl_VirtualPath (rqptr->PathInfoPtr, VirtualPath,
                          VirtualFileName, sizeof(VirtualFileName));

      FileName[0] = '\0';
      zptr = MapUrl (VirtualFileName, FileName, NULL, NULL);
      if (!zptr[0])
      {
         ErrorGeneral (rqptr, zptr+1, __FILE__, __LINE__);
         return (0);
      }
   }
   else
   {
      while (*sptr && *sptr != '=') sptr++;
      if (*sptr) sptr++;
      if (*sptr == '\"') sptr++;
      zptr = (fptr = FileName) + 256;
      while (*sptr && *sptr != ' ' && *sptr != '\"' && fptr < zptr)
      {
         if (*sptr == '\\') sptr++;
         *fptr++ = *sptr++;
      }
      *fptr = '\0';
      if (*sptr == '\"') sptr++;
   }

   if (Debug) fprintf (stdout, "FileName |%s|\n", FileName);
   return (sptr - String);
}

/*****************************************************************************/
/*
Retrieve and display the specified file's specified attribute (size,
modification time, etc.)  This function completes synchronously, that is it
waits for the single QIO on the file details to complete before assembling the
details for the listing, because of this the function introduces slight
granularity.  It is a candidate for future improvement.

This function uses the ACP-QIO interface detailed in the "OpenVMS I/O User's 
Reference Manual".
*/ 

boolean SsiFileDetails
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *FileName,
int DisplayThis,
char *FormatPtr
)
{
   static $DESCRIPTOR (AbbrevOneByteFaoDsc, "!UL byte");
   static $DESCRIPTOR (AbbrevBytesFaoDsc, "!UL bytes");
   static $DESCRIPTOR (AbbrevOnekByteFaoDsc, "!UL kbyte");
   static $DESCRIPTOR (AbbrevkBytesFaoDsc, "!UL kbytes");
   static $DESCRIPTOR (AbbrevOneMByteFaoDsc, "!UL Mbyte");
   static $DESCRIPTOR (AbbrevMBytesFaoDsc, "!UL Mbytes");
   static $DESCRIPTOR (OneBlockFaoDsc, "!UL block");
   static $DESCRIPTOR (BlocksFaoDsc, "!UL blocks");
   static $DESCRIPTOR (BytesFaoDsc, "!AZ bytes");
   static $DESCRIPTOR (NumberFaoDsc, "!UL");

   register char  *cptr, *sptr;

   int  status,
        NumBytes;

   unsigned short  AcpChannel,
                   Length;
   unsigned long  AllocatedVbn,
                  Bytes,
                  EndOfFileVbn;
   unsigned long  AtrCdt [2],
                  AtrRdt [2];

   char  ExpandedFileName [256],
         Scratch [256],
         String [256],
         TimeString [256];

   $DESCRIPTOR (DeviceDsc, "");
   $DESCRIPTOR (StringDsc, String);
   $DESCRIPTOR (ScratchDsc, Scratch);


   struct FAB  FileFab;
   struct NAM  FileNam;

   struct {
      unsigned long  OfNoInterest1;
      unsigned short  AllocatedVbnHi;
      unsigned short  AllocatedVbnLo;
      unsigned short  EndOfFileVbnHi;
      unsigned short  EndOfFileVbnLo;
      unsigned short  FirstFreeByte;
      unsigned short  OfNoInterest2;
      unsigned long  OfNoInterestLots [4];
   } AtrRecord;

   struct fibdef  FileFib;
   struct atrdef  FileAtr [] =
   {
      { sizeof(AtrRecord), ATR$C_RECATTR, &AtrRecord },
      { sizeof(AtrCdt), ATR$C_CREDATE, &AtrCdt },
      { sizeof(AtrRdt), ATR$C_REVDATE, &AtrRdt },
      { 0, 0, 0 }
   };

   struct {
      unsigned short  Length;
      unsigned short  Unused;
      unsigned long  Address;
   } FileNameAcpDsc,
     FileFibAcpDsc,
     FileAtrAcpDsc;

   struct {
      unsigned short  Status;
      unsigned short  Unused1;
      unsigned long  Unused2;
   } AcpIOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiFileDetails() |%d|%s|\n", DisplayThis, FileName);

   FileFab = cc$rms_fab;
   FileFab.fab$l_fna = FileName;
   FileFab.fab$b_fns = strlen(FileName);
   FileFab.fab$l_fop = FAB$M_NAM;
   FileFab.fab$l_nam = &FileNam;
   FileNam = cc$rms_nam;
   FileNam.nam$l_esa = ExpandedFileName;
   FileNam.nam$b_ess = sizeof(ExpandedFileName)-1;

   if (VMSok (status = sys$parse (&FileFab, 0, 0)))
   {
      /* assign a channel to the disk device containing the file */
      DeviceDsc.dsc$w_length = FileNam.nam$b_dev;
      DeviceDsc.dsc$a_pointer = FileNam.nam$l_dev;
      status = sys$assign (&DeviceDsc, &AcpChannel, 0, 0, 0);
      if (Debug) fprintf (stdout, "sys$assign() %%X%08.08X\n", status);
   }

   if (VMSok (status))
   {
      /*******************************************/
      /* queue an ACP I/O to get file attributes */
      /*******************************************/

      /* set up the File Information Block for the ACP interface */
      memset (&FileFib, 0, sizeof(struct fibdef));
      FileFibAcpDsc.Length = sizeof(FileFib);
      FileFibAcpDsc.Address = &FileFib;
#ifdef __DECC
      memcpy (&FileFib.fib$w_did, &FileNam.nam$w_did, 6);
#else
      memcpy (&FileFib.fib$r_did_overlay.fib$w_did, &FileNam.nam$w_did, 6);
#endif

      FileNameAcpDsc.Length = FileNam.nam$b_name +
                              FileNam.nam$b_type +
                              FileNam.nam$b_ver;
      FileNameAcpDsc.Address = FileNam.nam$l_name;

      status = sys$qiow (0, AcpChannel, IO$_ACCESS, &AcpIOsb, 0, 0, 
                         &FileFibAcpDsc, &FileNameAcpDsc, 0, 0,
                         &FileAtr, 0);

      sys$dassgn (AcpChannel);

      if (Debug)
         fprintf (stdout, "sys$qio() %%X%08.08X IOsb: %%X%08.08X\n",
                 status, AcpIOsb.Status);

      if (VMSok (status)) status = AcpIOsb.Status;
   }

   /* ensure parse internal data structures are released */
   FileFab.fab$l_fna = "a:[b]c.d;";
   FileFab.fab$b_fns = 9;
   FileFab.fab$b_dns = 0;
   FileNam.nam$b_nop = NAM$M_SYNCHK;
   sys$parse (&FileFab, 0, 0);

   if (VMSnok (status)) 
   {
      *FileNam.nam$l_ver = '\0';
      rqptr->ErrorTextPtr = "getting file details";
      rqptr->ErrorHiddenTextPtr = FileName;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }

   if (DisplayThis == FILE_FCREATED)
   {
      /*********************/
      /* date/time created */
      /*********************/

      if (!SsiTimeString (rqptr, &AtrCdt, FormatPtr, String, sizeof(String)))
      {
         SsiEnd (rqptr);
         return;
      }
      Length = strlen(String);
   }
   else
   if (DisplayThis == FILE_FLASTMOD)
   {
      /*********************/
      /* date/time revised */
      /*********************/

      if (!SsiTimeString (rqptr, &AtrRdt, FormatPtr, String, sizeof(String)))
      {
         SsiEnd (rqptr);
         return;
      }
      Length = strlen(String);
   }
   else
   if (DisplayThis == FILE_FSIZE)
   {
      /*************/
      /* file size */
      /*************/

      AllocatedVbn = AtrRecord.AllocatedVbnLo +
                     (AtrRecord.AllocatedVbnHi << 16);
      EndOfFileVbn = AtrRecord.EndOfFileVbnLo +
                     (AtrRecord.EndOfFileVbnHi << 16);

      if (Debug)
         fprintf (stdout,
         "AllocatedVbn: %d EndOfFileVbn: %d FirstFreeByte %d\n",
         AllocatedVbn, EndOfFileVbn, AtrRecord.FirstFreeByte);

      /* the "<< 9" is a bit more efficient than the equivalent "* 512" */
      if (EndOfFileVbn > 1)
         Bytes = (EndOfFileVbn << 9) + AtrRecord.FirstFreeByte;
      else
         Bytes = AtrRecord.FirstFreeByte;
      if (Debug) fprintf (stdout, "Bytes %d\n", Bytes);

      if (!FormatPtr[0]) FormatPtr = rqptr->SsiTaskPtr->SizeFmtPtr;
      if (toupper(FormatPtr[0]) == 'A')  /* "abbrev" */
      {
         if (Bytes < 1024)
         {
            if (Bytes == 1)
               sys$fao (&AbbrevOneByteFaoDsc, &Length, &StringDsc, Bytes);
            else
               sys$fao (&AbbrevBytesFaoDsc, &Length, &StringDsc, Bytes);
         }
         else
         if (Bytes < 1048576)
         {
            if ((NumBytes = Bytes / 1024) == 1)
               sys$fao (&AbbrevOnekByteFaoDsc, &Length, &StringDsc, NumBytes);
            else
               sys$fao (&AbbrevkBytesFaoDsc, &Length, &StringDsc, NumBytes);
         }
         else
         {
            if ((NumBytes = Bytes / 1048576) == 1)
               sys$fao (&AbbrevOneMByteFaoDsc, &Length, &StringDsc, NumBytes);
            else
               sys$fao (&AbbrevMBytesFaoDsc, &Length, &StringDsc, NumBytes);
         }
         String[Length] = '\0';
      }
      else
      if (toupper(FormatPtr[0]) == 'B' &&
          toupper(FormatPtr[1]) == 'Y')  /* "bytes" */
      {
         sys$fao (&NumberFaoDsc, &Length, &ScratchDsc, Bytes);
         Scratch[Length] = '\0';
         sptr = String;
         cptr = Scratch;
         while (Length--)
         {
            *sptr++ = *cptr++;
            if (Length && !(Length % 3)) *sptr++ = ',';
         }
         *sptr = '\0';
         for (cptr = " bytes"; *cptr; *sptr++ = *cptr++);
         *sptr = '\0';
         Length = sptr - String;
      }
      else
      if (toupper(FormatPtr[0]) == 'B' &&
          toupper(FormatPtr[1]) == 'L')  /* "blocks" */
      {
         if (EndOfFileVbn == 1)
            sys$fao (&OneBlockFaoDsc, &Length, &StringDsc, EndOfFileVbn);
         else
            sys$fao (&BlocksFaoDsc, &Length, &StringDsc, EndOfFileVbn);
         String[Length] = '\0';
      }
   }

   if (VMSnok (NetBufferOutput (rqptr, AstFunctionPtr, String, Length)))
   {
      SsiEnd (rqptr);
      return (false);
   }
   return (true);
}

/*****************************************************************************/
/*
Keep track of how many times an SSI file is accessed.  Do this by creating 
another file in the same directory, same name, extension modified by appending 
dollar symbol, containing a single longword record with the binary number of 
times the document has been accessed.  This can be reset by merely deleting 
the access count file.  Provide the access count and creation date of the 
access count file. 

This operation is fairly expensive in terms of I/O, and because it is atomic 
(all occurs during an AST routine, and opens-creates/writes/closes before 
returning to other processing) introduces a fair degree of granularity.  Its 
expense means the "#accesses" functionality should be used sparingly.  This 
functionality can be disabled.
*/ 

SsiAccessCount
(
struct RequestStruct *rqptr,
unsigned long *AccessCountPtr,
unsigned long *SinceBinTimePtr
)
{
   register char  *cptr, *sptr, *zptr;

   int  status,
        CreateStatus,
        SetPrvStatus;
   char  FileName [256];
   struct FAB  FileFab;
   struct RAB  FileRab;
   struct XABDAT  FileXabDat;
   struct XABPRO  FileXabPro;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SsiAccessCount |%s|\n",
               rqptr->SsiTaskPtr->FileName);

   zptr = (sptr = FileName) + sizeof(FileName);
   for (cptr = rqptr->SsiTaskPtr->FileName;
        *cptr && *cptr != ';' && sptr < zptr;
        *sptr++ = *cptr++);
   if (sptr < zptr) *sptr++ = '$';
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr,
         "Document file name too long to create access count file name.",
         __FILE__, __LINE__);
      return (status);
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "FileName: %s\n", FileName);

   FileFab = cc$rms_fab;
   FileFab.fab$l_fna = FileName;  
   FileFab.fab$b_fns = sptr-FileName;
   FileFab.fab$l_fop = FAB$M_CIF | FAB$M_SQO;
   FileFab.fab$b_fac = FAB$M_GET | FAB$M_PUT | FAB$M_UPD;
   FileFab.fab$w_mrs = sizeof(unsigned long);
   FileFab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT | FAB$M_SHRUPD;
   FileFab.fab$b_org = FAB$C_SEQ;
   FileFab.fab$b_rfm = FAB$C_FIX;

   /* initialize the date extended attribute block */
   FileFab.fab$l_xab = &FileXabDat;
   FileXabDat = cc$rms_xabdat;
   FileXabDat.xab$l_nxt = &FileXabPro;

   /* initialize the protection extended attribute block */
   FileXabPro = cc$rms_xabpro;
   FileXabPro.xab$w_pro = 0xaa00;  /* W:RE,G:RE,O:RWED,S:RWED */

   /* "temporary" file, automatic delete on closing it */
   if (rqptr->DeleteOnClose) FileFab.fab$l_fop |= FAB$M_DLT;

   /* turn on SYSPRV to allow access to the counter file */
   EnableSysPrv ();
   CreateStatus = sys$create (&FileFab, 0, 0);
   DisableSysPrv ();

   if (VMSnok (CreateStatus))
   {
      if (Debug) fprintf (stdout, "sys$create() %%X%08.08X\n", status);
      rqptr->ErrorTextPtr = "creating access count file";
      rqptr->ErrorHiddenTextPtr = FileName;
      ErrorVmsStatus (rqptr, CreateStatus, __FILE__, __LINE__);
      return (CreateStatus);
   }

   FileRab = cc$rms_rab;
   FileRab.rab$l_fab = &FileFab;
   FileRab.rab$l_rbf = AccessCountPtr;
   FileRab.rab$w_rsz = sizeof(*AccessCountPtr);

   if (VMSnok (status = sys$connect (&FileRab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
      rqptr->ErrorTextPtr = "connecting access count file";
      rqptr->ErrorHiddenTextPtr = FileName;
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
   }

   /* copy the creation date (since date) of the access count file */
   memcpy (SinceBinTimePtr, &FileXabDat.xab$q_cdt, 8);

   if (VMSok (status))
   {
      if (CreateStatus == RMS$_CREATED)
      {
         /***************************************/
         /* count file did not previously exist */
         /***************************************/
   
         *AccessCountPtr = 1;
         if (Debug) fprintf (stdout, "*AccessCountPtr: %d\n", *AccessCountPtr);

         if (VMSnok (status = sys$put (&FileRab, 0, 0)))
         {
            if (Debug) fprintf (stdout, "sys$put() %%X%08.08X\n", status);
            rqptr->ErrorTextPtr = "updating access count file";
            rqptr->ErrorHiddenTextPtr = FileName;
            ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
            return (status);
         }
      }
      else
      {
         /**********************/
         /* count file existed */
         /**********************/

         FileRab.rab$l_ubf = AccessCountPtr;
         FileRab.rab$w_usz = sizeof(*AccessCountPtr);
         if (VMSnok (status = sys$get (&FileRab, 0, 0))) 
         {
            if (Debug) fprintf (stdout, "sys$get() %%X%08.08X\n", status);
            rqptr->ErrorTextPtr = "reading access count file";
            rqptr->ErrorHiddenTextPtr = FileName;
            ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         }
         else
         {
            *AccessCountPtr += 1;
            if (Debug)
               fprintf (stdout, "*AccessCountPtr: %d\n", *AccessCountPtr);
         }

         if (VMSnok (status = sys$update (&FileRab, 0, 0)))
         {
            if (Debug) fprintf (stdout, "sys$update() %%X%08.08X\n", status);
            rqptr->ErrorTextPtr = "updating access count file";
            rqptr->ErrorHiddenTextPtr = FileName;
            ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         }
      }
   }

   if (rqptr->DeleteOnClose)
   {
      /* use SYSPRV to ensure deletion-on-close of file */
      EnableSysPrv ();
      sys$close (&FileFab, 0, 0);
      DisableSysPrv ();
   }
   else
      sys$close (&FileFab, 0, 0);

   return (status);
}

/*****************************************************************************/
/*
Generate a general error, with explanation about the pre-processor error.
*/ 

int SsiExplainError
(
struct RequestStruct *rqptr,
char *Explanation,
char *SourceFileName,
int SourceLineNumber
)
{
   static $DESCRIPTOR (ErrorMessageFaoDsc,
                       "!AZ (line !UL) ... ``<TT>!AZ</TT>''\n");

   int  status;
   unsigned short  Length;
   char  String [1024];
   $DESCRIPTOR (StringDsc, String);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SsiExplainError()\n");

   status = sys$fao (&ErrorMessageFaoDsc, &Length, &StringDsc,
                     Explanation, rqptr->SsiTaskPtr->DirectiveLineNumber,
                     rqptr->SsiTaskPtr->Directive);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$fao()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      SsiEnd (rqptr);
      return;
   }
   String[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", String);

   ErrorGeneral (rqptr, String, SourceFileName, SourceLineNumber);
}

/*****************************************************************************/

