/*****************************************************************************
/*
                                 Support.c

Miscellaneous support functions for HTTPd.


VERSION HISTORY
---------------
12-JUL-97  MGD  EnableSysPrv(), DisableSysPrv() to rationalize this requirement
07-JUL-97  MGD  attempt to reduce dynamic memory fragmentation within C RTL
                using HEAP_MIN_CHUNK_SIZE functionality,
                prevent request logical redefinition if keep-alive timeout 
16-JUN-97  MGD  generic linked list functions
12-MAR-97  MGD  HTTP header generation
01-FEB-97  MGD  HTTPd version 4
01-OCT-96  MGD  report menu
18-JUN-96  MGD  bugfix; HTTPD$GMT conversion to VMS delta time in TimeSetGmt()
06-APR-96  MGD  persistent connections ("keep-alive")
01-DEC-95  MGD  HTTPd version 3.0
27-SEP-95  MGD  provide GMT functions
20-DEC-94  MGD  initial development
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <time.h>

/* VMS related header files */
#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <lnmdef.h>
#include <prvdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>

/* application related header files */
#include "wasd.h"
#include "error.h"
#include "httpd.h"
#include "net.h"
#include "support.h"

/******************/
/* global storage */
/******************/

int  HeapOverhead = sizeof(struct HeapStruct*) +
                    sizeof(struct HeapStruct*) +
                    sizeof(unsigned int);

boolean TimeAheadOfGmt;
char  TimeGmtString [48],
      TimeGmtVmsString [50];
unsigned long  TimeGmtDeltaBinary [2];

char  HttpProtocol [] = "HTTP/1.0";

char  HtmlSgmlDoctype [] =
      "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n";

char  *DayName [] =
   { "", "Monday", "Tuesday", "Wednesday",
     "Thursday", "Friday", "Saturday", "Sunday" };

char  *MonthName [] =
   { "", "January", "February", "March", "April", "May", "June",
         "July", "August", "September", "October", "November", "December" };

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern boolean  AccountingZeroOnStartup;
extern int ServerPort;
extern char  CommentedSoftwareID[];
extern char  ErrorStringSize[];
extern char  ErrorUrlEncoding[];
extern char  ServerHostName[];
extern char  SoftwareID[];
extern struct AccountingStruct Accounting;
extern struct ConfigStruct  Config;

/*****************************************************************************/
/*
Allocate dynamic memory to an individual thread's heap.  Return a pointer to 
the start of new *usable* memory area if successful, return a NULL if not.

The heap is maintained as a doubly-linked list of separate calloc()ed chunks
of memory.  The size of each memory area is contained in an unsigned int
immediately preceding each returned pointer.  This can be used to provide the
length of dynamically stored, null-terminated string (without strlen()ing it,
remember to subtract one to exclude the null terminator), or to determine if a
given area is large enough to be re-used, etc.
*/ 

char* HeapAlloc
(
struct RequestStruct *rqptr,
int ChunkSize
)
{
   register struct HeapStruct  *NewPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HeapAlloc() %d\n", ChunkSize);

   /* try to avoid excessive fragmentation of dynamic memory */
   ChunkSize = ((ChunkSize / HEAP_MIN_CHUNK_SIZE) + 1) * HEAP_MIN_CHUNK_SIZE;

   /** if (Debug) fprintf (stdout, "ChunkSize: %d\n", ChunkSize); **/

   if ((NewPtr = (struct HeapStruct*)calloc (1, ChunkSize + HeapOverhead))
       == NULL)
   {
      Accounting.HeapCallocFailCount++;
      return (NULL);
   }

   /** if (Debug) fprintf (stdout, "NewPtr: %d\n", NewPtr); **/
   rqptr->HeapSize += ChunkSize + HeapOverhead;

   if (rqptr->HeapHeadPtr == NULL)
   {
      rqptr->HeapHeadPtr = NewPtr;
      NewPtr->PrevPtr = NULL;
   }
   else
   {
      rqptr->HeapTailPtr->NextPtr = NewPtr;
      NewPtr->PrevPtr = rqptr->HeapTailPtr;
   }
   NewPtr->NextPtr = NULL;
   NewPtr->ChunkSize = ChunkSize;
   rqptr->HeapTailPtr = NewPtr;

   /** if (Debug) HeapDebug (rqptr); **/

   return ((unsigned char*)NewPtr + HeapOverhead);
}

/*****************************************************************************/
/*
Expand (or even contract) an individual chunk.  See HeapAlloc().
*/ 

char* HeapRealloc
(
struct RequestStruct *rqptr,
unsigned char *ChunkPtr,
int ChunkSize
)
{
   register struct HeapStruct  *NewPtr,
                               *NextPtr,
                               *OldPtr,
                               *PrevPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HeapRealloc() %d %d\n", ChunkPtr, ChunkSize);

   if (ChunkPtr == NULL) return (HeapAlloc (rqptr, ChunkSize));

   /* point at the start of the chunk, not the usable area which was passed */
   OldPtr = (struct HeapStruct*)(ChunkPtr - HeapOverhead);

   if (OldPtr->ChunkSize >= ChunkSize) return (ChunkPtr);

   /* try to avoid excessive fragmentation of dynamic memory */
   ChunkSize = ((ChunkSize / HEAP_MIN_CHUNK_SIZE) + 1) * HEAP_MIN_CHUNK_SIZE;

   PrevPtr = OldPtr->PrevPtr;
   NextPtr = OldPtr->NextPtr;

   rqptr->HeapSize -= OldPtr->ChunkSize + HeapOverhead;

   if ((NewPtr = (struct HeapStruct*)
        realloc (OldPtr, ChunkSize + HeapOverhead)) == NULL)
   {
      Accounting.HeapReallocFailCount++;
      /* remove this former chunk from the linked list */
      if (PrevPtr == NULL)
         rqptr->HeapHeadPtr = NextPtr;
      else
         PrevPtr->NextPtr = NextPtr;
      if (NextPtr == NULL)
         rqptr->HeapTailPtr = PrevPtr;
      else
         NextPtr->PrevPtr = PrevPtr;
      return (NULL);
   }

   /** if (Debug) fprintf (stdout, "NewPtr: %d\n", NewPtr); **/
   rqptr->HeapSize += ChunkSize + HeapOverhead;

   if (NewPtr != OldPtr)
   {
      if (PrevPtr != NULL) PrevPtr->NextPtr = NewPtr;
      if (NextPtr != NULL) NextPtr->PrevPtr = NewPtr;
      NewPtr->PrevPtr = PrevPtr;
      NewPtr->NextPtr = NextPtr;
      if (rqptr->HeapHeadPtr == OldPtr) rqptr->HeapHeadPtr = NewPtr;
      if (rqptr->HeapTailPtr == OldPtr) rqptr->HeapTailPtr = NewPtr;
   }

   NewPtr->ChunkSize = ChunkSize;

   /** if (Debug) HeapDebug (rqptr); **/

   return ((unsigned char*)NewPtr + HeapOverhead);
}

/*****************************************************************************/
/*
Dispose of the thread's heap of dynamic memory by traversing the linked list
and disposing of each memory chunk.
*/ 

HeapFree (struct RequestStruct *rqptr)

{
   register struct HeapStruct  *cptr, *hptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HeapFree() %d bytes\n", rqptr->HeapSize);

   hptr = rqptr->HeapHeadPtr;
   while (hptr)
   {
      cptr = hptr;
      hptr = hptr->NextPtr;
/**
      if (Debug)
         fprintf (stdout, "free() %d|%d|%d|%d|%d|%30.30s|\n",
                  cptr, cptr->PrevPtr, cptr->NextPtr, cptr->ChunkSize,
                  (unsigned char*)cptr+HeapOverhead,
                  (unsigned char*)cptr+HeapOverhead);
**/
      free (cptr);
   }
   rqptr->HeapHeadPtr = rqptr->HeapTailPtr = NULL;
   rqptr->HeapSize = 0;
}

/*****************************************************************************/
/*
Return the size of the data storage portion of the specified heap chunk.
*/ 

HeapChunkSize (unsigned char *ChunkPtr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "HeapChunkSize() %d is %d bytes\n",
               ChunkPtr,
               ((struct HeapStruct*)(ChunkPtr - HeapOverhead))->ChunkSize);

   return (((struct HeapStruct*)(ChunkPtr - HeapOverhead))->ChunkSize);
}

/*****************************************************************************/
/*
Display the thread's heap of dynamic memory by traversing the linked list
and displaying the contents of each memory chunk.
*/ 

HeapDebug (struct RequestStruct *rqptr)

{
   register struct HeapStruct  *hptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HeapDebug() %d bytes\n", rqptr->HeapSize);

   hptr = rqptr->HeapHeadPtr;
   while (hptr)
   {
      fprintf (stdout, "%d|%d|%d|%d|%d|%30.30s|\n",
               hptr, hptr->PrevPtr, hptr->NextPtr, hptr->ChunkSize,
               (unsigned char*)hptr+HeapOverhead,
               (unsigned char*)hptr+HeapOverhead);
      hptr = hptr->NextPtr;
   }
}

/*****************************************************************************/
/*
Generic list handling function.  Add entry to head of list.
*/ 

void* ListAddHead
(
register struct GenericListStruct *lptr,
register struct GenericListEntryStruct *eptr,
void *DataPtr
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout, "ListAddHead() %d %d\n", lptr, eptr);
      ListDebug (lptr);
   }

   if (lptr->HeadPtr == NULL)
   {
      /* empty list */
      lptr->HeadPtr = lptr->TailPtr = eptr;
      eptr->PrevPtr = eptr->NextPtr = NULL;
      if (DataPtr != NULL) eptr->DataPtr = (void*)DataPtr;
      lptr->EntryCount++;
      if (Debug) ListDebug (lptr);
      return (eptr);
   }
   else
   {
      /* non-empty list */
      eptr->PrevPtr = NULL;
      eptr->NextPtr = lptr->HeadPtr;
      eptr->NextPtr->PrevPtr = lptr->HeadPtr = eptr;
      if (DataPtr != NULL) eptr->DataPtr = (void*)DataPtr;
      lptr->EntryCount++;
      if (Debug) ListDebug (lptr);
      return (eptr);
   }
}

/*****************************************************************************/
/*
Generic list handling function.  Add entry to tail of list.
*/ 

void* ListAddTail
(
register struct GenericListStruct *lptr,
register struct GenericListEntryStruct *eptr,
void *DataPtr
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout, "ListAddTail() %d %d\n", lptr, eptr);
      ListDebug (lptr);
   }

   if (lptr->HeadPtr == NULL)
   {
      /* empty list */
      lptr->HeadPtr = lptr->TailPtr = eptr;
      eptr->PrevPtr = eptr->NextPtr = NULL;
      if (DataPtr != NULL) eptr->DataPtr = (void*)DataPtr;
      lptr->EntryCount++;
      if (Debug) ListDebug (lptr);
      return (eptr);
   }
   else
   {
      /* non-empty list */
      eptr->NextPtr = NULL;
      eptr->PrevPtr = lptr->TailPtr;
      eptr->PrevPtr->NextPtr = lptr->TailPtr = eptr;
      if (DataPtr != NULL) eptr->DataPtr = (void*)DataPtr;
      lptr->EntryCount++;
      if (Debug) ListDebug (lptr);
      return (eptr);
   }
}

/*****************************************************************************/
/*
Generic list handling function.  Remove entry from list.  Check first that
entry looks legitimate, return NULL if not!
*/ 

void* ListRemove
(
register struct GenericListStruct *lptr,
register struct GenericListEntryStruct *eptr
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout, "ListRemove() %d %d\n", lptr, eptr);
      ListDebug (lptr);
   }

   if (eptr->PrevPtr == NULL)
   {                                                                        
      /* at head of list */
      if (eptr->NextPtr == NULL)
      {
         /* only entry in list */
         if (lptr->HeadPtr != eptr || lptr->TailPtr != eptr) return (NULL);
         lptr->HeadPtr = lptr->TailPtr = NULL;
         if (lptr->EntryCount) lptr->EntryCount--;
         if (Debug) ListDebug (lptr);
         return (eptr);
      }
      else
      {
         /* remove from head of list */
         if (lptr->HeadPtr != eptr) return (NULL);
         eptr->NextPtr->PrevPtr = NULL;
         lptr->HeadPtr = eptr->NextPtr;
         if (lptr->EntryCount) lptr->EntryCount--;
         if (Debug) ListDebug (lptr);
         return (eptr);
      }
   }
   else
   {
      /* not at head of list */
      if (eptr->NextPtr == NULL)
      {
         /* at tail of list */
         if (lptr->TailPtr != eptr) return (NULL);
         eptr->PrevPtr->NextPtr = NULL;
         lptr->TailPtr = eptr->PrevPtr;
         if (lptr->EntryCount) lptr->EntryCount--;
         if (Debug) ListDebug (lptr);
         return (eptr);
      }
      else
      {
         /* somewhere in the middle! */
         if (eptr->PrevPtr->NextPtr != eptr ||
             eptr->NextPtr->PrevPtr != eptr) return (NULL);
         eptr->PrevPtr->NextPtr = eptr->NextPtr;
         eptr->NextPtr->PrevPtr = eptr->PrevPtr;
         if (lptr->EntryCount) lptr->EntryCount--;
         if (Debug) ListDebug (lptr);
         return (eptr);
      }
   }
}

/*****************************************************************************/
/*
Generic list handling function.  Add entry to tail of list.
*/ 

ListDebug (register struct GenericListStruct* lptr)

{
   register struct GenericListEntryStruct  *eptr;

   /*********/
   /* begin */
   /*********/

   fprintf (stdout, "ListDebug() %d %d %d %d\n",
            lptr, lptr->HeadPtr, lptr->TailPtr, lptr->EntryCount);

   for (eptr = lptr->HeadPtr; eptr != NULL; eptr = eptr->NextPtr)
       fprintf (stdout, "%d <- %d -> %d\n",
                eptr->PrevPtr, eptr, eptr->NextPtr);
}

/*****************************************************************************/
/*
Create an HTTP format local time string in the storage pointed at by
'TimeString', e.g. "Fri, 25 Aug 1995 17:32:40" (RFC1123). If 'BinTimePtr' is
null the time string represents the current time. If it points to a quadword,
VMS time value the string represents that time. 'TimeString' must point to
storage large enough for 31 characters.
*/

int HttpLocalTimeString
(
char *TimeString,
unsigned long *BinTimePtr
)
{
   /* alternate formats for time string */
#ifdef DD_DASH_MMM_DASH_YYYY
   /* day, dd mmm yyyy hh:mm */
   static $DESCRIPTOR (HttpTimeFaoDsc, "!3AZ, !2ZW !3AZ !4ZW !2ZW:!2ZW:!2ZW");
#else
   /* weekday, dd-mmm-yyyy hh:mm */
   static $DESCRIPTOR (HttpTimeFaoDsc, "!3AZ, !2ZW-!3AZ-!4ZW !2ZW:!2ZW:!2ZW");
#endif

   static $DESCRIPTOR (TimeStringDsc, "");

   int  status;
   unsigned long  BinTime [2];
   unsigned short  Length;
   unsigned short  NumTime [7];
   unsigned long  DayOfWeek;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpLocalTimeString()\n");

   if (BinTimePtr == NULL)
      sys$gettim (&BinTime);
   else
   {
      BinTime[0] = BinTimePtr[0];
      BinTime[1] = BinTimePtr[1];
   }

   status = sys$numtim (&NumTime, &BinTime);
   if (Debug)
      fprintf (stdout, "sys$numtim() %%X%08.08X %d %d %d %d %d %d %d\n",
               status, NumTime[0], NumTime[1], NumTime[2],
               NumTime[3], NumTime[4], NumTime[5], NumTime[6]);

   if (VMSnok (status = lib$day_of_week (&BinTime, &DayOfWeek)))
      return (status);
   if (Debug)
      fprintf (stdout, "lib$day_of_week() %%X%08.08X is %d\n",
               status, DayOfWeek);

   /* set the descriptor address and size of the resultant time string */
   TimeStringDsc.dsc$w_length = 25;
   TimeStringDsc.dsc$a_pointer = TimeString;

   status = sys$fao (&HttpTimeFaoDsc, &Length, &TimeStringDsc,
                     DayName[DayOfWeek], NumTime[2], MonthName[NumTime[1]],
                     NumTime[0], NumTime[3], NumTime[4], NumTime[5]);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      if (Debug) fprintf (stdout, "sys$fao() %%X%08.08X\n", status);
      TimeString[0] = '\0';
   }
   else
      TimeString[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", TimeString);

   return (status);
}

/*****************************************************************************/
/*
Create an HTTP format Greenwich Mean Time (UTC) time string in the storage 
pointed at by 'TimeString', e.g. "Fri, 25 Aug 1995 17:32:40 GMT" (RFC 1123). 
This must be at least 30 characters capacity.  If 'BinTimePtr' is null the 
time string represents the current time.  If it points to a quadword, VMS time 
value the string represents that time.  'TimeString' must point to storage 
large enough for 31 characters.
*/

int HttpGmTimeString
(
char *TimeString,
unsigned long *BinTimePtr
)
{
   /* alternate formats for time string */
#ifdef DD_DASH_MMM_DASH_YYYY
   /* day, dd mmm yyyy hh:mm */
   static $DESCRIPTOR (HttpTimeFaoDsc,
          "!3AZ, !2ZW !3AZ !4ZW !2ZW:!2ZW:!2ZW GMT");
#else
   /* weekday, dd-mmm-yyyy hh:mm */
   static $DESCRIPTOR (HttpTimeFaoDsc,
          "!3AZ, !2ZW-!3AZ-!4ZW !2ZW:!2ZW:!2ZW GMT");
#endif

   static $DESCRIPTOR (TimeStringDsc, "");

   int  status;
   unsigned long  BinTime [2],
                  GmTime [2];
   unsigned short  Length;
   unsigned short  NumTime [7];
   unsigned long  DayOfWeek;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpGmTimeString()\n");

   if (BinTimePtr == NULL)
      sys$gettim (&GmTime);
   else
   {
      GmTime[0] = BinTimePtr[0];
      GmTime[1] = BinTimePtr[1];
   }
   if (VMSnok (status = TimeAdjustGMT (true, &GmTime)))
      return (status);

   status = sys$numtim (&NumTime, &GmTime);
   if (Debug)
      fprintf (stdout, "sys$numtim() %%X%08.08X %d %d %d %d %d %d %d\n",
               status, NumTime[0], NumTime[1], NumTime[2],
               NumTime[3], NumTime[4], NumTime[5], NumTime[6]);

   if (VMSnok (status = lib$day_of_week (&GmTime, &DayOfWeek)))
      return (status);
   if (Debug)
      fprintf (stdout, "lib$day_of_week() %%X%08.08X is %d\n",
               status, DayOfWeek);

   /* set the descriptor address and size of the resultant time string */
   TimeStringDsc.dsc$w_length = 29;
   TimeStringDsc.dsc$a_pointer = TimeString;

   status = sys$fao (&HttpTimeFaoDsc, &Length, &TimeStringDsc,
                     DayName[DayOfWeek], NumTime[2], MonthName[NumTime[1]],
                     NumTime[0], NumTime[3], NumTime[4], NumTime[5]);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      if (Debug) fprintf (stdout, "sys$fao() %%X%08.08X\n", status);
      TimeString[0] = '\0';
   }
   else
      TimeString[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", TimeString);

   return (status);
}

/*****************************************************************************/
/*
Given a string such as "Fri, 25 Aug 1995 17:32:40 GMT" (RFC 1123), or "Friday,
25-Aug-1995 17:32:40 GMT" (RFC 1036), create an internal, local, binary time
with the current GMT offset.  See complementary function HttpGmTimeString().
*/ 

int HttpGmTime
(
char *TimeString,
unsigned long *BinTimePtr
)
{
   static char  *MonthName [] =
      { "", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

   register char  *tptr;

   int  status;
   unsigned short  Length;
   unsigned short  NumTime [7] = { 0,0,0,0,0,0,0 };
   unsigned long  DayOfWeek;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpGmTime() |%s|\n", TimeString);

   tptr = TimeString;
   /* hunt straight for the comma after the weekday name! */
   while (*tptr && *tptr != ',') tptr++;
   if (*tptr) tptr++;
   /* span white space between weekday name and date */
   while (*tptr && isspace(*tptr)) tptr++;
   /** if (Debug) fprintf (stdout, "tptr |%s|\n", tptr); **/
   if (!*tptr) return (STS$K_ERROR);

   /* get the date and then skip to month name */
   if (isdigit(*tptr)) NumTime[2] = atoi (tptr);
   while (*tptr && isdigit(*tptr)) tptr++;
   while (*tptr && (*tptr == '-' || isspace(*tptr))) tptr++;
   /** if (Debug) fprintf (stdout, "tptr |%s|\n", tptr); **/
   if (!*tptr) return (STS$K_ERROR);

   /* get the month number from the name and skip to the year */
   for (NumTime[1] = 1; NumTime[1] <= 12; NumTime[1]++)
      if (strsame (tptr, MonthName[NumTime[1]], 3)) break;
   if (NumTime[1] > 12) return (STS$K_ERROR);
   while (*tptr && isalpha(*tptr)) tptr++;
   while (*tptr && (*tptr == '-' || isspace(*tptr))) tptr++;
   /** if (Debug) fprintf (stdout, "tptr |%s|\n", tptr); **/
   if (!*tptr) return (STS$K_ERROR);

   /* get the year and then skip to the hour */
   if (isdigit(*tptr))
   {
      NumTime[0] = atoi (tptr);
      if (NumTime[0] < 100) NumTime[0] += 1900;
   }
   while (*tptr && isdigit(*tptr)) tptr++;
   while (*tptr && isspace(*tptr)) tptr++;
   /** if (Debug) fprintf (stdout, "tptr |%s|\n", tptr); **/
   if (!*tptr) return (STS$K_ERROR);

   /* get the hour, minute and second */
   if (isdigit(*tptr)) NumTime[3] = atoi (tptr);
   while (*tptr && isdigit(*tptr)) tptr++;
   if (*tptr == ':') tptr++;
   if (isdigit(*tptr)) NumTime[4] = atoi (tptr);
   while (*tptr && isdigit(*tptr)) tptr++;
   if (*tptr == ':') tptr++;
   if (isdigit(*tptr)) NumTime[5] = atoi (tptr);
   while (*tptr && isdigit(*tptr)) tptr++;
   if (*tptr == ':') tptr++;
   if (!*tptr) return (STS$K_ERROR);

   /* the only thing remaining should be the "GMT" */
   while (*tptr && isspace(*tptr)) tptr++;
   /** if (Debug) fprintf (stdout, "tptr |%s|\n", tptr); **/
   if (!strsame (tptr, "GMT", 3)) return (STS$K_ERROR);

   /*******************************************/
   /* convert what looks like legitimate GMT! */
   /*******************************************/

   if (Debug)
      fprintf (stdout, "NumTime[] %d %d %d %d %d %d %d\n",
               NumTime[0], NumTime[1], NumTime[2], NumTime[3], 
               NumTime[4], NumTime[5], NumTime[6]); 
   status = lib$cvt_vectim (&NumTime, BinTimePtr);
   if (VMSnok (status)) return (status);
   if (Debug) fprintf (stdout, "lib$cvt_vectim() %%X%08.08X\n", status);

   return (TimeAdjustGMT (false, BinTimePtr));
}

/*****************************************************************************/
/*
Translate the logical HTTPD$GMT (defined to be something like "+10:30" or "-
01:15") and convert it into a delta time structure and store in 
'TimeGmtDeltaBinary'.  Store whether it is in advance or behind GMT in boolean 
'TimeAheadOfGmt'.  Store the logical string in 'TimeGmtString'.
*/

int TimeSetGMT ()

{
   register char  *cptr, *sptr;

   int  status;
   unsigned short  Length;
   $DESCRIPTOR (GmtLogicalNameDsc, "HTTPD$GMT");
   $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");
   $DESCRIPTOR (TimeGmtVmsStringDsc, "");
   struct {
      short int  buf_len;
      short int  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   } LnmItems [] =
   {
      { sizeof(TimeGmtString)-1, LNM$_STRING, TimeGmtString, &Length },
      { 0,0,0,0 }
   };

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TimeSetGMT()\n");

   status = sys$trnlnm (0, &LnmFileDevDsc, &GmtLogicalNameDsc, 0, &LnmItems);
   if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   TimeGmtString[Length] = '\0';
   if (Debug) fprintf (stdout, "TimeGmtString |%s|\n", TimeGmtString);
   if (*(cptr = TimeGmtString) == '-')
      TimeAheadOfGmt = false;
   else
      TimeAheadOfGmt = true;
   if (*cptr == '+' || *cptr == '-') cptr++;
   sptr = TimeGmtVmsString;
   *sptr++ = '0';
   *sptr++ = ' ';
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';
   if (Debug) fprintf (stdout, "TimeGmtVmsString |%s|\n", TimeGmtVmsString);

   TimeGmtVmsStringDsc.dsc$w_length = sptr - TimeGmtVmsString;
   TimeGmtVmsStringDsc.dsc$a_pointer = TimeGmtVmsString;

   if (VMSnok (status =
       sys$bintim (&TimeGmtVmsStringDsc, &TimeGmtDeltaBinary)))
      return (status);
   if (TimeGmtDeltaBinary[0] || TimeGmtDeltaBinary[1])
      return (status);
   /* time must have been zero, make it one, one-hundreth of a second */
   TimeGmtDeltaBinary[0] = -100000;
   TimeGmtDeltaBinary[1] = -1;
}

/*****************************************************************************/
/*
The GMT is generated by calculating using an offset using 
'TimeGmtDeltaOffset' and boolean 'TimeAheadOfGmt'.  Adjust either to or from 
GMT.
*/ 

int TimeAdjustGMT
(
boolean ToGmTime,
unsigned long *BinTimePtr
)
{
   int  status;
   unsigned long  AdjustedTime [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TimeAdjustGMT() ToGmTime: %d\n", ToGmTime);

   if ((ToGmTime && TimeAheadOfGmt) || (!ToGmTime && !TimeAheadOfGmt))
   {
      /* to GMT from local and ahead of GMT, or to local from GMT and behind */
      status = lib$sub_times (BinTimePtr, &TimeGmtDeltaBinary, &AdjustedTime);
      if (Debug) fprintf (stdout, "lib$sub_times() %%X%08.08X\n", status);
   }
   else
   {
      /* to GMT from local and behind GMT, or to local from GMT and ahead */
      status = lib$add_times (BinTimePtr, &TimeGmtDeltaBinary, &AdjustedTime);
      if (Debug) fprintf (stdout, "lib$add_times() %%X%08.08X\n", status);
   }

   if (Debug)
   {
      unsigned short  Length;
      char  String [64];
      $DESCRIPTOR (AdjustedTimeFaoDsc, "AdjustedTime: |!%D|\n");
      $DESCRIPTOR (StringDsc, String);

      sys$fao (&AdjustedTimeFaoDsc, &Length, &StringDsc, &AdjustedTime);
      String[Length] = '\0';
      fputs (String, stdout);
   }

   BinTimePtr[0] = AdjustedTime[0];
   BinTimePtr[1] = AdjustedTime[1];

   return (status);
}

/*****************************************************************************/
/*
Return a pointer to the date and time for the supplied binary time.
If a NULL pointer to the binary time then return current day of week.
As it returns a pointer to static area this will need to be copied out if the
contents must be preserved between calls.  'Precision' is the number of
characters in the time, '0' (or 23) for full hundreths, 20 to include seconds,
17 for minutes only.
*/

char* DateTime
(
unsigned long *BinTimePtr,
int Precision
)
{
   static char  DateTimeString [48];
   static $DESCRIPTOR (DateTimeFaoDsc, "!#%D\0");

   int  status;
   unsigned long  BinTime[2];
   $DESCRIPTOR (DateTimeStringDsc, DateTimeString);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DateTime()\n");

   if (BinTimePtr == NULL) sys$gettim (BinTimePtr = &BinTime);
   if (Precision <= 0 || Precision > 23) Precision = 23;
   if (!BinTimePtr[0] && !BinTimePtr[1])
      DateTimeString[0] = '\0';
   else
   if (VMSnok (sys$fao (&DateTimeFaoDsc, 0, &DateTimeStringDsc,
               Precision, BinTimePtr)))
      strcpy (DateTimeString, "*ERROR*");
   return (DateTimeString);
}

/*****************************************************************************/
/*
Return a pointer to the day name, date and time for the supplied binary time.
If a NULL pointer to the binary time then return current day of week.
As it returns a pointer to static area this will need to be copied out if the
contents must be preserved between calls.  'Precision' is the number of
characters in the time, '0' (or 23) for full hundreths, 20 to include seconds,
17 for minutes only.
*/

char* DayDateTime
(
unsigned long *BinTimePtr,
int Precision
)
{
   static char  DayDateTimeString [36];
   static $DESCRIPTOR (DayDateTimeFaoDsc, "!AZ, !#%D\0");

   int  status;
   unsigned long  DayOfWeek;
   unsigned long  BinTime[2];
   $DESCRIPTOR (DayDateTimeStringDsc, DayDateTimeString);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DayDateTime()\n");

   if (BinTimePtr == NULL) sys$gettim (BinTimePtr = &BinTime);
   if (Precision <= 0 || Precision > 23) Precision = 23;
   if (!BinTimePtr[0] && !BinTimePtr[1])
      DayDateTimeString[0] = '\0';
   else
   if (VMSnok (lib$day_of_week (BinTimePtr, &DayOfWeek)))
      strcpy (DayDateTimeString, "*ERROR*");
   else
   if (VMSnok (sys$fao (&DayDateTimeFaoDsc, 0, &DayDateTimeStringDsc,
               DayName[DayOfWeek], Precision, BinTimePtr)))
      strcpy (DayDateTimeString, "*ERROR*");
   return (DayDateTimeString);
}

/*****************************************************************************/
/*
Return a pointer to the name of the day of the week for the supplied binary
time.  If a NULL pointer to the binary time then return current day of week.
*/

char* DayOfWeekName (unsigned long *BinTimePtr)

{
   int  status;
   unsigned long  DayOfWeek;
   unsigned long  BinTime[2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DayOfWeekName()\n");

   if (BinTimePtr == NULL) sys$gettim (BinTimePtr = &BinTime);
   if (VMSnok (status = lib$day_of_week (BinTimePtr, &DayOfWeek)))
      DayOfWeek = 0;
   if (Debug)
      fprintf (stdout, "lib$day_of_week() %%X%08.08X is %d\n",
               status, DayOfWeek);
   return (DayName[DayOfWeek]);
}

/*****************************************************************************/
/*
If the content length has changed, or if it has been modified since the
specified date and time then return a normal status indicating that the object
has been modified.  If not modified then creat a 304 HTTP header and return a
LIB$_NEGTIM status to indicate the file should not be sent.  With VMS'
fractions of a second, add one second to the specified 'since' time to ensure a
reliable comparison.

Implements defacto HTTP/1.0 persistent connections.  Currently we only provide
"keep-alive" response if we're sending a file in binary mode and know its
precise length. An accurate "Content-Length:" field is vital for the client's
correct reception of the transmitted data.  The only other time a "keep-alive"
response can be provided is HERE, where a 304 header is returned (it has a
content length of precisely zero!)
*/ 
 
int HttpIfModifiedSince
(
struct RequestStruct *rqptr,
int ContentLength
)
{
   static unsigned long  OneSecondDelta [2] = { -10000000, -1 };

   register struct FileTaskStruct  *tkptr;

   int  status;
   unsigned long  AdjustedBinTime[2],
                  ScratchBinTime [2];
   char  *KeepAlivePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "HttpIfModifiedSince() %d %d\n",
               ContentLength, rqptr->IfModifiedSinceLength);

   tkptr = rqptr->FileTaskPtr;

   if (rqptr->IfModifiedSinceLength >= 0 &&
       ContentLength >= 0 &&
       rqptr->IfModifiedSinceLength != ContentLength)
      return (SS$_NORMAL);

   /* add one second to the modified time, ensures a negative time */
   if (VMSnok (status =
       lib$add_times (&rqptr->IfModifiedSinceBinaryTime,
                      &OneSecondDelta, &AdjustedBinTime)))
   {
      rqptr->ErrorTextPtr = "sys$add_times()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      return (status);
   }
   if (Debug) fprintf (stdout, "sys$add_times() %%X%08.08X\n", status);

   /* if a positive time results the file has been modified */
   if (VMSok (status =
       lib$sub_times (&tkptr->ModifiedBinaryTime,
                      &AdjustedBinTime, &ScratchBinTime)))
      return (status);

   if (Debug) fprintf (stdout, "sys$sub_times() %%X%08.08X\n", status);
   if (status != LIB$_NEGTIM)
   {
      rqptr->ErrorTextPtr = "sys$sub_times()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      return (status);
   }

   /*****************/
   /* not modified! */
   /*****************/

   rqptr->ResponseStatusCode = 304;

   /* return a "keep-alive" response field if configured and was requested */
   if (Config.KeepAliveTimeoutSeconds && rqptr->KeepAliveRequest)
   {
      rqptr->KeepAliveResponse = true;
      rqptr->KeepAliveCount++;
      KeepAlivePtr = KeepAliveHttpHeader;
   }
   else
      KeepAlivePtr = "";

   /* note the zero content length! */
   if ((rqptr->ResponseHeaderPtr =
        HttpHeader (rqptr, rqptr->ResponseStatusCode, NULL, 0,
                    NULL, KeepAlivePtr)) == NULL)
   {
      return (STS$K_ERROR);
      return;
   }

   return (LIB$_NEGTIM);
}

/*****************************************************************************/
/*
Returns a pointer to a HeapAlloc()ed string containing an HTTP response header.
'OtherHeaderPtr' is a catch-all parameter.  This string is directly included as
part of the HTTP header and so should contain correct carriage control, e.g.
"Keep-Alive:\r\n".
*/

char* HttpHeader
(
struct RequestStruct *rqptr,
int ResponseStatusCode,
char *ContentTypePtr,
int ContentLength,
unsigned long *ModifiedBinaryTimePtr,
char *OtherHeaderPtr
)
{
   static char  ContentTypeString [128] = "Content-Type: ",
                ExpiresString [64] = "Expires: ",
                ModifiedString [64] = "Last-Modified: ";

   static $DESCRIPTOR (ContentLengthFaoDsc, "Content-Length: !UL\r\n\0");

   static $DESCRIPTOR (HeaderFaoDsc,
"!AZ !UL !AZ\r\n\
Server: !AZ\r\n\
Date: !AZ\r\n\
!AZ\
!AZ\
!AZ\
!AZ\
!AZ\
\r\n");

   static $DESCRIPTOR (StringDsc, "");

   register char  *cptr, *sptr, *zptr;
   register unsigned long  *vecptr;

   int  status,
        ResponseStatusGroup;
   unsigned short  Length;
   unsigned long  FaoVector [16];
   char  *ExpiresStringPtr,
         *HeaderPtr,
         *ModifiedStringPtr;
   char  ContentLengthString [32],
         Modified [48],
         String [2048];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "HttpHeader() %d |%s| %d\n",
               ResponseStatusCode, ContentTypePtr, ContentLength);

   if (ContentTypePtr == NULL)
      ContentTypePtr = "";
   else
   {
      zptr = (sptr = ContentTypeString) + sizeof(ContentTypeString);
      sptr += 14;
      for (cptr = ContentTypePtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\r';
      if (sptr < zptr) *sptr++ = '\n';
      if (sptr >= zptr)
      {
         ErrorGeneral (rqptr, ErrorStringSize, __FILE__, __LINE__);
         return (NULL);
      }
      *sptr = '\0';
      ContentTypePtr = ContentTypeString;
   }

   if (ContentLength == -1)
      ContentLengthString[0] = '\0';
   else
   {
      StringDsc.dsc$a_pointer = ContentLengthString;
      StringDsc.dsc$w_length = sizeof(ContentLengthString) - 1;
      status = sys$fao (&ContentLengthFaoDsc, 0, &StringDsc, ContentLength);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         return (NULL);
      }
   }

   if (ModifiedBinaryTimePtr == NULL)
      ModifiedStringPtr = "";
   else
   {
      if (VMSnok (status = HttpGmTimeString (Modified, ModifiedBinaryTimePtr)))
         return (NULL);

      sptr = (ModifiedStringPtr = ModifiedString) + 15;
      for (cptr = Modified; *cptr; *sptr++ = *cptr++);
      *sptr++ = '\r';
      *sptr++ = '\n';
      *sptr = '\0';
   }

   if (rqptr->ResponsePreExpired)
   {
      if (ModifiedBinaryTimePtr == NULL)
      {
         sptr = (ModifiedStringPtr = ModifiedString) + 15;
         for (cptr = rqptr->GmDateTime; *cptr; *sptr++ = *cptr++);
         *sptr++ = '\r';
         *sptr++ = '\n';
         *sptr = '\0';
      }
      /* including the terminating null */
      memcpy ((ExpiresStringPtr = ExpiresString) + 9,
              "Thu, 01-Jan-1970 00:00:01 GMT\r\n", 32);
   }
   else
      ExpiresStringPtr = "";

   if (!ResponseStatusCode) ResponseStatusCode = 200;
   if (OtherHeaderPtr == NULL) OtherHeaderPtr = "";

   vecptr = FaoVector;

   *vecptr++ = HttpProtocol;
   *vecptr++ = ResponseStatusCode;
   if ((ResponseStatusGroup = ResponseStatusCode / 100) == 1)
      *vecptr++ = "Informational";
   else
   if (ResponseStatusGroup == 2)
      *vecptr++ = "Success";
   else
   if (ResponseStatusCode == 304)
      *vecptr++ = "Not Modified";
   else
   if (ResponseStatusGroup == 3)
      *vecptr++ = "Redirection";
   else
   if (ResponseStatusGroup == 4)
      *vecptr++ = "Client Error";
   else
   if (ResponseStatusGroup == 5)
      *vecptr++ = "Server Error";
   else
      *vecptr++ = "?";

   *vecptr++ = SoftwareID;
   *vecptr++ = rqptr->GmDateTime;
   *vecptr++ = ContentTypePtr;
   *vecptr++ = ContentLengthString;
   *vecptr++ = ModifiedStringPtr;
   *vecptr++ = ExpiresStringPtr;
   *vecptr++ = OtherHeaderPtr;

   StringDsc.dsc$a_pointer = String;
   StringDsc.dsc$w_length = sizeof(String) - 1;

   status = sys$faol (&HeaderFaoDsc, &Length, &StringDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      return (NULL);
   }
   String[Length] = '\0';

   if ((HeaderPtr = HeapAlloc (rqptr, Length+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      return (NULL);
   }
   memcpy (HeaderPtr, String, Length+1);
   rqptr->ResponseHeaderLength = Length;

   if (Debug) fprintf (stdout, "|%s|\n", String);

   return (HeaderPtr);
}

/*****************************************************************************/
/*
Append a string onto a HTTP header located in HTTPd heap allocated memory. 
This header string must already exist (created by HttpHeader200() or
equivalent).  The new string is appended to the current header, just before the
header terminating empty line.  Any string added must contain correct HTTP
header carriage control.
*/

char* HttpHeaderAppend
(
struct RequestStruct *rqptr,
char *HeaderPtr,
char *StringPtr,
int StringLength
)
{
   char  *cptr;
   int  HeaderLength;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpHeaderAppend()\n|%s|\n", HeaderPtr);

   if (HeaderPtr == NULL) return (NULL);

   HeaderLength = rqptr->ResponseHeaderLength;
   if (StringLength <= 0) StringLength = strlen(StringPtr);

   if ((HeaderPtr =
        HeapRealloc (rqptr, HeaderPtr, HeaderLength+StringLength)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      return (NULL);
   }
   /* point to just before the header terminating empty line */
   cptr = HeaderPtr + HeaderLength - 3;
   memcpy (cptr, StringPtr, StringLength);
   /* add the new header terminating empty line */
   cptr += StringLength;
   *cptr++ = '\r';
   *cptr++ = '\n';
   *cptr = '\0';

   rqptr->ResponseHeaderLength = HeaderLength + StringLength - 1;

   if (Debug)
      fprintf (stdout, "%d |%s|\n", rqptr->ResponseHeaderLength, HeaderPtr);

   return (HeaderPtr);
}

/*****************************************************************************/
/*
A heap string is one located in HTTPd heap allocated memory.  Append the
supplied, null-terminated string to that pointed to by the heap string pointer. 
If this is NULL then allocate new memory, if not reallocate sufficient memory
to append the new string.  Return a pointer the string.
*/

char* HeapStringAppend
(
struct RequestStruct *rqptr,
char *HeapStringPtr,
char *StringPtr,
int StringLength
)
{
   int  HeapStringLength;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HeapStringAppend()\n|%s|\n", HeapStringPtr);

   if (StringLength <= 0) StringLength = strlen(StringPtr);

   if (HeapStringPtr == NULL)
   {
      if ((HeapStringPtr = HeapAlloc (rqptr, StringLength+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         return (NULL);
      }
      memcpy (HeapStringPtr, StringPtr, StringLength+1);

      if (Debug) fprintf (stdout, "|%s|\n", HeapStringPtr);
      return (HeapStringPtr);
   }
   else
   {
      HeapStringLength = strlen(HeapStringPtr);

      if ((HeapStringPtr =
           HeapRealloc (rqptr, HeapStringPtr,
                        HeapStringLength+StringLength+1))
         == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         return (NULL);
      }
      /* copy just past the original end-of-string (incl. terminating null) */
      memcpy (HeapStringPtr+HeapStringLength, StringPtr, StringLength+1);

      if (Debug) fprintf (stdout, "|%s|\n", HeapStringPtr);
      return (HeapStringPtr);
   }
}

/*****************************************************************************/
/*
A heap string is one located in HTTPd heap allocated memory.  Prepend the
supplied, null-terminated string to that pointed to by the heap string pointer. 
If this is NULL then allocate new memory, if not reallocate sufficient memory
to append the new string.  Return a pointer the string.
*/

char* HeapStringPrepend
(
struct RequestStruct *rqptr,
char *HeapStringPtr,
char *StringPtr,
int StringLength
)
{
   int  HeapStringLength;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HeapStringPrepend()\n|%s|\n", HeapStringPtr);

   if (StringLength <= 0) StringLength = strlen(StringPtr);

   if (HeapStringPtr == NULL)
   {
      if ((HeapStringPtr = HeapAlloc (rqptr, StringLength+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         return (NULL);
      }
      memcpy (HeapStringPtr, StringPtr, StringLength+1);

      if (Debug) fprintf (stdout, "|%s|\n", HeapStringPtr);
      return (HeapStringPtr);
   }
   else
   {
      HeapStringLength = strlen(HeapStringPtr);

      if ((HeapStringPtr =
           HeapRealloc (rqptr, HeapStringPtr,
                        HeapStringLength+StringLength+1))
          == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         return (NULL);
      }
      /* shift the existing string along in memory creating a gap */
      memcpy (HeapStringPtr+HeapStringLength,
              HeapStringPtr,
              HeapStringLength+1);  /* includes terminating null */
      /* copy just past the original end-of-string */
      memcpy (HeapStringPtr, StringPtr, StringLength);

      if (Debug) fprintf (stdout, "|%s|\n", HeapStringPtr);
      return (HeapStringPtr);
   }
}

/*****************************************************************************/
/*
If we can't get SYSPRV when we need it, or more importantly, if we can't turn
it off when we're finished with it the exit the server!
*/ 

EnableSysPrv ()

{
   static unsigned long  SysPrvMask [2] = { PRV$M_SYSPRV, 0 };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "EnableSysPrv()\n");

   if (VMSok (status = sys$setprv (1, &SysPrvMask, 0, 0))) return;
   ErrorExitVmsStatus (status, "enabling SYSPRV", __FILE__, __LINE__);
}

/*****************************************************************************/
/*
If we can't get SYSPRV when we need it, or more importantly, if we can't turn
it off when we're finished with it the exit the server!
*/ 

DisableSysPrv ()

{
   static unsigned long  SysPrvMask [2] = { PRV$M_SYSPRV, 0 };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DisSysPrv() %d\n");

   if (VMSok (status = sys$setprv (0, &SysPrvMask, 0, 0))) return;
   ErrorExitVmsStatus (status, "disabling SYSPRV", __FILE__, __LINE__);
}

/*****************************************************************************/
/*
Returns a pointer to a HeapAlloc()ed string of <META ...> tags for an
internally generated HTML document.
*/

char* HtmlMetaInfo
(
struct RequestStruct *rqptr,
char *VmsInfoPtr
)
{
   static char  Port [16];

   register char  *cptr, *sptr, *zptr;

   char  *MetaPtr;
   char  String [1024];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HtmlMetaInfo()\n");

   if (!Port[0])
   {
      $DESCRIPTOR (PortFaoDsc, "!UL\0");
      $DESCRIPTOR (PortDsc, Port);

      sys$fao (&PortFaoDsc, 0, &PortDsc, ServerPort);
   }

   zptr = (sptr = String) + sizeof(String);

   for (cptr = "<META NAME=\"generator\" CONTENT=\"";
        *cptr && sptr < zptr;
        *sptr++ = *cptr++);
   for (cptr = SoftwareID; *cptr && sptr < zptr; *sptr++ = *cptr++);
   for (cptr = "\">\n"; *cptr && sptr < zptr; *sptr++ = *cptr++);

   for (cptr = "<META NAME=\"date\" CONTENT=\"";
        *cptr && sptr < zptr;
        *sptr++ = *cptr++);
   for (cptr = rqptr->GmDateTime; *cptr && sptr < zptr; *sptr++ = *cptr++);
   for (cptr = "\">\n"; *cptr && sptr < zptr; *sptr++ = *cptr++);

   if (rqptr->RemoteUser[0])
   {
      for (cptr = "<META NAME=\"author\" CONTENT=\"";
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      for (cptr = rqptr->RemoteUser; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '@';
   }
   else
   {
      for (cptr = "<META NAME=\"host\" CONTENT=\"";
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
   }
   for (cptr = ServerHostName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr < zptr) *sptr++ = ':';
   for (cptr = Port; *cptr && sptr < zptr; *sptr++ = *cptr++);
   for (cptr = "\">\n"; *cptr && sptr < zptr; *sptr++ = *cptr++);


   if (VmsInfoPtr != NULL)
      if (VmsInfoPtr[0])
         if (Config.IncludeCommentedInfo)
         {
            for (cptr = "<META NAME=\"VMS\" CONTENT=\"";
                 *cptr && sptr < zptr;
                 *sptr++ = *cptr++);
            for (cptr = VmsInfoPtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
            for (cptr = "\">\n"; *cptr && sptr < zptr; *sptr++ = *cptr++);
         }

   if (sptr >= zptr) return ("<!-- META overflow ERROR -->\n");
   *sptr++ = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", String);

   if ((MetaPtr = HeapAlloc (rqptr, sptr-String)) == NULL)
      return ("<!-- META HeapAlloc() ERROR -->\n");
   memcpy (MetaPtr, String, sptr-String);
   return (MetaPtr);
}

/*****************************************************************************/
/*
Return success if the specified file name exists in the specified directory, 
error otherwise.  This function completes synchronously.
*/ 

int FileExists
(
char *Directory,
char *FileName
)
{
   int  status;
   char  ExpandedFileName [256];
   struct FAB  SearchFab;
   struct NAM  SearchNam;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "FileExists() |%s|%s|\n", Directory, FileName);

   SearchFab = cc$rms_fab;
   SearchFab.fab$l_dna = Directory;
   SearchFab.fab$b_dns = strlen(Directory);
   SearchFab.fab$l_fna = FileName;
   SearchFab.fab$b_fns = strlen(FileName);
   SearchFab.fab$l_fop = FAB$M_NAM;
   SearchFab.fab$l_nam = &SearchNam;
   SearchNam = cc$rms_nam;
   SearchNam.nam$l_esa = ExpandedFileName;
   SearchNam.nam$b_ess = sizeof(ExpandedFileName)-1;

   status = sys$parse (&SearchFab, 0, 0);
   if (Debug) fprintf (stdout, "sys$parse() %%X%08.08X\n", status);

   if (VMSok (status))
   {
      status = sys$search (&SearchFab, 0, 0);
      if (Debug) fprintf (stdout, "sys$search() %%X%08.08X\n", status);
   }

   /* release parse and search internal data structures */
   SearchFab.fab$l_fna = "a:[b]c.d;";
   SearchFab.fab$b_fns = 9;
   SearchFab.fab$b_dns = 0;
   SearchNam.nam$b_nop = NAM$M_SYNCHK;
   sys$parse (&SearchFab, 0, 0);

   return (status);
}

/*****************************************************************************/
/*
There are enough exit points from the function that actually does the logical
name processing that this function wraps it with the appropriate privilege
changes to make sure we don't miss any!
*/ 

DefineMonitorLogicals (struct RequestStruct *rqptr)

{
   static unsigned long  SysNamMask [2] = { PRV$M_SYSNAM, 0 };

   int  status,
        SetPrvStatus;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DefineMonitorLogicals()\n");

   /* turn on SYSNAM to allow access to system table */
   if (VMSnok (status = sys$setprv (1, &SysNamMask, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);
      return (status);
   }

   status = DefineMonitorLogicalsHere (rqptr);

   /* turn off SYSNAM */
   if (VMSnok (SetPrvStatus = sys$setprv (0, &SysNamMask, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", SetPrvStatus);
      /* do NOT keep processing if there is a problem turning off SYSNAM! */
      ErrorExitVmsStatus (SetPrvStatus, "disabling SYSNAM", __FILE__, __LINE__);
   }

   return (status);
}

/*****************************************************************************/
/*
Define two/three logicals used by the HTTPd monitor utility.  The value of the 
count logical is the binary contents of the server counter data structure.  
The value of the request logical is some information on the latest request to 
complete.  The PID logical is defined only once, the first time this function 
is called.
*/ 

DefineMonitorLogicalsHere (struct RequestStruct *rqptr)

{
   static unsigned long  JpiPid;
   static char  CountLogicalName [32],
                RequestLogicalName [32],
                RequestLogicalValue [256];

   static $DESCRIPTOR (CountLogicalNameDsc, CountLogicalName);
   static $DESCRIPTOR (LnmSystemDsc, "LNM$SYSTEM");
   static $DESCRIPTOR (RequestLogicalNameDsc, RequestLogicalName);
   static $DESCRIPTOR (RequestLogicalValueDsc, RequestLogicalValue);
   /* day time HTTP-status client-host method path?query */
   static $DESCRIPTOR (RequestLogicalValueFaoDsc,
          "!2ZL !2ZL:!2ZL:!2ZL!AD!AZ!AD!UL!AD!UL!AD!UL!AD!AZ !AZ!AZ!AZ!AZ!AZ");

   static struct VmsItem LnmCountItem [] =
          {
             { sizeof(Accounting), LNM$_STRING, &Accounting, 0 },
             { 0,0,0,0 }
          };
   static struct VmsItem LnmRequestItem [] =
          {
             { 0, LNM$_STRING, RequestLogicalValue, 0 },
             { 0,0,0,0 }
          };

   register unsigned long  *vecptr;

   boolean  KeepAliveTimeout;
   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [64];
   char *SlashPtr,
        *PathInfoPtr,
        *QueryStringPtr,
        *QuestionMarkPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DefineMonitorLogicalsHere()\n");

   if (!JpiPid)
   {
      /**********************/
      /* first time through */
      /**********************/

      /* local storage */
      char  PidLogicalName [16];
      $DESCRIPTOR (PidLogicalNameDsc, PidLogicalName);
      $DESCRIPTOR (PidLogicalNameFaoDsc, "HTTPD!UL$PID");
      $DESCRIPTOR (CountLogicalNameFaoDsc, "HTTPD!UL$COUNT");
      $DESCRIPTOR (RequestLogicalNameFaoDsc, "HTTPD!UL$REQUEST");
      struct VmsItem  JpiPidItem [] =
      {
         { sizeof(JpiPid), JPI$_PID, &JpiPid, 0 },
         {0,0,0,0}
      };
      struct VmsItem  LnmPidItem [] =
      {
         { sizeof(JpiPid), LNM$_STRING, &JpiPid, 0 },
         { 0,0,0,0 }
      };

      /* define a logical name containing the PID of the server process */
      if (VMSnok (status = sys$getjpiw (0, 0, 0, &JpiPidItem, 0, 0, 0)))
         return (status);
      sys$fao (&PidLogicalNameFaoDsc, &Length, &PidLogicalNameDsc, ServerPort);
      PidLogicalName[PidLogicalNameDsc.dsc$w_length = Length] = '\0';
      if (Debug) fprintf (stdout, "PidLogicalName |%s|\n", PidLogicalName);
      if (VMSnok (status =
          sys$crelnm (0, &LnmSystemDsc, &PidLogicalNameDsc, 0, &LnmPidItem)))
         return (status);

      /* generate approriate names for the count and request logicals */
      sys$fao (&CountLogicalNameFaoDsc, &Length, &CountLogicalNameDsc,
               ServerPort);
      CountLogicalName[CountLogicalNameDsc.dsc$w_length = Length] = '\0';
      if (Debug) fprintf (stdout, "CountLogicalName |%s|\n", CountLogicalName);
      sys$fao (&RequestLogicalNameFaoDsc, &Length, &RequestLogicalNameDsc,
               ServerPort);
      RequestLogicalName[RequestLogicalNameDsc.dsc$w_length = Length] = '\0';
      if (Debug)
         fprintf (stdout, "RequestLogicalName |%s|\n", RequestLogicalName);

      if (!AccountingZeroOnStartup)
      {
         /* attempt to read the values of any previous count logical name */
         if (VMSok (status =
             sys$trnlnm (0, &LnmSystemDsc, &CountLogicalNameDsc, 0,
                         &LnmCountItem)))
            Accounting.StartupCount++;
         else
         {
            if (status != SS$_NOLOGNAM)
               return (status);
         }
      }
   }

   /*****************************************/
   /* (re)define count and request logicals */
   /*****************************************/

   if (rqptr == NULL)
   {
      KeepAliveTimeout = false;
      /* reset the request logical to all empty strings */
      RequestLogicalValue[0] = '\0';
      LnmRequestItem[0].buf_len = 1;
   }
   else
   if (!(KeepAliveTimeout = rqptr->KeepAliveTimeout))
   {
      if (rqptr->PathInfoPtr == NULL)
         PathInfoPtr = SlashPtr = "";
      else
      {
         PathInfoPtr = rqptr->PathInfoPtr;
         if (PathInfoPtr[0] == '/')
            SlashPtr = "";
         else
            SlashPtr = "/";
      }

      if (rqptr->QueryStringPtr == NULL)
      {
         QueryStringPtr = "";
         QuestionMarkPtr = "";
      }
      else
      {
         if ((QueryStringPtr = rqptr->QueryStringPtr)[0])
            QuestionMarkPtr = "?";
         else
            QuestionMarkPtr = "";
      }

      vecptr = FaoVector;
      *vecptr++ = rqptr->NumericTime[2];
      *vecptr++ = rqptr->NumericTime[3];
      *vecptr++ = rqptr->NumericTime[4];
      *vecptr++ = rqptr->NumericTime[5];
      *vecptr++ = 1;
      *vecptr++ = (unsigned long)"\0";
      *vecptr++ = rqptr->ClientHostName;
      *vecptr++ = 1;
      *vecptr++ = (unsigned long)"\0";
      *vecptr++ = rqptr->ResponseStatusCode;
      *vecptr++ = 1;
      *vecptr++ = (unsigned long)"\0";
      *vecptr++ = rqptr->BytesRx;
      *vecptr++ = 1;
      *vecptr++ = (unsigned long)"\0";
      *vecptr++ = rqptr->BytesTx;
      *vecptr++ = 1;
      *vecptr++ = (unsigned long)"\0";
      *vecptr++ = (unsigned long)rqptr->HttpMethodName;
      *vecptr++ = (unsigned long)rqptr->ScriptName;
      *vecptr++ = (unsigned long)SlashPtr;
      *vecptr++ = (unsigned long)PathInfoPtr;
      *vecptr++ = (unsigned long)QuestionMarkPtr;
      *vecptr++ = (unsigned long)QueryStringPtr;

      status = sys$faol (&RequestLogicalValueFaoDsc, &Length,
                         &RequestLogicalValueDsc, &FaoVector);

      if (Debug) fprintf (stdout, "sys$faol() %%X%08.08X\n", status);
      RequestLogicalValue[LnmRequestItem[0].buf_len = Length] = '\0';
      if (Debug)
         fprintf (stdout, "RequestLogicalValue |%s|\n", RequestLogicalValue);
   }

   if (VMSnok (status =
       sys$crelnm (0, &LnmSystemDsc, &CountLogicalNameDsc, 0, &LnmCountItem)))
      return (status);

   if (!KeepAliveTimeout)
   {
      if (VMSnok (status =
          sys$crelnm (0, &LnmSystemDsc, &RequestLogicalNameDsc, 0,
                      &LnmRequestItem)))
         return (status);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Convert a VMS quadword, binary time to a Unix-style time component structure.
*/

boolean TimeVmsToUnix
(
unsigned long *BinTimePtr,
struct tm *TmPtr
)
{
   static unsigned long  LibDayOfWeek = LIB$K_DAY_OF_WEEK;
   static unsigned long  LibDayOfYear = LIB$K_DAY_OF_YEAR;

   int  status;
   unsigned long  DayOfWeek,
                  DayOfYear;
   unsigned short  NumTime [7];
   unsigned long  BinTime [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TimeVmsToUnix()\n");

   if (BinTimePtr == NULL) sys$gettim (BinTimePtr = &BinTime);

   if (VMSnok (status = sys$numtim (&NumTime, BinTimePtr)))
      return (status);

   lib$cvt_from_internal_time (&LibDayOfWeek, &DayOfWeek, BinTimePtr);
   lib$cvt_from_internal_time (&LibDayOfYear, &DayOfYear, BinTimePtr);

   TmPtr->tm_sec = NumTime[5];
   TmPtr->tm_min = NumTime[4];
   TmPtr->tm_hour = NumTime[3];
   TmPtr->tm_mday = NumTime[2];
   TmPtr->tm_mon = NumTime[1] - 1;
   TmPtr->tm_year = NumTime[0] - 1900;
   TmPtr->tm_wday = DayOfWeek;
   TmPtr->tm_yday = DayOfYear;
   TmPtr->tm_isdst = 0;

   if (Debug) fprintf (stdout, "%s", asctime(TmPtr));

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Convert the 32/64 bit integer (depending on architecture) pointed to into an
ASCII number string containing commas.  The destination string should contain
capacity of a minimum 16 characters for 32 bits, or 32 characters for 64 bits.
*/

int CommaNumber
(
int Bits,
unsigned long Value,
char *String
)
{
   static $DESCRIPTOR (Value32FaoDsc, "!UL");
   static $DESCRIPTOR (Value64FaoDsc, "!@UJ");

   register int  cnt;
   register char  *cptr, *sptr;

   int  status;
   char  Scratch [32];
   unsigned short  Length;
   $DESCRIPTOR (ScratchDsc, Scratch);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommaNumber()\n");

   if (Bits > 32)
      status = sys$fao (&Value64FaoDsc, &Length, &ScratchDsc, Value);
   else
      status = sys$fao (&Value32FaoDsc, &Length, &ScratchDsc, Value);
   if (VMSnok (status))
   {
      strcpy (String, "*ERROR*");
      return;
   }
   Scratch[Length] = '\0';
   if (((Length-1) / 3) < 1)
   {
      strcpy (String, Scratch);
      return;
   }
   else
   if (!(cnt = Length % 3))
      cnt = 3;
   
   cptr = Scratch;
   sptr = String;
   while (*cptr)
   {
      if (!cnt--)
      {
         *sptr++ = ',';
         cnt = 2;
      }
      *sptr++ = *cptr++;
   }
   *sptr = '\0';
}

/*****************************************************************************/
/*
If not compiling with DEC C then we need to create our own strftime().  Make 
it a minimal implementation, no attention paid to the format string, etc.
*/ 

#ifndef __DECC

int strftime
(
char *TimeStringPtr,
int SizeOfTimeString,
char *TimeFormatPtr,
struct tm *TmPtr
)
{
   static $DESCRIPTOR (TimeFaoDsc, "!3AZ, !2ZL !3AZ !2ZL, !2ZL:!2ZL:!2ZL");

   unsigned short  Length;
   $DESCRIPTOR (TimeStringPtrDsc, TimeStringPtr);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "strftime() well ... sort of\n");

   TimeStringPtrDsc.dsc$a_pointer = TimeStringPtr;
   TimeStringPtrDsc.dsc$w_length = SizeOfTimeString - 1;

   Length = 0;
   sys$fao (&TimeFaoDsc, &Length, &TimeStringPtrDsc,
            DayName[TmPtr->tm_wday], TmPtr->tm_mday, MonthName[TmPtr->tm_mon],
            TmPtr->tm_year, TmPtr->tm_hour, TmPtr->tm_min, TmPtr->tm_sec);
   TimeStringPtr[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", TimeStringPtr);
   return (Length);
}

#endif

/*****************************************************************************/
/*
*** Will eventually convert all calls to the more robust CopyToHtml() ***

Copy text from one string to another, converting characters forbidden to 
appear as plain-text in HTML.  For example the '<', '&', etc.  Convert these 
to the corresponding HTML character entities.  Returns the number of characters
added to the buffer.
*/ 

int CopyTextIntoHtml
( 
register char *bptr,
register char *tptr,
register int  ccnt
)
{
    char *Buffer;

    Buffer = bptr;
    while (ccnt-- && *tptr)
    {
       switch (*tptr)
       {
          case '<' : memcpy (bptr, "&lt;", 4); bptr += 4; tptr++; break;
          case '>' : memcpy (bptr, "&gt;", 4); bptr += 4; tptr++; break;
          case '&' : memcpy (bptr, "&amp;", 5); bptr += 5; tptr++; break;
          case '\"' : memcpy (bptr, "&quot;", 6); bptr += 6; tptr++; break;
          default : *bptr++ = *tptr++;
       }
    }
    *bptr = '\0';
    return (bptr-Buffer);
}

/*****************************************************************************/
/*
Copy text from one string to another, converting characters forbidden to 
appear as plain-text in HTML.  For example the '<', '&', etc.  Convert these 
to the corresponding HTML character entities.  Returns the number of characters
added to the buffer or -1 to indicate buffer overflow.  Copies a maximum of
'Count' characters, or all id it is -1.
*/ 

int CopyToHtml
( 
char *Destination,
int SizeOfDestination,
char *Source,
int Count
)
{
    register char  *cptr, *sptr, *zptr;
    register int  ccnt;

    cptr = Source;
    zptr = (sptr = Destination) + SizeOfDestination;
    ccnt = Count;
    while (--ccnt && *cptr)
    {
       switch (*cptr)
       {
          case '<' :
             if (sptr+4 >= zptr) return (-1);
             memcpy (sptr, "&lt;", 4); sptr += 4; cptr++; break;
          case '>' :
             if (sptr+4 >= zptr) return (-1);
             memcpy (sptr, "&gt;", 4); sptr += 4; cptr++; break;
          case '&' :
             if (sptr+5 >= zptr) return (-1);
             memcpy (sptr, "&amp;", 5); sptr += 5; cptr++; break;
          case '\"' :
             if (sptr+6 >= zptr) return (-1);
             memcpy (sptr, "&quot;", 6); sptr += 6; cptr++; break;
          default :
             if (sptr >= zptr) return (-1);
             *sptr++ = *cptr++;
       }
    }
    if (sptr >= zptr) return (-1);
    *sptr = '\0';
    return (sptr-Destination);
}

/*****************************************************************************/
/*
A null terminated string is parsed for the next "fieldname=fieldvalue[&]"
pair.  If an error is encountered generated an error message and return NULL,
otherwise return a pointer to the next pair or end-of-string.
*/

char* ParseQueryField
(
struct RequestStruct *rqptr,
char *QueryString,
char *FieldName,
int SizeOfFieldName,
char *FieldValue,
int SizeOfFieldValue,
char *SourceCodeFile,
int SourceCodeLine
)
{
   register char  *cptr, *qptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ParseQueryField() |%s|\n", QueryString);

   qptr = QueryString;
   zptr = (sptr = FieldName) + SizeOfFieldName;
   while (*qptr && *qptr != '=')
   {
      while (*qptr == '\r' || *qptr == '\n') qptr++;
      while (*qptr && *qptr != '=' &&
             *qptr != '\r' && *qptr != '\n' && sptr < zptr) *sptr++ = *qptr++;
   }
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, ErrorStringSize, SourceCodeFile, SourceCodeLine);
      return (NULL);
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "FieldName |%s|\n", FieldName);

   if (*qptr++ != '=')
   {
      rqptr->ResponseStatusCode = 400;
      ErrorGeneral (rqptr, ErrorUrlEncoding, SourceCodeFile, SourceCodeLine);
      return (NULL);
   }

   zptr = (sptr = FieldValue) + SizeOfFieldValue;
   while (*qptr && *qptr != '&')
   {
      while (*qptr == '\r' || *qptr == '\n') qptr++;
      while (*qptr && *qptr != '&' &&
             *qptr != '\r' && *qptr != '\n' && sptr < zptr) *sptr++ = *qptr++;
   }
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, ErrorStringSize, SourceCodeFile, SourceCodeLine);
      return (NULL);
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "FieldValue |%s|\n", FieldValue);

   if (*qptr && *qptr != '&')
   {
      rqptr->ResponseStatusCode = 400;
      ErrorGeneral (rqptr, ErrorUrlEncoding, SourceCodeFile, SourceCodeLine);
      return (NULL);
   }
   if (*qptr) qptr++;

   if (UrlDecodeString (FieldValue) < 0)
   {
      rqptr->ResponseStatusCode = 400;
      ErrorGeneral (rqptr, ErrorUrlEncoding, SourceCodeFile, SourceCodeLine);
      return (NULL);
   }

   return (qptr);
}

/****************************************************************************/
/*
URL-decodes a string in place (can do this because URL-decoded text is always
the same or less than the length of the original).  Returns the number of
characters in the decoded string, or -1 to indicate a URL-encoding error.
*/ 
 
int UrlDecodeString (char *String)

{
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UrlDecodeString() |%s|\n", String);

   cptr = sptr = String;
   while (*cptr)
   {
      switch (*cptr)
      {
         case '=' :
         case '&' :
            /* URL-forbidden characters */
            return (-1);

         case '+' :
            *sptr++ = ' ';
            cptr++;
            break;

         case '%' :
            cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               *sptr = (*cptr - '0') << 4;
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               *sptr = (toupper(*cptr) - '7') << 4;
            else
               return (-1);
            cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               *sptr |= *cptr - '0';
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               *sptr |= toupper(*cptr) - '7';
            else
               return (-1);
            sptr++;
            cptr++;
            break;

         default :
            *sptr++ = *cptr++;
      }
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", String);
   return (sptr - String);
}
 
/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 
 
boolean strsame
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}
 
/****************************************************************************/

